/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
 */
#include <stdint.h>

#include "boards.h"
#include "debug.h"
#include "pru.h"
#include "softdevice_handler.h"
#include "pru_config.h"
#include "app_util_platform.h"
#include "app_print.h"
#include "nrf_nvic.h"


#ifdef DFU_SUPPORT
    #include "common_hal_buttons.h"
#endif

#define CENTRAL_LINK_COUNT               0  /**< Number of central links used by the application. When changing this number remember to adjust the RAM settings*/
#define PERIPHERAL_LINK_COUNT            1  /**< Number of peripheral links used by the application. When changing this number remember to adjust the RAM settings*/

/**@brief Dispatches a BLE stack event to all profiles.
 *
 * @details This function is called from the BLE Stack event interrupt handler after a BLE stack
 *          event has been received.
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 */
static void m_ble_evt_dispatch(ble_evt_t * p_ble_evt)
{
    pru_on_ble_evt(p_ble_evt);
}

/**@brief BLE stack initialization.
 *
 * @details Initializes the SoftDevice and the BLE event interrupt.
 */
static void m_ble_stack_init(void)
{
    uint32_t err_code;
    ble_conn_bw_counts_t count = {{1,0,0},{1,0,0}};

    // Initialize the SoftDevice handler module.
    nrf_clock_lf_cfg_t clock_lf_cfg = NRF_CLOCK_LFCLKSRC;
    SOFTDEVICE_HANDLER_INIT(&clock_lf_cfg, NULL);

    // Enable BLE stack.
    ble_enable_params_t ble_enable_params = {0};
    ble_enable_params.gap_enable_params.periph_conn_count = PERIPHERAL_LINK_COUNT;
    ble_enable_params.common_enable_params.p_conn_bw_counts = &count;

    ble_opt_t opt;
    opt.common_opt.conn_bw.conn_bw.conn_bw_rx = BLE_CONN_BW_HIGH;
    opt.common_opt.conn_bw.conn_bw.conn_bw_tx = BLE_CONN_BW_HIGH;
    opt.common_opt.conn_bw.role = BLE_GAP_ROLE_PERIPH;

    //Check the ram settings against the used number of links
    CHECK_RAM_START_ADDR(CENTRAL_LINK_COUNT,PERIPHERAL_LINK_COUNT);

    err_code = softdevice_enable(&ble_enable_params);
    APP_ERROR_CHECK(err_code);

    err_code = sd_ble_opt_set(BLE_COMMON_OPT_CONN_BW, &opt);
    APP_ERROR_CHECK(err_code);

    err_code = softdevice_ble_evt_handler_set(m_ble_evt_dispatch);
    APP_ERROR_CHECK(err_code);

    // It is important that this is set to the same priority as the
    // timers in the system.
    err_code = sd_nvic_SetPriority(SD_EVT_IRQn, APP_IRQ_PRIORITY_LOW);
    APP_ERROR_CHECK(err_code);
}

/**@brief Application main function.
 * @return Ignored.
 */
int main()
{
    APP_TIMER_INIT(APP_TIMER_PRESCALER, APP_TIMER_OP_QUEUE_SIZE, false);

    m_ble_stack_init();

    pru_init(NULL);

#ifdef DFU_SUPPORT
    // Set up button used for entering DFU mode
    common_hal_buttons_init(PRU_BUTTON_PRESS_LENGTH_DFU_MS, APP_TIMER_PRESCALER);
#endif

    pru_start();

    while(1)
    {
        #ifdef DEBUG_OUT_ENABLE
        debug();
        #endif
    }
}


/** @} */
//lint -restore
