/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
 *
 * @defgroup example_pru_sm PRU example
 * @{
 * @ingroup examples
 */
#include <stdint.h>

#include "debug.h"
#include "softdevice_handler.h"
#include "app_profile.h"
#include "pru_sensors.h"
#include "pru_hw_config.h"
#include "pru_config.h"
#include "nrf_assert.h"
#include "app_print.h"
#include "app_util_platform.h"
#include "nrf_nvic.h"

#ifdef DFU_SUPPORT
    #include "common_hal_buttons.h"
#endif


// Assert setup
#define DEAD_BEEF 0xDEADBEEF                                                            /**< Error code to signify a SoftDevice assert. */

#define VRECT_MEAS_INTERVAL         APP_TIMER_TICKS(50, APP_TIMER_PRESCALER)            /**< Heart rate measurement interval (ticks). */

#define VRECT_MEAS_INTERVAL                 APP_TIMER_TICKS(50, APP_TIMER_PRESCALER)    /**< Heart rate measurement interval (ticks). */
#define PRU_MULTI_PROFILE_TIMER_OP_QUEUE_SIZE (8 + PRU_MAX_APP_TIMERS)             /**< Max number of internal timers. */

static nrf_clock_lf_cfg_t clk_cfg = {
    .source = NRF_CLOCK_LF_SRC_XTAL,
    .xtal_accuracy = NRF_CLOCK_LF_XTAL_ACCURACY_20_PPM};  /**< Same clock source as 'softdevice_handler.c'. */

static bool m_connected = false;                                        /**< Is the device currently connected? */

APP_TIMER_DEF(m_vrect_read_timer);                                      /**< Timer used for measuring VRECT */

static bool m_profile_has_been_switched = false;                        /**< Do we have an unhandled profile switch? */

static const ble_gap_addr_t PRU_ADDR = {.addr_type = BLE_GAP_ADDR_TYPE_RANDOM_STATIC,
                                        .addr = {0xe4, 0x33, 0xba, 0xab, 0xea, 0xee}}; /**< Address used for PRU */

static const ble_gap_addr_t APP_PROFILE_ADDR = {.addr_type = BLE_GAP_ADDR_TYPE_RANDOM_STATIC,
                                        .addr = {0xc9, 0xff, 0xcc, 0xee, 0x11, 0xcc}}; /**< Address used for Application profile */

/**@brief BLE Enable params. */
static ble_enable_params_t m_ble_enable_params = {.common_enable_params.vs_uuid_count    = BLE_UUID_VS_COUNT_DEFAULT,
                                                  .gap_enable_params.periph_conn_count   = 1,
                                                  .gatts_enable_params.attr_tab_size     = BLE_GATTS_ATTR_TAB_SIZE_DEFAULT};


/**@brief Enum representing the current active profile. */
typedef enum {
    PROFILE_WPT,
    PROFILE_APPLICATION,
} device_profile_t;

device_profile_t m_current_profile = PROFILE_APPLICATION; /**< What profile is currently running? */


/**@brief Dispatches a BLE stack event to all profiles.
 *
 * @details This function is called from the BLE Stack event interrupt handler after a BLE stack
 *          event has been received.
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 */
static void m_ble_evt_dispatch(ble_evt_t * p_ble_evt)
{
    switch(p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
            m_connected = true;
            break;
    }

    switch(m_current_profile)
    {
        case PROFILE_WPT:
            pru_on_ble_evt(p_ble_evt);
            break;

        case PROFILE_APPLICATION:
            app_profile_on_ble_evt(p_ble_evt);
            break;
    }
}

/**@brief Dispatch system events to all profiles that require it. */
static void m_sys_evt_dispatch(uint32_t evt)
{
    app_profile_on_sys_evt(evt);
}

/**@brief BLE stack initialization.
 *
 * @details Initializes the SoftDevice and the BLE event interrupt.
 */
static void m_ble_stack_init(void)
{
    uint32_t err_code;

    SOFTDEVICE_HANDLER_INIT(&clk_cfg, NULL);

    err_code = softdevice_enable(&m_ble_enable_params);
    APP_ERROR_CHECK(err_code);

    err_code = softdevice_ble_evt_handler_set(m_ble_evt_dispatch);
    APP_ERROR_CHECK(err_code);

    err_code = softdevice_sys_evt_handler_set(m_sys_evt_dispatch);
    APP_ERROR_CHECK(err_code);

    // It is important that this is set to the same priority as the
    // timers in the system.
    err_code = sd_nvic_SetPriority(SD_EVT_IRQn, APP_IRQ_PRIORITY_LOW);
    APP_ERROR_CHECK(err_code);
}

/**@brief Handler for reading VRECT. */
static void vrect_read_timeout_handler(void * p_context)
{
    uint32_t err_code;
    const pru_sensor_data_t * p_data;

    pru_sensors_read_all();

    p_data = pru_sensors_data_get();
    if(p_data->vrect > PRU_VRECT_UVLO)
    {
        // Switch from application profile to PRU
        app_profile_stop();

        err_code = app_timer_stop_all();
        APP_ERROR_CHECK(err_code);

        err_code = sd_softdevice_disable();
        APP_ERROR_CHECK(err_code);


        m_current_profile = PROFILE_WPT;

        m_profile_has_been_switched = true;

        m_connected = false;

    }
}


/**@brief Init timers required by this module. */
static void m_timers_init(void)
{
    uint32_t err_code;

    APP_TIMER_INIT(APP_TIMER_PRESCALER, PRU_MULTI_PROFILE_TIMER_OP_QUEUE_SIZE, false);

    err_code = app_timer_create(&m_vrect_read_timer, APP_TIMER_MODE_REPEATED, vrect_read_timeout_handler);
    APP_ERROR_CHECK(err_code);
}

/**@brief PRU state machine event handler. Used to see when charging has stopped */
static void m_on_pru_sm_evt(pru_sm_signal_type_t signal, const pru_sm_state_vars_t * p_state_vars)
{
    uint32_t err_code;

    if( (p_state_vars->prev_state != p_state_vars->current_state) && (p_state_vars->current_state == PRU_SM_STATE_NULL))
    {
        // First stop all timers
        err_code = app_timer_stop_all();
        APP_ERROR_CHECK(err_code);

        // Disable softdevice to clear all settings
        err_code = sd_softdevice_disable();
        APP_ERROR_CHECK(err_code);

        // Start application profile
        err_code = app_timer_start(m_vrect_read_timer, VRECT_MEAS_INTERVAL, NULL);
        APP_ERROR_CHECK(err_code);

        m_current_profile = PROFILE_APPLICATION;

        m_profile_has_been_switched = true;

        m_connected = false;
    }
}

/**@brief Application main function.
 * @return Ignored.
 */
int main()
{
    uint32_t err_code;

    pru_sensors_init(NULL); // Initialize sensors with NULL value to allow sensor reading, event generation will not be enabled.
    m_ble_stack_init();
    m_timers_init();

#ifdef DFU_SUPPORT
    // Set up button used for entering DFU mode
    common_hal_buttons_init(PRU_BUTTON_PRESS_LENGTH_DFU_MS, APP_TIMER_PRESCALER);
#endif

    err_code = app_timer_start(m_vrect_read_timer, VRECT_MEAS_INTERVAL, NULL);
    APP_ERROR_CHECK(err_code);

    pru_init(m_on_pru_sm_evt);

    app_profile_init();
    app_profile_start();

    // We need to get to main context before switching profile in order to ensure that all
    // softdevice events has been handled.
    while(1)
    {
        if(m_profile_has_been_switched && !m_connected)
        {
            (void)sd_softdevice_enable(&clk_cfg, app_error_fault_handler);

            err_code = softdevice_enable(&m_ble_enable_params);
            APP_ERROR_CHECK(err_code);

            if(m_current_profile == PROFILE_WPT)
            {
                err_code = sd_ble_gap_address_set(BLE_GAP_ADDR_CYCLE_MODE_NONE, &PRU_ADDR);
                APP_ERROR_CHECK(err_code);

                pru_start();
            }
            else
            {
                err_code = sd_ble_gap_address_set(BLE_GAP_ADDR_CYCLE_MODE_NONE, &APP_PROFILE_ADDR);
                APP_ERROR_CHECK(err_code);

                app_profile_start();
            }
            m_profile_has_been_switched = false;
        }

    #ifdef DEBUG_OUT_ENABLE
        debug();
    #endif
    }
}


/** @} */
//lint -restore
