/* Copyright (c) 2012 Nordic Semiconductor. All Rights Reserved.
*
* The information contained herein is property of Nordic Semiconductor ASA.
* Terms and conditions of usage are described in detail in NORDIC
* SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
*
* Licensees are granted free, non-transferable use of the information. NO
* WARRANTY of ANY KIND is provided. This heading must NOT be removed from
* the file.s
*/

#include <stdio.h>
#include <stdint.h>
#include "nrf_assert.h"
#include "nrf_gzll.h"
#include "app_print.h"
#include "simple_uart.h"
#include "wireless_debug_config.h"
#include "nrf_gpio.h"
#include "boards.h"

/** @file
 *
 * @defgroup example_pru_wireless_debug Wireless debug dongle
 * @{
 * @ingroup examples
 * @brief
 */

#define DEBUG_POLL                              0               ///< Packed type used for polling opnly
#define DEBUG_PKT                               1               ///< Packed type used for packet with payload
#define DEBUG_STR                               2               ///< Packed type used for packet with string
#define PIPE_NUMBER                             0               ///< Use pipe 0 for the wireless debug interface (WDI)
#define ACK_PAYLOAD_LENGTH                      1               ///< We use 1 byte payload length when transmitting
#define DEBUG_PAYLOAD_LENGTH                   32               ///< Debug payload length
#define DEBUG_LINE_LENGTH                      80               ///< Debug string length

// Assert setup
#define DEAD_BEEF 0xDEADBEEF                                    ///< Error code for assert

static uint8_t data_payload[NRF_GZLL_CONST_MAX_PAYLOAD_LENGTH]; ///< GZLL data payload
static uint8_t ack_payload[ACK_PAYLOAD_LENGTH];                 ///< GZLL ack payload

static bool pkt_received;                                       ///< True when a packet with first byte = DEBUG_POLL received

/**@brief Local version of uart configure that take baud rate
 */
static void uart_config(uint32_t baudrate,
                        uint8_t rts_pin_number,
                        uint8_t txd_pin_number,
                        uint8_t cts_pin_number,
                        uint8_t rxd_pin_number,
                        bool    hwfc)
{
    nrf_gpio_cfg_output(txd_pin_number);
    nrf_gpio_cfg_input(rxd_pin_number, NRF_GPIO_PIN_NOPULL);

    NRF_UART0->PSELTXD = txd_pin_number;
    NRF_UART0->PSELRXD = rxd_pin_number;
    if (hwfc)
    {
        nrf_gpio_cfg_output(rts_pin_number);
        nrf_gpio_cfg_input(cts_pin_number, NRF_GPIO_PIN_NOPULL);
        NRF_UART0->PSELCTS = cts_pin_number;
        NRF_UART0->PSELRTS = rts_pin_number;
        NRF_UART0->CONFIG  = (UART_CONFIG_HWFC_Enabled << UART_CONFIG_HWFC_Pos);
    }

    NRF_UART0->BAUDRATE      = (baudrate << UART_BAUDRATE_BAUDRATE_Pos);
    NRF_UART0->ENABLE        = (UART_ENABLE_ENABLE_Enabled << UART_ENABLE_ENABLE_Pos);
    NRF_UART0->TASKS_STARTTX = 1;
    NRF_UART0->TASKS_STARTRX = 1;
    NRF_UART0->EVENTS_RXDRDY = 0;
}


/**@brief Checks if there are any characters available in the UART
 */
static bool m_uart_chars_available(void)
{
    return (NRF_UART0->EVENTS_RXDRDY == 1);
}

/**@brief Print a dashed line to the UART
 */
static void line_print(void)
{
    int i;

    for(i = 0; i < DEBUG_LINE_LENGTH; i++)
    {
        print_string("-");
    }
    print_string("\r\n");
}

/**@brief Convert state to string
 */
static char * state_string_get(uint8_t state)
{
    switch(state)
    {
        case 0:
            return "PRU Null";
        case 1:
            return "PRU Boot";
        case 2:
            return "PRU On";
        case 3:
            return "PRU System Error";
        default:
            return "Invalid state";
    }

}

/**@brief Print debug output
 *
 * @param debug_data    Pointer to the debug data
 */
static void debug_output_print(uint8_t *debug_data)
{
    char temp_string[DEBUG_LINE_LENGTH];                    /**< Debug output collection */
    uint16_t tmp16_1, tmp16_2, tmp16_3, tmp16_4, tmp16_5;
    int16_t temperature;

    // Status
    line_print();
    print_string("StatusStart\r\n");
    line_print();
    print_string("State\r\n");
    sprintf(temp_string, "%s\r\n", state_string_get(debug_data[0]));
    print_string(temp_string);

    // Dynamic parameters
    line_print();
    print_string("Dynamic parameters\r\n");
    line_print();
    tmp16_1 = (uint16_t)debug_data[2] << 8 | (uint16_t)debug_data[1];
    tmp16_2 = (uint16_t)debug_data[4] << 8 | (uint16_t)debug_data[3];
    tmp16_3 = (uint16_t)debug_data[6] << 8 | (uint16_t)debug_data[5];
    tmp16_4 = (uint16_t)debug_data[8] << 8 | (uint16_t)debug_data[7];
    temperature = (int16_t)((uint16_t)(debug_data[10]<<8 | (uint16_t)debug_data[9]));
    print_string("Vrect\tIrect\tVout\tIout\tTemp\tCharged\tP_adj\r\n");
    sprintf(temp_string, "%u\t%u\t%u\t%u\t%u\t%u\t%u\r\n", tmp16_1, tmp16_2, tmp16_3, tmp16_4, temperature, debug_data[11], debug_data[24]);
    print_string(temp_string);

    // Static parameters
    line_print();
    print_string("Static parameters\r\n");
    line_print();
    tmp16_1 = (uint16_t)debug_data[13] << 8 | (uint16_t)debug_data[12];
    tmp16_2 = (uint16_t)debug_data[15] << 8 | (uint16_t)debug_data[14];
    tmp16_3 = (uint16_t)debug_data[17] << 8 | (uint16_t)debug_data[16];
    tmp16_4 = (uint16_t)debug_data[20] << 8 | (uint16_t)debug_data[19];
    tmp16_5 = (uint16_t)debug_data[22] << 8 | (uint16_t)debug_data[21];
    print_string("Vrset\tVrmin\tVrmax\tTempmax\tVrmax\tPrmax\r\n");
    sprintf(temp_string, "%u\t%u\t%u\t%u\t%u\t%u\r\n", tmp16_1, tmp16_2, tmp16_3, debug_data[18], tmp16_4, tmp16_5);
    print_string(temp_string);

    // Previous received control data
    line_print();
    print_string("Control packet\r\n");
    line_print();

    print_string("En_out\tEn_Chin\tAdj_pow\tPerm\tTset\r\n");
    sprintf(temp_string, "%u\t%u\t%u\t%u\t%u\r\n", debug_data[25], debug_data[26], debug_data[27], debug_data[28], debug_data[29]);
    print_string(temp_string);

    if(debug_data[23] != 0)
    {
        print_string("\r\n");
        print_string("Alert sent\r\n");
    }
    print_string("StatusEnd\r\n");
}

/**@brief Error handler function, which is called when an error has occurred.
 */
void app_error_handler(uint32_t error_code, uint32_t line_num, const uint8_t * p_file_name)
{
    // This call can be used for debug purposes during development of an application.
    // @note CAUTION: Activating this code will write the stack to flash on an error.
    //                This function should NOT be used in a final product.
    //                It is intended STRICTLY for development/debugging purposes.
    //                The flash write will happen EVEN if the radio is active, thus interrupting
    //                any communication.
    //                Use with care. Un-comment the line below to use.

    // On assert, the system can only recover with a reset.;
    NVIC_SystemReset();
}


/**@brief Assert macro callback function.
 */
void assert_nrf_callback(uint16_t line_num, const uint8_t * p_file_name)
{
    app_error_handler(DEAD_BEEF, line_num, p_file_name);
}


/** @brief Dongle application.
 */
int main()
{
    bool res;
    uint8_t channels[NRF_GZLL_CHANNEL_TABLE_SIZE] = NRF_GZLL_CHANNELS;
    pkt_received = false;
    res = nrf_gzll_init(NRF_GZLL_MODE_HOST);
    ASSERT(res);
    res = nrf_gzll_set_base_address_0(NRF_GZLL_BASE_ADDRESS_0);
    ASSERT(res);
    res = nrf_gzll_set_channel_table(channels, NRF_GZLL_CHANNEL_TABLE_SIZE);
    ASSERT(res);
    res = nrf_gzll_enable();
    ASSERT(res);
    uart_config(UART_BAUDRATE_BAUDRATE_Baud460800, RTS_PIN_NUMBER, TX_PIN_NUMBER, CTS_PIN_NUMBER, RX_PIN_NUMBER, HWFC);
    while(m_uart_chars_available())
        (void)simple_uart_get();
    while(1)
    {
        if (pkt_received)
        {
            if (data_payload[0] == DEBUG_PKT)
                debug_output_print(&data_payload[1]);
            else if (data_payload[0] == DEBUG_STR)
                print_string((char *)&data_payload[1]);
            pkt_received = false;
        }
    }
}

/**
 * @brief GZLL data packet received callback
 */
void nrf_gzll_host_rx_data_ready(uint32_t pipe, nrf_gzll_host_rx_info_t rx_info)
{
    bool res;
    uint32_t data_payload_length = NRF_GZLL_CONST_MAX_PAYLOAD_LENGTH;
    res = nrf_gzll_fetch_packet_from_rx_fifo(pipe, data_payload, &data_payload_length);
    ASSERT(res);
    if (data_payload[0] != DEBUG_POLL)
        pkt_received = true;
    if (m_uart_chars_available())
    {
        ack_payload[0] = simple_uart_get();
        res = nrf_gzll_add_packet_to_tx_fifo(pipe, ack_payload, ACK_PAYLOAD_LENGTH);
        ASSERT(res);
    }
}

/**
 * @brief GZLL ACK received callback
 */
void nrf_gzll_device_tx_success(uint32_t pipe, nrf_gzll_device_tx_info_t tx_info)
{
}

/**
 * @brief GZLL Transmission failed callback
 */
void nrf_gzll_device_tx_failed(uint32_t pipe, nrf_gzll_device_tx_info_t tx_info)
{
}

/**
 * @brief GZLL Disabled callback.
 */
void nrf_gzll_disabled()
{
}

/** @} */
