/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
 *
 * @defgroup example_ptu_sm PTU example
 * @{
 * @ingroup examples
 * @brief
 */

#include <stdint.h>
#include <string.h>

#include "boards.h"
#include "app_print.h"
#include "debug.h"
#include "ptu.h"
#include "softdevice_handler.h"
#include "ptu_config.h"
#include "app_util_platform.h"
#include "nrf_nvic.h"


#ifdef DFU_SUPPORT
    #include "common_hal_buttons.h"
#endif

#define CENTRAL_LINK_COUNT               8  /**< Number of central links used by the application. When changing this number remember to adjust the RAM settings*/
#define PERIPHERAL_LINK_COUNT            0  /**< Number of peripheral links used by the application. When changing this number remember to adjust the RAM settings*/

/**@brief Dispatches BLE events to all profiles.
 *
 * @param[in]   p_evt   Bluetooth stack event.
 */
static void m_ble_evt_dispatch(ble_evt_t * p_evt)
{
    ptu_on_ble_evt(p_evt);
}

/**@brief Initialize the BLE sofdevice and IRQ handler.
 */
static void m_ble_stack_init(void)
{
    uint32_t err_code;
    ble_conn_bw_counts_t count          = {{0,0,CENTRAL_LINK_COUNT},{0,0,CENTRAL_LINK_COUNT}};

    // Initialize the SoftDevice handler module.
    nrf_clock_lf_cfg_t clock_lf_cfg = NRF_CLOCK_LFCLKSRC;
    SOFTDEVICE_HANDLER_INIT(&clock_lf_cfg, NULL);

    // Enable BLE stack.
    ble_enable_params_t ble_enable_params = {0};
    ble_enable_params.gap_enable_params.central_conn_count  = CENTRAL_LINK_COUNT;
    ble_enable_params.gap_enable_params.central_sec_count   = CENTRAL_LINK_COUNT;

    ble_enable_params.gatts_enable_params.attr_tab_size     = BLE_GATTS_ATTR_TAB_SIZE_DEFAULT;
    ble_enable_params.common_enable_params.p_conn_bw_counts = &count;

    ble_opt_t opt = {0};
    opt.common_opt.conn_bw.conn_bw.conn_bw_rx = BLE_CONN_BW_LOW;
    opt.common_opt.conn_bw.conn_bw.conn_bw_tx = BLE_CONN_BW_LOW;
    opt.common_opt.conn_bw.role = BLE_GAP_ROLE_CENTRAL;

    //Check the ram settings against the used number of links
    CHECK_RAM_START_ADDR(CENTRAL_LINK_COUNT, PERIPHERAL_LINK_COUNT);

    err_code = softdevice_enable(&ble_enable_params);
    APP_ERROR_CHECK(err_code);

    err_code = sd_ble_opt_set(BLE_COMMON_OPT_CONN_BW, &opt);
    APP_ERROR_CHECK(err_code);

    err_code = softdevice_ble_evt_handler_set(m_ble_evt_dispatch);
    APP_ERROR_CHECK(err_code);

    err_code = sd_ble_gap_tx_power_set(PTU_OUTPUT_TX_POWER);
    APP_ERROR_CHECK(err_code);

    // It is important that this is set to the same priority as the
    // timers in the system.
    err_code = sd_nvic_SetPriority(SD_EVT_IRQn, APP_IRQ_PRIORITY_LOW);
    APP_ERROR_CHECK(err_code);
}

/**@brief Application.
 */
int main()
{
    APP_TIMER_INIT(APP_TIMER_PRESCALER, APP_TIMER_OP_QUEUE_SIZE, false);

#ifdef DFU_SUPPORT
    // Set up button used for entering DFU mode
    common_hal_buttons_init(PTU_BUTTON_PRESS_LENGTH_DFU_MS, APP_TIMER_PRESCALER);
#endif

    m_ble_stack_init();

    // Initialize PTU state machine
    ptu_init(NULL);

    // Start PTU
    ptu_start();

    while(1)
    {
        #ifdef DEBUG_OUT_ENABLE
        debug();
        #endif
    }
}

//lint -restore
/** @} */
