/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
 */

#include <stdint.h>
#include <string.h>

#include "boards.h"
#include "app_util_platform.h"
#include "app_print.h"
#include "debug.h"
#include "ptu.h"
#include "app_timer.h"
#include "softdevice_handler.h"
#include "advertiser_beacon.h"
#include "ptu_config.h"
#include "nrf_nvic.h"

#ifdef DFU_SUPPORT
    #include "common_hal_buttons.h"
#endif

#define CENTRAL_LINK_COUNT                7           /**< Number of central links used by the application. When changing this number remember to adjust the RAM settings*/
#define PERIPHERAL_LINK_COUNT             1           /**< Number of peripheral links used by the application. When changing this number remember to adjust the RAM settings*/

// Assert setup
#define DEAD_BEEF 0xDEADBEEF                          /**< Error code to signify a SoftDevice assert. */
#ifdef BOARD_PCA10028
  #define NO_DEV_CONNECTED_LED_PATTERN    0x03        /**< Status LEDs when no devices connected */
  #define POWER_TRANSFER_LED_PATTERN      0x02        /**< Status LEDs during power transfer state */
  #define FAULT_LED_PATTERN               0x01        /**< Status LEDs when in any fault state */
#else
  #define NO_DEV_CONNECTED_LED_PATTERN    0x00        /**< Status LEDs when no devices connected */
  #define POWER_TRANSFER_LED_PATTERN      0x01        /**< Status LEDs during power transfer state */
  #define FAULT_LED_PATTERN               0x02        /**< Status LEDs when in any fault state */
#endif

#define BEACON_UUID 0xff, 0xfe, 0x2d, 0x12, 0x1e, 0x4b, 0x0f, 0xa4,\
                    0x99, 0x4e, 0xce, 0xb5, 0x31, 0xf4, 0x05, 0x45                      /**< The Beacon's UUID*/
#define BEACON_ADV_INTERVAL                  400                                        /**< The Beacon's advertising interval, in milliseconds*/
#define BEACON_MAJOR                         0x1234                                     /**< The Beacon's Major*/
#define BEACON_MINOR                         0x5678                                     /**< The Beacon's Minor*/
#define BEACON_RSSI                          0xC3                                       /**< The Beacon's measured RSSI at 1 meter distance in dBm. */

#define APP_COMPANY_IDENTIFIER               0x0059                                     /**< Company identifier for Nordic Semiconductor ASA. as per www.bluetooth.org. */

static ble_beacon_init_t beacon_init;                                                   /**< Beacon initializer parameters. */

/**@brief Function for initializing Beacon advertiser.
 */
static void beacon_adv_init(void)
{
    static uint8_t beacon_uuid[] = {BEACON_UUID};

    memcpy(beacon_init.uuid.uuid128, beacon_uuid, sizeof(beacon_uuid));
    beacon_init.adv_interval  = BEACON_ADV_INTERVAL;
    beacon_init.major         = BEACON_MAJOR;
    beacon_init.minor         = BEACON_MINOR;
    beacon_init.manuf_id      = APP_COMPANY_IDENTIFIER;
    beacon_init.rssi          = BEACON_RSSI;

    uint32_t err_code = sd_ble_gap_address_get(&beacon_init.beacon_addr);
    APP_ERROR_CHECK(err_code);

    app_beacon_init(&beacon_init);
}


/**@brief Dispatch system event, triggered by softdevice.
 * @param sys_evt The event to handle
 */
static void m_sys_evt_dispatch(uint32_t sys_evt)
{

}

/**@brief Dispatch BLE event, triggered by softdevice.
 * @param[in] p_evt The event to handle
 */
static void m_ble_evt_dispatch(ble_evt_t * p_evt)
{
    ptu_on_ble_evt(p_evt);
}

/**@brief Initialize the BLE sofdevice and IRQ handler.
 */
static void m_ble_stack_init(void)
{
    uint32_t err_code;
    ble_conn_bw_counts_t count          = {{0,0,8},{0,0,8}};

    // Initialize the SoftDevice handler module.
    nrf_clock_lf_cfg_t clock_lf_cfg = NRF_CLOCK_LFCLKSRC;
    SOFTDEVICE_HANDLER_INIT(&clock_lf_cfg, NULL);

    // Enable BLE stack.
    ble_enable_params_t ble_enable_params = {0};
    ble_enable_params.gap_enable_params.central_conn_count  = CENTRAL_LINK_COUNT;
    ble_enable_params.gap_enable_params.central_sec_count   = CENTRAL_LINK_COUNT;
    ble_enable_params.gap_enable_params.periph_conn_count = PERIPHERAL_LINK_COUNT;

    ble_enable_params.gatts_enable_params.attr_tab_size     = BLE_GATTS_ATTR_TAB_SIZE_DEFAULT;
    ble_enable_params.common_enable_params.p_conn_bw_counts = &count;

    ble_opt_t opt = {0};
    opt.common_opt.conn_bw.conn_bw.conn_bw_rx = BLE_CONN_BW_LOW;
    opt.common_opt.conn_bw.conn_bw.conn_bw_tx = BLE_CONN_BW_LOW;
    opt.common_opt.conn_bw.role = BLE_GAP_ROLE_CENTRAL;

    //Check the ram settings against the used number of links
    CHECK_RAM_START_ADDR(CENTRAL_LINK_COUNT, PERIPHERAL_LINK_COUNT);

    err_code = softdevice_enable(&ble_enable_params);
    APP_ERROR_CHECK(err_code);

    err_code = sd_ble_opt_set(BLE_COMMON_OPT_CONN_BW, &opt);
    APP_ERROR_CHECK(err_code);

    err_code = softdevice_ble_evt_handler_set(m_ble_evt_dispatch);
    APP_ERROR_CHECK(err_code);

    err_code = softdevice_sys_evt_handler_set(m_sys_evt_dispatch);
    APP_ERROR_CHECK(err_code);

    err_code = sd_ble_gap_tx_power_set(PTU_OUTPUT_TX_POWER);
    APP_ERROR_CHECK(err_code);

    // It is important that this is set to the same priority as the
    // timers in the system.
    err_code = sd_nvic_SetPriority(SD_EVT_IRQn, APP_IRQ_PRIORITY_LOW);
    APP_ERROR_CHECK(err_code);
}

/**@brief Application.
 */
int main()
{
    #ifndef INIT_APP_TIMERS_EXTERNALLY
    APP_TIMER_INIT(APP_TIMER_PRESCALER, APP_TIMER_OP_QUEUE_SIZE, false);
    #endif

#ifdef DFU_SUPPORT
    // Set up button used for entering DFU mode
    common_hal_buttons_init(PTU_BUTTON_PRESS_LENGTH_DFU_MS, APP_TIMER_PRESCALER);
#endif

    m_ble_stack_init();

    beacon_adv_init();

    // Initialize PTU state machine
    ptu_init(NULL);

    // Start PTU
    ptu_start();

    // Start beacons
    app_beacon_start();

    while(1)
    {
        #ifdef DEBUG_OUT_ENABLE
        debug();
        #endif
    }
}

//lint -restore
/** @} */
