/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
 *
 * @defgroup ble_wpts_server_api PRU BLE Service API (server)
 * @{
 * @ingroup ble_wpts
 */

#ifndef BLE_WPTS_H__
#define BLE_WPTS_H__

#include <stdint.h>

#include "ble.h"
#include "ble_srv_common.h"
#include "wpt.h"

/**< WPT Service event type. */
typedef enum
{
    BLE_WPTS_EVT_ALERT_ENABLED,                    /**< PRU Alert value notification/indication enabled. */
    BLE_WPTS_EVT_ALERT_DISABLED,                   /**< PRU Alert value notification/indication disabled. */
    BLE_WPTS_EVT_ALERT_INDICATION_CONFIRMED,       /**< PRU Alert confirmed received from PTU. */
    BLE_WPTS_EVT_PRU_CONTROL_COMMAND,              /**< PRU Control command received */
    BLE_WPTS_EVT_PTU_STATIC_PARAMETER,             /**< PTU Static parameters set */
    BLE_WPTS_EVT_PRU_DYNAMIC_READ,                 /**< PTU has read PRU dynamic characteristic. */
} ble_wpts_evt_type_t;

/**@brief WPT Service event. */
typedef struct
{
    ble_wpts_evt_type_t evt_type;                  /**< Type of event. */
    union
    {
        pru_control_t    pru_control;          /**< PRU Control command. */
        ptu_static_t     ptu_static;           /**< PTU Static parameters. */
    } data;                                         /**< Data union. */
} ble_wpts_evt_t;


/**< Forward declaration of the ble_wpts_t type. */
typedef struct ble_wpts_s ble_wpts_t;

/**@brief WPT Service event handler type.
 *
 * @param[in]   p_wpts     WPT Service structure.
 * @param[in]   p_evt       WPT Service event.
 */
typedef void (*ble_wpts_evt_handler_t) (ble_wpts_t * p_wpts, ble_wpts_evt_t * p_evt);

/**@brief WPT Service init structure. */
typedef struct
{
    ble_wpts_evt_handler_t       evt_handler;              /**< Event handler to be called for handling events in the WPT Service. */
    pru_static_t             pru_static;               /**< PRU Static Parameters. */
    ble_srv_security_mode_t       wpts_attr_md;            /**< Initial Security Setting for characteristics. */
} ble_wpts_init_t;

/**@brief WPT Service status structure. */
typedef struct ble_wpts_s
{
    ble_wpts_evt_handler_t      evt_handler;               /**< Event handler to be called for handling events in the WPT Service. */
    uint16_t                     service_handle;            /**< Handle of WPT Service (as provided by the BLE stack). */
    ble_gatts_char_handles_t     pru_control_handles;       /**< Handles related to the PRU Control characteristic. */
    ble_gatts_char_handles_t     ptu_static_handles;        /**< Handles related to the PTU Static Parameter characteristic. */
    ble_gatts_char_handles_t     pru_alert_handles;         /**< Handles related to the PRU Alert characteristic. */
    ble_gatts_char_handles_t     pru_static_handles;        /**< Handles related to the PRU Static Parameter characteristic. */
    ble_gatts_char_handles_t     pru_dynamic_handles;       /**< Handles related to the PRU Dynamic Parameter characteristic. */
    uint16_t                     conn_handle;               /**< Handle of the current connection (as provided by the BLE stack, is BLE_CONN_HANDLE_INVALID if not in a connection). */
    uint8_t                      reboot_bit:1;              /**< Reboot bit of device */
} ble_wpts_t;

/**@brief Initialize the WPT Service.
 *
 * @param[out]  p_wpts      WPT Service structure. This structure will have to be supplied by
 *                          the application. It will be initialized by this function, and will later
 *                          be used to identify this particular service instance.
 * @param[in]   p_wpts_init Information needed to initialize the service.
 *
 * @return      NRF_SUCCESS on successful initialization of service, otherwise an error code.
 */
uint32_t ble_wpts_init(ble_wpts_t * p_wpts, const ble_wpts_init_t * p_wpts_init);

/**@brief WPT Service BLE stack event handler.
 *
 * @details Handles all events from the BLE stack of interest to the WPT Service.
 *
 * @param[in]   p_wpts     WPT Service structure.
 * @param[in]   p_ble_evt  Event received from the BLE stack.
 */
void ble_wpts_on_ble_evt(ble_wpts_t * p_wpts, ble_evt_t * p_ble_evt);

/**@brief Sends PRU alert if notification has been enabled.
 *
 * @details The application calls this function to send a PRU alert.
 *          If notification has been enabled, the alert data is encoded and sent to
 *          the client.
 *
 * @param[in]   p_wpts                  WPT Service structure.
 * @param[in]   p_alert                  PRU alert value.
 *
 * @return      NRF_SUCCESS on success, otherwise an error code.
 */
uint32_t ble_wpts_alert_send(ble_wpts_t * p_wpts, pru_alert_t * p_alert);


/**@brief Sets value of the PRU Static Parameter characteristic.
 *
 * @param[in]   p_wpts                  WPT Service structure.
 * @param[in]   p_pru_static            PRU Static Parameter characteristic structure.
 *
 * @return      NRF_SUCCESS on success, otherwise an error code.
 */
uint32_t ble_wpts_pru_static_set(ble_wpts_t * p_wpts, pru_static_t * p_pru_static);


/**@brief Sets value of the PRU Dynamic Parameter characteristic.
 *
 * @param[in]   p_wpts                  WPT Service structure.
 * @param[in]   p_pru_dynamic            PRU dynamic data.
 *
 * @return      NRF_SUCCESS on success, otherwise an error code.
 */
uint32_t ble_wpts_pru_dynamic_set(ble_wpts_t * p_wpts, pru_dynamic_t * p_pru_dynamic);

#endif // BLE_WPTS_H__

/** @} */
