/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

 /** @file
 *
 * @defgroup ble_wpts_c PTU BLE Service API (client)
 * @{
 * @ingroup ble_wpts
 */

#ifndef BLE_WPTS_C_H__
#define BLE_WPTS_C_H__

#include <stdint.h>

#include "ble.h"
#include "ble_wpts_common.h"
#include "wpt.h"


/**@brief WPT Service Client event type. */
typedef enum
{
    BLE_WPTS_C_EVT_DISCOVERY_OK,                         /**< PRU characteristic and service discovery successful. */
    BLE_WPTS_C_EVT_DISCOVERY_FAILED,                     /**< PRU characteristic or service discovery failed. */
    BLE_WPTS_C_EVT_PRU_CONTROL_WRITE_RESP,               /**< PRU Control write response received. */
    BLE_WPTS_C_EVT_PTU_STATIC_WRITE_RESP,                /**< PTU Static Parameter write response received. */
    BLE_WPTS_C_EVT_PRU_STATIC_READ_RESP,                 /**< PRU Static Parameter read response received. */
    BLE_WPTS_C_EVT_PRU_DYNAMIC_READ_RESP,                /**< PRU Dynamic Parameter read response received. */
    BLE_WPTS_C_EVT_PRU_ALERT_ENABLE_WRITE_RESP,          /**< PRU Alert enable write response. */
    BLE_WPTS_C_EVT_PRU_ALERT                             /**< PRU Alert received. */
} ble_wpts_c_evt_type_t;

/**@brief WPT Service Client event. */
typedef struct
{
    ble_wpts_c_evt_type_t  type;                         /**< Type of event. */
    uint16_t                gatt_status;                  /**< GATT operation status. */
    union
    {
        pru_static_t   pru_static;                   /**< PRU Static parameters. */
        pru_dynamic_t  pru_dynamic;                  /**< PRU Dynamic parameters. */
        pru_alert_t    pru_alert;                    /**< PRU alert parameters. */
    } data;                                               /**< Data union */
} ble_wpts_c_evt_t;

/**@brief WPT Service Client structure. This contains various status information for the service. */
typedef struct
{
    ble_gattc_handle_range_t    service_handle_range;     /**< Discovered WPT Service handle range. */
    uint16_t                    pruc_handle;              /**< Handle of the PRU Control characteristic. */
    uint16_t                    ptusp_handle;             /**< Handle of the PTU Static Parameter characteristic. */
    uint16_t                    prua_handle;              /**< Handle of the PRU Alert characteristic. */
    uint16_t                    prua_cccd_handle;         /**< Handle of the PRU Alert characteristic CCCD. */
    uint16_t                    prusp_handle;             /**< Handle of the PRU Static Parameter characteristic. */
    uint16_t                    prudp_handle;             /**< Handle of the PRU Dynamic Parameter characteristic. */
    uint16_t                    conn_handle;              /**< Handle of the current connection (as provided by the BLE stack, is BLE_CONN_HANDLE_INVALID if not in a connection). */
} ble_wpts_c_t;

/**@brief WPT Service Client event handler type. */
typedef void (*ble_wpts_c_evt_handler_t) (ble_wpts_c_t * p_wpts_c, ble_wpts_c_evt_t * p_evt);

/**@brief WPT Service Client init structure. This contains all options and data needed for
 *        initialization of the service. */
typedef struct
{
    ble_wpts_c_evt_handler_t       evt_handler;          /**< Event handler to be called for handling events in the WPT Service Client. */
} ble_wpts_c_init_t;

/**@brief Initialize the WPT Service Client.
 *
 * @param[in]   p_wpts_c_init Information needed to initialize the service.
 *
 * @return      NRF_SUCCESS on successful initialization of service, otherwise an error code.
 */
uint32_t ble_wpts_c_init(const ble_wpts_c_init_t * p_wpts_c_init);

/**@brief WPT Service Client BLE stack event handler.
 *
 * @details Handles all events from the BLE stack of interest to the WPT Service Client.
 *
 * @param[in]   p_wpts_c    WPT Service Client structure.
 * @param[in]   p_ble_evt    Event received from the BLE stack.
 */
void ble_wpts_c_on_ble_evt(ble_wpts_c_t * p_wpts_c, ble_evt_t * p_ble_evt);

/**@brief Send a PRU Control message to the PRU server.
 *
 * @details The application calls this function to send a PRU Control
 *          message to the PRU server.
 *
 * @param[in]   p_wpts_c            WPT Service Client structure.
 * @param[in]   p_wpts_pru_control  PRU control value.
 *
 * @return      NRF_SUCCESS on success, otherwise an error code.
 */
uint32_t ble_wpts_c_send_pru_control(ble_wpts_c_t * p_wpts_c, pru_control_t * p_wpts_pru_control);

/**@brief Enable alert notifications to be sent from PRU.
 *
 * @param[in]   p_wpts_c            WPT Service Client structure.
 *
 * @return      NRF_SUCCESS on success, otherwise an error code.
 */
uint32_t ble_wpts_c_enable_pru_alert_notification(ble_wpts_c_t * p_wpts_c);

/**@brief Read the PRU Static Parameter characterisic value from the PRU server.
 *
 * @details The application calls this function to read the PRU Static
 *          Parameter characterisic value from the PRU server.
 *
 * @param[in]   p_wpts_c            WPT Service Client structure.
 *
 * @return      NRF_SUCCESS on success, otherwise an error code.
 */
uint32_t ble_wpts_c_read_pru_static(ble_wpts_c_t * p_wpts_c);

/**@brief Read the PRU Dynamic characterisic value from the PRU server.
 *
 * @details The application calls this function to read the PRU Dynamic
 *          Parameter characterisic value from the PRU server.
 *
 * @param[in]   p_wpts_c            WPT Service Client structure.
 *
 * @return      NRF_SUCCESS on success, otherwise an error code.
 */
uint32_t ble_wpts_c_read_pru_dynamic(ble_wpts_c_t * p_wpts_c);

/**@brief Write the PTU Static characteristic to the PRU server.
 *
 * @details The application calls this function to write the PTU Static
 *          Parameter characteristic to the PRU server.
 *
 * @param[in]   p_wpts_c            WPT Service Client structure.
 * @param[in]   p_wpts_ptu_static   PTU Static characteristic value.
 *
 * @return      NRF_SUCCESS on success, otherwise an error code.
 */
uint32_t ble_wpts_c_write_ptu_static(ble_wpts_c_t * p_wpts_c, ptu_static_t * p_wpts_ptu_static);

/** Read advertising packet, search for Service Data AD Type and return contents
 *
 * @param[in]  p_adv_report     Advertising report to parse.
 * @param[out] p_service_data   Struct in which to return service data contents.
 *
 * @return
 * NRF_SUCCESS if advertising packet contained Service Name AD Type and value was written.
 */
uint32_t ble_wpts_service_data_read(ble_gap_evt_adv_report_t * p_adv_report, ble_wpts_service_data_t * p_service_data);

/** Derive and update characteristic handles from WPT primary service handle.
 *@param[in]  prim_service_handle  Primary service handle from which to derive
                                        characteristic handles.
 *@param[out] p_wpts_c                 Struct holding handles to be updated. Only the
 *                                      characteristic handles in this struct will be updated.
 */
void ble_wpts_char_handles_update(uint16_t prim_service_handle, ble_wpts_c_t * p_wpts_c);

#endif // BLE_WPTS_C_H__

/** @} */

