/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
 *
 * @defgroup ble_wpts_common Common service API (PRU/PTU)
 * @{
 * @ingroup ble_wpts
 */

#ifndef BLE_WPTS_COMMON_H__
#define BLE_WPTS_COMMON_H__

#include <stdint.h>

#define BLE_WPTS_SEC_MODE_1_LEVEL_1 BLE_GAP_CONN_SEC_MODE_SET_OPEN          /**< Redefinition of security modees for mapping to spec */
#define BLE_WPTS_SEC_MODE_1_LEVEL_2 BLE_GAP_CONN_SEC_MODE_SET_ENC_NO_MITM   /**< Redefinition of security modees for mapping to spec */

/**@brief Advertising data */
#define BLE_WPTS_ADV_UUID_POS                  0           /**< 16-bit UUID byte offset */
#define BLE_WPTS_ADV_UUID_LEN                  2           /**< 16-bit UUID length in bytes */

/**@brief Advertising Service Data */
#define BLE_WPTS_ADV_SERVICE_HANDLE_POS        0           /**< Service handle position */
#define BLE_WPTS_ADV_SERVICE_HANDLE_LEN        2           /**< Service handle length */

#define BLE_WPTS_ADV_PRU_RSSI_PARAMS_POS       (BLE_WPTS_ADV_SERVICE_HANDLE_POS + BLE_WPTS_ADV_SERVICE_HANDLE_LEN)   /**< PRU RSSI parameters position */
#define BLE_WPTS_ADV_PRU_RSSI_PARAMS_LEN       1                                                                       /**< PRU RSSI parameters length */

#define BLE_WPTS_ADV_ANTENNA_GAIN_OFFSET          5                                                /**< PRU antenna gain offset in [dB] for val in adv packet */
#define BLE_WPTS_ADV_ANTENNA_GAIN_UNKNOWN_VAL     0xFF                                             /**< Value indicating "antenna gain unknown" in adv packet */
#define BLE_WPTS_ADV_PRU_TX_POWER_LEVEL_OFFSET    20                                               /**< PRU tx level offset in [dBm] for val in adv packet */
#define BLE_WPTS_ADV_POWER_UNKNOWN_VAL            0xFF                                             /**< Value indicating "TX power unknown" in adv packet */

#define BLE_WPTS_ADV_PRU_RSSI_PRU_GAIN_BITPOS     0                                                /**< PRU antenna gain bit position offset */
#define BLE_WPTS_ADV_PRU_RSSI_PRU_GAIN_BITMSK     0x07                                             /**< PRU antenna gain bitmask */
#define BLE_WPTS_ADV_PRU_RSSI_PRU_PWR_BITPOS      3                                                /**< PRU output power bit position offset */
#define BLE_WPTS_ADV_PRU_RSSI_PRU_PWR_BITMSK      (0x1F << BLE_WPTS_ADV_PRU_RSSI_PRU_PWR_BITPOS)  /**< PRU output power bitmask */

#define BLE_WPTS_ADV_FLAGS_POS                 (BLE_WPTS_ADV_PRU_RSSI_PARAMS_POS + BLE_WPTS_ADV_PRU_RSSI_PARAMS_LEN) /**< WPT advertising flags position offset */
#define BLE_WPTS_ADV_FLAGS_LEN                 1                                                                       /**< WPT advertising flags bitmask */

/** Total service data length */
#define BLE_WPTS_ADV_SERVICE_DATA_LENGTH (BLE_WPTS_ADV_UUID_LEN + BLE_WPTS_ADV_SERVICE_HANDLE_LEN + BLE_WPTS_ADV_PRU_RSSI_PARAMS_LEN + BLE_WPTS_ADV_FLAGS_LEN)

#define BLE_WPTS_ADV_FLAG_IMPEDANCE_SHIFT_BITPOS     5     /**< Impedance shift bit position offset */
#define BLE_WPTS_ADV_FLAG_IMPEDANCE_SHIFT_BITMSK     (0x07 << BLE_WPTS_ADV_FLAG_IMPEDANCE_SHIFT_BITPOS) /**< Impedance shift bitmask */
#define BLE_WPTS_ADV_FLAG_REBOOT_INDICATOR_BITPOS    4     /**< Reboot indicator bit position offset */
#define BLE_WPTS_ADV_FLAG_REBOOT_INDICATOR_BITMSK    (1 << BLE_WPTS_ADV_FLAG_REBOOT_INDICATOR_BITPOS) /**< Reboot indicator bitmask */
#define BLE_WPTS_ADV_FLAG_OVP_STATUS_BITPOS          3     /**< OVP status bit position offset */
#define BLE_WPTS_ADV_FLAG_OVP_STATUS_BITMSK          (1 << BLE_WPTS_ADV_FLAG_OVP_STATUS_BITPOS) /**< OVP status bitmask */
#define BLE_WPTS_ADV_FLAG_TIME_SET_SUPPORT_BITPOS    2     /**< Time set support bit position offset */
#define BLE_WPTS_ADV_FLAG_TIME_SET_SUPPORT_BITMSK    (1 << BLE_WPTS_ADV_FLAG_TIME_SET_SUPPORT_BITPOS) /**< Time set support bitmask */

#define BLE_WPTS_ADV_FLAG_REBOOT_INDICATOR_REBOOTED_VAL 0      /**< Value indicating that the PRU has rebooted */
#define BLE_WPTS_ADV_FLAG_REBOOT_INDICATOR_NOT_REBOOTED_VAL 1  /**< Value indicating that the PRU has not rebooted */

#define BLE_WPTS_ADV_SERVICE_DATA_LEN          (BLE_WPTS_ADV_FLAGS_POS + BLE_WPTS_ADV_FLAGS_LEN) /**< WPT Service data length*/

/**@brief PRU Control characteristic fields byte positions */
#define BLE_WPTS_PRU_CONTROL_ENABLES_POS        0  /**< PTU turn on, PTU on indication etc. */
#define BLE_WPTS_PRU_CONTROL_PERMISSION_POS     1  /**< PRU is permitted in PTU. */
#define BLE_WPTS_PRU_CONTROL_TIME_SET_POS       2  /**< PTU set up time [ms]. */

/**@brief PRU Control characteristic Enables bit field definitions */
#define BLE_WPTS_ENABLE_PRU_OUTPUT_BITPOS 7                                                /**< PRU output enable bit position */
#define BLE_WPTS_ENABLE_PRU_OUTPUT_BITMSK (0x01 << BLE_WPTS_ENABLE_PRU_OUTPUT_BITPOS)     /**< PRU output enable bit mask */
#define BLE_WPTS_ENABLE_PRU_OUTPUT_ENABLE_VAL 0x01                                         /**< PRU output enable value */
#define BLE_WPTS_ENABLE_PRU_OUTPUT_DISABLE_VAL 0x00                                        /**< PRU output disable value */

#define BLE_WPTS_ENABLE_PRU_CHARGE_INDICATOR_BITPOS 6                                      /**< PRU charge indicator bit position */
#define BLE_WPTS_ENABLE_PRU_CHARGE_INDICATOR_BITMSK (0x01 << BLE_WPTS_ENABLE_PRU_CHARGE_INDICATOR_BITPOS) /**< PRU charge indicator bit mask */
#define BLE_WPTS_ENABLE_PRU_CHARGE_INDICATOR_ENABLE_VAL 0x01                               /**< PRU charge indicator enable value */
#define BLE_WPTS_ENABLE_PRU_CHARGE_INDICATOR_DISABLE_VAL 0x00                              /**< PRU charge indicator disable value */

#define BLE_WPTS_ENABLE_PRU_ADJ_POWER_BITPOS 4                                                 /**< PRU adjust power field posistion */
#define BLE_WPTS_ENABLE_PRU_ADJ_POWER_BITMSK (0x03 << BLE_WPTS_ENABLE_PRU_ADJ_POWER_BITPOS)   /**< PRU adjust power field mask  */
#define BLE_WPTS_ENABLE_PRU_ADJ_POWER_MAX_VAL 0x00                                             /**< PRU adjust power value */
#define BLE_WPTS_ENABLE_PRU_ADJ_POWER_66_PCT_VAL 0x01                                          /**< PRU adjust power value */
#define BLE_WPTS_ENABLE_PRU_ADJ_POWER_33_PCT_VAL 0x02                                          /**< PRU adjust power value */
#define BLE_WPTS_ENABLE_PRU_ADJ_POWER_2_5_W_VAL 0x03                                           /**< PRU adjust power value */

/**@brief PRU Control characteristic Permission field values */
#define BLE_WPTS_PERMISSION_PERMITTED          0x00        /**< Permitted without reason */
#define BLE_WPTS_PERMISSION_PERMITTED_WAITING  0x01        /**< Permitted with waiting time due to limited available power */
#define BLE_WPTS_PERMISSION_DENIED_CC          0x80        /**< Denied due to cross connection */
#define BLE_WPTS_PERMISSION_DENIED_POWER       0x81        /**< Denied due to limited affordable power */
#define BLE_WPTS_PERMISSION_DENIED_DEVICES     0x82        /**< Denied due to limited PTU Number of Devices */
#define BLE_WPTS_PERMISSION_DENIED_CLASS       0x83        /**< Denied due to limited PTU Class support */
#define BLE_WPTS_PERMISSION_DENIED_HIGH_TEMP   0x84        /**< Denied due to high temperature on PTU */

/**@brief PRU Control characteristic Time set values */
#define BLE_WPTS_TIME_SET_DENIED               0x00        /**< Do not perform Time Set */
#define BLE_WPTS_TIME_SET_10MS                 0x01        /**< 10ms */
#define BLE_WPTS_TIME_SET_20MS                 0x02        /**< 20ms */
#define BLE_WPTS_TIME_SET_30MS                 0x03        /**< 30ms */
#define BLE_WPTS_TIME_SET_40MS                 0x04        /**< 40ms */
#define BLE_WPTS_TIME_SET_50MS                 0x05        /**< 50ms */
#define BLE_WPTS_TIME_SET_60MS                 0x06        /**< 60ms */
#define BLE_WPTS_TIME_SET_70MS                 0x07        /**< 70ms */
#define BLE_WPTS_TIME_SET_80MS                 0x08        /**< 80ms */

/**@brief PTU Static Parameter characteristic byte positions */
#define BLE_WPTS_PTU_STATIC_OPTIONAL_FIELDS_POS                0  /**< Optional fields byte position. */
#define BLE_WPTS_PTU_STATIC_PTU_POWER_POS                      1  /**< PTU Power byte position. */
#define BLE_WPTS_PTU_STATIC_PTU_MAX_SRC_IMPED_POS              2  /**< PTU maximum source impedance byte position. */
#define BLE_WPTS_PTU_STATIC_PTU_MAX_LOAD_RES_POS               3  /**< PTU maximum load resistance  byte position. */
#define BLE_WPTS_PTU_STATIC_PTU_CLASS_POS                      6  /**< PTU Class byte position. */
#define BLE_WPTS_PTU_STATIC_HW_REV_POS                         7  /**< Hardware revision  byte position. */
#define BLE_WPTS_PTU_STATIC_FW_REV_POS                         8  /**< Firmware revision byte position. */
#define BLE_WPTS_PTU_STATIC_PROTOCOL_REV_POS                   9   /**< Protocol revision byte position. */
#define BLE_WPTS_PTU_STATIC_NUM_DEVICES_SUPPORTED_POS          10  /**< Number of devices supported byte position. */

/**@brief Common bitifeld valus */
#define BLE_WPTS_OPTION_INVALID_VAL            0                   /**< Optional field invalid */
#define BLE_WPTS_OPTION_VALID_VAL              1                   /**< Optional field valid */

/**@brief PTU Static characteristic Option bitfield definitions */
#define BLE_WPTS_PTU_STATIC_OPTION_MAX_IMPEDANCE_BITPOS   0x07     /**< Max source impedance option bit position */
#define BLE_WPTS_PTU_STATIC_OPTION_MAX_IMPEDANCE_BITMSK (0x01 << BLE_WPTS_PTU_STATIC_OPTION_MAX_IMPEDANCE_BITPOS) /**< Max source impedance option bit mask */

#define BLE_WPTS_PTU_STATIC_OPTION_MAX_RESISTANCE_BITPOS  0x06     /**< Max load resistance option bit position */
#define BLE_WPTS_PTU_STATIC_OPTION_MAX_RESISTANCE_BITMSK (0x01 << BLE_WPTS_PTU_STATIC_OPTION_MAX_RESISTANCE_BITPOS) /**< Max load resistance option bit mask */

/**@brief PTU Static characteristic PTU Max Source Impedance bitfield */
#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_BITPOS   3  /**< PTU Max Source Impedance value posistion. */
#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_BITMSK (0x1F << BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_BITPOS) /**< PTU Max Source Impedance value mask. */
#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_50_VAL   0           /**< 50  */
#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_60_VAL   1           /**< 60  */
#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_70_VAL   2           /**< 70  */
#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_80_VAL   3           /**< 80  */
#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_90_VAL   4           /**< 90  */
#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_100_VAL  5           /**< 100 */
#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_110_VAL  6           /**< 110 */
#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_120_VAL  7           /**< 120 */
#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_130_VAL  8           /**< 130 */
#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_140_VAL  9           /**< 140 */
#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_150_VAL  10          /**< 150 */
#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_175_VAL  11          /**< 175 */
#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_200_VAL  12          /**< 200 */
#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_225_VAL  13          /**< 225 */
#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_250_VAL  14          /**< 250 */
#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_275_VAL  15          /**< 275 */
#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_300_VAL  16          /**< 300 */
#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_350_VAL  17          /**< 350 */
#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_375_VAL  18          /**< 375 */

/**@brief PTU Static characteristic PTU Max Load Resistance bitfield */
#define BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_BITPOS  3  /**< PTU Max Load Resistance value posistion. */
#define BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_BITMSK (0x1F << BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_BITPOS) /**< PTU Max Load Resistance value mask. */
#define BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_5_VAL    0           /**< 5  */
#define BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_10_VAL   1           /**< 10 */
#define BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_15_VAL   2           /**< 15 */
#define BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_20_VAL   3           /**< 20 */
#define BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_25_VAL   4           /**< 25 */
#define BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_30_VAL   5           /**< 30 */
#define BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_35_VAL   6           /**< 35 */
#define BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_40_VAL   7           /**< 40 */
#define BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_45_VAL   8           /**< 45 */
#define BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_50_VAL   9           /**< 50 */
#define BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_55_VAL   10          /**< 55 */

/**@brief PTU Static characteristic PTU MAx Number of Devices Supported field */
#define BLE_WPTS_PTU_STATIC_NUM_DEVICES_BITPOS  0  /**< Max number of devices value posistion. */
#define BLE_WPTS_PTU_STATIC_NUM_DEVICES_BITMSK (0x0F << BLE_WPTS_PTU_STATIC_NUM_DEVICES_BITPOS)   /**< Max number of devices value mask. */
#define BLE_WPTS_PTU_STATIC_NUM_DEVICES_1_VAL             0           /**< 1 */
#define BLE_WPTS_PTU_STATIC_NUM_DEVICES_2_VAL             1           /**< 2 */
#define BLE_WPTS_PTU_STATIC_NUM_DEVICES_3_VAL             2           /**< 3 */
#define BLE_WPTS_PTU_STATIC_NUM_DEVICES_4_VAL             3           /**< 4 */
#define BLE_WPTS_PTU_STATIC_NUM_DEVICES_5_VAL             4           /**< 5 */
#define BLE_WPTS_PTU_STATIC_NUM_DEVICES_6_VAL             5           /**< 6 */
#define BLE_WPTS_PTU_STATIC_NUM_DEVICES_7_VAL             6           /**< 7 */
#define BLE_WPTS_PTU_STATIC_NUM_DEVICES_8_VAL             7           /**< 8 */

/**@brief PTU Static characteristic PTU Class field */
#define BLE_WPTS_PTU_CLASS_1                   0           /**< Class 1 */
#define BLE_WPTS_PTU_CLASS_2                   1           /**< Class 2 */
#define BLE_WPTS_PTU_CLASS_3                   2           /**< Class 3 */
#define BLE_WPTS_PTU_CLASS_4                   3           /**< Class 4 */
#define BLE_WPTS_PTU_CLASS_5                   4           /**< Class 5 */
#define BLE_WPTS_PTU_CLASS_6                   5           /**< Class 6 */

/**@brief PRU Static Parameter characteristic byte positions */
#define BLE_WPTS_PRU_STATIC_OPTIONAL_FIELDS_POS    0  /**< Optional fields byte position. */
#define BLE_WPTS_PRU_STATIC_PROTOCOL_REV_POS       1  /**< Protocol revision byte position. */
#define BLE_WPTS_PRU_STATIC_PRU_CATEGORY_POS       3  /**< PRU Category byte position. */
#define BLE_WPTS_PRU_STATIC_PRU_INFORMATION_POS    4  /**< PRU Information byte position. */
#define BLE_WPTS_PRU_STATIC_HW_REV_POS             5  /**< Hardware revision byte position. */
#define BLE_WPTS_PRU_STATIC_FW_REV_POS             6  /**< Firmware revision  byte position. */
#define BLE_WPTS_PRU_STATIC_PRECT_MAX_POS          7  /**< PRECT_MAX byte position. */
#define BLE_WPTS_PRU_STATIC_VRECT_MIN_STATIC_POS   8  /**< VRECT_MIN_STATIC byte position. */
#define BLE_WPTS_PRU_STATIC_VRECT_HIGH_STATIC_POS 10  /**< VRECT_HIGH_STATIC byte position. */
#define BLE_WPTS_PRU_STATIC_VRECT_SET_POS         12  /**< VRECT SET byte position. */
#define BLE_WPTS_PRU_STATIC_DELTA_R1_POS          14  /**< Delta R1 byte position. */

/**@brief PRU Static characteristic Option field */
#define BLE_WPTS_PRU_STATIC_OPTION_DELTA_R1_BITPOS   0x07  /**< Max impedance value position */
#define BLE_WPTS_PRU_STATIC_OPTION_DELTA_R1_BITMSK   (0x01 << BLE_WPTS_PRU_STATIC_OPTION_DELTA_R1_BITPOS) /**< Max impedance value mask */

/**@brief PRU Static characteristic PRU Category field */
#define BLE_WPTS_PRU_CATEGORY_NO_IMP_SHIFT     0           /**< Can never create impedance shift */
#define BLE_WPTS_PRU_CATEGORY_1                1           /**< Category 1 */
#define BLE_WPTS_PRU_CATEGORY_2                2           /**< Category 2 */
#define BLE_WPTS_PRU_CATEGORY_3                3           /**< Category 3 */
#define BLE_WPTS_PRU_CATEGORY_4                4           /**< Category 4 */
#define BLE_WPTS_PRU_CATEGORY_5                5           /**< Category 5 */
#define BLE_WPTS_PRU_CATEGORY_6                6           /**< Category 6 */
#define BLE_WPTS_PRU_CATEGORY_7                7           /**< Category 7 */

/**@brief PRU Static characteristic PRU Information field */
#define BLE_WPTS_PRU_INFO_NFC_BITPOS   0x07                                    /**< NFC supported bit position. */
#define BLE_WPTS_PRU_INFO_NFC_BITMSK (0x01 << BLE_WPTS_PRU_INFO_NFC_BITPOS)   /**< NFC supported bit mask. */
#define BLE_WPTS_PRU_INFO_NFC_SUPPORTED_VAL 1                                  /**< NFC supported value. */
#define BLE_WPTS_PRU_INFO_NFC_NOT_SUPPORTED_VAL 0                              /**< NFC not supported value. */

#define BLE_WPTS_PRU_INFO_BLE_BITPOS  0x06                                     /**< Separate BTLE radio in PRU bit position. */
#define BLE_WPTS_PRU_INFO_BLE_BITMSK (0x01 << BLE_WPTS_PRU_INFO_BLE_BITPOS)   /**< Separate BTLE radio in PRU bit mask. */
#define BLE_WPTS_PRU_INFO_BLE_SUPPORTED_VAL 0x01                               /**< Separate BTLE radio in PRU supported value. */
#define BLE_WPTS_PRU_INFO_BLE_NOT_SUPPORTED_VAL 0x00                           /**< Separate BTLE radio in PRU not supported value. */

#define BLE_WPTS_PRU_INFO_POWER_CONTROL_PREF_BITPOS  0x05                                                  /**< Power control algorithm preference value position. */
#define BLE_WPTS_PRU_INFO_POWER_CONTROL_PREF_BITMSK (0x01 << BLE_WPTS_PRU_INFO_POWER_CONTROL_PREF_BITPOS) /**< Power control algorithm preference value mask. */
#define BLE_WPTS_PRU_INFO_POWER_CONTROL_PREF_VRECT_MIN_ERROR_VAL 0x00                                      /**< Vrect min errror value. */
#define BLE_WPTS_PRU_INFO_POWER_CONTROL_PREF_MAX_SYS_EFC_VAL 0x01                                          /**< Max system efficiency value. */

#define BLE_WPTS_PRU_INFO_ADJUST_POWER_BITPOS  0x04                                            /**< Adjust power capability value position. */
#define BLE_WPTS_PRU_INFO_ADJUST_POWER_BITMSK (0x01 << BLE_WPTS_PRU_INFO_ADJUST_POWER_BITPOS) /**< Adjust power capability value mask. */
#define BLE_WPTS_PRU_INFO_ADJUST_POWER_SUPPORTED_VAL 0x01                                      /**< Adjust PRU power supported value. */
#define BLE_WPTS_PRU_INFO_ADJUST_POWER_NOT_SUPPORTED_VAL 0x00                                  /**< Adjust PRU power not supported value. */

#define BLE_WPTS_PRU_INFO_CHARGE_COMPLETE_CONN_MODE_BITPOS  0x03                                                           /**< Charge Completed Connected Mode value position. */
#define BLE_WPTS_PRU_INFO_CHARGE_COMPLETE_CONN_MODE_BITMSK (0x01 << BLE_WPTS_PRU_INFO_CHARGE_COMPLETE_CONN_MODE_BITPOS)   /**< Charge Completed Connected Mode value mask. */
#define BLE_WPTS_PRU_INFO_CHARGE_COMPLETE_CONN_MODE_SUPPORTED_VAL 0x01                                                     /**< Charge Completed Connected Mode supported value. */
#define BLE_WPTS_PRU_INFO_CHARGE_COMPLETE_CONN_MODE_NOT_SUPPORTED_VAL 0x00                                                 /**< Charge Completed Connected Mode not supported value. */

#define BLE_WPTS_PRU_INFO_PTU_TEST_MODE_BITPOS  0x02                                               /**< PTU Test Mode value position. */
#define BLE_WPTS_PRU_INFO_PTU_TEST_MODE_BITMSK (0x01 << BLE_WPTS_PRU_INFO_PTU_TEST_MODE_BITPOS)   /**< PTU Test Mode value mask. */
#define BLE_WPTS_PRU_INFO_PTU_TEST_MODE_ENABLED_VAL 0x01                                           /**< PTU Test Mode enabled value. */
#define BLE_WPTS_PRU_INFO_PTU_TEST_MODE_DISABLED_VAL 0x00                                          /**< PTU Test Mode disabled value. */

/**@brief PRU dynamic characteristic Alert bitfield.
 * Remaining bitifields is the same as PRU Alert characteristic bitfields.
*/
#define BLE_WPTS_PRU_DYNAMIC_ALERT_CHARGE_PORT_BITPOS        0x01  /**< Mode Transition bits*/
#define BLE_WPTS_PRU_DYNAMIC_ALERT_CHARGE_PORT_BITMSK        (0x01 << BLE_WPTS_PRU_DYNAMIC_ALERT_CHARGE_PORT_BITPOS) /**< Wired Charger Detect */
#define BLE_WPTS_PRU_DYNAMIC_ALERT_ADJ_POWER_RESP_BITPOS     0x00  /**< Adjust power bit position */
#define BLE_WPTS_PRU_DYNAMIC_ALERT_ADJ_POWER_RESP_BITMSK     (0x01 << BLE_WPTS_PRU_DYNAMIC_ALERT_ADJ_POWER_RESP_BITPOS) /**< Adjust power bitmask */

/**@brief PRU dynamic characteristic byte positions */
#define BLE_WPTS_PRU_DYNAMIC_OPTIONAL_FIELDS_POS           0  /**< Optional fields byte position. */
#define BLE_WPTS_PRU_DYNAMIC_VRECT_POS                     1  /**< VRECT byte position. */
#define BLE_WPTS_PRU_DYNAMIC_IRECT_POS                     3  /**< IRECT byte position. */
#define BLE_WPTS_PRU_DYNAMIC_VOUT_POS                      5  /**< VOUT byte position. */
#define BLE_WPTS_PRU_DYNAMIC_IOUT_POS                      7  /**< IOUT byte position. */
#define BLE_WPTS_PRU_DYNAMIC_TEMPERATURE_POS               9  /**< TEMPERATURE byte position. */
#define BLE_WPTS_PRU_DYNAMIC_VRECT_MIN_DYN_POS             10  /**< VRECT_MIN_DYN byte position. */
#define BLE_WPTS_PRU_DYNAMIC_VRECT_SET_DYN_POS             12  /**< VRECT_SET_DYN byte position. */
#define BLE_WPTS_PRU_DYNAMIC_VRECT_HIGH_DYN_POS            14  /**< VRECT_HIGH_DYN byte position. */
#define BLE_WPTS_PRU_DYNAMIC_PRU_ALERT_POS                 16  /**< PRU_ALERT byte position.  */
#define BLE_WPTS_PRU_DYNAMIC_PTU_TESTER_COMMAND_POS        17  /**< PTU_TESTER_COMMAND byte position. */

///**@brief PRU Dynamic Parameter characteristic Option field */
#define BLE_WPTS_PRU_DYNAMIC_OPTION_VOUT_BITPOS              7  /**< Vout */
#define BLE_WPTS_PRU_DYNAMIC_OPTION_IOUT_BITPOS              6  /**< Iout */
#define BLE_WPTS_PRU_DYNAMIC_OPTION_TEMPERATURE_BITPOS       5  /**< Temperature */
#define BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_MIN_DYN_BITPOS     4  /**< Vrect_min_dyn */
#define BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_SET_DYN_BITPOS     3  /**< Vrect_set_dyn */
#define BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_HIGH_DYN_BITPOS    2  /**< Vrect_high_dyn */

#define BLE_WPTS_PRU_DYNAMIC_OPTION_VOUT_BITMSK              (1 << BLE_WPTS_PRU_DYNAMIC_OPTION_VOUT_BITPOS)           /**< Vout */
#define BLE_WPTS_PRU_DYNAMIC_OPTION_IOUT_BITMSK              (1 << BLE_WPTS_PRU_DYNAMIC_OPTION_IOUT_BITPOS)           /**< Iout */
#define BLE_WPTS_PRU_DYNAMIC_OPTION_TEMPERATURE_BITMSK       (1 << BLE_WPTS_PRU_DYNAMIC_OPTION_TEMPERATURE_BITPOS)    /**< Temperature */
#define BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_MIN_DYN_BITMSK     (1 << BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_MIN_DYN_BITPOS)  /**< Vrect_min_dyn */
#define BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_SET_DYN_BITMSK     (1 << BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_SET_DYN_BITPOS)  /**< Vrect_set_dyn */
#define BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_HIGH_DYN_BITMSK    (1 << BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_HIGH_DYN_BITPOS)  /**< Vrect_high_dyn */

/**@brief PRU alert characteristic byte positions */
#define BLE_WPTS_PRU_ALERT_POS                             0  /**< Alert bitfield. */
#define BLE_WPTS_PRU_ADDRESS_POS                           1  /**< Device address. */

/**@brief PRU Alert characteristic value bitfield. */
#define BLE_WPTS_PRU_ALERT_OVER_VOLTAGE_BITPOS                0x07                                                       /**< Overvoltage bit posistion */
#define BLE_WPTS_PRU_ALERT_OVER_VOLTAGE_BITMSK                (0x01 << BLE_WPTS_PRU_ALERT_OVER_VOLTAGE_BITPOS)          /**< Overvoltage value mask */
#define BLE_WPTS_PRU_ALERT_OVER_CURRENT_BITPOS                0x06                                                       /**< Overcurrent bit posistion */
#define BLE_WPTS_PRU_ALERT_OVER_CURRENT_BITMSK                (0x01 << BLE_WPTS_PRU_ALERT_OVER_CURRENT_BITPOS)          /**< Overcurrent bit mask*/
#define BLE_WPTS_PRU_ALERT_OVER_TEMPERATURE_BITPOS            0x05                                                       /**< Overtemp bit position */
#define BLE_WPTS_PRU_ALERT_OVER_TEMPERATURE_BITMSK            (0x01 << BLE_WPTS_PRU_ALERT_OVER_TEMPERATURE_BITPOS)      /**< Overtemp bit mask */
#define BLE_WPTS_PRU_ALERT_SELF_PROTECTION_BITPOS             0x04                                                       /**< Self protection bit position */
#define BLE_WPTS_PRU_ALERT_SELF_PROTECTION_BITMSK             (0x01 << BLE_WPTS_PRU_ALERT_SELF_PROTECTION_BITPOS)       /**< Self protection bit mask */
#define BLE_WPTS_PRU_ALERT_CHARGE_COMPLETE_BITPOS             0x03                                                       /**< Charge Complete bit position */
#define BLE_WPTS_PRU_ALERT_CHARGE_COMPLETE_BITMSK             (0x01 << BLE_WPTS_PRU_ALERT_CHARGE_COMPLETE_BITPOS)       /**< Charge Complete bit mask */
#define BLE_WPTS_PRU_ALERT_WIRED_CHARGER_DETECT_BITPOS        0x02                                                       /**< Wired Charger Detect bit position */
#define BLE_WPTS_PRU_ALERT_WIRED_CHARGER_DETECT_BITMSK        (0x01 << BLE_WPTS_PRU_ALERT_WIRED_CHARGER_DETECT_BITPOS)  /**< Wired Charger Detect bit mask */
#define BLE_WPTS_PRU_ALERT_MODE_TRANSITION_BITPOS             0x00                                                       /**< Mode Transition bit position */
#define BLE_WPTS_PRU_ALERT_MODE_TRANSITION_BITMSK             (0x03 << BLE_WPTS_PRU_ALERT_MODE_TRANSITION_BITPOS)       /**< Mode Transition bit mask */
#define BLE_WPTS_PRU_ALERT_NO_SELF_PROTECTION_VAL             0x00                                                       /**< No self protection*/

///**@brief PRU Alert Mode transition */
#define BLE_WPTS_PRU_ALERT_MODE_NO_MODE_TRANSITION_VAL      0x00  /**< No Mode Transition */
#define BLE_WPTS_PRU_ALERT_MODE_2S_MODE_TRANSITION_VAL      0x01  /**< 2s Mode Transition */
#define BLE_WPTS_PRU_ALERT_MODE_3S_MODE_TRANSITION_VAL      0x02  /**< 3s Mode Transition */
#define BLE_WPTS_PRU_ALERT_MODE_6S_MODE_TRANSITION_VAL      0x03  /**< 6s Mode Transition */

/**@brief WPT Service UUIDs */
#define BLE_WPTS_UUID16                                    0xFFFE /**< 16 bit WPT Service. */
#define BLE_WPTS_UUID16_LEN                                2 /**< Length in bytes of 16-bit UUID. */
#define BLE_WPTS_UUID128_CHAR_BASE                         {0x67, 0x9A, 0x0C, 0x20, 0x00, 0x08, 0x96, 0x9E, 0xE2, 0x11, 0x46, 0xA1, 0x70, 0xE6, 0x55, 0x64} /**< Full 128 bit WPT Service. */
#define BLE_WPTS_UUID128_LEN                               16 /**< Length in bytes of 128-bit UUID. */
#define BLE_WPTS_UUID_PRU_CONTROL_CHAR_OFFSET              0 /**< UUID byte 0 offset relative to 128-bit base byte 0. */
#define BLE_WPTS_UUID_PTU_STATIC_PARAMETER_CHAR_OFFSET     1 /**< UUID byte 0 offset relative to 128-bit base byte 0. */
#define BLE_WPTS_UUID_PRU_ALERT_CHAR_OFFSET                2 /**< UUID byte 0 offset relative to 128-bit base byte 0. */
#define BLE_WPTS_UUID_PRU_STATIC_PARAMETER_CHAR_OFFSET     3 /**< UUID byte 0 offset relative to 128-bit base byte 0. */
#define BLE_WPTS_UUID_PRU_DYNAMIC_PARAMETER_CHAR_OFFSET    4 /**< UUID byte 0 offset relative to 128-bit base byte 0. */
#define BLE_WPTS_MODIFIED_BYTE                             0 /**< Byte number of WPT Service UUID modified to generate other 128 bit UUIDs  */

/**@brief WPT Service Characteristic value handles offset relative to primary service handle */
#define BLE_WPTS_HANDLE_OFFSET_PRU_CONTROL              2  /**< Offset of PRU Control characteristic relative to WPT Service handle. */
#define BLE_WPTS_HANDLE_OFFSET_PTU_STATIC               4  /**< Offset of PTU Static characteristic relative to WPT Service handle. */
#define BLE_WPTS_HANDLE_OFFSET_PRU_ALERT                6  /**< Offset of PRU Alert characteristic relative to WPT Service handle. */
#define BLE_WPTS_HANDLE_OFFSET_PRU_ALERT_CCCD           7  /**< Offset of PRU Alert CCCD relative to WPT Service handle. */
#define BLE_WPTS_HANDLE_OFFSET_PRU_STATIC               9  /**< Offset of PRU Static characteristic relative to WPT Service handle. */
#define BLE_WPTS_HANDLE_OFFSET_PRU_DYNAMIC             11  /**< Offset of PRU Dynamic characteristic relative to WPT Service handle. */
#define BLE_WPTS_HANDLE_OFFSET_MIN                      2  /**< Minimum offset of WPT Service characteristics relative to WPT Service handle. */
#define BLE_WPTS_HANDLE_OFFSET_MAX                     11  /**< Maximum offset of WPT Service characteristics relative to WPT Service handle. */

/**@brief Characteristic value lengths */
#define BLE_WPTS_PRU_CONTROL_LEN                        5   /**< PRU Control characteristic length. */
#define BLE_WPTS_PTU_STATIC_LEN                        17   /**< PTU Static characteristic length. */
#define BLE_WPTS_PRU_ALERT_MIN_LEN                      1   /**< PRU Alert characteristic minimum length. */
#define BLE_WPTS_PRU_ALERT_MAX_LEN                      7   /**< PRU Alert characteristic maximum length. */
#define BLE_WPTS_PRU_STATIC_LEN                        20   /**< PRU Static characteristic  length. */
#define BLE_WPTS_PRU_DYNAMIC_LEN                       20   /**< PRU Dynamic characteristiclength. */

/**@brief PTU Static characteristic PTU Power field */
#define BLE_WPTS_PTU_POWER_0_1W                0            /**< Undocumented */
#define BLE_WPTS_PTU_POWER_0_2W                1            /**< Undocumented */
#define BLE_WPTS_PTU_POWER_0_3W                2            /**< Undocumented */
#define BLE_WPTS_PTU_POWER_0_4W                3            /**< Undocumented  */
#define BLE_WPTS_PTU_POWER_0_5W                4            /**< Undocumented */
#define BLE_WPTS_PTU_POWER_0_6W                5            /**< Undocumented */
#define BLE_WPTS_PTU_POWER_0_7W                6            /**< Undocumented */
#define BLE_WPTS_PTU_POWER_0_8W                7            /**< Undocumented */
#define BLE_WPTS_PTU_POWER_0_9W                8            /**< Undocumented */
#define BLE_WPTS_PTU_POWER_1_0W                10           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_1_1W                11           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_1_2W                12           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_1_3W                13           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_1_4W                14           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_1_5W                15           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_1_6W                16           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_1_7W                17           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_1_8W                18           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_1_9W                19           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_2_0W                20           /**< Undocumented */

#define BLE_WPTS_PTU_POWER_2_2W                21           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_2_4W                22           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_2_6W                23           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_2_8W                24           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_3_0W                25           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_3_2W                26           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_3_4W                27           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_3_6W                28           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_3_8W                29           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_4_0W                30           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_4_2W                31           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_4_4W                32           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_4_6W                33           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_4_8W                34           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_5_0W                35           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_5_2W                36           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_5_4W                37           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_5_6W                38           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_5_8W                39           /**< Undocumented */

#define BLE_WPTS_PTU_POWER_6_0W                40           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_6_3W                41           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_6_6W                42           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_6_9W                43           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_7_2W                44           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_7_5W                45           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_7_8W                46           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_8_1W                47           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_8_4W                48           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_8_7W                49           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_9_0W                50           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_9_3W                51           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_9_6W                52           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_9_9W                53           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_10_2W               54           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_10_5W               55           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_10_8W               56           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_11_1W               57           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_11_4W               58           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_11_7W               59           /**< Undocumented */

#define BLE_WPTS_PTU_POWER_12_0W               60           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_12_4W               61           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_12_8W               62           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_13_2W               63           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_13_6W               64           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_13_0W               65           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_14_4W               66           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_14_8W               67           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_15_2W               68           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_15_6W               69           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_16_0W               70           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_16_4W               71           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_16_8W               72           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_17_2W               73           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_17_6W               74           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_18_0W               75           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_18_4W               76           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_18_8W               77           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_19_2W               78           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_19_6W               79           /**< Undocumented */

#define BLE_WPTS_PTU_POWER_20_0W               80           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_20_5W               81           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_21_0W               82           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_21_5W               83           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_22_0W               84           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_22_5W               85           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_23_0W               86           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_23_5W               87           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_24_0W               88           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_24_5W               89           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_25_0W               90           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_25_5W               91           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_26_0W               92           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_26_5W               93           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_27_0W               94           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_27_5W               95           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_28_0W               96           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_28_5W               97           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_29_0W               98           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_29_5W               99           /**< Undocumented */

#define BLE_WPTS_PTU_POWER_30_0W               100           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_30_6W               101           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_31_2W               102           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_31_8W               103           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_32_4W               104           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_33_0W               105           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_33_6W               106           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_34_2W               107           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_34_8W               108           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_35_4W               109           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_36_0W               110           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_36_6W               111           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_37_2W               112           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_37_8W               113           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_38_4W               114           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_39_0W               115           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_39_6W               116           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_40_2W               117           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_40_8W               118           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_41_4W               119           /**< Undocumented */

#define BLE_WPTS_PTU_POWER_42_0W               120           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_43_0W               121           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_44_0W               122           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_45_0W               123           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_46_0W               124           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_47_0W               125           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_48_0W               126           /**< Undocumented */
#define BLE_WPTS_PTU_POWER_49_0W               127           /**< Undocumented */

/**@brief Struct holding contents of "Service data AD type". */
typedef struct
{
    uint16_t uuid16;                        /**< 16-bit WPT Service UUID. */
    uint16_t gatt_prim_srv_handle;          /**< PRU primary service handle. */
    int8_t pru_pwr;                         /**< PRU Power value. [dBm] */
    int8_t pru_gain;                        /**< PRU Gain value. [dBm] */
    uint8_t adv_flags;                      /**< ADV flags. */
} ble_wpts_service_data_t;

/**@brief Set of available timer operation types. */
typedef enum
{
    TIMER_USER_OP_TYPE_NONE,                /**< No timer operation */
    TIMER_USER_OP_TYPE_START,               /**< Timer operation type Start. */
    TIMER_USER_OP_TYPE_STOP,                /**< Timer operation type Stop. */
    TIMER_USER_OP_TYPE_STOP_ALL             /**< Timer operation type Stop All. */
} timer_user_op_type_t;

/**@brief Identification of 128 bit UUIDs. */
typedef enum
{
    BLE_WPTS_UUID128_PRU_SERVICE = 0,                  /**< WPT Service UUID. */
    BLE_WPTS_UUID128_PRU_CONTROL_CHAR,                 /**< PRU Control Characteristic UUID. */
    BLE_WPTS_UUID128_PTU_STATIC_PARAMETER_CHAR,        /**< PTU Static Parameter Characteristic UUID. */
    BLE_WPTS_UUID128_PRU_ALERT_CHAR,                   /**< PRU Alert Characteristic UUID. */
    BLE_WPTS_UUID128_PRU_STATIC_PARAMETER_CHAR,        /**< PRU Static Parameter Characteristic UUID. */
    BLE_WPTS_UUID128_PRU_DYNAMIC_PARAMETER_CHAR,       /**< PRU Dynamic Parameter Characteristic UUID. */
    BLE_WPTS_UUID128_CNT
} ble_wpts_uuid128_t;

/**@brief Prepare softdevice for using the WPTS 128 bit UUIDs.
 * @return      NRF_SUCCESS on successful initialization of UUIDs, otherwise an error code.
 */
uint32_t ble_wpts_uuid_init(void);

/**@brief Get ble_uuid representation of 128 bit UUID.
 *
 * @param[in]   uuid128
 * @param[out]  ble_uuid
 * @return      NRF_SUCCESS on successful initialization of UUIDs, otherwise an error code.
 */
void ble_wpts_uuid_get(ble_wpts_uuid128_t uuid128, ble_uuid_t * ble_uuid);


/**@brief Write bitfield
 *
 * @param[in]   bitfield bitfield to be modified
 * @param[in]   msk mask for bitfield
 * @param[in]   pos position for bitfield
 * @param[in]   val value to be written
 */
#define BLE_WPTS_BITFIELD_WRITE(bitfield, msk, pos, val) do{   \
  bitfield = (bitfield & ~msk) | ((val << pos) & msk);                  \
} while(0);

/**@brief Read bitfield
 *
 * @param[in]   bitfield bitfield to be read
 * @param[in]   msk mask for bit/value to be read
 * @param[in]   pos position for bit/value to be read
 *
 * @return Value of field.
 */
#define BLE_WPTS_BITFIELD_READ(bitfield, msk, pos) ((bitfield & msk) >> pos)


#endif // BLE_WPTS_COMMON_H__

/** @} */
