/* Copyright (c) Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
 *
 * @defgroup ptu_config PTU profile configuration
 * @{
 * @ingroup profile
 */

#ifndef __PTU_CONFIG_H__
#define __PTU_CONFIG_H__

#include "ptu_hw_config.h"

/******************************************************************************
 * Misc. parameters
 ******************************************************************************/

#define PTU_MAX_CONNECTIONS               8       /**< Maximum number of simultaneous connections supported by PTU [#]. */
#define PTU_MAX_DEVICES                   8       /**< Maximum number of devices that can be charged simultaneously [#]. */
#define PTU_OUTPUT_TX_POWER               0       /**< Ouput TX power [dBm]. */
#define PTU_HW_REVISION                   0       /**< Revision of the PTU HW. */
#define PTU_FW_REVISION                   0       /**< Revision of the PTU SW. */
#define PTU_PROTOCOL_REVISION             1       /**< Protocol Revision */
#define PTU_MAX_APP_TIMERS                3       /**< Maximium timers to be instansiated in application [#].  */
#define PTU_SENSORS_READ_INTERVAL_MS      30      /**< Local sensor read interval [ms]. */
#define PTU_PRU_DYN_READ_INTERVAL_MS      50      /**< PRU dynamic parameters read interval [ms]. */
#define PTU_CONFIG_TIMEOUT_MS             10      /**< Duration of configuruation state [ms]. */
#define PTU_SEC_MIN_KEYSIZE               7       /**< Minimum keysize [7,16] to be used for encryption [bytes]. */
#define PTU_SEC_MAX_KEYSIZE               16      /**< Minimum keysize [7,16] to be used for encryption [bytes]. */
#define PTU_P_ADJ_CMD_COOLDOWN_MS         2000    /**< Time required between power adjust commands to a PRU [ms]. */
#define PTU_ADJUSTED_POWER_BIT_TIMEOUT_MS 2000    /**< Timeout value for PRU setting power adjusted bit [ms]. */
#define PTU_POWER_TRANSFER_HYSTERISIS     100     /**< Set slack for adjusting vrect. Will not be adjusted before value exceeds VRECT +/- (1/2 * slack) [mV]*/
#define PTU_RECONNECT_TIMEOUT_MS          1100    /**< Timeout of PTU reconnection procedure. [ms]*/
#define PTU_DISTANT_LIST_LEN              30      /**< How many entries should the distant list hold. */
#define PTU_MS_FROM_BEACON_STOP_TO_ACTIVATE_DISTANT_LIST            100  /**< How long should PTU wait after Long Beacon end before
                                                                              starting to add devices to distant list? [ms] */
#define PTU_ALLOWED_REDUCED_POSITIVE_ITX_COIL_ADJUSTMENT_MULTIPLIER  0.95 /**< Multiply with PRU VRECT to see if positive
                                                                               ITX_COIL adjustment step size can be reduced. [scalar] */
#define PTU_ALLOWED_REDUCED_NEGATIVE_ITX_COIL_ADJUSTMENT_MULTIPLIER  1.05 /**< Multiply with PRU VRECT to see if negative
                                                                               ITX_COIL adjustment step size can be reduced. [scalar] */
#define PTU_DOMINANT_PRU_UPPER_LIMIT_VRECTSET_MULTIPLIER_FOR_REDUCTION_IN_STEP_SIZE 1.0/0.95 /**< Multiply with dominant PRU VRECT to see if
                                                                                                  PTU qualifies for smaller ITX_COIL step size,
                                                                                                  in conjunction with lower limit.*/
#define PTU_DOMINANT_PRU_LOWER_LIMIT_VRECTSET_MULTIPLIER_FOR_REDUCTION_IN_STEP_SIZE 1.0/1.05 /**< Multiply with dominant PRU VRECT to see if
                                                                                                  PTU qualifies for smaller ITX_COIL step size,
                                                                                                  in conjunction with upper limit.*/

/******************************************************************************
 * Load detection
 ******************************************************************************/

#define PTU_LOAD_VARIATION_DETECT_LIFETIME_MS  1500     /**< Time a load variation detect is assumed valid [ms] */
#define PTU_Z_TX_IN_NO_LOAD                    100      /**< Measured impedance above this limit [ohms] are seen as "no load" */
#define PTU_Z_TX_IN_LOAD_DETECT                50       /**< Impedenacce threshold [ohms] use for "load variation detect" */
#define PTU_P_TX_IN_LOAD_DETECT                50       /**< 1/100 W change required to generate load variation */

/** Number of impedance samples to store for load variation detection.
 * Time over which to calculate load variation will be given as:
 * PTU_SENSORS_READ_INTERVAL_MS * PTU_Z_MEAS_BUF_LENGTH.
 */
#define PTU_Z_MEAS_BUF_LENGTH                  ((100 / PTU_SENSORS_READ_INTERVAL_MS) + 1)

/******************************************************************************
 * Cross connection
 ******************************************************************************/

#define PTU_CCA_RSSI_ENABLE                true     /**< Enable filtering of adv packets based on receive power */
#define PTU_CCA_TIME_SET_ENABLE            true     /**< Expect load variation after sending "time set value" in control packet */
#define PTU_CCA_REBOOT_BIT_ENABLE          false    /**< Reboot bit handling + "distant list" */
#define PTU_TIME_SET_VAL_MS            80       /**< Time set value [ms]. */
#define PTU_TIME_SET_VAL_MS_MARGIN      4       /**< Allowed offset margin for time set load variation [ms]. */
#define PTU_TIME_SET_CHECK_INTERVAL_MS 2        /**< How often will samples of P_TX be taken when performing Time Set procedure [ms]. */
#define PTU_TIME_SET_500_MS_OF_SAMPLES \
        (500 / PTU_TIME_SET_CHECK_INTERVAL_MS)  /**< How many time set samples makes up 500ms. */
#define PTU_TIME_SET_LOWER_LIMIT_MS    \
        (PTU_TIME_SET_VAL_MS -         \
         PTU_TIME_SET_VAL_MS_MARGIN -  \
         PTU_TIME_SET_CHECK_INTERVAL_MS)        /**< Lower limit of duration of load variation during timeset. */
#define PTU_TIME_SET_UPPER_LIMIT_MS    \
        (PTU_TIME_SET_VAL_MS +         \
         PTU_TIME_SET_VAL_MS_MARGIN)            /**< Upper limit of duration of load variation during timeset. */
#define PTU_TIME_SET_VAL_LOW_LOAD_MS   20       /**< Period load shall remain off after PRU has enabled dummy load [ms].*/
#define PTU_MAX_DISTANT_PRUS           3        /**< Maximum number of PRUs in distant list [#]. */
#define PTU_ADV_PRE_CONNECT_ALLOW      11       /**< Number of adv packets received from a PRU only fulfilling one of two connection
                                                          conditions before connection is granted anyway 7(6.6.2). */
#define PTU_ADV_IGNORE_TIMEOUT_MS      1700     /**< Time elapsed after 1 of 2 conditions are fulfilled until connection may be
                                                          sent anyway (6.6.2). */

/******************************************************************************
 * Scan parameters
 ******************************************************************************/

#define PTU_POWER_SAVE_STATE_SCAN_INTERVAL     160 /**< Scan interval when 0 devices connected. 100 ms (0.625 ms units)  */
#define PTU_POWER_SAVE_STATE_SCAN_WINDOW       PTU_POWER_SAVE_STATE_SCAN_INTERVAL  /**< Scan window when 0 devices connected.Equals tp scan interval */
#ifdef WDBGPTU
#define PTU_NORMAL_SCAN_INTERVAL              100  /**< Scan interval when > 0 devices connected. 62.5 ms (0.625 ms units) */
#else
#define PTU_NORMAL_SCAN_INTERVAL               80  /**< Scan interval when > 0 devices connected. 50 ms (0.625 ms units) */
#endif
#define PTU_NORMAL_SCAN_WINDOW                 48  /**< Scan window when > 0 devices connected. 30 ms (0.625 ms units) */

/******************************************************************************
 * Connection parameters
 ******************************************************************************/

#define PTU_MIN_CONN_INTERVAL                       40      /**< Minimum acceptable connection interval, Connection interval uses 1.25 ms units. */
#define PTU_MAX_CONN_INTERVAL                       40      /**< Maximum acceptable connection interval, Connection interval uses 1.25 ms units. */
#define PTU_SLAVE_LATENCY                           0       /**< Slave latency. */
#define PTU_CONN_SUP_TIMEOUT                        100     /**< Connection supervisory timeout, Supervision Timeout uses 10 ms units. */
#define PTU_MAX_BLE_CONN_ATTEMPTS                   2       /**< Max BLE connection attempts per registration */
#define PTU_MAX_REG_ATTEMPTS                        2       /**< Number of connection attempts before entering latching fault */
#define PTU_REG_TIMEOUT_MS                          500     /**< Registration timeout [ms]. */

/** Decides whether PTU should revoke acceptance of PRUs which were fully accepted after fulfilling
 *  the criteria for partially accepted devices mentioned in the "Acceptance of Advertisement" chapter.
 *
 *  If this value is set to false, no revocation will take place, and the PRU will stay in state
 *  'FULLY_ACCEPTED' until registry is filled up, and the PRU will be removed to make room for
 *  new devices.
  *
 *  If this value is set to true the PRU will be revoked acceptance after a configurable amount of time
 *  after being set to 'FULLY_ACCEPTED'.
 *  ! - Note that the specification does not mention any upper limit
 *      for the validity of a PRU which has been set to 'FULLY_ACCEPTED'
 *
 *  If Revocation of Fully Accepted PRUs is used, the revocation timeouts should be configured. I.E. if
 *  the PTU is in Power Save, the PTU should wait until at least one Long Beacon has been performed.
 */
#define PTU_REVOKE_ACCEPTED_STATE                   true

/** Timeout value for revoking the 'FULLY_ACCEPTED' PRU when the PTU is in
 *  any state BUT power save.
 */
#define PTU_REVOKE_ACCEPTED_STATE_NON_POWER_SAVE_MS \
    (PTU_REVOKE_ACCEPTED_STATE_RETRIES * PTU_BEACON_DURATION_LONG_MS)

/** Timeout value for revoking the 'FULLY_ACCEPTED' state from PRU when the PTU is in
 *  power save state.
 */
#define PTU_REVOKE_ACCEPTED_STATE_POWER_SAVE_MS \
    (PTU_REVOKE_ACCEPTED_STATE_RETRIES * ((PTU_BEACON_CYCLES_LONG * PTU_BEACON_CYCLE_PERIOD_MS) + PTU_BEACON_DURATION_LONG_MS))

/** Both values PTU_REVOKE_ACCEPTED_STATE_POWER_SAVE_MS and PTU_REVOKE_ACCEPTED_STATE_NON_POWER_SAVE_MS
 *  are multiplied with this value to get the timeout value of revoking the accepted PRU.
 *  The two abovementioned values will typically reflect the time required for allowing the PRU one more
 *  chance to advertise, in which case this value configures how many "chances" the PRU should get.
 */
#define PTU_REVOKE_ACCEPTED_STATE_RETRIES           3


/******************************************************************************
 * Rogue object detection
 ******************************************************************************/

#define PTU_ROGUE_OBJECT_DETECT_THRESHOLD     30      /**< Rougue object detected threshold [1/10 W]. */
#define PTU_CCA_ROGUE_DETECT_ENABLE           true    /**< Rogue object detection enable */

/******************************************************************************
 * Beacon parameters (Power Save and Latching Fault state)
 ******************************************************************************/

#define PTU_BEACON_CYCLE_PERIOD_MS                 250     /**< Beacon cycle (t_cycle) [ms] */
#define PTU_BEACON_CYCLES_LONG                     10      /**< Number of beacon cycles for every long beacon. */
#define PTU_BEACON_DURATION_SHORT_MS               10      /**< Duration of short beacon [ms]. */

/* Verify that beacon cycles and durations are BSS compliant (section 5.2.3.3 and 5.2.3.4) */
#if PTU_BEACON_CYCLE_PERIOD_MS != 250
    #error PTU_BEACON_CYCLE_PERIOD_MS MUST be 250
#endif

#if (PTU_BEACON_CYCLES_LONG * PTU_BEACON_CYCLE_PERIOD_MS) < 850
    #error PTU_BEACON_CYCLES_LONG * PTU_BEACON_CYCLE_PERIOD_MS can not be smaller than 850
#elif (PTU_BEACON_CYCLES_LONG * PTU_BEACON_CYCLE_PERIOD_MS) > 3000
    #error PTU_BEACON_CYCLES_LONG * PTU_BEACON_CYCLE_PERIOD_MS can not be larger than 3000
#endif

#if PTU_BEACON_DURATION_SHORT_MS > 30
    #error PTU_BEACON_DURATION_SHORT_MS can not be larger than 30
#endif

/******************************************************************************
 * Latching fault parameters
 ******************************************************************************/

#define PTU_LATCH_FAULT_ENTRY_DELAY_MS             1000      /**< Delay after entering Latching Fault starting short beacon. [ms] */
#define PTU_LATCH_FAULT_EXIT_ATTEMPT_DELAY_MS      5000      /**< Delay after entering Latching Fault before attempting to clear fault. [ms] */
#define PTU_LATCH_FAULT_MAX_AUTO_CLR_ATTEMPTS      3         /**< Maximum number of successive latching faults that can be auto cleared. */
#define PTU_LATCH_FAULT_NO_ERR_CLR_TIMEOUT_SEC     (5 * 60)  /**< Clear successive latching fault count after this period without errors. */
#define PTU_LATCH_FAULT_NO_ERR_CLR_TIMEOUT_MS      PTU_LATCH_FAULT_NO_ERR_CLR_TIMEOUT_SEC * 1000  /**< Clear successive latching fault count after this period without errors. */
#define PTU_LATCH_FAULT_NO_LOAD_VAR_NO_DEVICES_MS  5000      /**< How many milliseconds needs to pass without seeing a load to imply that no devices are on pad. */


/******************************************************************************
 * PTU Static characteristic default values
 ******************************************************************************/

#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_VAL    BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_100_VAL  /**< Default MAX Source Impedance */
#define BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_VALID  0                                                  /**< 1 if MAX Source Impedance is valid, else 0 */
#define BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_VAL     BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_55_VAL    /**< Default MAX Load Resistance */
#define BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_VALID   0                                                  /**< 1 if MAX Load Resistance is valid, else 0 */
#define BLE_WPTS_PTU_POWER                              BLE_WPTS_PTU_POWER_9_9W                           /**< Default PTU Power */


/******************************************************************************
 * PTU DFU parameters
 ******************************************************************************/
 #define PTU_BUTTON_PRESS_LENGTH_DFU_MS  5 * 1000       /**< Button needs to be pressed for 5 seconds in order to start DFU mode. */



/******************************************************************************
 * PTU Beacon parameters
 ******************************************************************************/
#define PTU_BEACON_DURATION_LONG_MS                105     /**< Duration of long beacon [ms]. */
#define PTU_BEACON_DURATION_EXTENDED_MS            3000    /**< Duration of extended beacon [ms]. */
#define PTU_POWER_SAVE_LONG_BEACON_PTX_READ_INTERVAL_MS 2  /**< How often is PTX value read. Changing this value will break LBE functionality. */


#if(PTU_MAX_DEVICES == 0)
    #error PTU_MAX_DEVICES can not be 0
#elif(PTU_MAX_DEVICES == 1)
    #define BLE_WPTS_PTU_STATIC_NUM_DEVICES_BITVAL BLE_WPTS_PTU_STATIC_NUM_DEVICES_1_VAL /**< Internal definition */
#elif(PTU_MAX_DEVICES == 2)
    #define BLE_WPTS_PTU_STATIC_NUM_DEVICES_BITVAL BLE_WPTS_PTU_STATIC_NUM_DEVICES_2_VAL /**< Internal definition */
#elif(PTU_MAX_DEVICES == 3)
    #define BLE_WPTS_PTU_STATIC_NUM_DEVICES_BITVAL BLE_WPTS_PTU_STATIC_NUM_DEVICES_3_VAL /**< Internal definition */
#elif(PTU_MAX_DEVICES == 4)
    #define BLE_WPTS_PTU_STATIC_NUM_DEVICES_BITVAL BLE_WPTS_PTU_STATIC_NUM_DEVICES_4_VAL /**< Internal definition */
#elif(PTU_MAX_DEVICES == 5)
    #define BLE_WPTS_PTU_STATIC_NUM_DEVICES_BITVAL BLE_WPTS_PTU_STATIC_NUM_DEVICES_5_VAL /**< Internal definition */
#elif(PTU_MAX_DEVICES == 6)
    #define BLE_WPTS_PTU_STATIC_NUM_DEVICES_BITVAL BLE_WPTS_PTU_STATIC_NUM_DEVICES_6_VAL /**< Internal definition */
#elif(PTU_MAX_DEVICES == 7)
    #define BLE_WPTS_PTU_STATIC_NUM_DEVICES_BITVAL BLE_WPTS_PTU_STATIC_NUM_DEVICES_7_VAL /**< Internal definition */
#elif(PTU_MAX_DEVICES == 8)
    #define BLE_WPTS_PTU_STATIC_NUM_DEVICES_BITVAL BLE_WPTS_PTU_STATIC_NUM_DEVICES_8_VAL /**< Internal definition */
#else
    #error BLE_WPTS_PTU_STATIC_NUM_DEVICES_BITVAL must be in range [1, 8]
#endif

#if(PTU_TIME_SET_VAL_MS == 0)
    #error PTU_MAX_CONNECTIONS can not be 0
#elif(PTU_TIME_SET_VAL_MS == 10)
    #define BLE_WPTS_TIME_SET_BITVAL CTL_TIME_SET_10MS /**< Internal definition */
#elif(PTU_TIME_SET_VAL_MS == 20)
    #define BLE_WPTS_TIME_SET_BITVAL CTL_TIME_SET_20MS /**< Internal definition */
#elif(PTU_TIME_SET_VAL_MS == 30)
    #define BLE_WPTS_TIME_SET_BITVAL CTL_TIME_SET_30MS /**< Internal definition */
#elif(PTU_TIME_SET_VAL_MS == 40)
    #define BLE_WPTS_TIME_SET_BITVAL CTL_TIME_SET_40MS /**< Internal definition */
#elif(PTU_TIME_SET_VAL_MS == 50)
    #define BLE_WPTS_TIME_SET_BITVAL CTL_TIME_SET_50MS /**< Internal definition */
#elif(PTU_TIME_SET_VAL_MS == 60)
    #define BLE_WPTS_TIME_SET_BITVAL CTL_TIME_SET_60MS /**< Internal definition */
#elif(PTU_TIME_SET_VAL_MS == 70)
    #define BLE_WPTS_TIME_SET_BITVAL CTL_TIME_SET_70MS /**< Internal definition */
#elif(PTU_TIME_SET_VAL_MS == 80)
    #define BLE_WPTS_TIME_SET_BITVAL CTL_TIME_SET_80MS /**< Internal definition */
#else
    #error PTU_TIME_SET_VAL_MS must be in range [10, 80] in steps of 10.
#endif

#if(PTU_CLASS == BLE_WPTS_PTU_CLASS_1)
   #if(PTU_PTX_IN_MAX < 20)
       #warning PTU Class 1 requires PTU_PTX_IN_MAX >= 20
   #endif
#elif(PTU_CLASS == BLE_WPTS_PTU_CLASS_2)
   #if(PTU_PTX_IN_MAX < 100)
       #warning PTU Class 2 requires PTU_PTX_IN_MAX >= 100
   #endif
#elif(PTU_CLASS == BLE_WPTS_PTU_CLASS_3)
   #if(PTU_PTX_IN_MAX < 160)
       #warning PTU Class 3 requires PTU_PTX_IN_MAX >= 160
   #endif
#elif(PTU_CLASS == BLE_WPTS_PTU_CLASS_4)
   #if(PTU_PTX_IN_MAX < 330)
       #warning PTU Class 4 requires PTU_PTX_IN_MAX >= 330
   #endif
#elif(PTU_CLASS == BLE_WPTS_PTU_CLASS_5)
   #if(PTU_PTX_IN_MAX < 500)
       #warning PTU Class 5 requires PTU_PTX_IN_MAX >= 500
   #endif
#elif(PTU_CLASS == BLE_WPTS_PTU_CLASS_6)
   #if(PTU_PTX_IN_MAX < 700)
       #warning PTU Class 6 requires PTU_PTX_IN_MAX >= 700
   #endif
#endif

#if(PTU_SENSORS_READ_INTERVAL_MS > (PTU_TIME_SET_VAL_MS / 2))
    #error PTU_TIME_SET_VAL_MS must be <= (PTU_TIME_SET_VAL_MS / 2)
#endif

#endif /* __PTU_CONFIG_H__ */
/** @} */
