/* Copyright (c) Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
 *
 * @defgroup ptu_hw_config PTU HW dependent configuration
 * @{
 * @ingroup hal
 */

#ifndef __PTU_HW_CONFIG_H__
#define __PTU_HW_CONFIG_H__

#include <stdint.h>
#include <stdbool.h>

/******************************************************************************
 * PTU classification
 ******************************************************************************/

#define PTU_CLASS BLE_WPTS_PTU_CLASS_2                        /**< PTU class. */
#define PTU_PTX_IN_MAX 100                                     /**< Max resonator input power [100 mW]. */

/******************************************************************************
 * Cross connection
 ******************************************************************************/

#define PTU_ADV_PWR_MIN  -55                                       /**< Minimum RSSI strength of ADV packet [dBm]. */

/******************************************************************************
 * Rogue object detection
 ******************************************************************************/

#define PTU_POWER_EFFICIENCY   1.0 /**< Portion of power amplifier input power that is output to resonator [fraction [0.0 - 1.0]]. */

/******************************************************************************
 * Enable / disable cross connection algorithms.
 * When an algorithm is disabled, registration of device will be allowed
 * even if the given registration criteria is not fulfilled.
 ******************************************************************************/

#define PTU_CCA_ADV_IMP_SHIFT_ENABLE               true    /**< An impedance shift must have been registered close to the time of advertisement */

/******************************************************************************
 * Power save state parameters
******************************************************************************/

#define POWERAMP_TEST_VAR_STEP_SIZE                10       /**< Power amplifier step size for test increment / decrement. */
#define PTU_POWERAMP_INPUT_TEST_STEP_SIZE     5        /**< Power amplifier step size when in test mode [%]. */
#define PTU_POWERAMP_INPUT_BEACON_SHORT       30       /**< Power amplifier input level for short beacon [%]. */
#define PTU_POWERAMP_INPUT_BEACON_LONG        40       /**< Power amplifier input level for long beacon  [%]. */

/******************************************************************************
 * Low power state parameters
 ******************************************************************************/

#define PTU_POWERAMP_INPUT_START              50       /**< Power amplifier when entering low power state [mA] */

/******************************************************************************
 * Power transfer state
 ******************************************************************************/

#define PTU_POWERAMP_INPUT_NOMINAL              60       /**< Power amplifier input level when entering power transfer state */
#define PTU_POWERAMP_INPUT_MAX                  100      /**< Maximum level of power amplifier input. */
#define PTU_POWERAMP_INPUT_MIN                  0        /**< Minimum level of power amplifier input. */
#define PTU_POWERAMP_INPUT_STEP_SIZE            2        /**< Power amplifier input step size. */
#define PTU_POWERAMP_INPUT_STEP_SIZE_REDUCED    1        /**< Power amplifier input step size when criteria for reduced step size is fulfilled.
                                                              NOTE: To disable variable step size, set to 'PTU_POWERAMP_INPUT_STEP_SIZE' */
#define PTU_POWERAMP_INPUT_STEP_SIZE_INCREASED  10        /**< Power amplifier input step size when criteria for increased step size is fulfilled. */
#define PTU_VRECT_HIGH_OFFSET                   100      /**< VRECT_HIGH threshold is lowered by this amount in order not to put PRU in the high voltage substate [mV]. */

/******************************************************************************
 * Fault condition parameters
 ******************************************************************************/

#define PTU_ITX_MAX                           (1000)   /**< Maximum ITX before triggering fault condition [mA] */
#define PTU_ITX_MIN                           (10)     /**< Minimum ITX before before triggering fault condition [mA]  */
#define PTU_TEMPERATURE_MAX                   (60)     /**< Maximum PTU temperature before triggering fault condition [Deg C] */
#define PTU_TEMPERATURE_MIN                   (-10)    /**< Minimum PTU temperature before triggering fault condition [Deg C]  */
#define PTU_PTX_IN_TEMP_WARNING_DEC           50       /**< Factor in [%] the PTU MAX is reduced during local temperature warning . */
#define PTU_TEMPERATURE_WARNING_SET           50       /**< Threshold for setting temperature warning condition [Deg C] */
#define PTU_TEMPERATURE_WARNING_CLEAR         40       /**< Threshold for clearing temperature warning condition [Deg C] */

/******************************************************************************
 * Sensor data
 ******************************************************************************/

/**@brief Definition of PTU sensor data */
typedef struct
{
    uint8_t     nof_extend_beacon_load_var_detected;    /**< Extended Beacon load variation detected */
    uint16_t    i_ina;                                  /**< Resonator coil current [mA]. */
    uint16_t    v_ina;                                  /**< Resonator coil voltage [mV]. */
    uint16_t    p_tx_in;                                /**< Input power to the PTU resonator [mW]. */
    int16_t     temperature;                            /**< Temperature of PTU */
    uint16_t    res_impedance;                          /**< Measured impedance [ohms]. */
    bool        temperature_warning;                    /**< Temperature warning. */
    bool        load_var_detected;                      /**< Load variation detected. */
    bool        load_detected;                          /**< Load detected. */
    bool        rogue_obj_detected;                     /**< Rogue object detected. */
} ptu_sensor_data_t;

/******************************************************************************
 * Suggested default values for e.g. emulated HAL libraries.
 ******************************************************************************/

#define PTU_ITX_DEFAULT                (PTU_ITX_MIN+1)                                                /**< [mA] */
#define PTU_VTX_DEFAULT                ((90 * PTU_ITX_DEFAULT))                                       /**< [mV] */
#define PTU_TEMPERATUR_DEFAULT         ((PTU_TEMPERATURE_MAX + PTU_TEMPERATURE_MIN) / 2)         /**< [deg C] */
#define PTU_POWERAMP_INPUT_DEFAULT     ((PTU_POWERAMP_INPUT_MAX + PTU_POWERAMP_INPUT_MIN) / 2)   /**< [unitless] */
#define PTU_PTX_IN_DEFAULT             50                                                                  /**< [1/100 W] */

#endif /* __PTU_HW_CONFIG_H__ */
/** @} */
