/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
 *
 * @defgroup pru_hal PRU HAL API
 * @{
 * @ingroup hal
 */

#ifndef __PRU_HAL_H__
#define __PRU_HAL_H__

#include <stdbool.h>
#include <stdint.h>
#include "wpt.h"

/******************************************************************************
 * Main API
 ******************************************************************************/

/**@brief Initialize the PRU's Hardware abstraction layer.
 *
 * @retval NRF_SUCCESS         Status was read successfully.
 */
uint32_t pru_hal_init(void);


/**@brief Read the value of VRECT.
 *
 * @param[out]  p_vrect      Voltage at rectifier diode output [mV].
 *
 * @retval      NRF_SUCCESS  VRECT was read successfully.
 */
uint32_t pru_hal_vrect_get(uint16_t * p_vrect);


/**@brief Read the value of IRECT.
 *
 * @param[out]  p_irect      Current at rectifier diode output [mA].
 *
 * @retval      NRF_SUCCESS  IRECT was read successfully.
 */
uint32_t pru_hal_irect_get(uint16_t * p_irect);


/**@brief Enable output to battery.
 *
 * @param[in] enable set to true enables output to battery.
 *
 * @retval      NRF_SUCCESS If VOUT was enabled.
 */
uint32_t pru_hal_vout_enable_set(bool enable);


/**@brief Get status of VOUT enable.
 *
 * @param[out] enable enable status of VOUT enable.
 *
 * @retval  NRF_SUCCESS If VOUT was enabled.
 */
uint32_t pru_hal_vout_enable_get(bool * enable);


/**@brief Read the voltage output to battery.
 *
 * @param[out]  p_vout       Voltage at charge/battery port [mV].
 *
 * @retval      NRF_SUCCESS  VOUT was read successfully.
 */
uint32_t pru_hal_vout_get(uint16_t * p_vout);


/**@brief Read the value of IOUT.
 *
 * @param[out]  p_iout       Current at charge/battery port [mA].
 *
 * @retval      NRF_SUCCESS  IOUT was read successfully.
 */
uint32_t pru_hal_iout_get(uint16_t * p_iout);


/**@brief Read the temperature.
 *
 * @param[out]  p_temperature   Temperature of PRU [Deg C].
 *
 * @retval      NRF_SUCCESS     Temperature was read successfully.
 */
uint32_t pru_hal_temperature_get(int16_t * p_temperature);


/**@brief Get the status of whether the PRU is charged by wire.
 *
 * @param[out]  p_wired         True if PRU is connected to a wired charger, false otherwise.
 *
 * @retval      NRF_SUCCESS  Status was read successfully.
 */
uint32_t pru_hal_wired_charge_detect_status_get(bool * p_wired);


/**@brief Get the status of whether the PRU is charged.
 *
 * @param[out]  p_charge_complete  True if PRU has completed charging, false otherwise.
 *
 * @retval      NRF_SUCCESS        Status was read successfully.
 */
uint32_t pru_hal_charge_complete_status_get(bool * p_charge_complete);


/**@brief Limit maximum power consumed by PRU.
 *
 *        This function is invoked by the PRU Profile when the PRU receives a power
 *        adjust command from the PTU.
 *
 *        Once the PRU has successfully adjusted its power consumption, it should
 *        set the power adjusted value to 'true' by invoking
 *        'pru_hal_power_adjusted_set'.
 *
 * @param[in]  adj_setting  specifies the maximium amount of power the PRU can consume.
 *
 * @retval     NRF_SUCCESS  Value was set successfully.
 */
uint32_t pru_hal_adj_power_set(ctl_adj_power_t adj_setting);


/**@brief Get adjust power setting.
 *
 * @param[out] adj_setting  Power adjust setting.
 *
 * @retval     NRF_SUCCESS  Value was returned successfully.
 */
uint32_t pru_hal_adj_power_get(ctl_adj_power_t * adj_setting);


/**@brief Set power adjusted value.
 *
 *        When power adjusted value is set to 'true' the PRU will set its
 *        'Adjusted Power Response' bit in the 'PRU Dynamic' characteristic
 *        to '1' for 1 second, informing the PTU that the PRU has adjusted
 *        its power draw as instructed.
 *
 *        The PRU profile will automatically invoke this function to set the
 *        power adjusted value to 'false' after reading it as 'true'.
 *
 *        Note: If 'STATIC_POWER_ADJUSTED_BIT' is defined, the PRU profile
 *        will not reset the power adjusted value as described above.
 *
 * @param[in] adjusted New power adjusted value.
 *
 * @retval  NRF_SUCCESS the value was successfully set.
 */
uint32_t pru_hal_power_adjusted_set(bool adjusted);

/**@brief This function is called by the PRU profile to see if the PRU has limited its
 *        power consumption as instructed by previous call to 'pru_hal_adj_power_set()'.
 *
 *        When *p_adjusted is set to 'true', the PRU will set its 'Adjusted Power Response'
 *        bit in the 'PRU Dynamic' characteristic to '1' for 1 second, informing the PTU
 *        that the PRU has adjusted its power draw as instructed.
 *
 * @param[out] p_adjusted Set to 'true' when power draw has been adjusted as specified by
 *                        previous call to 'pru_hal_adj_power_set()'.
 *
 * @retval     NRF_SUCCESS  Value was returned successfully.
 */
uint32_t pru_hal_power_adjusted_get(bool * p_adjusted);


#endif /* __PRU_HAL_H__ */
/** @} */
