/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 */

/** @file
 *
 * @defgroup pru_sensors PRU sensors API
 * @{
 * @ingroup profile
 */

#ifndef __PRU_SENSORS_H__
#define __PRU_SENSORS_H__

#include "pru.h"


/**@brief Definition of PRU reporting data */
typedef struct
{
    uint16_t vrect;                     /**< Voltage at diode output [mV]. */
    uint16_t irect;                     /**< Current at diode output [mA]. */
    uint16_t vout;                      /**< Voltage at charge/battery port [mV]. */
    uint16_t iout;                      /**< Current at charge/battery port [mA]. */
    int16_t  temperature;               /**< Temperature of PRU [Deg C above -40C]. */
    uint8_t  over_voltage           :1; /**< Indicates that VRECT at the PRU has exceeded the OVP limit. */
    uint8_t  over_current           :1; /**< Indicates that IRECT at the PRU has exceeded the PRUs current limit. */
    uint8_t  over_temperature       :1; /**< Indicates that measured temperature at the PRU has exceeded the PRUs temperature limit. */
    uint8_t  sustained_high_voltage :1; /**< Indicates that VRECT has exceeded VRECT_HIGH for over 5 seconds. */
    uint8_t  wired_charge_detect    :1; /**< A wired charger has been detected. */
    uint8_t  charge_port            :1; /**< Charge port has been detected. */
    uint8_t  charge_complete        :1; /**< Charge has completed. */
    uint8_t power_adjusted          :1; /**< Power is adjusted according to setting in control packet. */
} pru_sensor_data_t;

/******************************************************************************/
/** @name General API functions
 *  @{ */
/******************************************************************************/


/**@brief Initialize sensors and read default values.
 * @param pru_sm_handler State machine event handler
 * @return      NRF_SUCCESS on success, otherwise an error code.
 */
void pru_sensors_init(pru_sm_handler_t pru_sm_handler);


/**@brief Read PRU sensors.
 */
void pru_sensors_read_all(void);

/**@brief Get the latest data from the PRU sensors.
 *
 * @returns Pointer to the report data.
 */
const pru_sensor_data_t * pru_sensors_data_get(void);


/**@brief Extract alerts from the latest data read from PRU sensors.
 *
 * @param[out] new_alert_occured returns true if an alert should be sent to PTU.
 * @param[out] p_pru_alert pointer to where to write alert data.

 * @returns NRF_SUCCESS on success, otherwise an error code.
 */
uint32_t pru_sensor_alerts_get(bool * new_alert_occured, pru_alert_t * p_pru_alert);


/**@brief Read PRU HW specific sensors.
 *
 * Function that can be customized for HW.
 *
 * @param[out] p_sensor_data pointer in which to return sensor data.
 *
 * @return      NRF_SUCCESS on success, otherwise an error code.
 */
uint32_t pru_sensors_board_read_all(pru_sensor_data_t * p_sensor_data);


/**@brief Generate PRU state machine events from sensor data.
 *
 */
void pru_sensors_gen_evts(void);

/** @} */

#endif /* __PRU_SENSORS_H__ */
/** @} */
