/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
 *
 * @defgroup pru_tmux PRU test multiplexer API
 * @{
 * @ingroup test_debug
 */

#ifndef __PRU_TEST_MUX_H__
#define __PRU_TEST_MUX_H__

#include <stdbool.h>
#include <stdint.h>
#include "test_mux_common.h"
#include "pru_hal.h"

/******************************************************************************
 * Types
 ******************************************************************************/

 /**
 * @struct pru_tmux_override_t
 * @brief Collection of PRU HAL variables that can be overrided by test multiplexer.
 */
 typedef struct
{
    bool     vout_enable;                 /**< Test multiplexer override value */
    bool     charge_complete;             /**< Test multiplexer override value */
    bool     wired_charge_detect;         /**< Test multiplexer override value */
    bool     power_adjusted;              /**< Test multiplexer override value */
    ctl_adj_power_t adj_setting;     /**< Test multiplexer override value */
    uint16_t vrect;                       /**< Test multiplexer override value */
    uint16_t irect;                       /**< Test multiplexer override value */
    uint16_t vout;                        /**< Test multiplexer override value */
    uint16_t iout;                        /**< Test multiplexer override value */
    uint16_t temperature;                 /**< Test multiplexer override value */
} pru_tmux_override_t;

/******************************************************************************
 * Profile API
 ******************************************************************************/

/**@brief Test wrapper for pru_hal_vrect_get().
 *
 * @param[out]  p_vrect return value.
 *
 * @retval      NRF_SUCCESS  Value was read successfully.
 */
uint32_t pru_tmux_vrect_get(uint16_t * p_vrect);


/**@brief Test wrapper for pru_hal_irect_get().
 *
 * @param[out]  p_irect return value.
 *
 * @retval      NRF_SUCCESS  Value was read successfully.
 */
uint32_t pru_tmux_irect_get(uint16_t * p_irect);


/**@brief Test wrapper for pru_hal_vout_enable_get().
 *
 * @param[out]  enable return value.
 *
 * @retval      NRF_SUCCESS  Value was read successfully.
 */
uint32_t pru_tmux_vout_enable_get(bool * enable);


/**@brief Test wrapper for pru_hal_vout_get().
 *
 * @param[out]  p_vout return value.
 *
 * @retval      NRF_SUCCESS  Value was read successfully.
 */
uint32_t pru_tmux_vout_get(uint16_t * p_vout);


/**@brief Test wrapper for pru_hal_iout_get().
 *
 * @param[out]  p_iout return value.
 *
 * @retval      NRF_SUCCESS  Value was read successfully.
 */
uint32_t pru_tmux_iout_get(uint16_t * p_iout);


/**@brief Test wrapper for pru_hal_temperature_get().
 *
 * @param[out]  p_temperature return value.
 *
 * @retval      NRF_SUCCESS  Value was read successfully.
 */
uint32_t pru_tmux_temperature_get(int16_t * p_temperature);


/**@brief Test wrapper for pru_hal_wired_charge_detect_status_get().
 *
 * @param[out]  p_wired return value.
 *
 * @retval      NRF_SUCCESS  Value was read successfully.
 */
uint32_t pru_tmux_wired_charge_detect_status_get(bool * p_wired);


/**@brief Test wrapper for pru_hal_charge_complete_status_get().
 *
 * @param[out]  p_charge_complete return value.
 *
 * @retval      NRF_SUCCESS  Value was read successfully.
 */
uint32_t pru_tmux_charge_complete_status_get(bool * p_charge_complete);


/**@brief Test wrapper for pru_hal_vout_enable_set().
 *
 * @param[in]   enable value to be set.
 *
 * @retval      NRF_SUCCESS  Enabled successfully.
 */
uint32_t pru_tmux_vout_enable_set(bool enable);


/**@brief Test wrapper for pru_hal_adj_power_set().
 *
 * @param[in]   adj_setting value to be set.
 *
 * @retval      NRF_SUCCESS  Value was set successfully.
 */
uint32_t pru_tmux_adj_power_set(ctl_adj_power_t adj_setting);


/**@brief Test wrapper for pru_hal_adj_power_get().
 *
 * @param[out]  adj_setting return value.
 *
 * @retval      NRF_SUCCESS  Value was read successfully.
 */
uint32_t pru_tmux_adj_power_get(ctl_adj_power_t * adj_setting);

/**@brief Test wrapper for pru_hal_power_adjusted_get().
 *
 * @param[out]  adjusted return value.
 *
 * @retval      NRF_SUCCESS  Value was read successfully..
 */
uint32_t pru_tmux_power_adjusted_get(bool * adjusted);

/**@brief Test wrapper for pru_hal_power_adjusted_set().
 *
 * @param[in]   adjusted value.
 *
 * @retval      NRF_SUCCESS    Success.
 */
uint32_t pru_tmux_power_adjusted_set(bool adjusted);


/******************************************************************************
 * Test API
 ******************************************************************************/

/**@brief Set override values.
 *
 * @param[in]  values values to be written.
 */
void pru_tmux_override_values_set(pru_tmux_override_t * values);


/**@brief Get ovveride values.
 *
 * @param[out]  values returned override values.
 */
void pru_tmux_override_values_get(pru_tmux_override_t * values);

#endif /* __PRU_TEST_MUX_H__ */
/** @} */
