/* Copyright (c) Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
 */

#ifndef PTU_DISTANT_LIST_HANDLER_H__
#define PTU_DISTANT_LIST_HANDLER_H__

#include <stdbool.h>
#include <stdint.h>

#include "ble.h"
#include "ble_gap.h"

/**< Initialization struct for the distant list handler. */
typedef struct {
    uint32_t ignore_latency_ms; /**< How many ms should pass from notification of beacon being disabled to new devices are added to the distant list? */
} ptu_dlh_init_t;

/**@brief Notify distant list handler that the beacon is now active.
 *
 * @details When this function is called, the distant list handler will not add new devices
 * discovered to the distant list.
 */
void ptu_dlh_notify_beacon_active(void);

/**@brief Notify distant list handler that the beacon is now disabled.
 *
 * @details When this function is called, the distant list handler will not add new devices
 * discovered to the distant list.
 */
void ptu_dlh_notify_beacon_disabled(void);

/**@brief Distance list handler WPT Advertisement report handler.
 *
 * @details Will check address of advertisement and depending on the current state of
 *  the beacon, it will either add it to the distant list, or not.
 *
 * @param p_addr Address of device advertising
 * @param adv_flags WPT Advertisement Flags found in advertisement
 */
void ptu_dlh_on_wpt_adv_report(ble_gap_addr_t const * p_addr, uint8_t adv_flags);

/**@brief Check if device with provided address is in the distant list.
 *
 * @details Will check address of advertisement and depending on the current state of
 *  the beacon, it will either add it to the distant list, or not.
 *
 * @param p_addr Address to check
 * @return true if given address is in distant list, false otherwise
 */
bool ptu_dlh_device_is_in_distant_list(ble_gap_addr_t const * p_addr);


/**@brief Check if device with provided address is in the distant list.
 *
 * @param p_init Initializing parameters
 *
 * @return NRF_SUCCESS if success, otherwise error code.
 */
uint32_t ptu_dlh_init(ptu_dlh_init_t const * p_init);


#endif //PTU_DISTANT_LIST_HANDLER_H__
