/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
 *
 * @defgroup ptu_hal PTU HAL API
 * @{
 * @ingroup hal
 */

#ifndef __PTU_HAL_H__
#define __PTU_HAL_H__

#include <stdbool.h>
#include <stdint.h>

/******************************************************************************
 * Main API
 ******************************************************************************/

/**@brief Initialize the PTU's Hardware abstraction layer.
 *
 * @retval NRF_SUCCESS         Initialization was successful.
 */
uint32_t ptu_hal_init(void);

/**@brief Enable power amplifier output (resonator power on).
 *
 * @param[in]  enable   Set to true enables power amplifier output.
 *
 * @retval      NRF_SUCCESS    Resonator enable was set successfully.
 */
uint32_t ptu_hal_poweramp_enable_set(bool enable);

/**@brief Get status of power amplifier output enable (resonator power on).
 *
 * @param[out]  enable   Returns status.
 *
 * @retval      NRF_SUCCESS    Resonator power was read successfully.
 */
uint32_t ptu_hal_poweramp_enable_get(bool * enable);

/**@brief Set output level of power amplifier (resonator power).
 *
 * @param[in]  level output level of power amplifier. Unit is given by implementation.
 *
 * @retval      NRF_SUCCESS    Resonator level was set successfully.
 */
uint32_t ptu_hal_poweramp_level_set(uint16_t level);

/**@brief Get output level setting of power amplifier.
 *
 * @param[out]  level   Output level set for power amplifier. Unit is given by implementation.
 *
 * @retval      NRF_SUCCESS    Resonator level setting was written to level.
 */
uint32_t ptu_hal_poweramp_level_get(uint16_t * level);

/**@brief Read the value of ITX.
 *
 * @param[out]  p_itx      Current being delivered to resonator.
 *
 * @retval      NRF_SUCCESS  ITX was read successfully.
 */
uint32_t ptu_hal_itx_get(uint16_t * p_itx);

/**@brief Read resonator voltage.
 *
 * @param[out]  p_vtx          Resonator voltage [mV].
 *
 * @retval      NRF_SUCCESS    Resonator power was read successfully.
 */
uint32_t ptu_hal_vtx_get(uint16_t * p_vtx);

/**@brief Read resonator voltage.
 *
 * @param[out]  p_ptx          Resonator voltage [mW].
 *
 * @retval      NRF_SUCCESS    Resonator power was read successfully.
 */
uint32_t ptu_hal_ptx_get(uint16_t * p_ptx);


/**@brief Read the board temperature.
 *
 * @param[out]  p_temperature   Temperature of PTU [Deg C].
 *
 * @retval      NRF_SUCCESS     Temperature was read successfully.
 */
uint32_t ptu_hal_temperature_get(int16_t * p_temperature);

#endif
/** @} */

