/* Copyright (c) Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
 *
 * @defgroup ptu_registry PTU registry API
 * @{
 * @ingroup profile
 */

#ifndef __PTU_REGISTRY_H__
#define __PTU_REGISTRY_H__

#include <stdbool.h>
#include <stdint.h>

#include "ble_wpts_c.h"
#include "ptu_config.h"
#include "app_timer.h"

/**@brief Registry item status */
typedef enum
{
    REG_ITEM_STATE_UNUSED = 0,                  /**< Item not used */
    REG_ITEM_STATE_PRE_CONNECT,                 /**< Adv. packet received, but only 1 of 2 criteria for connection fulfilled */
    REG_ITEM_STATE_FULLY_ACCEPTED,              /**< State of previously partially accepted device that is now fully accepted (11 ADV packets or 1700ms) */
    REG_ITEM_STATE_WAITING_TO_CONNECT,          /**< Item granted connection. Waiting for scan response after which BLE connection shall start */
    REG_ITEM_STATE_CONNECTING,                  /**< In process of establishing BLE connection */
    REG_ITEM_STATE_REGISTERING_EN_ALERT,        /**< In process of completing registration */
    REG_ITEM_STATE_REGISTERING_PRU_STAT_RD,     /**< PRU static read while completing registration, first connected state */
    REG_ITEM_STATE_REGISTERING_PTU_STAT_WR,     /**< PRU static write while completing registration */
    REG_ITEM_STATE_REGISTERING_PRU_DYN_RD,      /**< PRU dynamic read while completing registration */
    REG_ITEM_STATE_PRU_CTL_SEND,                /**< Send control packet completing registration */
    REG_ITEM_STATE_REGISTERED,                  /**< Registration completed */
} ptu_reg_item_state_t;

/**@brief Registry item */
typedef struct
{
    ptu_reg_item_state_t   state;                      /**< Item state */
    uint8_t                     adv_flags;                  /**< ADV flags */
    ble_gap_addr_t              address;                    /**< Device peer address */
    ble_wpts_c_t               ble_wpts_c;                /**< Collection of handles */
    pru_static_t           prev_pru_static;            /**< Previous received PRU static parameters */
    pru_dynamic_t          prev_pru_dynamic;           /**< Previous received PRU dynamic parameters */
    pru_control_t          prev_ctl;                   /**< Previous control packet being sent to PRU */
    app_timer_id_t              timer_id;                   /**< Registration / reconnection timer for this item */
    ctl_adj_power_t        prev_adj_power;             /**< Previous power adjustment level. */
    uint8_t                     age;                        /**< Number of devices items being added after this one */
    uint8_t                     pre_connect_cnt;            /**< Number of adv packets ignored due to only 1 of 2 criterie fullfilled.  */
    uint32_t                    p_adj_time;                 /**< Current TICKS value of RTC at time of sending power adjust command.  */
    int32_t                     p_rect_before_p_adj;        /**< PRECT in the PRU before we send the power adjust command. */
    uint8_t                     active_p_adj_cooldown:1;    /**< Does this PRU have an active cooldown from a previous power adjust command? */
    uint8_t                     adjusted_power:1;           /**< Has this PRU responded to the previous adjust power command? */
    uint8_t                     pending_dyn_read:1;         /**< Waiting for response on dyn param read */
    uint8_t                     pending_ctl_write:1;        /**< Waiting for response on control write */
    uint8_t                     charged:1;                  /**< True when PRU charged, false otherwise */
    uint8_t                     pending_charge_disable:1;   /**< We are waiting on a Charge disable command to be acknowledged. */
    uint8_t                     link_encrypted:1;           /**< Link is encrypted. */
    uint8_t                     p_adj_disabled:1;           /**< No more power adjust commands should be sent to PRU. */
    uint8_t                     reconnect:1;                /**< PTU is reconnecting to this device. */
} ptu_reg_item_t;

/**@brief Registry */
typedef struct
{
    uint8_t n_entries;                                     /**< Total number of entries in registry. (Registered device + devices in process og being registered */
    ptu_reg_item_t items[PTU_MAX_CONNECTIONS];   /**< Items in registry. */
} ptu_reg_t;

/******************************************************************************/
/** @name General API functions
 *  @{ */
/******************************************************************************/

/**@brief Initialize registry database.
 */
void ptu_reg_init(void);

/**@brief Get handles for all registered devices.
 *
 * @param[out]      registered_devices      Placeholder to write the registry item pointers of all registered devices to.
 * @return Number of registered devices
 */
uint8_t ptu_reg_registered_devices_get(ptu_reg_item_t** registered_devices);

/**@brief Get the number of devices currently in registry.
 *        This will include all connected devices, as well as all devices which is currently being registered.
 *
 * @return pointer to the registry variable holding the number of connected devices.
 */
uint8_t ptu_reg_n_entries_get(void);

/**@brief Get registry item from index in database. Index must be < PTU_MAX_CONNECTIONS.
 *
 * @param[in] index
 * @param[out] item_p pointer to where pointer to registry item will be written.
 * @return error code if index is not < PTU_MAX_CONNECTIONS
 */
uint32_t ptu_reg_item_get_from_index(uint8_t index, ptu_reg_item_t ** item_p);

/**@brief Get registry item from connection handle.
 *
 * @param[in] conn_handle
 * @return pointer to registry item. NULL if item not found.
 */
ptu_reg_item_t * ptu_reg_item_get_from_conn_handle(uint16_t conn_handle);

/**@brief Get registry item from GAP address.
 *
 * @param[in] address
 * @return pointer to registry item. NULL if item not found.
 */
ptu_reg_item_t  * ptu_reg_item_get_from_address(ble_gap_addr_t * address);

/**@brief Get oldest registry item having a specific state.
 *
 * @param[in] state state of registry item. Note: REG_ITEM_STATE_UNUSED is not a legal value.
 * @return pointer to registry item. NULL if item not found.
 */
ptu_reg_item_t * ptu_reg_item_oldest_get(ptu_reg_item_state_t state);

/**@brief Initiate registration of new device.
 *
 * @param p_adv_report pointer to adv report from device from which to register.
 * @param init_state initial state for the added registry item.
 * @return pointer to registry item. NULL if no item available.
 */
ptu_reg_item_t  * ptu_reg_item_add(ble_gap_evt_adv_report_t * p_adv_report, ptu_reg_item_state_t init_state);

/**@brief Set the state of registry item to "ready to connect" state.
 *          Associates the address with the registry item.
 *          Should only be called with non-zero address values.
 * @param reg_item_p The registry item to update.
 * @param address The non-zero BLE address.
 * @return NRF_SUCCESS if successful.
 */
uint32_t ptu_reg_set_device_waiting_to_connect(ptu_reg_item_t * reg_item_p, uint8_t * address);

/**@brief Delete item that has previoulsy been added to registry.
 * @param item is the index in registry of the item to be deleted.
 */
void ptu_reg_item_delete(ptu_reg_item_t * item);

/**@brief Function returning true if all items in registry are charged or there are not items in registry.
 * @return true if all items in registry are charged.
 */
bool ptu_reg_all_charged(void);

/** @} */

#endif /* __PTU_H__ */
/** @} */
