/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

 /** @file
 *
 * @defgroup ptu_sensors PTU sensors API
 * @{
 * @ingroup profile
 */

#ifndef PTU_SENSORS_H__
#define PTU_SENSORS_H__

#include "ptu.h"
#include "ptu_hw_config.h"

#include <stdbool.h>
#include <stdint.h>

/******************************************************************************/
/** @name API.
 *  @{ */
/******************************************************************************/

/**@brief Initialize PTU Sensors. This function must be called before any other PTU Sensors
 * function can be called.
 *
 * @param sm_handler State machine event handler
 *
 * @return NRF_ERROR_INVALID_PARAM if sm_handler is NULL, otherwise NRF_SUCCESS.
 */
uint32_t ptu_sensors_init(ptu_sm_handler_t sm_handler);

/**@brief Read sensors. This function should typically be called regularly by a timer and can
 * generate events by calling ptu_evt_handler().
 *
 * @retval  NRF_SUCCESS     Sensor readings updated successfully.
 */
uint32_t ptu_sensors_read(void);

/**@brief Get the latest data read from the PTU sensors.
 *
 * @param[out] sensors_data is a pointer to where a pointer to the latest read sensor data shall be written.
 *
 * @returns NRF_SUCCESS if sensor data was successully read.
 */
uint32_t ptu_sensors_data_get(const ptu_sensor_data_t ** sensors_data);


/**@brief Empty buffer containing load variation sample data for Long Beacon Extension.
 */
void ptu_sensors_clear_long_beacon_extension_load_variation_buffer(void);

/**@brief Check if a valid PTX load variation has occured.
 *
 *  Once called, this function will store a touple containing the current ticks value, and the current PTX value to a list of samples.
 *  This list of samples is then checked to see whether or not a valid load variation can be found.
 *  The overall algorithm is: Find the first top, find the bottom, find the second top, is the time between the two tops OK?
 *
 *  Finding the first top: This is done by simply comparing the value at the a point with the next sample value, if the next value is smaller
 *  we have a valid top.
 *
 *  Finding the bottom: We jump to the next sample as long as the difference between the value at the current sample and the first top is
 *  NOT sufficient (smaller than 0.5W). Note that when we have found the bottom, we will still check whether or not a LOWER bottom exists,
 *  in which case the bottom-pointer will be updated.
 *
 *  Finding the second top: The second top does not need to be a top, just a point that is sufficiently higher than bottom (0.5W above).
 *  NOTE: This implementation will accept a load where the second top is lower than the first top, as long as its delta from the bottom is
 *  sufficient.
 *
 *  Checking the time is a simple check of the time delta between the first top and the second top. If its accepted, the function will
 *  report that a valid PTX load variation has been found.
 *
 * @returns true if valid PTX load variation has been seen.
 */
bool ptu_sensors_valid_long_beacon_extension_load_variation_found(void);

/** @} */

#endif /* PTU_SENSORS_H__ */
/** @} */
