/* Copyright (c) Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
 *
 * @defgroup pru_tmux_common PTU / PRU common test multiplexer API
 * @{
 * @ingroup test_debug
 */

#ifndef __TEST_MUX_COMMON_H__
#define __TEST_MUX_COMMON_H__

#include <stdint.h>


/******************************************************************************
 * Definitions
 ******************************************************************************/

/**@brief Macro for multiplexing **hal**set() functions.
 *
 * @param[in]   hal_fun is the HAL function to be written.
 * @param[in]   arg is the argument of the HAL function to be written.
 */
#define TMUX_SET(hal_fun, arg) do                              \
{                                                                   \
    if(set_mux_select == TMUX_SELECT_NO_OVERRIDE)              \
    {                                                               \
        return hal_fun(arg);                                        \
    }                                                               \
    else                                                            \
    {                                                               \
        ASSERT(set_mux_select == TMUX_SELECT_OVERRIDE)         \
        return NRF_SUCCESS;                                         \
    }                                                               \
} while(0)

/**@brief Macro for multiplexing **hal**get() functions.
 *
 * @param[in]   hal_fun is the HAL function to be read.
 * @param[out]  arg is the variable where the read value shall be
 *              written.
 * @param[in]   override_val is the corresponding override value
 *              for the HAL function.
 */
#define TMUX_GET(hal_fun, arg, override_val) do                \
{                                                                   \
    if(get_mux_select == TMUX_SELECT_NO_OVERRIDE)              \
    {                                                               \
        return hal_fun(arg);                                        \
    }                                                               \
    else                                                            \
    {                                                               \
        ASSERT(get_mux_select == TMUX_SELECT_OVERRIDE)         \
        *arg = override_val;                                        \
        return NRF_SUCCESS;                                         \
    }                                                               \
} while(0)

/******************************************************************************
 * Types
 ******************************************************************************/

/**
 * @enum tmux_select_t
 * @brief Enumerator used for configuring test multiplexer.
 */
 typedef enum
 {
     TMUX_SELECT_NO_OVERRIDE = 0,
     TMUX_SELECT_OVERRIDE,
 } tmux_select_t;

/******************************************************************************
 * Profile API
 ******************************************************************************/

 /**@brief Initialize the PTU's test multiplexer.
 *
 * @retval NRF_SUCCESS Initialization succeeded.
 */
uint32_t tmux_init(void);

/******************************************************************************
 * Test API
 ******************************************************************************/

/**@brief Select source for all **tmux**get() functions.
 *
 * @param[in]  select selects the source for values returned by **tmux**get() functions.
 * @arg TMUX_SELECT_NO_OVERRIDE **tmux**get() functions will return values from **hal**get() functions.
 * @arg TMUX_SELECT_OVERRIDE **tmux**get() functions will return values set by **tmux**override_set() functions.
 */
void tmux_get_select(tmux_select_t select);

/**@brief Select whether override values shall be used when writing to **hal**set() functions.
 *
 * @param[in]  select selects the source for values written to HAL.
 * @arg TMUX_SELECT_NO_OVERRIDE values given in **tmux**set() functions will be written to **hal**set() functions.
 * @arg TMUX_SELECT_OVERRIDE values given in **tmux**override_set() functions will be written to **hal**set() functions.
 */
void tmux_set_select(tmux_select_t select);

#endif /* __TMUX_H__ */
/** @} */
