/* Copyright (c) Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
 *
 * @defgroup common PTU / PRU common profile API
 * @{
 * @ingroup profile
 */

#ifndef H__
#define H__

#include <stdint.h>
#include <stdbool.h>
#include "ble_gap.h"
#include "ble_wpts_common.h"

// SDK version

#define SDK_VERSION "2.1.0"                                                /**< SDK version. */

#define TIMER_START(timer_id, ms, p_ctx) \
        app_timer_start( timer_id, APP_TIMER_TICKS(ms, APP_TIMER_PRESCALER), p_ctx); /**< Start app_timer with correct prescaler. */
/******************************************************************************/
/** @name WPT Service characteristics structs.
 *  @{ */
/******************************************************************************/

/**@brief Power adjust settings in control packet.
 */
typedef enum
{
    CTL_ADJ_POWER_MAX_VAL = BLE_WPTS_ENABLE_PRU_ADJ_POWER_MAX_VAL,       /**< Allow maximium PRU power consumption */
    CTL_ADJ_POWER_66_PCT_VAL = BLE_WPTS_ENABLE_PRU_ADJ_POWER_66_PCT_VAL, /**< Allow PRU to consume 66% of Prect_max */
    CTL_ADJ_POWER_33_PCT_VAL = BLE_WPTS_ENABLE_PRU_ADJ_POWER_33_PCT_VAL, /**< Allow PRU to consume 33% of Prect_max */
    CTL_ADJ_POWER_2_5_W_VAL = BLE_WPTS_ENABLE_PRU_ADJ_POWER_2_5_W_VAL    /**< Allow PRU to consume up to 2.5 W */
} ctl_adj_power_t;

/**@brief Permissions fields in control packet.
 */
typedef enum
{
    CTL_PERMISSION_PERMITTED = BLE_WPTS_PERMISSION_PERMITTED,                   /**< Permitted without reason */
    CTL_PERMISSION_PERMITTED_WAITING = BLE_WPTS_PERMISSION_PERMITTED_WAITING,   /**< Permitted with waiting time due to limited available power */
    CTL_PERMISSION_DENIED_CROSS_CONNECTION = BLE_WPTS_PERMISSION_DENIED_CC,      /**< Denied due to cross connection */
    CTL_PERMISSION_DENIED_POWER = BLE_WPTS_PERMISSION_DENIED_POWER,             /**< Denied due to limited available power */
    CTL_PERMISSION_DENIED_DEVICES = BLE_WPTS_PERMISSION_DENIED_DEVICES,         /**< Denied due to limited PTU Number of Devices */
    CTL_PERMISSION_DENIED_CLASS = BLE_WPTS_PERMISSION_DENIED_CLASS,             /**< Denied due to limited PTU Class support */
    CTL_PERMISSION_DENIED_HIGH_TEMP = BLE_WPTS_PERMISSION_DENIED_HIGH_TEMP,      /**< Denied due to high temperature at PTU */
} ctl_perm_t;

/**@brief Time set field in control packet.
 */
typedef enum
{
    CTL_TIME_SET_DENIED = BLE_WPTS_TIME_SET_DENIED,   /**< Do not perform Time Set */
    CTL_TIME_SET_10MS = BLE_WPTS_TIME_SET_10MS,       /**< 10ms */
    CTL_TIME_SET_20MS = BLE_WPTS_TIME_SET_20MS,       /**< 20ms */
    CTL_TIME_SET_30MS = BLE_WPTS_TIME_SET_30MS,       /**< 30ms */
    CTL_TIME_SET_40MS = BLE_WPTS_TIME_SET_40MS,       /**< 40ms */
    CTL_TIME_SET_50MS = BLE_WPTS_TIME_SET_50MS,       /**< 50ms */
    CTL_TIME_SET_60MS = BLE_WPTS_TIME_SET_60MS,       /**< 60ms */
    CTL_TIME_SET_70MS = BLE_WPTS_TIME_SET_70MS,       /**< 70ms */
    CTL_TIME_SET_80MS = BLE_WPTS_TIME_SET_80MS        /**< 80ms */
} ctl_time_set_t;

/**@brief PRU Control structure. */
typedef struct
{
    uint8_t               enable_pru_output:1;            /**< PRU can enable charge output. */
    uint8_t               enable_pru_charge_indicator:1;  /**< PTU charge indicator enabled. */
    ctl_adj_power_t  adj_power;                      /**< Adjust power command. */
    ctl_perm_t       permissions;                    /**< Power permitted/denied reason code. */
    ctl_time_set_t   time_set;                       /**< PTU sets up time. */
} pru_control_t;

/**@brief PRU Static Parameter structure. */
typedef struct
{
    uint8_t     protocol_rev;               /**< Revision of the protocol. */
    uint8_t     pru_category;               /**< Category of PRU. */
    uint8_t     hw_revision;                /**< Revision of the PRU HW. */
    uint8_t     fw_revision;                /**< Revision of the PRU SW. */
    uint8_t     prect_max;                  /**< PRECT_MAX of the PRU [mW * 100]. */
    uint16_t    vrect_min_static;           /**< VRECT_MIN (static, first estimate). */
    uint16_t    vrect_high_static;          /**< VRECT_HIGH (static, first estimate). */
    uint16_t    vrect_set;                  /**< VRECT_SET. */
    uint16_t    delta_r1;                   /**< Delta R1 caused by PRU. */
    uint8_t     option_fields_r1_valid:1;   /**< Delta R1 field is valid. */
    uint8_t     info_nfc_receiver:1;        /**< NFC receiver supported. */
    uint8_t     info_sep_btle_radio:1;      /**< Separate BTLE radio in PRU. */
    uint8_t     info_pc_alg_pref:1;         /**< Power control algorithm preference. */
    uint8_t     info_adj_power_capability:1;/**< Adjust power capability. */
    uint8_t     info_ccp_connected_mode:1;  /**< Charge complete connected mode supported. */
    uint8_t     info_ptu_test_mode:1;       /**< PTU test mode enabled. */
} pru_static_t;

/**@brief PTU Static Parameter structure. */
typedef struct
{
    uint8_t     option_valid_max_imp:1;     /**< PTU max source impedance field is valid. */
    uint8_t     option_valid_max_res:1;     /**< PTU max load resistance field is valid. */
    uint8_t     ptu_power;                  /**< Power of PTU. */
    uint8_t     ptu_max_source_impedance;   /**< Maximum source impedance of the PTU. */
    uint8_t     ptu_max_load_resistance;    /**< Maximum load resistance of the PTU. */
    uint8_t     ptu_class;                  /**< PTU class  Class 1-5. */
    uint8_t     hardware_rev;               /**< Revision of the PTU HW. */
    uint8_t     firmware_rev;               /**< Revision of the PTU SW. */
    uint8_t     protocol_rev;               /**< Revision of the protocol. */
    uint8_t     devices_supported_bitval;   /**< Max devices bitcode. */
} ptu_static_t;

/**@brief Definition of PRU alert type */
typedef enum
{
    ALERT_TYPE_NOTIFICATION = 0,       /**< Send alert as notification. */
    ALERT_TYPE_INDICATION   = 1,       /**< Send alert as indication. */
} pru_alert_type_t;

/**@brief Alert field used in both Alert Characteristic and Dynamic Parameters Characteristic. */
typedef struct
{
    uint8_t     pru_over_voltage      :1;   /**< Indicates that VRECT at the PRU has exceeded the OVP limit. */
    uint8_t     pru_over_current      :1;   /**< Indicates that IRECT at the PRU has exceeded the PRUs current limit. */
    uint8_t     pru_over_temperature  :1;   /**< Indicates that measured temperature at the PRU has exceeded the PRUs temperature limit. */
    uint8_t     pru_self_protection   :1;   /**< Indicates that the PRU is protecting itself by reducing power to its load.*/
    uint8_t     charge_complete       :1;   /**< Indicates that the PRU does not require charging.*/
    uint8_t     wired_charge_detect   :1;   /**< Indicates that the PRU is powered by external wired power. */
} pru_alert_bits_t;



/**@brief PRU Alert Parameter structure. */
typedef struct
{
    pru_alert_bits_t alerts;           /**< Alerts bitfields */
    uint8_t     mode_transition       :2;   /**< The Mode Transition bits shall be set to a non-zero value to indicate to the PTU the duration of the pend-4 ing Mode Transition procedure. */
    uint8_t     device_address[BLE_GAP_ADDR_LEN];          /**< The Device address is only valid when the Mode transition bits are set to a non-zero value */
    pru_alert_type_t           type;   /**< '1' indicates 'Indication', 0 indicates 'Notification'. */
    uint8_t     pru_charge_port       :1;   /**< PRU Dynamic characteristic only. Indicate that the PRU charge port output is activated. */
} pru_alert_t;

/**@brief PTU Tester commands. */
typedef enum
{
    WPTS_PTU_TEST_CMD_NORMAL_OPERATION_POWER_CTL    = 0, /**< No adjustment requested */
    WPTS_PTU_TEST_CMD_INCREASE_ITX                  = 1, /**< Increase ITX one step size */
    WPTS_PTU_TEST_CMD_DECREASE_ITX                  = 2, /**< Decrease ITX one step size */
    WPTS_PTU_TEST_CMD_NO_ADJ_ITX                    = 3, /**< Disable ITX adjustment by any other way than test commands. */
}  ptu_tester_command_t;

/**@brief PRU Dynamic Parameter structure. */
typedef struct
{
    uint8_t             optional_fields;            /**< Defines which optional fields are populated. */
    uint16_t            vrect;                      /**< Voltage at diode output. [mV]*/
    uint16_t            irect;                      /**< Current at diode output. [mA]*/
    uint16_t            vout;                       /**< Voltage at charge/battery port. [mV]*/
    uint16_t            iout;                       /**< Current at charge/battery port. [mA]*/
    int16_t             temperature;                /**< Temperature of PRU. [Deg C]*/
    uint16_t            vrect_min_dyn;              /**< VRECT_LOW_LIMIT (dynamic value). */
    uint16_t            vrect_set_dyn;              /**< Desired VRECT (dynamic value). */
    uint16_t            vrect_high_dyn;             /**< VRECT_HIGH_LIMIT (dynamic value). */
    pru_alert_bits_t alerts;                   /**< Alerts bitfields */
    uint8_t             pru_charge_port:1;          /**< PRU charge port activated */
    uint8_t             adjust_power_response :1;   /**< PRU has limited it's power draw in response to Adjust Power Command. */
    uint16_t            prect;                      /**< Vrect power. Derived value: vrect * irect. [mW * 100] */
    uint8_t             putil;                      /**< Utilization of rated output [(prect / prectmax) */
    ptu_tester_command_t tester_command;       /**< PTU tester command */
} pru_dynamic_t;

/** @} */

/******************************************************************************/
/** @name Common functions and macros.
 *  @{ */
/******************************************************************************/

/**@brief Power manager.
 */
void power_manage(void);

/**@brief Function to put system in SYSTEMOFF mode.
 */
void system_off(void);

/**
 * Calculate difference between two ticks values. Takes care of overflow.
 * @param ticks_now The most recent number of ticks.
 * @param ticks_old Oldest sample number of ticks
 *
 * @return Difference in ticks with respect to overflow.
 */
uint32_t ticks_diff(uint32_t ticks_now, uint32_t ticks_old);

#ifdef DFU_SUPPORT
/**@brief Starts DFU mode.
 *        The function writes a special value to retention register to tell bootloader to start DFU mode, disables SoftDevice and performs a reset.
 */
void dfu_start(void);

/**@brief Needs to be implemented by application. Check if it is OK to jump to the DFU application.
 * @return true if it is OK to jump to DFU application. False otherwise.
 */
bool dfu_check(void);

/**@brief Needs to be implemented by application. Perform all requried actions before jump to DFU application can be done.
 */
void terminate(void);
#endif //DFU_SUPPORT

/**@brief Macro equivalent to APP_ERROR_CHECK, except that error codes
 * returned if connection is lost will be allowed.
 *
 * @param[in] ERR_CODE Error code supplied to the error handler.
 */
#define APP_ERROR_CHECK_ALLOW_DISCONNECT(ERR_CODE)          \
    do                                                      \
    {                                                       \
        if ((ERR_CODE != BLE_ERROR_INVALID_CONN_HANDLE) &&  \
            (ERR_CODE != NRF_ERROR_INVALID_STATE) &&        \
            (ERR_CODE != NRF_ERROR_BUSY))                   \
        {                                                   \
            APP_ERROR_CHECK(ERR_CODE);                      \
        }                                                   \
    } while (0)

/** @} */

#endif /* H__ */
/** @} */

