/* packet-nordic_ble.c
 * Routines for nordic ble sniffer dissection
 * Copyright 201x, YOUR_NAME <YOUR_EMAIL_ADDRESS>
 * Nordic Semiconductor
 *
 *
 * $Id$
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "config.h"

#include <glib.h>

#include <epan/packet.h>
#include <epan/prefs.h>
#include <epan/expert.h>
#include <wiretap/pcap-encap.h>
#include <epan/wmem/wmem.h>

/* definition to expand macro then apply to pragma message */
#define VALUE_TO_STRING(x) #x
#define VALUE(x) VALUE_TO_STRING(x)
#define VAR_NAME_VALUE(var) #var "="  VALUE(var)

#pragma message(VAR_NAME_VALUE(VERSION_MAJOR))
// #pragma message(VERSION_MAJOR)
#pragma message(VAR_NAME_VALUE(VERSION_MINOR))
// #pragma message(VERSION_MINOR)

#define IS_VERSION_1_10 (VERSION_MAJOR == 1 && VERSION_MINOR == 10)
#define IS_VERSION_1_11 (VERSION_MAJOR == 1 && VERSION_MINOR == 11)

#define BLE_CID_INDEX               (8)
#define BLE_CID_ATTRIBUTE_PROTOCOL  (4)

// Spec V3 part F 3.3.4
#define BLE_BTATT_OPCODE_INDEX  (10)
#define BLE_BTATT_VALUE_INDEX   (11)
#define BLE_BTATT_READ_REQUEST  (0xa)
#define BLE_BTATT_READ_RESPONSE (0xb)
#define BLE_BTATT_HANDLE_NOTIFY (0x1b)
#define BLE_BTATT_WRITE_COMMAND (0x52)

#define BLE_HEADER_INDEX    (4)
#define BLE_UUID_INDEX (17)
#define BLE_PRIMARY_SERVICE_HANDLE (19)
#define BLE_HEADER_ADV_IND  (0x0)

#define WPT_OPCODE_INDEX                (0)
#define WPT_VALUE_INDEX	    		    (1)
#define WPT_READ_RESPONSE_VALUE_INDEX   (1)
#define WPT_READ_REQUEST_HANDLE_INDEX   (1)
#define WPT_READ_REQUEST_VALUE_INDEX    (3)

#define WPT_PRU_CONTROL_VAL                 (2)
#define WPT_PTU_STATIC_VAL                  (4)
#define WPT_PRU_ALERT                       (6)
#define WPT_CLIENT_CHAR_CONF_UUID_PRU_ALERT (7)
#define WPT_PRU_STATIC_VAL                  (9)
#define WPT_PRU_DYN_VAL                     (11)

#define WPT_UUID                            (0xfffe)


/* Forward declaration that is needed below if using the
 * proto_reg_handoff_btle_wpt function as a callback for when protocol
 * preferences get changed. */
void proto_reg_handoff_btle_wpt(void);
void proto_register_btle_wpt(void);

/* Initialize the protocol and registered fields */
static int proto_btle_wpt = -1;
static int proto_wpt = -1;
static int proto_wpt_enables = -1;
static int proto_wpt_optional_fields_validity = -1;
static int proto_wpt_pru_information = -1;
static int proto_wpt_pru_alert = -1;
static int proto_wpt_pru_alert_notification = -1;

/* Initialize the subtree pointers */
static gint ett_btle_wpt = -1;
static gint ett_wpt = -1;
static gint ett_wpt_optional_fields_validity = -1;
static gint ett_wpt_enables = -1;
static gint ett_wpt_pru_information = -1;
static gint ett_wpt_pru_alert = -1;
static gint ett_wpt_pru_alert_notification = -1;
static gint ett_flags = -1;


static int hf_nordic_ble_wpt_field_vout = -1;
static int hf_nordic_ble_wpt_field_iout = -1;
static int hf_nordic_ble_wpt_field_temp = -1;
static int hf_nordic_ble_wpt_field_vrect_min_dyn = -1;
static int hf_nordic_ble_wpt_field_vrect_set_dyn = -1;
static int hf_nordic_ble_wpt_field_vrect_high_dyn = -1;
static int hf_nordic_ble_wpt_vout = -1;
static int hf_nordic_ble_wpt_iout = -1;
static int hf_nordic_ble_wpt_vrect = -1;
static int hf_nordic_ble_wpt_irect = -1;
static int hf_nordic_ble_wpt_temp = -1;
static int hf_nordic_ble_wpt_vrect_md = -1;
static int hf_nordic_ble_wpt_vrect_sd = -1;
static int hf_nordic_ble_wpt_vrect_hd = -1;
static int hf_nordic_ble_wpt_pru_alert = -1;
static int hf_nordic_ble_wpt_handles = -1;
static int hf_nordic_ble_wpt_field_max_impetance = -1;
static int hf_nordic_ble_wpt_field_max_resistance = -1;
static int hf_nordic_ble_wpt_protocol_rev = -1;
static int hf_nordic_ble_wpt_pru_category = -1;
static int hf_nordic_ble_wpt_hardware_rev = -1;
static int hf_nordic_ble_wpt_firmware_rev = -1;
static int hf_nordic_ble_wpt_p_rect_max = -1;
static int hf_nordic_ble_wpt_v_rect_min_static = -1;
static int hf_nordic_ble_wpt_v_rect_high_static = -1;
static int hf_nordic_ble_wpt_v_rect_set = -1;
static int hf_nordic_ble_wpt_delta_r1 = -1;
static int hf_nordic_ble_wpt_field_delta_r1 = -1;
static int hf_nordic_ble_wpt_ptu_max_source_impetance = -1;
static int hf_nordic_ble_wpt_ptu_max_load_resistance = -1;
static int hf_nordic_ble_wpt_ptu_class = -1;
static int hf_nordic_ble_wpt_ptu_nof_devs_supported = -1;
static int hf_nordic_ble_wpt_ptu_power = -1;
static int hf_nordic_ble_wpt_enable_pru_output = -1;
static int hf_nordic_ble_wpt_enable_pru_charge_indicator = -1;
static int hf_nordic_ble_wpt_adjust_power_command = -1;
static int hf_nordic_ble_wpt_permissions = -1;
static int hf_nordic_ble_wpt_time_set = -1;
static int hf_nordic_ble_wpt_rfu = -1;
static int hf_nordic_ble_wpt_pru_information_nfc_receiver = -1;
static int hf_nordic_ble_wpt_pru_information_separate_btle_radio = -1;
static int hf_nordic_ble_wpt_pru_information_power_algorithm_pref = -1;
static int hf_nordic_ble_wpt_pru_information_adjust_power_cap = -1;
static int hf_nordic_ble_wpt_pru_information_charge_complete_connected_mode = -1;
static int hf_nordic_ble_wpt_pru_alert_over_voltage = -1;
static int hf_nordic_ble_wpt_pru_alert_over_temp = -1;
static int hf_nordic_ble_wpt_pru_alert_over_current = -1;
static int hf_nordic_ble_wpt_pru_alert_pru_self_protection = -1;
static int hf_nordic_ble_wpt_pru_alert_charge_complete = -1;
static int hf_nordic_ble_wpt_pru_alert_wired_charger_detect = -1;
static int hf_nordic_ble_wpt_pru_alert_pru_change_port = -1;
static int hf_nordic_ble_wpt_pru_alert_adjust_power_response = -1;
static int hf_nordic_ble_wpt_pru_alert_notify_over_voltage = -1;
static int hf_nordic_ble_wpt_pru_alert_notify_over_temp = -1;
static int hf_nordic_ble_wpt_pru_alert_notify_over_current = -1;
static int hf_nordic_ble_wpt_pru_alert_notify_pru_self_protection = -1;
static int hf_nordic_ble_wpt_pru_alert_notify_charge_complete = -1;
static int hf_nordic_ble_wpt_pru_alert_notify_wired_charger_detect = -1;
static int hf_nordic_ble_wpt_pru_alert_notify_mode_transition_bit_1 = -1;
static int hf_nordic_ble_wpt_pru_alert_notify_mode_transition_bit_2 = -1;
static int hf_nordic_ble_wpt_pru_alert_notify_device_address = -1;
static int hf_nordic_ble_wpt_pru_dynamic_tester_command = -1;
static int hf_nordic_ble_wpt_pru_information_ptu_test_mode = -1;

static guint16 gatt_primary_service_handle = 0;

static void
reverse_byte_order(guint8* dest, const guint8* src, guint size)
{
    guint i;
    for (i = 0; i < size; i++)
    {
        dest[i] = src[size-1-i];
    }
}

static const true_false_string encrypted_tfs =
{
    "Yes",
    "No"
};

static const true_false_string enable_disable_tfs =
{
    "Enable",
    "Disable"
};

static guint32 adv_aa = 0x8e89bed6;

static value_string wpt_handles[] = {
    { 0, "Start of WPT Service" },
    { 1, "PRU Control Characteristic declaration" },
    { 2, "PRU Control Characteristic value" },
    { 3, "PTU Static Parameter Characteristic declaration" },
    { 4, "PTU Static Parameter Characteristic value" },
    { 5, "PRU Alert Parameter Characteristic declaration" },
    { 6, "PRU Alert Parameter Characteristic value" },
    { 7, "Client Characteristic Configuration UUID for PRU Alert" },
    { 8, "PRU Static Parameter Characteristic declaration" },
    { 9, "PRU Static Parameter Characteristic value" },
    { 10, "PRU Dynamic Parameter Characteristic declaration" },
    { 11, "PRU Dynamic Parameter Characteristic value" },
    { 0, NULL }
};

static const value_string wpt_protocol_rev_tfs[] = {
    { 0, "BSS v1.2.1" },
    { 1, "BSS v1.3" },
    { 0, NULL }
};
static const value_string wpt_pru_category_tfs[] = {
    { 0, "Undefined" },
    { 1, "Category 1" },
    { 2, "Category 2" },
    { 3, "Category 3" },
    { 4, "Category 4" },
    { 5, "Category 5" },
    { 0, NULL }
};

static const value_string wpt_ptu_max_source_impetance_tfs[] = {
    { 0, "50" },
    { 1, "60" },
    { 2, "70" },
    { 3, "80" },
    { 4, "90" },
    { 5, "100" },
    { 6, "110" },
    { 7, "120" },
    { 8, "130" },
    { 9, "140" },
    { 10, "150" },
    { 11, "175" },
    { 12, "200" },
    { 13, "225" },
    { 14, "250" },
    { 15, "275" },
    { 16, "300" },
    { 17, "350" },
    { 18, "375" },
    { 0, NULL }
};

static const value_string wpt_ptu_max_load_resistance_tfs[] = {
    { 0, "5" },
    { 1, "10" },
    { 2, "15" },
    { 3, "20" },
    { 4, "25" },
    { 5, "30" },
    { 6, "35" },
    { 7, "40" },
    { 8, "45" },
    { 9, "50" },
    { 10, "55" },
    { 0, NULL }
};

static const value_string wpt_ptu_tester_command_tfs[] = {
    { 0, "Normal operation power control" },
    { 1, "Increase I_TX by one step"},
    { 2, "Decrease I_TX by one step"},
    { 3, "I_TX_COIL shall not be adjusted"},
    { 0, NULL}
};

static const value_string wpt_ptu_class_tfs[] = {
    { 0, "Class 1" },
    { 1, "Class 2" },
    { 2, "Class 3" },
    { 3, "Class 4" },
    { 4, "Class 5" },
    { 0, NULL }
};

static const true_false_string wpt_pru_information_power_alorithm_preferance =
{
    "Max System Efficiency",
    "V_RECT_MIN_ERROR"
};

static const true_false_string supported_tfs =
{
    "Supported",
    "Not Supported"
};

static const true_false_string oz_tfs =
{
    "1",
    "0"
};

static const value_string wpt_adjust_power_command_tfs[] = {
    { 0, "Maximum Power" },
    { 1, "66% * P_RECT_MAX" },
    { 2, "33% * P_RECT_MAX" },
    { 3, "2.5W" },
    { 0, NULL }
};

static const value_string wpt_permissions_tfs[] = {
    { 0x0,	"Permitted without reason" },
    { 0x1, "Permitted with waiting time due to limited available power" },
    { 0x80, "Denied due to cross connection" },
    { 0x81, "Denied due to limited available power" },
    { 0x82, "Denied due to limited PTU Number of Devices" },
    { 0x83, "Denied due to limited PTU Class support" },
    { 0x84, "Denied due to high temperature at PTU" },
    { 0, NULL }
};

static const value_string wpt_time_set_tfs[] = {
    { 0x0, "RFU" },
    { 0x01, "10ms" },
    { 0x02, "20ms" },
    { 0x03, "30ms" },
    { 0x04, "40ms" },
    { 0x05, "50ms" },
    { 0x06, "60ms" },
    { 0x07, "70ms" },
    { 0x08, "80ms" },
    { 0, NULL }
};


static void
dissect_optional_fields_pru_dynamic(tvbuff_t *tvb, proto_tree *tree, guint8 offset)
{
    proto_item *ti;
    proto_tree *optional_tree = NULL;
    ti = proto_tree_add_item(tree, proto_wpt_optional_fields_validity, tvb, offset, 1, ENC_NA);
    optional_tree = proto_item_add_subtree(ti, ett_wpt_optional_fields_validity);
    proto_tree_add_bits_item(optional_tree, hf_nordic_ble_wpt_field_vout, tvb, (offset * 8) + 0, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(optional_tree, hf_nordic_ble_wpt_field_iout, tvb, (offset * 8) + 1, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(optional_tree, hf_nordic_ble_wpt_field_temp, tvb, (offset * 8) + 2, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(optional_tree, hf_nordic_ble_wpt_field_vrect_min_dyn, tvb, (offset * 8) + 3, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(optional_tree, hf_nordic_ble_wpt_field_vrect_set_dyn, tvb, (offset * 8) + 4, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(optional_tree, hf_nordic_ble_wpt_field_vrect_high_dyn, tvb,(offset * 8) + 5, 1, ENC_LITTLE_ENDIAN);
}

static void
dissect_wpt_pru_dynamic_pru_alert(tvbuff_t *tvb, proto_tree *tree, guint8 offset)
{
    proto_item *ti;
    proto_tree *pru_alert_tree = NULL;
    ti = proto_tree_add_item(tree, proto_wpt_pru_alert, tvb, offset, 1, ENC_NA);
    pru_alert_tree = proto_item_add_subtree(ti, ett_wpt_pru_alert);

    proto_tree_add_bits_item(pru_alert_tree, hf_nordic_ble_wpt_pru_alert_over_voltage , tvb, (offset * 8) + 0, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(pru_alert_tree, hf_nordic_ble_wpt_pru_alert_over_current , tvb, (offset * 8) + 1, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(pru_alert_tree, hf_nordic_ble_wpt_pru_alert_over_temp , tvb, (offset * 8) + 2, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(pru_alert_tree, hf_nordic_ble_wpt_pru_alert_pru_self_protection , tvb, (offset * 8) + 3, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(pru_alert_tree, hf_nordic_ble_wpt_pru_alert_charge_complete , tvb, (offset * 8) + 4, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(pru_alert_tree, hf_nordic_ble_wpt_pru_alert_wired_charger_detect , tvb, (offset * 8) + 5, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(pru_alert_tree, hf_nordic_ble_wpt_pru_alert_pru_change_port , tvb, (offset * 8) + 6, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(pru_alert_tree, hf_nordic_ble_wpt_pru_alert_adjust_power_response , tvb, (offset * 8) + 7, 1, ENC_LITTLE_ENDIAN);
}

static void
dissect_wpt_pru_alert_notification(tvbuff_t *tvb, proto_tree *tree)
{
    guint16 handle = tvb_get_letohs(tvb, WPT_READ_REQUEST_HANDLE_INDEX) - gatt_primary_service_handle;
    guint8 offset = 3; /* Opcode, Handle2, Handle1, VALUE0 */
    guint8 *adv_addr_le, *adv_addr_be;
    wmem_allocator_t* epan_scope_mem_pool;
    proto_item *ti;
    proto_tree *pru_alert_tree = NULL;

    epan_scope_mem_pool = wmem_epan_scope();
    adv_addr_le = wmem_alloc(epan_scope_mem_pool, 6);
    adv_addr_be = wmem_alloc(epan_scope_mem_pool, 6);

    proto_tree_add_uint(tree, hf_nordic_ble_wpt_handles, tvb, WPT_READ_REQUEST_HANDLE_INDEX, 2, handle);
    ti = proto_tree_add_item(tree, proto_wpt_pru_alert_notification, tvb, offset, 1, ENC_NA);
    pru_alert_tree = proto_item_add_subtree(ti, ett_wpt_pru_alert_notification);

    proto_tree_add_bits_item(pru_alert_tree, hf_nordic_ble_wpt_pru_alert_notify_over_voltage , tvb, (offset * 8) + 0, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(pru_alert_tree, hf_nordic_ble_wpt_pru_alert_notify_over_current , tvb, (offset * 8) + 1, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(pru_alert_tree, hf_nordic_ble_wpt_pru_alert_notify_over_temp , tvb, (offset * 8) + 2, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(pru_alert_tree, hf_nordic_ble_wpt_pru_alert_notify_pru_self_protection , tvb, (offset * 8) + 3, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(pru_alert_tree, hf_nordic_ble_wpt_pru_alert_notify_charge_complete , tvb, (offset * 8) + 4, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(pru_alert_tree, hf_nordic_ble_wpt_pru_alert_notify_wired_charger_detect , tvb, (offset * 8) + 5, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(pru_alert_tree, hf_nordic_ble_wpt_pru_alert_notify_mode_transition_bit_1 , tvb, (offset * 8) + 6, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(pru_alert_tree, hf_nordic_ble_wpt_pru_alert_notify_mode_transition_bit_2 , tvb, (offset * 8) + 7, 1, ENC_LITTLE_ENDIAN);
    offset++;

    if(tvb_length(tvb) > 4) // More than just opcode, handle1, handle2, notify_alert
    {
        tvb_memcpy(tvb, adv_addr_le, offset, 6);
        reverse_byte_order(adv_addr_be, adv_addr_le, 6);
        proto_tree_add_ether(tree, hf_nordic_ble_wpt_pru_alert_notify_device_address, tvb, offset, 6, adv_addr_be);
    }
}

static void
dissect_wpt_pru_dynamic_params(tvbuff_t *tvb, proto_tree *tree, guint8 offset)
{
    dissect_optional_fields_pru_dynamic(tvb,tree, offset++);
    proto_tree_add_item(tree, hf_nordic_ble_wpt_vrect, tvb, offset, 2, ENC_LITTLE_ENDIAN); offset+=2;
    proto_tree_add_item(tree, hf_nordic_ble_wpt_irect, tvb, offset, 2, ENC_LITTLE_ENDIAN); offset+=2;
    proto_tree_add_item(tree, hf_nordic_ble_wpt_vout, tvb, offset, 2, ENC_LITTLE_ENDIAN); offset+=2;
    proto_tree_add_item(tree, hf_nordic_ble_wpt_iout, tvb, offset, 2, ENC_LITTLE_ENDIAN); offset+=2;
    proto_tree_add_int(tree, hf_nordic_ble_wpt_temp, tvb, offset, 1, tvb_get_guint8(tvb, offset) - 40); offset++;
    proto_tree_add_item(tree, hf_nordic_ble_wpt_vrect_md, tvb, offset, 2, ENC_LITTLE_ENDIAN); offset+=2;
    proto_tree_add_item(tree, hf_nordic_ble_wpt_vrect_sd, tvb, offset, 2, ENC_LITTLE_ENDIAN); offset+=2;
    proto_tree_add_item(tree, hf_nordic_ble_wpt_vrect_hd, tvb, offset, 2, ENC_LITTLE_ENDIAN); offset+=2;
    dissect_wpt_pru_dynamic_pru_alert(tvb, tree, offset++);
    proto_tree_add_item(tree, hf_nordic_ble_wpt_pru_dynamic_tester_command, tvb, offset++, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_item(tree, hf_nordic_ble_wpt_rfu, tvb, offset , 2, ENC_LITTLE_ENDIAN);

}

static void
dissect_optional_fields_ptu_static(tvbuff_t *tvb, proto_tree *tree, guint8 offset)
{
    proto_item *ti;
    proto_tree *optional_tree = NULL;
    ti = proto_tree_add_item(tree, proto_wpt_optional_fields_validity, tvb, offset, 1, ENC_NA);
    optional_tree = proto_item_add_subtree(ti, ett_wpt_optional_fields_validity);
    proto_tree_add_bits_item(optional_tree, hf_nordic_ble_wpt_field_max_impetance, tvb, (offset * 8) + 0, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(optional_tree, hf_nordic_ble_wpt_field_max_resistance, tvb, (offset * 8) + 1, 1, ENC_LITTLE_ENDIAN);
}

static void
dissect_pru_information(tvbuff_t *tvb, proto_tree *tree, guint8 offset)
{
    proto_item *ti;
    proto_tree *pru_information_tree = NULL;
    ti = proto_tree_add_item(tree, proto_wpt_pru_information, tvb, offset, 1, ENC_NA);
    pru_information_tree = proto_item_add_subtree(ti, ett_wpt_pru_information);
    proto_tree_add_bits_item(pru_information_tree, hf_nordic_ble_wpt_pru_information_nfc_receiver, tvb, (offset * 8) + 0, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(pru_information_tree, hf_nordic_ble_wpt_pru_information_separate_btle_radio, tvb, (offset * 8) + 1, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(pru_information_tree, hf_nordic_ble_wpt_pru_information_power_algorithm_pref, tvb, (offset * 8) + 2, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(pru_information_tree, hf_nordic_ble_wpt_pru_information_adjust_power_cap, tvb, (offset * 8) + 3, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(pru_information_tree, hf_nordic_ble_wpt_pru_information_charge_complete_connected_mode, tvb, (offset * 8) + 4, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(pru_information_tree, hf_nordic_ble_wpt_pru_information_ptu_test_mode, tvb, (offset * 8) + 5, 1, ENC_LITTLE_ENDIAN);

}

static void
dissect_optional_fields_pru_static(tvbuff_t *tvb, proto_tree *tree, guint8 offset)
{
    proto_item *ti;
    proto_tree *optional_tree = NULL;
    ti = proto_tree_add_item(tree, proto_wpt_optional_fields_validity, tvb, offset, 1, ENC_NA);
    optional_tree = proto_item_add_subtree(ti, ett_wpt_optional_fields_validity);
    proto_tree_add_bits_item(optional_tree, hf_nordic_ble_wpt_field_delta_r1, tvb, (offset * 8) + 0, 1, ENC_LITTLE_ENDIAN);
}

static void
dissect_ptu_power(tvbuff_t *tvb, proto_tree *tree, guint8 index)
{
    gfloat pwr = 0;
    guint8 val = tvb_get_guint8(tvb, index);
    guint8 i = 0;
    /* A4WP spec (v1.3) 6.5.4.4-1 */
    for(i = 0; i < val && i < 120; i++)
        pwr += (gfloat)(0.1*((i/20)+1));
    for(i = 120; i < val && i < 139; i++)
        pwr += 1;

    proto_tree_add_float(tree, hf_nordic_ble_wpt_ptu_power, tvb, index, 1, pwr);

}


static void
dissect_wpt_ptu_static_params(tvbuff_t *tvb, proto_tree *tree, guint8 offset)
{
    dissect_optional_fields_ptu_static(tvb,tree,offset++);
    dissect_ptu_power(tvb,tree,offset++);
    proto_tree_add_uint(tree, hf_nordic_ble_wpt_ptu_max_source_impetance, tvb, offset, 1, (tvb_get_guint8(tvb, offset++)>>3));
    proto_tree_add_uint(tree, hf_nordic_ble_wpt_ptu_max_load_resistance, tvb, offset, 1, (tvb_get_guint8(tvb, offset++)>>3));
    proto_tree_add_item(tree, hf_nordic_ble_wpt_rfu, tvb, offset , 2, ENC_LITTLE_ENDIAN);
    offset += 2;
    proto_tree_add_item(tree, hf_nordic_ble_wpt_ptu_class, tvb, offset++ , 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_item(tree, hf_nordic_ble_wpt_hardware_rev, tvb, offset++ , 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_item(tree, hf_nordic_ble_wpt_firmware_rev, tvb, offset++ , 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_item(tree, hf_nordic_ble_wpt_protocol_rev, tvb, offset++ , 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_item(tree, hf_nordic_ble_wpt_ptu_nof_devs_supported, tvb, offset++ , 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_item(tree, hf_nordic_ble_wpt_rfu, tvb, offset , 6, ENC_LITTLE_ENDIAN);

}

static void
dissect_wpt_pru_static_params(tvbuff_t *tvb, proto_tree *tree, guint8 offset)
{
    dissect_optional_fields_pru_static(tvb,tree, offset++);
    proto_tree_add_item(tree, hf_nordic_ble_wpt_protocol_rev, tvb, offset++ , 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_item(tree, hf_nordic_ble_wpt_rfu, tvb, offset++ , 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_item(tree, hf_nordic_ble_wpt_pru_category, tvb, offset++ , 1, ENC_LITTLE_ENDIAN);
    dissect_pru_information(tvb, tree, offset++);
    proto_tree_add_item(tree, hf_nordic_ble_wpt_hardware_rev, tvb, offset++ , 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_item(tree, hf_nordic_ble_wpt_firmware_rev, tvb, offset++ , 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_uint(tree, hf_nordic_ble_wpt_p_rect_max, tvb, offset , 1, (tvb_get_guint8(tvb, offset)*100)); offset++;
    proto_tree_add_item(tree, hf_nordic_ble_wpt_v_rect_min_static, tvb, offset , 2, ENC_LITTLE_ENDIAN); 	offset += 2;
    proto_tree_add_item(tree, hf_nordic_ble_wpt_v_rect_high_static, tvb, offset , 2, ENC_LITTLE_ENDIAN); 	offset += 2;
    proto_tree_add_item(tree, hf_nordic_ble_wpt_v_rect_set, tvb, offset , 2, ENC_LITTLE_ENDIAN); 			offset += 2;
    proto_tree_add_float(tree, hf_nordic_ble_wpt_delta_r1, tvb, offset , 2, ((gfloat)(tvb_get_letohs(tvb, offset)/100))); 				offset += 2;
    proto_tree_add_item(tree, hf_nordic_ble_wpt_rfu, tvb, offset , 4, ENC_LITTLE_ENDIAN);
}

static void
dissect_wpt_pru_control(tvbuff_t *tvb, proto_tree *tree, guint8 offset)
{
    proto_item *ti;
    proto_tree *enables_tree = NULL;
    ti = proto_tree_add_item(tree, proto_wpt_enables, tvb, offset, 1, ENC_NA);
    enables_tree = proto_item_add_subtree(ti, ett_wpt_enables);
    proto_tree_add_bits_item(enables_tree, hf_nordic_ble_wpt_enable_pru_output, tvb, (offset * 8) + 0, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(enables_tree, hf_nordic_ble_wpt_enable_pru_charge_indicator, tvb, (offset * 8) + 1, 1, ENC_LITTLE_ENDIAN);
    proto_tree_add_bits_item(enables_tree, hf_nordic_ble_wpt_adjust_power_command, tvb, (offset * 8) + 2, 2, ENC_LITTLE_ENDIAN);
    offset++;
    proto_tree_add_item(tree, hf_nordic_ble_wpt_permissions, tvb, offset++ , 1, ENC_NA);
    proto_tree_add_item(tree, hf_nordic_ble_wpt_time_set, tvb, offset++ , 1, ENC_NA);
    proto_tree_add_item(tree, hf_nordic_ble_wpt_rfu, tvb, offset , 2, ENC_LITTLE_ENDIAN);
}

static void
dissect_wpt_read_response(tvbuff_t *tvb, proto_tree *tree, guint16 last_handle)
{
    guint8 offset = WPT_READ_RESPONSE_VALUE_INDEX;
    if      (last_handle == WPT_PRU_DYN_VAL)        dissect_wpt_pru_dynamic_params(tvb, tree, offset);
    else if (last_handle == WPT_PRU_STATIC_VAL)     dissect_wpt_pru_static_params(tvb, tree, offset);
    else if (last_handle == WPT_PTU_STATIC_VAL)     dissect_wpt_ptu_static_params(tvb, tree, offset);
    else if (last_handle == WPT_PRU_CONTROL_VAL)    dissect_wpt_pru_control(tvb, tree, offset);
}

/* Returns the handle value */
static guint16
dissect_wpt_read_request(tvbuff_t *tvb, proto_tree *tree)
{
    guint16 handle = 0;
    /* As the service handle may change, use the statically defined offset (Table 6.5.2.2-1) */
    handle = tvb_get_letohs(tvb, WPT_READ_REQUEST_HANDLE_INDEX) - gatt_primary_service_handle;
    proto_tree_add_uint(tree, hf_nordic_ble_wpt_handles, tvb, WPT_READ_REQUEST_HANDLE_INDEX, 2, handle);

    return handle;
}

static void
dissect_wpt_write_command(tvbuff_t *tvb, proto_tree *tree)
{
    guint16 handle = 0;
    guint8 offset = WPT_READ_REQUEST_VALUE_INDEX; /* Write Request has value after 2 byte handle */

    handle = tvb_get_letohs(tvb, WPT_READ_REQUEST_HANDLE_INDEX) - gatt_primary_service_handle;
    proto_tree_add_uint(tree, hf_nordic_ble_wpt_handles, tvb, WPT_READ_REQUEST_HANDLE_INDEX, 2, handle);

    if (handle == WPT_PTU_STATIC_VAL) dissect_wpt_ptu_static_params(tvb, tree, offset);
    else if(handle == WPT_PRU_CONTROL_VAL) dissect_wpt_pru_control(tvb, tree, offset);
}

static guint16 last_handle = 0;

static void
dissect_wpt(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
    guint8 opcode;
    opcode = tvb_get_guint8(tvb, WPT_OPCODE_INDEX);

    if( opcode == BLE_BTATT_READ_REQUEST)
        last_handle = dissect_wpt_read_request(tvb, tree);

    else if(opcode == BLE_BTATT_READ_RESPONSE)
        dissect_wpt_read_response(tvb, tree, last_handle);

    else if(opcode == BLE_BTATT_WRITE_COMMAND)
        dissect_wpt_write_command(tvb, tree);

    else if(opcode == BLE_BTATT_HANDLE_NOTIFY)    // Only PRU Alert which uses this opcode
        dissect_wpt_pru_alert_notification(tvb, tree);
}

/* Entry point. Entire BTLE packet (including access address and crc */
static void
dissect_btle_wpt(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree) //, void *data _U_)
{
    /* Set up structures needed to add the protocol subtree and manage it */
    proto_item *ti;
    proto_tree *wpt_tree;
    static gboolean last_adv_packet_was_wpt = FALSE;

    /* Packet is ADV_IND */
    if((tvb_get_guint8(tvb, BLE_HEADER_INDEX) & 0x0F ) == BLE_HEADER_ADV_IND)

        if((tvb_get_letohs(tvb, BLE_UUID_INDEX) == WPT_UUID))
        {
            last_adv_packet_was_wpt = TRUE;
            gatt_primary_service_handle = tvb_get_letohs(tvb, BLE_PRIMARY_SERVICE_HANDLE);
        }

        else
            last_adv_packet_was_wpt = FALSE;

    if (last_adv_packet_was_wpt && tvb_get_letohl(tvb, 0) != adv_aa  && tvb_get_guint8(tvb, BLE_CID_INDEX) == BLE_CID_ATTRIBUTE_PROTOCOL)
    {
        ti = proto_tree_add_item(tree, proto_btle_wpt, tvb, 0, -1, ENC_NA);
        wpt_tree = proto_item_add_subtree(ti, ett_btle_wpt);
        /* Send entire BTATT payload to wpt */
        dissect_wpt(tvb_new_subset(tvb, BLE_BTATT_OPCODE_INDEX, tvb_length(tvb)- BLE_BTATT_OPCODE_INDEX - 3, tvb_length(tvb)- BLE_BTATT_OPCODE_INDEX - 3), pinfo, wpt_tree);
    }
}

/* Register the protocol with Wireshark.
 *
 * This format is require because a script is used to build the C function that
 * calls all the protocol registration.
 */
void
proto_register_btle_wpt(void)
{
    static hf_register_info hf[] = {
                 /* Begin WPT */
        { &hf_nordic_ble_wpt_vrect,
           { "V_RECT", "nordic_ble.vrect",
           FT_UINT16, BASE_DEC, NULL, 0x0,
          "DC voltage at the output of the rectifier (mV)", HFILL }
        },
        { &hf_nordic_ble_wpt_irect,
            { "I_RECT", "nordic_ble.irect",
               FT_UINT16, BASE_DEC, NULL, 0x0,
              "DC current at the output of the rectifier (mA)", HFILL }
        },
        { &hf_nordic_ble_wpt_field_max_impetance,
            { "Max Impetance", "nordic_ble.wpt_max_impetance",
               FT_BOOLEAN, BASE_NONE, TFS(&encrypted_tfs), 0x0,
              "Is this field valid", HFILL }
        },
        { &hf_nordic_ble_wpt_field_max_resistance,
            { "Max Resistance", "nordic_ble.wpt_max_resistance",
               FT_BOOLEAN, BASE_NONE, TFS(&encrypted_tfs), 0x0,
              "Is this field valid", HFILL }
        },
        { &hf_nordic_ble_wpt_field_vout,
            { "vout", "nordic_ble.vout_val",
               FT_BOOLEAN, BASE_NONE, TFS(&encrypted_tfs), 0x0,
              "Is this field valid", HFILL }
        },
        { &hf_nordic_ble_wpt_field_delta_r1,
            { "Delta R1", "nordic_ble.wpt_valid_delta_r1",
               FT_BOOLEAN, BASE_NONE, TFS(&encrypted_tfs), 0x0,
              "Is this field valid", HFILL }
        },
        { &hf_nordic_ble_wpt_protocol_rev,
            { "Protocol Revision", "nordic_ble.wpt_protocol_rev",
               FT_UINT8, BASE_DEC, VALS(&wpt_protocol_rev_tfs), 0x0,
              "A4WP Supported Revision", HFILL }
        },
        { &hf_nordic_ble_wpt_pru_category,
            { "PRU Category", "nordic_ble.wpt_pru_category",
               FT_UINT8, BASE_DEC, VALS(&wpt_pru_category_tfs), 0x0,
              "Category of PRU", HFILL }
        },
        { &hf_nordic_ble_wpt_hardware_rev,
            { "Hardware Revision", "nordic_ble.wpt_hardware_rev",
               FT_UINT8, BASE_HEX, NULL, 0x0,
              "Revision of the PRU HW", HFILL }
        },
        { &hf_nordic_ble_wpt_firmware_rev,
            { "Firmware Revision", "nordic_ble.wpt_firmware_rev",
               FT_UINT8, BASE_HEX, NULL, 0x0,
              "Revision of the PRU SW", HFILL }
        },
        { &hf_nordic_ble_wpt_p_rect_max,
            { "P_RECT_MAX (mW)", "nordic_ble.wpt_p_rect_max",
               FT_UINT8, BASE_DEC, NULL, 0x0,
              "P_RECT_MAX of the PRU", HFILL }
        },
        { &hf_nordic_ble_wpt_v_rect_min_static,
            { "V_RECT_MIN_STATIC (mV)", "nordic_ble.wpt_v_rect_min_static",
               FT_UINT16, BASE_DEC, NULL, 0x0,
              "V_RECT_MIN, static, first estimate (mV)", HFILL }
        },
        { &hf_nordic_ble_wpt_v_rect_high_static,
            { "V_RECT_HIGH_STATIC (mV)", "nordic_ble.wpt_v_rect_high_static",
               FT_UINT16, BASE_DEC, NULL, 0x0,
              "V_RECT_HIGH_STATIC, static, first estimate", HFILL }
        },
        { &hf_nordic_ble_wpt_v_rect_set,
            { "V_RECT_SET (mV)", "nordic_ble.wpt_v_rect_set",
               FT_UINT16, BASE_DEC, NULL, 0x0,
              "V_RECT_SET, static, first estimate", HFILL }
        },
        { &hf_nordic_ble_wpt_delta_r1,
            { "Delta R1 Value (ohms)", "nordic_ble.wpt_delta_r1",
               FT_FLOAT, BASE_NONE, NULL, 0x0,
              "Delta R1 caused by PRU", HFILL }
        },
        { &hf_nordic_ble_wpt_field_iout,
            { "iout", "nordic_ble.iout_val",
               FT_BOOLEAN, BASE_NONE, TFS(&encrypted_tfs), 0x0,
              "Is this field valid", HFILL }
        },
        { &hf_nordic_ble_wpt_field_temp,
            { "temp", "nordic_ble.temp_val",
               FT_BOOLEAN, BASE_NONE, TFS(&encrypted_tfs), 0x0,
              "Is this field valid", HFILL }
        },
        { &hf_nordic_ble_wpt_field_vrect_min_dyn,
            { "vrect_min_dyn", "nordic_ble.vrect_md_val",
               FT_BOOLEAN, BASE_NONE, TFS(&encrypted_tfs), 0x0,
              "Is this field valid", HFILL }
        },
        { &hf_nordic_ble_wpt_field_vrect_set_dyn,
            { "vrect_set_dyn", "nordic_ble.vrect_sd_val",
               FT_BOOLEAN, BASE_NONE, TFS(&encrypted_tfs), 0x0,
              "Is this field valid", HFILL }
        },
        { &hf_nordic_ble_wpt_field_vrect_high_dyn,
            { "vrect_high_dyn", "nordic_ble.vrect_hd_val",
               FT_BOOLEAN, BASE_NONE, TFS(&encrypted_tfs), 0x0,
              "Is this field valid", HFILL }
        },
        { &hf_nordic_ble_wpt_handles,
            { "Handle", "nordic_ble.wpt_handle",
            FT_UINT8, BASE_HEX, VALS(wpt_handles), 0x0,
            NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_vout,
            { "V_OUT", "nordic_ble.vout",
               FT_UINT16, BASE_DEC, NULL, 0x0,
              "Voltage at charge/battery port (mV)", HFILL }
        },
        { &hf_nordic_ble_wpt_iout,
            { "I_OUT", "nordic_ble.iout",
               FT_UINT16, BASE_DEC, NULL, 0x0,
              "Current at charge/battery port (mA).", HFILL }
        },
        { &hf_nordic_ble_wpt_temp,
            { "temperature", "nordic_ble.temp",
               FT_INT8, BASE_DEC, NULL, 0x0,
              "Temperature of PRU", HFILL }
        },
        { &hf_nordic_ble_wpt_vrect_md,
            { "v_rect_min_dynamic", "nordic_ble.vrect_md",
               FT_UINT16, BASE_DEC, NULL, 0x0,
              "The current dynamic minimum rectifier voltage desired (mV)", HFILL }
        },
        { &hf_nordic_ble_wpt_vrect_sd,
            { "v_rect_set_dynamic", "nordic_ble.vrect_sd",
               FT_UINT16, BASE_DEC, NULL, 0x0,
              "Desired V_RECT (dynamic value) (mV)", HFILL }
        },
        { &hf_nordic_ble_wpt_vrect_hd,
            { "v_rect_high_dynamic", "nordic_ble.vrect_hd",
               FT_UINT16, BASE_DEC, NULL, 0x0,
              "The current dynamic maximum rectifier voltage desired (mV)", HFILL }
        },
        { &hf_nordic_ble_wpt_pru_alert,
            { "PRU Alert", "nordic_ble.pru",
               FT_UINT8, BASE_DEC, NULL, 0x0,
              "Warnings", HFILL }
        },
        { &hf_nordic_ble_wpt_ptu_power,
            { "PTU Power (Watts)", "nordic_ble.wpt_ptu_power",
               FT_FLOAT, BASE_NONE, NULL, 0x0,
              "Power of PTU", HFILL }
        },
        { &hf_nordic_ble_wpt_ptu_max_source_impetance,
            { "PTU Max Source Impedance (ohms)", "nordic_ble.wpt_ptu_max_source_impetance",
               FT_UINT8, BASE_DEC, VALS(wpt_ptu_max_source_impetance_tfs), 0x0,
              "Maximum source impetance of the PTU", HFILL }
        },
        { &hf_nordic_ble_wpt_ptu_max_load_resistance,
            { "PTU Max Load Resistance (ohms)", "nordic_ble.wpt_max_source_resiscance",
               FT_UINT8, BASE_DEC, VALS(wpt_ptu_max_load_resistance_tfs), 0x0,
              "Maximum load resistance of the PTU", HFILL }
        },
        { &hf_nordic_ble_wpt_ptu_class,
            { "PTU Class", "nordic_ble.wpt_ptu_class",
               FT_UINT8, BASE_DEC, VALS(wpt_ptu_class_tfs), 0x0,
              "PTU Class", HFILL }
        },
        { &hf_nordic_ble_wpt_ptu_nof_devs_supported,
            { "PTU Number of Devices Supported", "nordic_ble.wpt_nof_devs_supported",
               FT_UINT8, BASE_DEC, NULL, 0x0,
              "Max Number of Devices", HFILL }
        },
        { &hf_nordic_ble_wpt_enable_pru_output,
            { "Enable PRU output", "nordic_ble.wpt_enable_pru_output",
               FT_BOOLEAN, BASE_NONE, TFS(&enable_disable_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_enable_pru_charge_indicator,
            { "Enable PRU charge indicator", "nordic_ble.wpt_enable_pru_charge_indicator",
               FT_BOOLEAN, BASE_NONE, TFS(&enable_disable_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_adjust_power_command,
            { "Adjust Power Command", "nordic_ble.wpt_adjust_power_command",
               FT_UINT8, BASE_HEX, VALS(wpt_adjust_power_command_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_permissions,
            { "Permissions", "nordic_ble.wpt_permissions",
               FT_UINT8, BASE_HEX, VALS(&wpt_permissions_tfs), 0x0,
              "PRU is permitted in PTU", HFILL }
        },
        { &hf_nordic_ble_wpt_time_set,
            { "Time Set", "nordic_ble.wpt_time_set",
               FT_UINT8, BASE_HEX, VALS(&wpt_time_set_tfs), 0x0,
              "PTU Sets up time", HFILL }
        },
        { &hf_nordic_ble_wpt_rfu,
            { "RFU", "nordic_ble.wpt_rfu",
               FT_STRING , BASE_NONE, NULL, 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_information_nfc_receiver,
            { "NFC Receiver", "nordic_ble.wpt_pru_info_nfc",
               FT_BOOLEAN , BASE_NONE, TFS(&supported_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_information_separate_btle_radio,
            { "Separate BTLE Radio in PRU", "nordic_ble.wpt_pru_information_separate_btle_radio",
               FT_BOOLEAN , BASE_NONE, TFS(&supported_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_information_power_algorithm_pref,
            { "Power Control Algorithm Preference", "nordic_ble.wpt_pru_info_pwr_ctrl_alg_pref",
               FT_BOOLEAN , BASE_NONE, TFS(&wpt_pru_information_power_alorithm_preferance), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_information_adjust_power_cap,
            { "Adjust power capability", "nordic_ble.wpt_pru_information_adjust_power",
               FT_BOOLEAN , BASE_NONE, TFS(&supported_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_information_charge_complete_connected_mode,
            { "Charge Complete Connected Mode", "nordic_ble.wpt_pru_information_charge_complete",
               FT_BOOLEAN , BASE_NONE, TFS(&supported_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_information_ptu_test_mode,
            { "PTU Test Mode", "nordic_ble.wpt_pru_information_ptu_test_mode",
               FT_BOOLEAN , BASE_NONE, TFS(&encrypted_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_alert_over_voltage,
            { "Over-voltage", "nordic_ble.wpt_pru_alert_over_voltage",
               FT_BOOLEAN , BASE_NONE, TFS(&oz_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_alert_over_current,
            { "Over-current", "nordic_ble.wpt_pru_alert_over_current",
               FT_BOOLEAN , BASE_NONE, TFS(&oz_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_alert_over_temp,
            { "Over-temp", "nordic_ble.wpt_pru_alert_over_temp",
               FT_BOOLEAN , BASE_NONE, TFS(&oz_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_alert_pru_self_protection,
            { "PRU Self Protection", "nordic_ble.wpt_pru_alert_self_protection",
               FT_BOOLEAN , BASE_NONE, TFS(&oz_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_alert_charge_complete,
            { "Charge Complete", "nordic_ble.wpt_pru_alert_charge_complete",
               FT_BOOLEAN , BASE_NONE, TFS(&oz_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_alert_wired_charger_detect,
            { "Wired Charger Detect", "nordic_ble.wpt_pru_alert_wired_charger_detect",
               FT_BOOLEAN , BASE_NONE, TFS(&oz_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_alert_pru_change_port,
            { "PRU Charge Port", "nordic_ble.wpt_pru_alert_pru_change_port",
               FT_BOOLEAN , BASE_NONE, TFS(&oz_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_alert_adjust_power_response,
            { "Adjust Power Response", "nordic_ble.wpt_pru_alert_adjust_power_response",
               FT_BOOLEAN , BASE_NONE, TFS(&oz_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_alert_notify_over_voltage,
            { "PRU Over-voltage", "nordic_ble.wpt_pru_alert_notify_over_voltage",
               FT_BOOLEAN , BASE_NONE, TFS(&oz_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_alert_notify_over_current,
            { "PRU Over-current", "nordic_ble.wpt_pru_alert_notify_over_current",
               FT_BOOLEAN , BASE_NONE, TFS(&oz_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_alert_notify_over_temp,
            { "PRU Over-temp", "nordic_ble.wpt_pru_alert_notify_over_temp",
               FT_BOOLEAN , BASE_NONE, TFS(&oz_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_alert_notify_pru_self_protection,
            { "PRU Self Protection", "nordic_ble.wpt_pru_alert_notify_self_protection",
               FT_BOOLEAN , BASE_NONE, TFS(&oz_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_alert_notify_charge_complete,
            { "Charge Complete", "nordic_ble.wpt_pru_alert_notify_charge_complete",
               FT_BOOLEAN , BASE_NONE, TFS(&oz_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_alert_notify_wired_charger_detect,
            { "Wired Charger Detect", "nordic_ble.wpt_pru_alert_notify_wired_charger_detect",
               FT_BOOLEAN , BASE_NONE, TFS(&oz_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_alert_notify_mode_transition_bit_1,
            { "Mode Transition Bit 1", "nordic_ble.wpt_pru_alert_notify_mode_transition_bit1",
               FT_BOOLEAN , BASE_NONE, TFS(&oz_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_alert_notify_mode_transition_bit_2,
            { "Mode Transition Bit 2", "nordic_ble.wpt_pru_alert_notify_mode_transition_bit2",
               FT_BOOLEAN , BASE_NONE, TFS(&oz_tfs), 0x0,
              NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_alert_notify_device_address,
            { "Device Address (Optional)", "nordic_ble.wpt_pru_alert_notify_device_address",
            FT_ETHER, BASE_NONE, NULL, 0x0,
            NULL, HFILL }
        },
        { &hf_nordic_ble_wpt_pru_dynamic_tester_command,
            { "Tester Command", "nordic_ble.tester_command",
               FT_UINT8, BASE_DEC,  VALS(wpt_ptu_tester_command_tfs), 0x0,
              "PTU Tester Command", HFILL }
        },

    };

    static gint *ett[] = {
        &ett_btle_wpt,
        &ett_wpt,
        &ett_wpt_optional_fields_validity,
        &ett_wpt_enables,
        &ett_wpt_pru_information,
            &ett_wpt_pru_alert,
        &ett_wpt_pru_alert_notification
    };

    /* Register the protocol name and description */
    proto_btle_wpt = proto_register_protocol("BTLE A4WP WPT","btle_wpt", "btle_wpt");
    proto_wpt_enables = proto_register_protocol("Enables", "wpt_enables", "wpt_enables");
    proto_wpt_optional_fields_validity = proto_register_protocol("Optional Fields Validity", "wpt_optfields", "wpt_optfields");
    proto_wpt_pru_information = proto_register_protocol("PRU Information", "pru_information", "pru_information");
    proto_wpt_pru_alert = proto_register_protocol("PRU Alert", "pru_alert", "pru_alert");
    proto_wpt_pru_alert_notification = proto_register_protocol("PRU Alert Notification", "pru_alert_notification", "pru_alert_notification");
    register_dissector("nordic_debug", dissect_btle_wpt, proto_btle_wpt);

    proto_register_field_array(proto_btle_wpt, hf, array_length(hf));
    proto_register_subtree_array(ett, array_length(ett));
}

void
proto_reg_handoff_btle_wpt(void)
{
    /* We are called explicitly by nordic_ble dissector, so nothing to do here. */
}


/*
 * Editor modelines  -  http://www.wireshark.org/tools/modelines.html
 *
 * Local variables:
 * c-basic-offset: 4
 * tab-width: 8
 * indent-tabs-mode: nil
 * End:
 *
 * vi: set shiftwidth=4 tabstop=8 expandtab:
 * :indentSize=4:tabSize=8:noTabs=true:
 */
