/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file */

#include <stdint.h>
#include <string.h>

#include "ble_wpts.h"
#include "ble_wpts_common.h"
#include "nordic_common.h"

// Likely access out of bonds
//lint -save -e415 // Out of bonds
//lint -save -e416
//lint -save -e661 // Creating out of bond point
//lint -save -e662 // Creating out of bond point
//lint -save -e485 // Duplicate initialization of object element

/******************************************************************************/
/** @name Static functions and variables
 *  @{ */
/******************************************************************************/

const ble_gatts_rw_authorize_reply_params_t m_auth_reply =  \
    {.type = BLE_GATTS_AUTHORIZE_TYPE_READ,                 \
     .params.read.gatt_status = BLE_GATT_STATUS_SUCCESS,    \
     .params.read.update = 0}; /**< Constant authorize response, we only need authorization to get a hold of the read event. */


/**@brief Connect event handler.
 *
 * @param[in]   p_wpt      PRU Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_connect(ble_wpts_t * p_wpt, ble_evt_t * p_ble_evt)
{
    p_wpt->conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
}

/**@brief Disconnect event handler.
 *
 * @param[in]   p_wpt       PRU Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_disconnect(ble_wpts_t * p_wpt, ble_evt_t * p_ble_evt)
{
    UNUSED_PARAMETER(p_ble_evt);
    p_wpt->conn_handle = BLE_CONN_HANDLE_INVALID;
}

/**@brief Handle write events to the PRU Control characteristic.
 *
 * @param[in]   p_wpts        PRU Service structure.
 * @param[in]   p_evt_write   Write event received from the BLE stack.
 */
static void on_pru_control_write(ble_wpts_t * p_wpts, ble_gatts_evt_write_t * p_evt_write)
{
    if (p_evt_write->len == BLE_WPTS_PRU_CONTROL_LEN)
    {
        ble_wpts_evt_t  evt;

        // Decode/parse control data
        evt.data.pru_control.enable_pru_output = BLE_WPTS_BITFIELD_READ(p_evt_write->data[BLE_WPTS_PRU_CONTROL_ENABLES_POS],    \
                                                                         BLE_WPTS_ENABLE_PRU_OUTPUT_BITMSK,    \
                                                                         BLE_WPTS_ENABLE_PRU_OUTPUT_BITPOS);
        evt.data.pru_control.enable_pru_charge_indicator = BLE_WPTS_BITFIELD_READ(p_evt_write->data[BLE_WPTS_PRU_CONTROL_ENABLES_POS],    \
                                                                                   BLE_WPTS_ENABLE_PRU_CHARGE_INDICATOR_BITMSK,    \
                                                                                   BLE_WPTS_ENABLE_PRU_CHARGE_INDICATOR_BITPOS);
        evt.data.pru_control.adj_power = (ctl_adj_power_t)BLE_WPTS_BITFIELD_READ(p_evt_write->data[BLE_WPTS_PRU_CONTROL_ENABLES_POS],    \
                                                                                   BLE_WPTS_ENABLE_PRU_ADJ_POWER_BITMSK,    \
                                                                                   BLE_WPTS_ENABLE_PRU_ADJ_POWER_BITPOS);
        evt.data.pru_control.permissions = (ctl_perm_t)p_evt_write->data[BLE_WPTS_PRU_CONTROL_PERMISSION_POS];
        evt.data.pru_control.time_set = (ctl_time_set_t)p_evt_write->data[BLE_WPTS_PRU_CONTROL_TIME_SET_POS];

        // Generate service event
        evt.evt_type = BLE_WPTS_EVT_PRU_CONTROL_COMMAND;
        p_wpts->evt_handler(p_wpts, &evt);
    }
}

/**@brief Handle write events to the PTU Static Parameter characteristic.
 *
 * @param[in]   p_wpts        PRU Service structure.
 * @param[in]   p_evt_write   Write event received from the BLE stack.
 */
static void on_ptu_static_parameter_write(ble_wpts_t * p_wpts, ble_gatts_evt_write_t * p_evt_write)
{
    if (p_evt_write->len == BLE_WPTS_PTU_STATIC_LEN)
    {
        ble_wpts_evt_t  evt;
        uint8_t         * p = p_evt_write->data;

        evt.data.ptu_static.option_valid_max_imp = BLE_WPTS_BITFIELD_READ(p[BLE_WPTS_PTU_STATIC_OPTIONAL_FIELDS_POS],      \
                                                                           BLE_WPTS_PTU_STATIC_OPTION_MAX_IMPEDANCE_BITMSK, \
                                                                           BLE_WPTS_PTU_STATIC_OPTION_MAX_IMPEDANCE_BITPOS);

        evt.data.ptu_static.option_valid_max_res = BLE_WPTS_BITFIELD_READ(p[BLE_WPTS_PTU_STATIC_OPTIONAL_FIELDS_POS],      \
                                                                           BLE_WPTS_PTU_STATIC_OPTION_MAX_RESISTANCE_BITMSK, \
                                                                           BLE_WPTS_PTU_STATIC_OPTION_MAX_RESISTANCE_BITPOS);

        evt.data.ptu_static.ptu_max_source_impedance = BLE_WPTS_BITFIELD_READ(p[BLE_WPTS_PTU_STATIC_PTU_MAX_SRC_IMPED_POS],    \
                                                                               BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_BITMSK, \
                                                                               BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_BITPOS);

        evt.data.ptu_static.ptu_max_load_resistance = BLE_WPTS_BITFIELD_READ(p[BLE_WPTS_PTU_STATIC_PTU_MAX_LOAD_RES_POS],    \
                                                                              BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_BITMSK, \
                                                                              BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_BITPOS);

        evt.data.ptu_static.devices_supported_bitval = BLE_WPTS_BITFIELD_READ(p[BLE_WPTS_PTU_STATIC_NUM_DEVICES_SUPPORTED_POS], \
                                                                               BLE_WPTS_PTU_STATIC_NUM_DEVICES_BITMSK, \
                                                                               BLE_WPTS_PTU_STATIC_NUM_DEVICES_BITPOS);

        evt.data.ptu_static.ptu_power = p[BLE_WPTS_PTU_STATIC_PTU_POWER_POS];
        evt.data.ptu_static.ptu_class = p[BLE_WPTS_PTU_STATIC_PTU_CLASS_POS];
        evt.data.ptu_static.hardware_rev = p[BLE_WPTS_PTU_STATIC_HW_REV_POS];
        evt.data.ptu_static.firmware_rev = p[BLE_WPTS_PTU_STATIC_FW_REV_POS];
        evt.data.ptu_static.protocol_rev = p[BLE_WPTS_PTU_STATIC_PROTOCOL_REV_POS];


        evt.evt_type = BLE_WPTS_EVT_PTU_STATIC_PARAMETER;
        p_wpts->evt_handler(p_wpts, &evt);
    }
}

/**@brief Handle write events to the PRU Alert CCCD.
 *
 * @param[in]   p_wpts        PRU Service structure.
 * @param[in]   p_evt_write   Write event received from the BLE stack.
 */
static void on_pru_alert_cccd_write(ble_wpts_t * p_wpts, ble_gatts_evt_write_t * p_evt_write)
{
    // Ensure that CCCD data length is at least 16 bit
    //(Needed for ble_srv_is_notification_enabled())
    if (p_evt_write->len >= 2)
    {
        // CCCD written, update notification state
        if (p_wpts->evt_handler != NULL)
        {
            ble_wpts_evt_t evt;

            if (ble_srv_is_notification_enabled(p_evt_write->data))
            {
                evt.evt_type = BLE_WPTS_EVT_ALERT_ENABLED;
            }
            else
            {
                evt.evt_type = BLE_WPTS_EVT_ALERT_DISABLED;
            }

            p_wpts->evt_handler(p_wpts, &evt);
        }
    }
}


/**@brief Write event handler.
 *
 * @param[in]   p_wpt      PRU Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_write(ble_wpts_t * p_wpt, ble_evt_t * p_ble_evt)
{
    ble_gatts_evt_write_t * p_evt_write = &p_ble_evt->evt.gatts_evt.params.write;

    if (p_evt_write->handle == p_wpt->pru_control_handles.value_handle)
    {
        on_pru_control_write(p_wpt, p_evt_write);
    }
    else if (p_evt_write->handle == p_wpt->ptu_static_handles.value_handle)
    {
        on_ptu_static_parameter_write(p_wpt, p_evt_write);
    }
    else if (p_evt_write->handle == p_wpt->pru_alert_handles.cccd_handle)
    {
        on_pru_alert_cccd_write(p_wpt, p_evt_write);
    }
}

/** @} */

/******************************************************************************/
/** @name API implementation
 *  @{ */
/******************************************************************************/

void ble_wpts_on_ble_evt(ble_wpts_t * p_wpt, ble_evt_t * p_ble_evt)
{
    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
            on_connect(p_wpt, p_ble_evt);
            break;

        case BLE_GAP_EVT_DISCONNECTED:
            on_disconnect(p_wpt, p_ble_evt);
            break;

        case BLE_GATTS_EVT_WRITE:
            on_write(p_wpt, p_ble_evt);
            break;

        default:
            break;
    }
}


/**@brief Encode PRU Static Parameter structure.
 *
 * @param[in]   p_wpt             PRU Service structure.
 * @param[in]   p_pru_static       Static parameter structure to be encoded.
 * @param[out]  p_encoded_buffer   Buffer where the encoded data will be written.
 *
 * @return      Size of encoded data.
 */
static void pru_static_encode(ble_wpts_t * p_wpt, const pru_static_t * p_pru_static,
                                 uint8_t * p_encoded_buffer)
{
    memset(p_encoded_buffer,0,BLE_WPTS_PRU_STATIC_LEN);
    p_encoded_buffer[BLE_WPTS_PRU_STATIC_PROTOCOL_REV_POS] = p_pru_static->protocol_rev;
    p_encoded_buffer[BLE_WPTS_PRU_STATIC_PRU_CATEGORY_POS] = p_pru_static->pru_category;
    p_encoded_buffer[BLE_WPTS_PRU_STATIC_HW_REV_POS] = p_pru_static->hw_revision;
    p_encoded_buffer[BLE_WPTS_PRU_STATIC_FW_REV_POS] = p_pru_static->fw_revision;
    p_encoded_buffer[BLE_WPTS_PRU_STATIC_PRECT_MAX_POS] = p_pru_static->prect_max;
    (void)uint16_encode(p_pru_static->vrect_min_static, &p_encoded_buffer[BLE_WPTS_PRU_STATIC_VRECT_MIN_STATIC_POS]);
    (void)uint16_encode(p_pru_static->vrect_high_static, &p_encoded_buffer[BLE_WPTS_PRU_STATIC_VRECT_HIGH_STATIC_POS]);
    (void)uint16_encode(p_pru_static->vrect_set, &p_encoded_buffer[BLE_WPTS_PRU_STATIC_VRECT_SET_POS]);
    (void)uint16_encode(p_pru_static->delta_r1, &p_encoded_buffer[BLE_WPTS_PRU_STATIC_DELTA_R1_POS]);

    BLE_WPTS_BITFIELD_WRITE(p_encoded_buffer[BLE_WPTS_PTU_STATIC_OPTIONAL_FIELDS_POS], \
                             BLE_WPTS_PRU_STATIC_OPTION_DELTA_R1_BITMSK, \
                             BLE_WPTS_PRU_STATIC_OPTION_DELTA_R1_BITPOS, \
                             p_pru_static->option_fields_r1_valid);

    BLE_WPTS_BITFIELD_WRITE(p_encoded_buffer[BLE_WPTS_PRU_STATIC_PRU_INFORMATION_POS], \
                             BLE_WPTS_PRU_INFO_NFC_BITMSK, \
                             BLE_WPTS_PRU_INFO_NFC_BITPOS, \
                             p_pru_static->info_nfc_receiver);

    BLE_WPTS_BITFIELD_WRITE(p_encoded_buffer[BLE_WPTS_PRU_STATIC_PRU_INFORMATION_POS], \
                             BLE_WPTS_PRU_INFO_BLE_BITMSK, \
                             BLE_WPTS_PRU_INFO_BLE_BITPOS, \
                             p_pru_static->info_sep_btle_radio);

    BLE_WPTS_BITFIELD_WRITE(p_encoded_buffer[BLE_WPTS_PRU_STATIC_PRU_INFORMATION_POS], \
                             BLE_WPTS_PRU_INFO_POWER_CONTROL_PREF_BITMSK, \
                             BLE_WPTS_PRU_INFO_POWER_CONTROL_PREF_BITPOS, \
                             p_pru_static->info_pc_alg_pref);

    BLE_WPTS_BITFIELD_WRITE(p_encoded_buffer[BLE_WPTS_PRU_STATIC_PRU_INFORMATION_POS], \
                             BLE_WPTS_PRU_INFO_ADJUST_POWER_BITMSK, \
                             BLE_WPTS_PRU_INFO_ADJUST_POWER_BITPOS, \
                             p_pru_static->info_adj_power_capability);

    BLE_WPTS_BITFIELD_WRITE(p_encoded_buffer[BLE_WPTS_PRU_STATIC_PRU_INFORMATION_POS], \
                             BLE_WPTS_PRU_INFO_CHARGE_COMPLETE_CONN_MODE_BITMSK, \
                             BLE_WPTS_PRU_INFO_CHARGE_COMPLETE_CONN_MODE_BITPOS, \
                             p_pru_static->info_ccp_connected_mode);

    BLE_WPTS_BITFIELD_WRITE(p_encoded_buffer[BLE_WPTS_PRU_STATIC_PRU_INFORMATION_POS], \
                             BLE_WPTS_PRU_INFO_PTU_TEST_MODE_BITMSK, \
                             BLE_WPTS_PRU_INFO_PTU_TEST_MODE_BITPOS, \
                             p_pru_static->info_ptu_test_mode);
}

/**@brief Encode PRU Dynamic Parameter structure.
 *
 * @param[in]   p_wpt             PRU Service structure.
 * @param[in]   p_pru_dynamic      Dynamic parameter structure to be encoded.
 * @param[out]  p_encoded_buffer   Buffer where the encoded data will be written.
 *
 * @return      Size of encoded data.
 */
static void pru_dynamic_encode(ble_wpts_t * p_wpt, const pru_dynamic_t * p_pru_dynamic,
                                  uint8_t * p_encoded_buffer)
{

    memset(p_encoded_buffer,0,BLE_WPTS_PRU_DYNAMIC_LEN);
    p_encoded_buffer[BLE_WPTS_PRU_DYNAMIC_OPTIONAL_FIELDS_POS] = p_pru_dynamic->optional_fields;
    (void)uint16_encode(p_pru_dynamic->vrect, &p_encoded_buffer[BLE_WPTS_PRU_DYNAMIC_VRECT_POS]);
    (void)uint16_encode(p_pru_dynamic->irect, &p_encoded_buffer[BLE_WPTS_PRU_DYNAMIC_IRECT_POS]);

    if(BLE_WPTS_BITFIELD_READ(p_pru_dynamic->optional_fields,  \
       BLE_WPTS_PRU_DYNAMIC_OPTION_VOUT_BITMSK,                \
       BLE_WPTS_PRU_DYNAMIC_OPTION_VOUT_BITPOS) == BLE_WPTS_OPTION_VALID_VAL)
    {
        (void)uint16_encode(p_pru_dynamic->vout, &p_encoded_buffer[BLE_WPTS_PRU_DYNAMIC_VOUT_POS]);
    }

    if(BLE_WPTS_BITFIELD_READ(p_pru_dynamic->optional_fields,  \
       BLE_WPTS_PRU_DYNAMIC_OPTION_IOUT_BITMSK,                \
       BLE_WPTS_PRU_DYNAMIC_OPTION_IOUT_BITPOS) == BLE_WPTS_OPTION_VALID_VAL)
    {
        (void)uint16_encode(p_pru_dynamic->iout, &p_encoded_buffer[BLE_WPTS_PRU_DYNAMIC_IOUT_POS]);
    }

    if(BLE_WPTS_BITFIELD_READ(p_pru_dynamic->optional_fields,     \
       BLE_WPTS_PRU_DYNAMIC_OPTION_TEMPERATURE_BITMSK,            \
       BLE_WPTS_PRU_DYNAMIC_OPTION_TEMPERATURE_BITPOS) == BLE_WPTS_OPTION_VALID_VAL)
    {
        p_encoded_buffer[BLE_WPTS_PRU_DYNAMIC_TEMPERATURE_POS] = (uint8_t)(p_pru_dynamic->temperature + 40);
    }

    if(BLE_WPTS_BITFIELD_READ(p_pru_dynamic->optional_fields,     \
       BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_MIN_DYN_BITMSK,          \
       BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_MIN_DYN_BITPOS) == BLE_WPTS_OPTION_VALID_VAL)
    {
        (void)uint16_encode(p_pru_dynamic->vrect_min_dyn, &p_encoded_buffer[BLE_WPTS_PRU_DYNAMIC_VRECT_MIN_DYN_POS]);
    }

    if(BLE_WPTS_BITFIELD_READ(p_pru_dynamic->optional_fields,     \
       BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_SET_DYN_BITMSK,          \
       BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_SET_DYN_BITPOS) == BLE_WPTS_OPTION_VALID_VAL)
    {
        (void)uint16_encode(p_pru_dynamic->vrect_set_dyn, &p_encoded_buffer[BLE_WPTS_PRU_DYNAMIC_VRECT_SET_DYN_POS]);
    }

    if(BLE_WPTS_BITFIELD_READ(p_pru_dynamic->optional_fields,     \
       BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_HIGH_DYN_BITMSK,         \
       BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_HIGH_DYN_BITPOS) == BLE_WPTS_OPTION_VALID_VAL)
    {
        (void)uint16_encode(p_pru_dynamic->vrect_high_dyn, &p_encoded_buffer[BLE_WPTS_PRU_DYNAMIC_VRECT_HIGH_DYN_POS]);
    }


    BLE_WPTS_BITFIELD_WRITE(p_encoded_buffer[BLE_WPTS_PRU_DYNAMIC_PRU_ALERT_POS], \
                             BLE_WPTS_PRU_ALERT_OVER_VOLTAGE_BITMSK,               \
                             BLE_WPTS_PRU_ALERT_OVER_VOLTAGE_BITPOS,               \
                             p_pru_dynamic->alerts.pru_over_voltage);

    BLE_WPTS_BITFIELD_WRITE(p_encoded_buffer[BLE_WPTS_PRU_DYNAMIC_PRU_ALERT_POS], \
                            BLE_WPTS_PRU_ALERT_OVER_CURRENT_BITMSK,                \
                            BLE_WPTS_PRU_ALERT_OVER_CURRENT_BITPOS,                \
                            p_pru_dynamic->alerts.pru_over_current);

    BLE_WPTS_BITFIELD_WRITE(p_encoded_buffer[BLE_WPTS_PRU_DYNAMIC_PRU_ALERT_POS], \
                            BLE_WPTS_PRU_ALERT_OVER_TEMPERATURE_BITMSK,            \
                            BLE_WPTS_PRU_ALERT_OVER_TEMPERATURE_BITPOS,            \
                            p_pru_dynamic->alerts.pru_over_temperature);

    BLE_WPTS_BITFIELD_WRITE(p_encoded_buffer[BLE_WPTS_PRU_DYNAMIC_PRU_ALERT_POS], \
                            BLE_WPTS_PRU_ALERT_SELF_PROTECTION_BITMSK,             \
                            BLE_WPTS_PRU_ALERT_SELF_PROTECTION_BITPOS,             \
                            p_pru_dynamic->alerts.pru_self_protection);

    BLE_WPTS_BITFIELD_WRITE(p_encoded_buffer[BLE_WPTS_PRU_DYNAMIC_PRU_ALERT_POS], \
                            BLE_WPTS_PRU_ALERT_CHARGE_COMPLETE_BITMSK,             \
                            BLE_WPTS_PRU_ALERT_CHARGE_COMPLETE_BITPOS,             \
                            p_pru_dynamic->alerts.charge_complete);

    BLE_WPTS_BITFIELD_WRITE(p_encoded_buffer[BLE_WPTS_PRU_DYNAMIC_PRU_ALERT_POS], \
                            BLE_WPTS_PRU_ALERT_WIRED_CHARGER_DETECT_BITMSK,        \
                            BLE_WPTS_PRU_ALERT_WIRED_CHARGER_DETECT_BITPOS,        \
                            p_pru_dynamic->alerts.wired_charge_detect);

    BLE_WPTS_BITFIELD_WRITE(p_encoded_buffer[BLE_WPTS_PRU_DYNAMIC_PRU_ALERT_POS], \
                            BLE_WPTS_PRU_DYNAMIC_ALERT_ADJ_POWER_RESP_BITMSK,      \
                            BLE_WPTS_PRU_DYNAMIC_ALERT_ADJ_POWER_RESP_BITPOS,      \
                            p_pru_dynamic->adjust_power_response);

    BLE_WPTS_BITFIELD_WRITE(p_encoded_buffer[BLE_WPTS_PRU_DYNAMIC_PRU_ALERT_POS], \
                            BLE_WPTS_PRU_DYNAMIC_ALERT_CHARGE_PORT_BITMSK,         \
                            BLE_WPTS_PRU_DYNAMIC_ALERT_CHARGE_PORT_BITPOS,         \
                            p_pru_dynamic->pru_charge_port);

    p_encoded_buffer[BLE_WPTS_PRU_DYNAMIC_PTU_TESTER_COMMAND_POS] = (uint8_t)(p_pru_dynamic->tester_command);
}

/**@brief Add PRU Control characteristic.
 *
 * @param[in]   p_wpt        PRU Service structure.
 * @param[in]   p_wpts_init   Information needed to initialize the service.
 *
 * @return      NRF_SUCCESS on success, otherwise an error code.
 */
static uint32_t pru_control_char_add(ble_wpts_t * p_wpt, const ble_wpts_init_t * p_wpts_init)
{
    ble_gatts_char_md_t char_md;
    ble_gatts_attr_t    attr_char_value;
    ble_uuid_t          ble_uuid;
    ble_gatts_attr_md_t attr_md;

    memset(&char_md, 0, sizeof(char_md));

    char_md.char_props.write = 1;
    char_md.char_props.read = 1;
    char_md.char_props.write_wo_resp = 1; // Required for "GATT write without response" procedure
    char_md.p_char_user_desc = NULL;
    char_md.p_char_pf        = NULL;
    char_md.p_user_desc_md   = NULL;
    char_md.p_cccd_md        = NULL;
    char_md.p_sccd_md        = NULL;

    ble_wpts_uuid_get(BLE_WPTS_UUID128_PRU_CONTROL_CHAR, &ble_uuid);

    memset(&attr_md, 0, sizeof(attr_md));

    attr_md.write_perm = p_wpts_init->wpts_attr_md.write_perm;
    attr_md.read_perm = p_wpts_init->wpts_attr_md.read_perm;
    attr_md.vloc       = BLE_GATTS_VLOC_STACK;
    attr_md.rd_auth    = 0;
    attr_md.wr_auth    = 0;
    attr_md.vlen       = 0;

    memset(&attr_char_value, 0, sizeof(attr_char_value));

    attr_char_value.p_uuid       = &ble_uuid;
    attr_char_value.p_attr_md    = &attr_md;
    attr_char_value.init_len     = 0;
    attr_char_value.init_offs    = 0;
    attr_char_value.max_len      = BLE_WPTS_PRU_CONTROL_LEN;
    attr_char_value.p_value      = NULL;

    return sd_ble_gatts_characteristic_add(p_wpt->service_handle,
                                           &char_md,
                                           &attr_char_value,
                                           &p_wpt->pru_control_handles);
}

/**@brief Add PTU Static Parameter characteristic.
 *
 * @param[in]   p_wpt        PRU Service structure.
 * @param[in]   p_wpts_init   Information needed to initialize the service.
 *
 * @return      NRF_SUCCESS on success, otherwise an error code.
 */
static uint32_t ptu_static_char_add(ble_wpts_t * p_wpt, const ble_wpts_init_t * p_wpts_init)
{
    ble_gatts_char_md_t char_md;
    ble_gatts_attr_t    attr_char_value;
    ble_uuid_t          ble_uuid;
    ble_gatts_attr_md_t attr_md;

    memset(&char_md, 0, sizeof(char_md));

    char_md.char_props.write  = 1;
    char_md.char_props.read = 1;
    char_md.char_props.write_wo_resp = 1; // Required for "GATT write without response" procedure
    char_md.p_char_user_desc = NULL;
    char_md.p_char_pf        = NULL;
    char_md.p_user_desc_md   = NULL;
    char_md.p_cccd_md        = NULL;
    char_md.p_sccd_md        = NULL;

    ble_wpts_uuid_get(BLE_WPTS_UUID128_PTU_STATIC_PARAMETER_CHAR, &ble_uuid);

    memset(&attr_md, 0, sizeof(attr_md));

    attr_md.write_perm = p_wpts_init->wpts_attr_md.write_perm;
    attr_md.read_perm = p_wpts_init->wpts_attr_md.read_perm;

    attr_md.vloc       = BLE_GATTS_VLOC_STACK;
    attr_md.rd_auth    = 0;
    attr_md.wr_auth    = 0;
    attr_md.vlen       = 1;

    memset(&attr_char_value, 0, sizeof(attr_char_value));

    attr_char_value.p_uuid       = &ble_uuid;
    attr_char_value.p_attr_md    = &attr_md;
    attr_char_value.init_len     = 0;
    attr_char_value.init_offs    = 0;
    attr_char_value.max_len      = BLE_WPTS_PTU_STATIC_LEN;
    attr_char_value.p_value      = NULL;

    return sd_ble_gatts_characteristic_add(p_wpt->service_handle,
                                           &char_md,
                                           &attr_char_value,
                                           &p_wpt->ptu_static_handles);
}

/**@brief Add PRU Alert characteristic.
 *
 * @param[in]   p_wpt        PRU Service structure.
 * @param[in]   p_wpts_init   Information needed to initialize the service.
 *
 * @return      NRF_SUCCESS on success, otherwise an error code.
 */
static uint32_t pru_alert_char_add(ble_wpts_t * p_wpt, const ble_wpts_init_t * p_wpts_init)
{
    ble_gatts_char_md_t char_md;
    ble_gatts_attr_md_t cccd_md;
    ble_gatts_attr_t    attr_char_value;
    ble_uuid_t          ble_uuid;
    ble_gatts_attr_md_t attr_md;
    uint8_t             init_value = 0;

    memset(&char_md, 0, sizeof(char_md));
    char_md.char_props.notify   = 1;
    char_md.char_props.read = 1;
    char_md.p_char_user_desc    = NULL;
    char_md.p_char_pf           = NULL;
    char_md.p_user_desc_md      = NULL;
    char_md.p_cccd_md           = &cccd_md;
    char_md.p_sccd_md           = NULL;

    memset(&cccd_md, 0, sizeof(cccd_md));
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.read_perm);
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.write_perm);
    cccd_md.vloc = BLE_GATTS_VLOC_STACK;

    ble_wpts_uuid_get(BLE_WPTS_UUID128_PRU_ALERT_CHAR, &ble_uuid);

    memset(&attr_md, 0, sizeof(attr_md));

    BLE_GAP_CONN_SEC_MODE_SET_NO_ACCESS(&attr_md.write_perm);
    attr_md.read_perm = p_wpts_init->wpts_attr_md.read_perm;
    attr_md.vloc       = BLE_GATTS_VLOC_STACK;
    attr_md.rd_auth    = 0;
    attr_md.wr_auth    = 0;
    attr_md.vlen       = 0;

    memset(&attr_char_value, 0, sizeof(attr_char_value));

    attr_char_value.p_uuid       = &ble_uuid;
    attr_char_value.p_attr_md    = &attr_md;
    attr_char_value.init_len     = 1;
    attr_char_value.init_offs    = 0;
    attr_char_value.max_len      = BLE_WPTS_PRU_ALERT_MAX_LEN;
    attr_char_value.p_value      = &init_value;

    return sd_ble_gatts_characteristic_add(p_wpt->service_handle,
                                           &char_md,
                                           &attr_char_value,
                                           &p_wpt->pru_alert_handles);
}

/**@brief Add PRU Static Parameter characteristic.
 *
 * @param[in]   p_wpt        PRU Service structure.
 * @param[in]   p_wpts_init   Information needed to initialize the service.
 *
 * @return      NRF_SUCCESS on success, otherwise an error code.
 */
static uint32_t pru_static_char_add(ble_wpts_t * p_wpt, const ble_wpts_init_t * p_wpts_init)
{
    ble_gatts_char_md_t char_md;
    ble_gatts_attr_t    attr_char_value;
    ble_uuid_t          ble_uuid;
    ble_gatts_attr_md_t attr_md;
    uint8_t             encoded_initial_val[BLE_WPTS_PRU_STATIC_LEN];

    memset(&char_md, 0, sizeof(char_md));

    char_md.char_props.read   = 1;
    char_md.p_char_user_desc  = NULL;
    char_md.p_char_pf         = NULL;
    char_md.p_user_desc_md    = NULL;
    char_md.p_cccd_md         = NULL;
    char_md.p_sccd_md         = NULL;

    ble_wpts_uuid_get(BLE_WPTS_UUID128_PRU_STATIC_PARAMETER_CHAR, &ble_uuid);

    memset(&attr_md, 0, sizeof(attr_md));

    attr_md.read_perm  = p_wpts_init->wpts_attr_md.read_perm;
    BLE_GAP_CONN_SEC_MODE_SET_NO_ACCESS(&attr_md.write_perm);
    attr_md.vloc       = BLE_GATTS_VLOC_STACK;
    attr_md.rd_auth    = 0;
    attr_md.wr_auth    = 0;
    attr_md.vlen       = 1;

    memset(&attr_char_value, 0, sizeof(attr_char_value));

    attr_char_value.p_uuid       = &ble_uuid;
    attr_char_value.p_attr_md    = &attr_md;
    attr_char_value.init_len     = BLE_WPTS_PRU_STATIC_LEN;

    pru_static_encode(p_wpt, &p_wpts_init->pru_static, encoded_initial_val);

    attr_char_value.init_offs    = 0;
    attr_char_value.max_len      = BLE_WPTS_PRU_STATIC_LEN;
    attr_char_value.p_value      = encoded_initial_val;

    return sd_ble_gatts_characteristic_add(p_wpt->service_handle,
                                           &char_md,
                                           &attr_char_value,
                                           &p_wpt->pru_static_handles);
}

/**@brief Add PRU Dynamic Parameter characteristic.
 *
 * @param[in]   p_wpt        PRU Service structure.
 * @param[in]   p_wpts_init   Information needed to initialize the service.
 *
 * @return      NRF_SUCCESS on success, otherwise an error code.
 */
static uint32_t pru_dynamic_char_add(ble_wpts_t * p_wpt, const ble_wpts_init_t * p_wpts_init)
{
    ble_gatts_char_md_t char_md;
    ble_gatts_attr_t    attr_char_value;
    ble_uuid_t          ble_uuid;
    ble_gatts_attr_md_t attr_md;
    uint8_t             encoded_initial_val[BLE_WPTS_PRU_DYNAMIC_LEN];

    memset(&char_md, 0, sizeof(char_md));

    char_md.char_props.read  = 1;
    char_md.p_char_user_desc = NULL;
    char_md.p_char_pf        = NULL;
    char_md.p_user_desc_md   = NULL;
    char_md.p_cccd_md        = NULL;
    char_md.p_sccd_md        = NULL;

    ble_wpts_uuid_get(BLE_WPTS_UUID128_PRU_DYNAMIC_PARAMETER_CHAR, &ble_uuid);

    memset(&attr_md, 0, sizeof(attr_md));

    attr_md.read_perm  = p_wpts_init->wpts_attr_md.read_perm;
    BLE_GAP_CONN_SEC_MODE_SET_NO_ACCESS(&attr_md.write_perm);
    attr_md.vloc       = BLE_GATTS_VLOC_STACK;
    attr_md.rd_auth    = 0;
    attr_md.wr_auth    = 0;
    attr_md.vlen       = 1;

    memset(&attr_char_value, 0, sizeof(attr_char_value));
    memset(encoded_initial_val, 0, sizeof(encoded_initial_val));

    attr_char_value.p_uuid       = &ble_uuid;
    attr_char_value.p_attr_md    = &attr_md;
    attr_char_value.init_len     = sizeof(encoded_initial_val);
    attr_char_value.init_offs    = 0;
    attr_char_value.max_len      = BLE_WPTS_PRU_DYNAMIC_LEN;
    attr_char_value.p_value      = encoded_initial_val;

    return sd_ble_gatts_characteristic_add(p_wpt->service_handle,          \
                                           &char_md,                        \
                                           &attr_char_value,                \
                                           &p_wpt->pru_dynamic_handles);
}

uint32_t ble_wpts_init(ble_wpts_t * p_wpt, const ble_wpts_init_t * p_wpts_init)
{
    uint32_t   err_code;
    ble_uuid_t ble_uuid;

    // Initialize service structure
    p_wpt->evt_handler                 = p_wpts_init->evt_handler;
    p_wpt->conn_handle                 = BLE_CONN_HANDLE_INVALID;

    err_code = ble_wpts_uuid_init();
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    }

    BLE_UUID_BLE_ASSIGN(ble_uuid, BLE_WPTS_UUID16);

    err_code = sd_ble_gatts_service_add(BLE_GATTS_SRVC_TYPE_PRIMARY, &ble_uuid, &p_wpt->service_handle);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    }

    // Add PRU Control characteristic
    err_code = pru_control_char_add(p_wpt, p_wpts_init);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    }

    // Add PTU Static Parameter characteristic
    err_code = ptu_static_char_add(p_wpt, p_wpts_init);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    }

    // Add PRU Alert characteristic
    err_code = pru_alert_char_add(p_wpt, p_wpts_init);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    }

    // Add PRU Static Parameter characteristic
    err_code = pru_static_char_add(p_wpt, p_wpts_init);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    }

    // Add PRU Dynamic Parameter characteristic
    err_code = pru_dynamic_char_add(p_wpt, p_wpts_init);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    }

    return NRF_SUCCESS;
}

uint32_t ble_wpts_alert_send(ble_wpts_t * p_wpt, pru_alert_t * p_alert)
{
    uint32_t err_code;

    // Send value if connected and notifying
    if (p_wpt->conn_handle != BLE_CONN_HANDLE_INVALID)
    {
        uint16_t               hvx_len;
        ble_gatts_hvx_params_t hvx_params;
        uint8_t alert[BLE_WPTS_PRU_ALERT_MAX_LEN];

        memset(&hvx_params, 0, sizeof(hvx_params));
        memset(alert, 0, BLE_WPTS_PRU_ALERT_MAX_LEN);
        hvx_params.p_data = alert;
        hvx_len = BLE_WPTS_PRU_ALERT_MIN_LEN;

        hvx_params.handle   = p_wpt->pru_alert_handles.value_handle;
        hvx_params.type     = BLE_GATT_HVX_NOTIFICATION; // As Mode Transition is not supported, we only send Alerts as notifications.
        hvx_params.offset   = 0;
        hvx_params.p_len    = &hvx_len;

        BLE_WPTS_BITFIELD_WRITE(hvx_params.p_data[BLE_WPTS_PRU_ALERT_POS], \
                                BLE_WPTS_PRU_ALERT_OVER_VOLTAGE_BITMSK,     \
                                BLE_WPTS_PRU_ALERT_OVER_VOLTAGE_BITPOS,     \
                                p_alert->alerts.pru_over_voltage );

        BLE_WPTS_BITFIELD_WRITE(hvx_params.p_data[BLE_WPTS_PRU_ALERT_POS], \
                                BLE_WPTS_PRU_ALERT_OVER_CURRENT_BITMSK,     \
                                BLE_WPTS_PRU_ALERT_OVER_CURRENT_BITPOS,     \
                                p_alert->alerts.pru_over_current);

        BLE_WPTS_BITFIELD_WRITE(hvx_params.p_data[BLE_WPTS_PRU_ALERT_POS], \
                                BLE_WPTS_PRU_ALERT_OVER_TEMPERATURE_BITMSK, \
                                BLE_WPTS_PRU_ALERT_OVER_TEMPERATURE_BITPOS, \
                                p_alert->alerts.pru_over_temperature);

        BLE_WPTS_BITFIELD_WRITE(hvx_params.p_data[BLE_WPTS_PRU_ALERT_POS], \
                                BLE_WPTS_PRU_ALERT_SELF_PROTECTION_BITMSK,  \
                                BLE_WPTS_PRU_ALERT_SELF_PROTECTION_BITPOS,  \
                                p_alert->alerts.pru_self_protection);

        BLE_WPTS_BITFIELD_WRITE(hvx_params.p_data[BLE_WPTS_PRU_ALERT_POS], \
                                BLE_WPTS_PRU_ALERT_CHARGE_COMPLETE_BITMSK,  \
                                BLE_WPTS_PRU_ALERT_CHARGE_COMPLETE_BITPOS,  \
                                p_alert->alerts.charge_complete);

        BLE_WPTS_BITFIELD_WRITE(hvx_params.p_data[BLE_WPTS_PRU_ALERT_POS],    \
                                BLE_WPTS_PRU_ALERT_WIRED_CHARGER_DETECT_BITMSK,\
                                BLE_WPTS_PRU_ALERT_WIRED_CHARGER_DETECT_BITPOS,\
                                p_alert->alerts.wired_charge_detect);

        BLE_WPTS_BITFIELD_WRITE(hvx_params.p_data[BLE_WPTS_PRU_ALERT_POS],\
                                BLE_WPTS_PRU_ALERT_MODE_TRANSITION_BITMSK, \
                                BLE_WPTS_PRU_ALERT_MODE_TRANSITION_BITPOS, \
                                BLE_WPTS_PRU_ALERT_MODE_NO_MODE_TRANSITION_VAL);

        err_code = sd_ble_gatts_hvx(p_wpt->conn_handle, &hvx_params);
    }
    else
    {
        err_code = NRF_ERROR_INVALID_STATE;
    }

    return err_code;
}

uint32_t ble_wpts_pru_static_set(ble_wpts_t * p_wpt, pru_static_t * p_pru_static)
{
    uint8_t     encoded_val[BLE_WPTS_PRU_STATIC_LEN];
    ble_gatts_value_t val;

    val.len = BLE_WPTS_PRU_STATIC_LEN;
    val.offset = 0;
    val.p_value = encoded_val;

    // encode characteristic value
    pru_static_encode(p_wpt, p_pru_static, encoded_val);

    // set value
    return sd_ble_gatts_value_set(p_wpt->conn_handle,
                                  p_wpt->pru_static_handles.value_handle,
                                  &val);
}


/**@brief Transfer latest read sensor data to characteristic attributes.
 */
uint32_t ble_wpts_pru_dynamic_set(ble_wpts_t * p_wpts, pru_dynamic_t * p_pru_dynamic)
{
    uint8_t     encoded_val[BLE_WPTS_PRU_DYNAMIC_LEN];
    ble_gatts_value_t val;

    val.len = BLE_WPTS_PRU_DYNAMIC_LEN;
    val.offset = 0;
    val.p_value = encoded_val;

    // encode characteristic value
    pru_dynamic_encode(p_wpts, p_pru_dynamic, encoded_val);

    // set value
    return sd_ble_gatts_value_set(p_wpts->conn_handle,
                                  p_wpts->pru_dynamic_handles.value_handle,
                                  &val);
}

/** @} */

//lint -restore
