/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file */

#include <stdint.h>
#include <string.h>

#include "ble_wpts_c.h"
#include "ble_srv_common.h"

/******************************************************************************/
/** @name Static funcitons and variables.
 *  @{ */
/******************************************************************************/

static ble_wpts_c_evt_handler_t m_wpts_c_evt_handler;     /**< Event handler */
static uint16_t m_pending_write_cmd_char_handle;            /**< Write WO response temp variable */
static uint16_t m_pending_write_conn_handle;                /**< Write WO response temp conn_handle */


//lint -save -e415 -e416 -e661 -e662 // Possible out-of-bounds errors.
/**@brief Process a PRU Dynamic Parameter read response.
 *
 * @param[in]   p_wpts_c        PRU Service Client structure.
 * @param[in]   p_ble_evt       Read response event received from the BLE stack.
 */
static void proc_pru_dynamic_read_rsp(ble_wpts_c_t * p_wpts_c, ble_evt_t * p_ble_evt)
{
    ble_wpts_c_evt_t    evt;
    uint8_t           * p_data;
    uint16_t            len;

    p_data = p_ble_evt->evt.gattc_evt.params.read_rsp.data;
    len = p_ble_evt->evt.gattc_evt.params.read_rsp.len;
    evt.gatt_status = p_ble_evt->evt.gattc_evt.gatt_status;
    memset(&evt.data.pru_dynamic, 0, sizeof(pru_dynamic_t));

    if(evt.gatt_status == BLE_GATT_STATUS_SUCCESS)
    {
        if (len == BLE_WPTS_PRU_DYNAMIC_LEN)
        {
            evt.data.pru_dynamic.optional_fields = p_data[BLE_WPTS_PRU_DYNAMIC_OPTIONAL_FIELDS_POS];
            evt.data.pru_dynamic.vrect = uint16_decode(&p_data[BLE_WPTS_PRU_DYNAMIC_VRECT_POS]);
            evt.data.pru_dynamic.irect = uint16_decode(&p_data[BLE_WPTS_PRU_DYNAMIC_IRECT_POS]);
            evt.data.pru_dynamic.vout = uint16_decode(&p_data[BLE_WPTS_PRU_DYNAMIC_VOUT_POS]);
            evt.data.pru_dynamic.iout = uint16_decode(&p_data[BLE_WPTS_PRU_DYNAMIC_IOUT_POS]);
            evt.data.pru_dynamic.temperature = (int16_t)p_data[BLE_WPTS_PRU_DYNAMIC_TEMPERATURE_POS]-40;
            evt.data.pru_dynamic.vrect_min_dyn  = uint16_decode(&p_data[BLE_WPTS_PRU_DYNAMIC_VRECT_MIN_DYN_POS]);
            evt.data.pru_dynamic.vrect_set_dyn  = uint16_decode(&p_data[BLE_WPTS_PRU_DYNAMIC_VRECT_SET_DYN_POS]);
            evt.data.pru_dynamic.vrect_high_dyn = uint16_decode(&p_data[BLE_WPTS_PRU_DYNAMIC_VRECT_HIGH_DYN_POS]);

            evt.data.pru_dynamic.alerts.pru_over_voltage = BLE_WPTS_BITFIELD_READ( p_data[BLE_WPTS_PRU_DYNAMIC_PRU_ALERT_POS], \
                                                                                        BLE_WPTS_PRU_ALERT_OVER_VOLTAGE_BITMSK, \
                                                                                        BLE_WPTS_PRU_ALERT_OVER_VOLTAGE_BITPOS );

            evt.data.pru_dynamic.alerts.pru_over_current = BLE_WPTS_BITFIELD_READ( p_data[BLE_WPTS_PRU_DYNAMIC_PRU_ALERT_POS], \
                                                                                      BLE_WPTS_PRU_ALERT_OVER_CURRENT_BITMSK,   \
                                                                                      BLE_WPTS_PRU_ALERT_OVER_CURRENT_BITPOS );

            evt.data.pru_dynamic.alerts.pru_over_temperature = BLE_WPTS_BITFIELD_READ( p_data[BLE_WPTS_PRU_DYNAMIC_PRU_ALERT_POS],            \
                                                                                                  BLE_WPTS_PRU_ALERT_OVER_TEMPERATURE_BITMSK,  \
                                                                                                  BLE_WPTS_PRU_ALERT_OVER_TEMPERATURE_BITPOS );

            evt.data.pru_dynamic.alerts.pru_self_protection = BLE_WPTS_BITFIELD_READ( p_data[BLE_WPTS_PRU_DYNAMIC_PRU_ALERT_POS],     \
                                                                                           BLE_WPTS_PRU_ALERT_SELF_PROTECTION_BITMSK,  \
                                                                                           BLE_WPTS_PRU_ALERT_SELF_PROTECTION_BITPOS );

            evt.data.pru_dynamic.alerts.charge_complete = BLE_WPTS_BITFIELD_READ( p_data[BLE_WPTS_PRU_DYNAMIC_PRU_ALERT_POS],     \
                                                                                      BLE_WPTS_PRU_ALERT_CHARGE_COMPLETE_BITMSK,   \
                                                                                      BLE_WPTS_PRU_ALERT_CHARGE_COMPLETE_BITPOS );

            evt.data.pru_dynamic.alerts.wired_charge_detect = BLE_WPTS_BITFIELD_READ( p_data[BLE_WPTS_PRU_DYNAMIC_PRU_ALERT_POS],        \
                                                                                          BLE_WPTS_PRU_ALERT_WIRED_CHARGER_DETECT_BITMSK, \
                                                                                          BLE_WPTS_PRU_ALERT_WIRED_CHARGER_DETECT_BITPOS );

            evt.data.pru_dynamic.pru_charge_port = BLE_WPTS_BITFIELD_READ( p_data[BLE_WPTS_PRU_DYNAMIC_PRU_ALERT_POS], \
                                                                                      BLE_WPTS_PRU_DYNAMIC_ALERT_CHARGE_PORT_BITMSK,  \
                                                                                      BLE_WPTS_PRU_DYNAMIC_ALERT_CHARGE_PORT_BITPOS );

            evt.data.pru_dynamic.adjust_power_response = BLE_WPTS_BITFIELD_READ( p_data[BLE_WPTS_PRU_DYNAMIC_PRU_ALERT_POS], \
                                                                                      BLE_WPTS_PRU_DYNAMIC_ALERT_ADJ_POWER_RESP_BITMSK, \
                                                                                      BLE_WPTS_PRU_DYNAMIC_ALERT_ADJ_POWER_RESP_BITPOS );

            evt.data.pru_dynamic.tester_command = (ptu_tester_command_t) p_data[BLE_WPTS_PRU_DYNAMIC_PTU_TESTER_COMMAND_POS];

        }
        else
        {
            // Invalid length, report error to application
            evt.gatt_status = BLE_GATT_STATUS_ATTERR_CPS_OUT_OF_RANGE;
        }
    }
    evt.type = BLE_WPTS_C_EVT_PRU_DYNAMIC_READ_RESP;
    m_wpts_c_evt_handler(p_wpts_c, &evt);
}
//lint -restore

//lint -save -e415 -e416 -e661 -e662 // Possible out-of-bounds errors.
/**@brief Process a PRU Static Parameter read response.
 *
 * @param[in]   p_wpts_c        PRU Service Client structure.
 * @param[in]   p_ble_evt       Read response event received from the BLE stack.
 */
static void proc_pru_static_read_rsp(ble_wpts_c_t * p_wpts_c, ble_evt_t * p_ble_evt)
{
    ble_wpts_c_evt_t   evt;
    uint8_t           * p_data;
    uint16_t            len;

    p_data = p_ble_evt->evt.gattc_evt.params.read_rsp.data;
    len = p_ble_evt->evt.gattc_evt.params.read_rsp.len;
    evt.gatt_status = p_ble_evt->evt.gattc_evt.gatt_status;

    if (evt.gatt_status == BLE_GATT_STATUS_SUCCESS)
    {
        if (len == BLE_WPTS_PRU_STATIC_LEN )
        {
            evt.data.pru_static.protocol_rev = p_data[BLE_WPTS_PRU_STATIC_PROTOCOL_REV_POS];
            evt.data.pru_static.pru_category = p_data[BLE_WPTS_PRU_STATIC_PRU_CATEGORY_POS];
            evt.data.pru_static.hw_revision = p_data[BLE_WPTS_PRU_STATIC_HW_REV_POS];
            evt.data.pru_static.fw_revision = p_data[BLE_WPTS_PRU_STATIC_FW_REV_POS];
            evt.data.pru_static.prect_max = p_data[BLE_WPTS_PRU_STATIC_PRECT_MAX_POS];
            evt.data.pru_static.vrect_min_static = uint16_decode(&p_data[BLE_WPTS_PRU_STATIC_VRECT_MIN_STATIC_POS]);
            evt.data.pru_static.vrect_high_static = uint16_decode(&p_data[BLE_WPTS_PRU_STATIC_VRECT_HIGH_STATIC_POS]);
            evt.data.pru_static.vrect_set = uint16_decode(&p_data[BLE_WPTS_PRU_STATIC_VRECT_SET_POS]);
            evt.data.pru_static.delta_r1 = uint16_decode(&p_data[BLE_WPTS_PRU_STATIC_DELTA_R1_POS]);

            evt.data.pru_static.option_fields_r1_valid = BLE_WPTS_BITFIELD_READ(p_data[BLE_WPTS_PTU_STATIC_OPTIONAL_FIELDS_POS], \
                                                                           BLE_WPTS_PRU_STATIC_OPTION_DELTA_R1_BITMSK, \
                                                                           BLE_WPTS_PRU_STATIC_OPTION_DELTA_R1_BITPOS);

            evt.data.pru_static.info_nfc_receiver = BLE_WPTS_BITFIELD_READ(p_data[BLE_WPTS_PRU_STATIC_PRU_INFORMATION_POS], \
                                         BLE_WPTS_PRU_INFO_NFC_BITMSK, \
                                         BLE_WPTS_PRU_INFO_NFC_BITPOS);

            evt.data.pru_static.info_sep_btle_radio = BLE_WPTS_BITFIELD_READ(p_data[BLE_WPTS_PRU_STATIC_PRU_INFORMATION_POS], \
                                         BLE_WPTS_PRU_INFO_BLE_BITMSK, \
                                         BLE_WPTS_PRU_INFO_BLE_BITPOS);

            evt.data.pru_static.info_pc_alg_pref = BLE_WPTS_BITFIELD_READ(p_data[BLE_WPTS_PRU_STATIC_PRU_INFORMATION_POS], \
                                         BLE_WPTS_PRU_INFO_POWER_CONTROL_PREF_BITMSK, \
                                         BLE_WPTS_PRU_INFO_POWER_CONTROL_PREF_BITPOS);


            evt.data.pru_static.info_adj_power_capability = BLE_WPTS_BITFIELD_READ(p_data[BLE_WPTS_PRU_STATIC_PRU_INFORMATION_POS], \
                                         BLE_WPTS_PRU_INFO_ADJUST_POWER_BITMSK, \
                                         BLE_WPTS_PRU_INFO_ADJUST_POWER_BITPOS);

            evt.data.pru_static.info_ccp_connected_mode = BLE_WPTS_BITFIELD_READ(p_data[BLE_WPTS_PRU_STATIC_PRU_INFORMATION_POS], \
                                         BLE_WPTS_PRU_INFO_CHARGE_COMPLETE_CONN_MODE_BITMSK, \
                                         BLE_WPTS_PRU_INFO_CHARGE_COMPLETE_CONN_MODE_BITPOS);

            evt.data.pru_static.info_ptu_test_mode = BLE_WPTS_BITFIELD_READ(p_data[BLE_WPTS_PRU_STATIC_PRU_INFORMATION_POS], \
                                         BLE_WPTS_PRU_INFO_PTU_TEST_MODE_BITMSK, \
                                         BLE_WPTS_PRU_INFO_PTU_TEST_MODE_BITPOS);
        }
        else
        {
            // Invalid length, report error to application
            evt.gatt_status = BLE_GATT_STATUS_ATTERR_CPS_OUT_OF_RANGE;
        }
        evt.type = BLE_WPTS_C_EVT_PRU_STATIC_READ_RESP;
        m_wpts_c_evt_handler(p_wpts_c, &evt);
    }
}
//lint -restore


/**@brief Handle read response events.
 *
 * @param[in]   p_wpts_c        PRU Service Client structure.
 * @param[in]   p_ble_evt       Read response event received from the BLE stack.
 */
static void on_read_rsp(ble_wpts_c_t * p_wpts_c, ble_evt_t * p_ble_evt)
{
    if(p_ble_evt->evt.gattc_evt.gatt_status == BLE_GATT_STATUS_SUCCESS)
    {
        if (p_ble_evt->evt.gattc_evt.params.read_rsp.handle == p_wpts_c->prudp_handle)
        {
            proc_pru_dynamic_read_rsp(p_wpts_c, p_ble_evt);
        }
        else if (p_ble_evt->evt.gattc_evt.params.read_rsp.handle == p_wpts_c->prusp_handle)
        {
            proc_pru_static_read_rsp(p_wpts_c, p_ble_evt);
        }
    }
}

/**@brief Handle write response events.
 *
 * @param[in]   p_wpts_c        PRU Service Client structure.
 * @param[in]   p_ble_evt       Write response event received from the BLE stack.
 */
static void on_write_rsp(ble_wpts_c_t * p_wpts_c, ble_evt_t * p_ble_evt)
{
    ble_wpts_c_evt_t    evt;
    evt.gatt_status = p_ble_evt->evt.gattc_evt.gatt_status;

    if(p_ble_evt->evt.gattc_evt.gatt_status == BLE_GATT_STATUS_SUCCESS)
    {
        if (p_ble_evt->evt.gattc_evt.params.write_rsp.handle == p_wpts_c->prua_cccd_handle)
        {
            evt.type = BLE_WPTS_C_EVT_PRU_ALERT_ENABLE_WRITE_RESP;
            m_wpts_c_evt_handler(p_wpts_c, &evt);
        }
    }
}

/**@brief Handle TX complete event.
 *
 * @param[in]   p_wpts_c       PRU Service Client structure.
 * @param[in]   p_ble_evt       TX complete event received from the BLE stack.
 */
static void on_tx_complete(ble_wpts_c_t * p_wpts_c, ble_evt_t * p_ble_evt)
{
    ble_wpts_c_evt_t    evt;

    // Here, we always assume success, even though we didn't get a response
    evt.gatt_status = BLE_GATT_STATUS_SUCCESS;

    // Here we use a local variable as char handle is not included in the
    // TX_COMPLETE event
    if ((m_pending_write_cmd_char_handle == p_wpts_c->pruc_handle) && (m_pending_write_conn_handle  == p_wpts_c->conn_handle))
    {
        evt.type = BLE_WPTS_C_EVT_PRU_CONTROL_WRITE_RESP;
        m_wpts_c_evt_handler(p_wpts_c, &evt);
    }
    else if ((m_pending_write_cmd_char_handle == p_wpts_c->ptusp_handle) && (m_pending_write_conn_handle  == p_wpts_c->conn_handle))
    {
        evt.type = BLE_WPTS_C_EVT_PTU_STATIC_WRITE_RESP;
        m_wpts_c_evt_handler(p_wpts_c, &evt);
    }
}

/**@brief Handle value indication/notification events.
 *
 * @param[in]   p_wpts_c       PRU Service Client structure.
 * @param[in]   p_ble_evt       Indication/notification received from the BLE stack.
 */
static void on_hvx(ble_wpts_c_t * p_wpts_c, ble_evt_t * p_ble_evt)
{
    ble_wpts_c_evt_t    evt;


    evt.gatt_status = p_ble_evt->evt.gattc_evt.gatt_status;

    if ((p_ble_evt->evt.gattc_evt.params.hvx.handle == p_wpts_c->prua_handle) &&
        ( (p_ble_evt->evt.gattc_evt.params.hvx.type == BLE_GATT_HVX_NOTIFICATION) ||
          (p_ble_evt->evt.gattc_evt.params.hvx.type == BLE_GATT_HVX_INDICATION))
       )
    {

        uint8_t data = p_ble_evt->evt.gattc_evt.params.hvx.data[0];

        evt.type = BLE_WPTS_C_EVT_PRU_ALERT;

        memset(&evt.data.pru_alert,0,sizeof(pru_alert_t));

        evt.data.pru_alert.alerts.pru_over_voltage = BLE_WPTS_BITFIELD_READ(data, BLE_WPTS_PRU_ALERT_OVER_VOLTAGE_BITMSK, BLE_WPTS_PRU_ALERT_OVER_VOLTAGE_BITPOS);
        evt.data.pru_alert.alerts.pru_over_current = BLE_WPTS_BITFIELD_READ(data, BLE_WPTS_PRU_ALERT_OVER_CURRENT_BITMSK, BLE_WPTS_PRU_ALERT_OVER_CURRENT_BITPOS);
        evt.data.pru_alert.alerts.pru_over_temperature = BLE_WPTS_BITFIELD_READ(data, BLE_WPTS_PRU_ALERT_OVER_TEMPERATURE_BITMSK, BLE_WPTS_PRU_ALERT_OVER_TEMPERATURE_BITPOS);
        evt.data.pru_alert.alerts.pru_self_protection = BLE_WPTS_BITFIELD_READ(data, BLE_WPTS_PRU_ALERT_SELF_PROTECTION_BITMSK, BLE_WPTS_PRU_ALERT_SELF_PROTECTION_BITPOS);
        evt.data.pru_alert.alerts.charge_complete = BLE_WPTS_BITFIELD_READ(data, BLE_WPTS_PRU_ALERT_CHARGE_COMPLETE_BITMSK, BLE_WPTS_PRU_ALERT_CHARGE_COMPLETE_BITPOS);
        evt.data.pru_alert.alerts.wired_charge_detect = BLE_WPTS_BITFIELD_READ(data, BLE_WPTS_PRU_ALERT_WIRED_CHARGER_DETECT_BITMSK, BLE_WPTS_PRU_ALERT_WIRED_CHARGER_DETECT_BITPOS);
        evt.data.pru_alert.mode_transition = BLE_WPTS_BITFIELD_READ(data, BLE_WPTS_PRU_ALERT_MODE_TRANSITION_BITMSK, BLE_WPTS_PRU_ALERT_MODE_TRANSITION_BITPOS);

        if(evt.data.pru_alert.mode_transition != BLE_WPTS_PRU_ALERT_MODE_NO_MODE_TRANSITION_VAL)
        {
            if(p_ble_evt->evt.gattc_evt.params.hvx.len == (BLE_GAP_ADDR_LEN + 1))
            {
                //lint -save -e420
                memcpy(evt.data.pru_alert.device_address,&p_ble_evt->evt.gattc_evt.params.hvx.data[1],BLE_GAP_ADDR_LEN);
                //lint -restore
            }
        }

        if(p_ble_evt->evt.gattc_evt.params.hvx.type == BLE_GATT_HVX_NOTIFICATION)
        {
            evt.data.pru_alert.type = ALERT_TYPE_NOTIFICATION;
            m_wpts_c_evt_handler(p_wpts_c, &evt);
        }
        else if (p_ble_evt->evt.gattc_evt.params.hvx.type == BLE_GATT_HVX_INDICATION)
        {
            evt.data.pru_alert.type = ALERT_TYPE_INDICATION;

            if(sd_ble_gattc_hv_confirm(p_wpts_c->conn_handle, p_wpts_c->prua_handle) == NRF_SUCCESS)
            {
                m_wpts_c_evt_handler(p_wpts_c, &evt);
            }
        }
    }
}

/** @} */

/******************************************************************************/
/** @name API implementation
 *  @{ */
/******************************************************************************/

void ble_wpts_c_on_ble_evt(ble_wpts_c_t * p_wpts_c, ble_evt_t * p_ble_evt)
{
    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
            // Handled on profile level outside service
            break;

        case BLE_GAP_EVT_DISCONNECTED:
            // Handled on profile level outside service
            break;

        case BLE_GATTC_EVT_PRIM_SRVC_DISC_RSP:
            //on_prim_srvc_disc_rsp(p_wpts_c, p_ble_evt);
            break;

        case BLE_GATTC_EVT_CHAR_DISC_RSP:
            //on_char_disc_rsp(p_wpts_c, p_ble_evt);
            break;

        case BLE_GATTC_EVT_DESC_DISC_RSP:
            // on_desc_disc_rsp(p_wpts_c, p_ble_evt);
            break;

        case BLE_GATTC_EVT_READ_RSP:
            on_read_rsp(p_wpts_c, p_ble_evt);
            break;

        case BLE_GATTC_EVT_WRITE_RSP:
            on_write_rsp(p_wpts_c, p_ble_evt);
            break;
        case BLE_EVT_TX_COMPLETE:   // Used for Write Without Response
            on_tx_complete(p_wpts_c, p_ble_evt);
            break;
        case BLE_GATTC_EVT_HVX:
            on_hvx(p_wpts_c, p_ble_evt);
            break;

        default:
            break;
    }
}

uint32_t ble_wpts_c_init(const ble_wpts_c_init_t * p_wpts_c_init)
{
    if (p_wpts_c_init->evt_handler == NULL)
    {
        return NRF_ERROR_INVALID_PARAM;
    }
    m_wpts_c_evt_handler = p_wpts_c_init->evt_handler;
    m_pending_write_cmd_char_handle = BLE_GATT_HANDLE_INVALID;
    m_pending_write_conn_handle = BLE_CONN_HANDLE_INVALID;
    return NRF_SUCCESS;
}

uint32_t ble_wpts_c_send_pru_control(ble_wpts_c_t * p_wpts_c, pru_control_t * p_wpts_pru_control)
{
    ble_gattc_write_params_t    params;
    uint8_t                     buf[BLE_WPTS_PRU_CONTROL_LEN];
    memset(buf, 0, BLE_WPTS_PRU_CONTROL_LEN);

    // Encode control data
    BLE_WPTS_BITFIELD_WRITE(buf[BLE_WPTS_PRU_CONTROL_ENABLES_POS], \
                             BLE_WPTS_ENABLE_PRU_OUTPUT_BITMSK, \
                             BLE_WPTS_ENABLE_PRU_OUTPUT_BITPOS, \
                             p_wpts_pru_control->enable_pru_output);

    BLE_WPTS_BITFIELD_WRITE(buf[BLE_WPTS_PRU_CONTROL_ENABLES_POS], \
                           BLE_WPTS_ENABLE_PRU_CHARGE_INDICATOR_BITMSK, \
                           BLE_WPTS_ENABLE_PRU_CHARGE_INDICATOR_BITPOS, \
                           p_wpts_pru_control->enable_pru_charge_indicator);

    BLE_WPTS_BITFIELD_WRITE(buf[BLE_WPTS_PRU_CONTROL_ENABLES_POS], \
                           BLE_WPTS_ENABLE_PRU_ADJ_POWER_BITMSK, \
                           BLE_WPTS_ENABLE_PRU_ADJ_POWER_BITPOS, \
                           p_wpts_pru_control->adj_power);

    buf[BLE_WPTS_PRU_CONTROL_PERMISSION_POS] = p_wpts_pru_control->permissions;
    buf[BLE_WPTS_PRU_CONTROL_TIME_SET_POS] = p_wpts_pru_control->time_set;

    // Initiate write without response
    params.write_op = BLE_GATT_OP_WRITE_CMD;
    m_pending_write_cmd_char_handle = p_wpts_c->pruc_handle;
    m_pending_write_conn_handle = p_wpts_c->conn_handle;

    params.handle = p_wpts_c->pruc_handle;
    params.offset = 0;
    params.len = BLE_WPTS_PRU_CONTROL_LEN;
    params.p_value = buf;
    return sd_ble_gattc_write(p_wpts_c->conn_handle, &params);
}

uint32_t ble_wpts_c_read_pru_static(ble_wpts_c_t * p_wpts_c)
{
    return sd_ble_gattc_read(p_wpts_c->conn_handle, p_wpts_c->prusp_handle, 0);
}

uint32_t ble_wpts_c_read_pru_dynamic(ble_wpts_c_t * p_wpts_c)
{
    return sd_ble_gattc_read(p_wpts_c->conn_handle, p_wpts_c->prudp_handle, 0);
}

uint32_t ble_wpts_c_write_ptu_static(ble_wpts_c_t * p_wpts_c, ptu_static_t * p_wpts_ptu_static)
{
    ble_gattc_write_params_t    params;
    uint8_t                     buf[BLE_WPTS_PTU_STATIC_LEN];

    memset(buf, 0, BLE_WPTS_PTU_STATIC_LEN);

    // Encode data

    BLE_WPTS_BITFIELD_WRITE(buf[BLE_WPTS_PTU_STATIC_OPTIONAL_FIELDS_POS],                    \
                            BLE_WPTS_PTU_STATIC_OPTION_MAX_IMPEDANCE_BITMSK, \
                            BLE_WPTS_PTU_STATIC_OPTION_MAX_IMPEDANCE_BITPOS, \
                            p_wpts_ptu_static->option_valid_max_imp);

    BLE_WPTS_BITFIELD_WRITE(buf[BLE_WPTS_PTU_STATIC_OPTIONAL_FIELDS_POS],                    \
                            BLE_WPTS_PTU_STATIC_OPTION_MAX_RESISTANCE_BITMSK, \
                            BLE_WPTS_PTU_STATIC_OPTION_MAX_RESISTANCE_BITPOS, \
                            p_wpts_ptu_static->option_valid_max_res);

    BLE_WPTS_BITFIELD_WRITE(buf[BLE_WPTS_PTU_STATIC_PTU_MAX_SRC_IMPED_POS],             \
                             BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_BITMSK,  \
                             BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_BITPOS,  \
                             p_wpts_ptu_static->ptu_max_source_impedance);

    BLE_WPTS_BITFIELD_WRITE(buf[BLE_WPTS_PTU_STATIC_PTU_MAX_LOAD_RES_POS],              \
                             BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_BITMSK,   \
                             BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_BITPOS,   \
                             p_wpts_ptu_static->ptu_max_load_resistance);

    BLE_WPTS_BITFIELD_WRITE(buf[BLE_WPTS_PTU_STATIC_NUM_DEVICES_SUPPORTED_POS],            \
                             BLE_WPTS_PTU_STATIC_NUM_DEVICES_BITMSK,   \
                             BLE_WPTS_PTU_STATIC_NUM_DEVICES_BITPOS,   \
                             p_wpts_ptu_static->devices_supported_bitval);

    buf[BLE_WPTS_PTU_STATIC_PTU_POWER_POS] = p_wpts_ptu_static->ptu_power;
    buf[BLE_WPTS_PTU_STATIC_PTU_CLASS_POS] = p_wpts_ptu_static->ptu_class;
    buf[BLE_WPTS_PTU_STATIC_HW_REV_POS] = p_wpts_ptu_static->hardware_rev;
    buf[BLE_WPTS_PTU_STATIC_FW_REV_POS] = p_wpts_ptu_static->firmware_rev;
    buf[BLE_WPTS_PTU_STATIC_PROTOCOL_REV_POS] = p_wpts_ptu_static->protocol_rev;

    // Initiate write without response
    params.write_op = BLE_GATT_OP_WRITE_CMD;
    m_pending_write_cmd_char_handle = p_wpts_c->ptusp_handle;
    m_pending_write_conn_handle = p_wpts_c->conn_handle;

    params.handle = p_wpts_c->ptusp_handle;
    params.offset = 0;
    params.len = BLE_WPTS_PTU_STATIC_LEN;
    params.p_value = buf;
    return sd_ble_gattc_write(p_wpts_c->conn_handle, &params);
}

uint32_t ble_wpts_c_enable_pru_alert_notification(ble_wpts_c_t * p_wpts_c)
{
    ble_gattc_write_params_t    params;
    uint8_t                     buf[2];

    //lint -save -e534  // Function always returns '2'. Ignoring.
    uint16_encode(BLE_GATT_HVX_NOTIFICATION | BLE_GATT_HVX_INDICATION, buf);
    //lint -restore

    params.write_op = BLE_GATT_OP_WRITE_REQ;

    params.handle = p_wpts_c->prua_cccd_handle;
    params.offset = 0;
    params.len = sizeof(buf);
    params.p_value = buf;
    return sd_ble_gattc_write(p_wpts_c->conn_handle, &params);
}

uint32_t ble_wpts_service_data_read(ble_gap_evt_adv_report_t * p_adv_report, ble_wpts_service_data_t * p_service_data)
{
    uint8_t   i = 0;
    uint8_t   len;
    uint8_t * p_adv_service_data;

    while (i < p_adv_report->dlen)
    {
        if (p_adv_report->data[i] == 0)
        {
          return NRF_ERROR_INTERNAL;
        }

        if (p_adv_report->data[i + 1] == BLE_GAP_AD_TYPE_SERVICE_DATA)
        {
            len = p_adv_report->data[i];
            if(len >= BLE_WPTS_ADV_SERVICE_DATA_LENGTH)
            {
                p_service_data -> uuid16 = uint16_decode(&p_adv_report->data[i + 2]);

                p_adv_service_data = &p_adv_report->data[i + 4];
                p_service_data -> gatt_prim_srv_handle = uint16_decode(&p_adv_service_data[BLE_WPTS_ADV_SERVICE_HANDLE_POS]);

                p_service_data -> pru_pwr = BLE_WPTS_BITFIELD_READ(p_adv_service_data[BLE_WPTS_ADV_PRU_RSSI_PARAMS_POS], BLE_WPTS_ADV_PRU_RSSI_PRU_PWR_BITMSK, BLE_WPTS_ADV_PRU_RSSI_PRU_PWR_BITPOS);
                if(p_service_data -> pru_pwr != (int8_t)BLE_WPTS_ADV_POWER_UNKNOWN_VAL)
                {
                    p_service_data -> pru_pwr -= BLE_WPTS_ADV_PRU_TX_POWER_LEVEL_OFFSET;
                }

                p_service_data -> pru_gain = BLE_WPTS_BITFIELD_READ(p_adv_service_data[BLE_WPTS_ADV_PRU_RSSI_PARAMS_POS], BLE_WPTS_ADV_PRU_RSSI_PRU_GAIN_BITMSK, BLE_WPTS_ADV_PRU_RSSI_PRU_GAIN_BITPOS);
                if(p_service_data -> pru_gain != (int8_t)BLE_WPTS_ADV_ANTENNA_GAIN_UNKNOWN_VAL)
                {
                   p_service_data -> pru_gain -= BLE_WPTS_ADV_ANTENNA_GAIN_OFFSET;
                }

                p_service_data -> adv_flags = p_adv_service_data[BLE_WPTS_ADV_FLAGS_POS];
                return NRF_SUCCESS;
            }
        }

        i += (p_adv_report->data[i] + 1);
    }

    return NRF_ERROR_INTERNAL;
}

void ble_wpts_char_handles_update(uint16_t prim_service_handle, ble_wpts_c_t * wpts_c)
{
    wpts_c -> pruc_handle = prim_service_handle +  BLE_WPTS_HANDLE_OFFSET_PRU_CONTROL;
    wpts_c -> ptusp_handle = prim_service_handle + BLE_WPTS_HANDLE_OFFSET_PTU_STATIC;
    wpts_c -> prua_handle = prim_service_handle + BLE_WPTS_HANDLE_OFFSET_PRU_ALERT;
    wpts_c -> prua_cccd_handle = prim_service_handle + BLE_WPTS_HANDLE_OFFSET_PRU_ALERT_CCCD;
    wpts_c -> prusp_handle = prim_service_handle + BLE_WPTS_HANDLE_OFFSET_PRU_STATIC;
    wpts_c -> prudp_handle = prim_service_handle + BLE_WPTS_HANDLE_OFFSET_PRU_DYNAMIC;
    wpts_c -> service_handle_range.start_handle = prim_service_handle + BLE_WPTS_HANDLE_OFFSET_MIN;
    wpts_c -> service_handle_range.end_handle = prim_service_handle + BLE_WPTS_HANDLE_OFFSET_MAX;
}

/** @} */

//lint -restore
