/* Copyright (c) Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */


/** @file */

#include <stdint.h>

#include "nrf_soc.h"
#include "app_error.h"
#include "nrf_sdm.h"

#ifdef DFU_SUPPORT
    #include "bootloader_types.h"
    #include "bootloader_util.h"
    #define NOF_IRQS    32                  /**< Max number of interrupts. */
#endif // DFU_SUPPORT

#define MAX_RTC_COUNTER_VAL     0x00FFFFFF /**< Maximum value of the RTC counter. */

/******************************************************************************/
/** @name Static functions and variables
 *  @{ */
/******************************************************************************/


#ifdef DFU_SUPPORT
/**@brief Function for disabling all interrupts before jumping from bootloader to application.
 */
static void interrupts_disable(void)
{
    uint32_t interrupt_setting_mask;
    uint32_t irq = 0; // We start from first interrupt, i.e. interrupt 0.

    // Fetch the current interrupt settings.
    interrupt_setting_mask = NVIC->ISER[0];

    for (; irq < NOF_IRQS; irq++)
    {
        if (interrupt_setting_mask & (0x01 << irq))
        {
            // The interrupt was enabled, and hence disable it.
            NVIC_DisableIRQ((IRQn_Type)irq);
        }
    }
}
#endif // DFU_SUPPORT

/** @} */

/******************************************************************************/
/** @name API implementation
 *  @{ */
/******************************************************************************/

void power_manage(void)
{
    uint32_t err_code = sd_app_evt_wait();
    APP_ERROR_CHECK(err_code);
}


void system_off(void)
{
    uint32_t err_code = sd_power_system_off();
    APP_ERROR_CHECK(err_code);
}

uint32_t ticks_diff(uint32_t ticks_now, uint32_t ticks_old)
{
    if(ticks_old>ticks_now) //Overflow occured
    {
        return (((MAX_RTC_COUNTER_VAL-ticks_old)+ticks_now) & MAX_RTC_COUNTER_VAL);
    }
    else
    {
        return ((ticks_now-ticks_old) & MAX_RTC_COUNTER_VAL);
    }
}


#ifdef DFU_SUPPORT

void dfu_start(void)
{
    uint32_t err_code = sd_power_gpregret_set(BOOTLOADER_DFU_START);
    APP_ERROR_CHECK(err_code);

    err_code = sd_softdevice_disable();
    APP_ERROR_CHECK(err_code);

    err_code = sd_softdevice_vector_table_base_set(NRF_UICR->NRFFW[0]);
    APP_ERROR_CHECK(err_code);

    NVIC_ClearPendingIRQ(SWI2_IRQn);
    interrupts_disable();
    bootloader_util_app_start(NRF_UICR->NRFFW[0]);
}
#endif // DFU_SUPPORT

/** @} */

//lint -restore
