/* Copyright (c) Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/**
 * @file
 * @brief Button HAL.
 */

#ifdef DFU_SUPPORT

#include <stdbool.h>
#include <stdint.h>
#include "nrf_gpio.h"
#include "nrf_error.h"
#include "nrf_soc.h"
#include "nrf_gpiote.h"
#include "common_hal_buttons.h"
#include "app_timer.h"
#include "app_util_platform.h"
#include "wpt.h"

#define GPIOTE_CHANNEL_NUMBER_BUTTON_PRESS      0           /**< GPIOTE channel number for button pressed event */
#define GPIOTE_CHANNEL_NUMBER_BUTTON_RELEASE    1           /**< GPIOTE channel number for button released event */
#define DFU_BUTTON_PIN                          17          /**< Button 1 pin */
#define GPIOTE_IRQ_INTENCLR_ALL_VAL             0xffffffff  /**< Value used for clearing all interrupts for GPIOTE */
#define GPIOTE_IRQ_ENABLE_VAL              3           /**< Value used for enabling both GPIOTE interrupts */

static uint32_t m_button_press_duration_ms;                 /**< For how long does the button need to be pressed before an attempt to switch to DFU mode is made */
static uint8_t m_app_timer_prescaler;                       /**< The prescaler used for initializing the app_timer module */

void common_hal_buttons_init(uint32_t button_press_duration_ms, uint8_t app_timer_prescaler)
{
    m_button_press_duration_ms = button_press_duration_ms;
    m_app_timer_prescaler = app_timer_prescaler;
    nrf_gpio_cfg_input(DFU_BUTTON_PIN, NRF_GPIO_PIN_PULLUP);
    nrf_gpiote_event_configure(GPIOTE_CHANNEL_NUMBER_BUTTON_PRESS, DFU_BUTTON_PIN, NRF_GPIOTE_POLARITY_HITOLO);
    nrf_gpiote_event_configure(GPIOTE_CHANNEL_NUMBER_BUTTON_RELEASE, DFU_BUTTON_PIN, NRF_GPIOTE_POLARITY_LOTOHI);

    NRF_GPIOTE->INTENCLR = GPIOTE_IRQ_INTENCLR_ALL_VAL;
    NRF_GPIOTE->INTENSET = GPIOTE_IRQ_ENABLE_VAL;
    NVIC_ClearPendingIRQ(GPIOTE_IRQn);
    NVIC_SetPriority(GPIOTE_IRQn, APP_IRQ_PRIORITY_LOW);
    NVIC_EnableIRQ(GPIOTE_IRQn);
}

/**@brief Handler for GPIOTE interrupts
 */
void GPIOTE_IRQHandler(void)
{
    static uint32_t ticks_at_button_release, ticks_at_button_press;
    uint32_t diff, err_code;

    if(NRF_GPIOTE->EVENTS_IN[GPIOTE_CHANNEL_NUMBER_BUTTON_PRESS] == 1)
    {
        // Button has been pressed down
        err_code = app_timer_cnt_get(&ticks_at_button_press);
        APP_ERROR_CHECK(err_code);

        NRF_GPIOTE->EVENTS_IN[GPIOTE_CHANNEL_NUMBER_BUTTON_PRESS] = 0;
    }

    else if(NRF_GPIOTE->EVENTS_IN[GPIOTE_CHANNEL_NUMBER_BUTTON_RELEASE] == 1)
    {
        // Button has been released
        err_code = app_timer_cnt_get(&ticks_at_button_release);
        APP_ERROR_CHECK(err_code);

        diff = ticks_diff(ticks_at_button_release, ticks_at_button_press);

        // If enough time has passed, start bootloader in DFU mode
        if(diff > APP_TIMER_TICKS(m_button_press_duration_ms, m_app_timer_prescaler))
        {
            if(dfu_check())
            {
                terminate();
                dfu_start();
            }
        }

        NRF_GPIOTE->EVENTS_IN[GPIOTE_CHANNEL_NUMBER_BUTTON_RELEASE] = 0;
    }
}

#endif // DFU_SUPPORT
