/* Copyright (c) Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/**
 * @file
 * @brief Emulated version of PRU hardware.
 */

#include <stdbool.h>
#include <stdint.h>

#include "nrf_error.h"
#include "pru_hw_config.h"
#include "pru_hal.h"

#define VRECT_DEFAULT_VAL   0                           /**< [mV] */
#define IRECT_DEFAULT_VAL   100                         /**< [mA] */
#define VOUT_DEFAULT_VAL    2000                        /**< [mV] */
#define IOUT_DEFAULT_VAL    100                         /**< [mA] */
#define TEMP_DEFAULT_VAL    20                          /**< [deg C] */

/******************************************************************************/
/** @name Static functions and variables
 *  @{ */
/******************************************************************************/

static uint32_t m_vrect;                      /**< VRECT value */
static uint32_t m_irect;                      /**< IRECT value */
static uint32_t m_vout;                       /**< VOUT value */
static uint32_t m_iout;                       /**< IOUT value */
static int16_t m_temp;                        /**< Temperature value */
static bool m_wired_charge_detected;          /**< Wired charge detect status */
static bool m_charge_complete;                /**< Charge complete status */
static bool m_vout_enable;                    /**< VOUT enabled*/
static ctl_adj_power_t m_power_adj;      /**< Power adjust setting */
static bool m_power_adjusted;                 /**< Power adjusted */

/** @} */

/******************************************************************************/
/** @name API implementation
 *  @{ */
/******************************************************************************/

uint32_t pru_hal_init(void)
{
    // Set default values
    m_vrect = VRECT_DEFAULT_VAL;
    m_irect = IRECT_DEFAULT_VAL;
    m_vout = VOUT_DEFAULT_VAL;
    m_iout = IOUT_DEFAULT_VAL;
    m_temp = TEMP_DEFAULT_VAL;
    m_wired_charge_detected = false;
    m_charge_complete = false;
    m_vout_enable = false;
    m_power_adj = CTL_ADJ_POWER_MAX_VAL;
    m_power_adjusted = false;

    return NRF_SUCCESS;
}

uint32_t pru_hal_vrect_get(uint16_t * p_vrect)
{
    *p_vrect = m_vrect;
    return NRF_SUCCESS;
}

uint32_t pru_hal_irect_get(uint16_t * p_irect)
{
    *p_irect = m_irect;
    return NRF_SUCCESS;
}

uint32_t pru_hal_vout_enable_set(bool enable)
{
    m_vout_enable = enable;
    return NRF_SUCCESS;
}

uint32_t pru_hal_vout_enable_get(bool * enable)
{
    *enable = m_vout_enable;
    return NRF_SUCCESS;
}

uint32_t pru_hal_vout_get(uint16_t * p_vout)
{
    *p_vout = m_vout;
    return NRF_SUCCESS;
}

uint32_t pru_hal_iout_get(uint16_t * p_iout)
{
    *p_iout = m_iout;
    return NRF_SUCCESS;
}

uint32_t pru_hal_temperature_get(int16_t * p_temperature)
{
    *p_temperature = m_temp;
    return NRF_SUCCESS;
}

uint32_t pru_hal_wired_charge_detect_status_get(bool * p_wired)
{
    *p_wired = m_wired_charge_detected;
    return NRF_SUCCESS;
}

uint32_t pru_hal_charge_complete_status_get(bool * p_charge_complete)
{
    *p_charge_complete = m_charge_complete;
    return NRF_SUCCESS;
}

uint32_t pru_hal_adj_power_set(ctl_adj_power_t adj_setting)
{
    m_power_adj = adj_setting;
    return NRF_SUCCESS;
}

uint32_t pru_hal_adj_power_get(ctl_adj_power_t * adj_setting)
{
    *adj_setting = m_power_adj;
    return NRF_SUCCESS;
}

uint32_t pru_hal_power_adjusted_get(bool * adjusted)
{
    *adjusted = m_power_adjusted;
    return NRF_SUCCESS;
}

uint32_t pru_hal_power_adjusted_set(bool adjusted)
{
    m_power_adjusted = adjusted;
    return NRF_SUCCESS;
}

/** @} */
