/* Copyright (c) 2012 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file */

#include <stdint.h>
#include <string.h>
#include "nordic_common.h"
#include "nrf_assert.h"
#include "pru.h"
#include "pru_sensors.h"
#include "pru_hw_config.h"
#include "pru_test_mux.h"
#include "pru_sm.h"
#include "pru_adv.h"
#include "pru_config.h"
#include "ble_hci.h"
#ifdef PRU_PSTORAGE_ENABLE
#include "pru_pstorage.h"
#endif


/******************************************************************************/
/** @name Static functions and variables
 *  @{ */
/******************************************************************************/

static uint16_t   m_conn_handle = BLE_CONN_HANDLE_INVALID; /**< Handle of the current connection. */
static ble_wpts_t m_wpts;                                  /**< Structure used to identify the PRU service. */
APP_TIMER_DEF(m_pru_sensors_timer_single_shot_id);         /**< App timer ID of single shot measurement. */
APP_TIMER_DEF(m_p_pru_sensors_timer_id);                   /**< App timer ID of repeating sensor measurement. */
static uint32_t             m_vrect_high_duration_ms;      /**< For how long has VRECT been above limit. */
static pru_sm_state_vars_t  m_state_variables;             /**< Previous, current and next state. */
static bool                 m_adv_allowed = false;         /**< Is advertising allowed. */
static app_sm_evt_handler_t m_app_sm_evt_handler = NULL;   /**< Optional application state machine event handler. */
static bool                 m_advertising;                 /**< True when advertising */


/******************************************************************************
 * Static functions
 ******************************************************************************/

/**@brief Start advertising.
 *
 * @details Initialize advertisement data, and start advertising.
 *
 */
static void m_advertising_start(void)
{
    uint32_t err_code;
    ble_gap_adv_params_t     adv_params;

    err_code = pru_advdata_set(&m_wpts);
    APP_ERROR_CHECK(err_code);

    memset(&adv_params, 0, sizeof(adv_params));

    // We always use indirect advertising
    adv_params.type        = BLE_GAP_ADV_TYPE_ADV_IND;
    adv_params.p_peer_addr = NULL;
    adv_params.fp          = BLE_GAP_ADV_FP_ANY;
    adv_params.interval    = PRU_ADV_INTERVAL;
    adv_params.timeout     = BLE_GAP_ADV_TIMEOUT_LIMITED_MAX;

    err_code = sd_ble_gap_adv_start(&adv_params);
    m_advertising = true;

    if(err_code != NRF_ERROR_INVALID_STATE && err_code != NRF_ERROR_BUSY)
    {
        APP_ERROR_CHECK(err_code);
    }
}

#ifdef PRU_PSTORAGE_ENABLE
/**@brief Dispatches a System stack event to flash storage module.
 *
 * @details This function is called after a system event has been received.
 *
 * @param[in]   event   Bluetooth stack event.
 */
static void m_sys_evt_dispatch(uint32_t event)
{
    pru_pstorage_sys_evt_handler(event);
}
#endif

/**@brief GAP initialization.
 *
 * @details This function shall be used to setup all the necessary GAP (Generic Access Profile)
 *          parameters of the device. It also sets the permissions and appearance.
 */
static void m_gap_params_init(void)
{
    uint32_t                err_code;
    ble_gap_conn_params_t   gap_conn_params;
    ble_gap_conn_sec_mode_t sec_mode;

    BLE_GAP_CONN_SEC_MODE_SET_NO_ACCESS(&sec_mode);
    err_code = sd_ble_gap_device_name_set(&sec_mode, (const uint8_t *)PRU_ADV_DEVICE_NAME, strlen(PRU_ADV_DEVICE_NAME));
    APP_ERROR_CHECK(err_code);

    err_code = sd_ble_gap_appearance_set(BLE_APPEARANCE_UNKNOWN);
    APP_ERROR_CHECK(err_code);

    err_code = sd_ble_gap_tx_power_set(PRU_TX_POWER_LEVEL);
    APP_ERROR_CHECK(err_code);

    memset(&gap_conn_params, 0, sizeof(gap_conn_params));

    gap_conn_params.min_conn_interval = PRU_MIN_CONN_INTERVAL;
    gap_conn_params.max_conn_interval = PRU_MAX_CONN_INTERVAL;
    gap_conn_params.slave_latency     = PRU_SLAVE_LATENCY;
    gap_conn_params.conn_sup_timeout  = PRU_CONN_SUP_TIMEOUT;

    err_code = sd_ble_gap_ppcp_set(&gap_conn_params);
    APP_ERROR_CHECK(err_code);
}


/**@brief Handle PRU SM events that require profile handling
 * @param evt The event
 */
static void m_handle_event(pru_sm_signal_type_t evt)
{
    uint32_t err_code;

    switch(evt)
    {
        case PRU_SM_SIGNAL_POWER_APPLIED:
            if(m_state_variables.current_state == PRU_SM_STATE_BOOT || m_state_variables.current_state == PRU_SM_STATE_NULL)
            {
                m_vrect_high_duration_ms += PRU_SENSORS_TIMER_PERIOD_MS;
                m_adv_allowed = (!pru_sensors_data_get()->charge_complete) || (m_vrect_high_duration_ms >= PRU_CHARGE_COMPLETE_ADV_WAIT_TIME_MS);

                if(m_adv_allowed && !m_advertising)
                    m_advertising_start();
            }
            break;

        case PRU_SM_SIGNAL_SYSTEM_ERROR:
            m_adv_allowed = false;
            if(m_state_variables.current_state == PRU_SM_STATE_PRU_ON)
            {
                    err_code = pru_tmux_vout_enable_set(false);
                    APP_ERROR_CHECK(err_code);
                    ASSERT(m_advertising);
                    (void) sd_ble_gap_adv_stop(); // We do not care about the return code.
                    m_advertising = false;
            }
            break;

        case PRU_SM_SIGNAL_DISCONNECTED:
            if(m_state_variables.current_state == PRU_SM_STATE_SYSTEM_ERROR && pru_sensors_data_get()->vrect >= PRU_VRECT_UVLO && !m_advertising)
                m_advertising_start();
            break;

        default:
            break;
    }
}

/**@brief Forward events to the PRU state machine. Also handles the resulting request event.
 *
 * @param[in]   evt   PRU state machine event.
 */
static void m_on_sm_event(pru_sm_signal_type_t evt)
{
    uint32_t err_code;

    m_handle_event(evt);

    (void)pru_sm_execute(evt, &m_state_variables);

    if(m_state_variables.current_state != m_state_variables.prev_state)
    {
        switch(m_state_variables.current_state)
        {
            case PRU_SM_STATE_COUNT:
                break;

            case PRU_SM_STATE_NULL:
                if (m_advertising)
                {
                    (void) sd_ble_gap_adv_stop(); // We do not care about the return code.
                    m_advertising = false;
                }
                err_code = sd_ble_gap_disconnect(m_wpts.conn_handle, BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);

                if(err_code != NRF_ERROR_INVALID_STATE && err_code != BLE_ERROR_INVALID_CONN_HANDLE && err_code != BLE_ERROR_NOT_ENABLED)
                    APP_ERROR_CHECK(err_code);

                err_code = pru_tmux_vout_enable_set(false);
                APP_ERROR_CHECK(err_code);
                break;

            case PRU_SM_STATE_BOOT:
                err_code = pru_tmux_vout_enable_set(false);
                APP_ERROR_CHECK(err_code);

                m_vrect_high_duration_ms = 0;
                break;

            case PRU_SM_STATE_PRU_ON:
                err_code = pru_tmux_vout_enable_set(true);
                APP_ERROR_CHECK(err_code);
                break;

            case PRU_SM_STATE_SYSTEM_ERROR:
                err_code = pru_tmux_vout_enable_set(false);
                APP_ERROR_CHECK(err_code);
                break;
        }
    }

    // Notify application of state machine signal
    // To avoid application from setting softdevice in illegal state before we need it,
    // invoke at the end of the handling.
    if(m_app_sm_evt_handler != NULL)
        m_app_sm_evt_handler(evt, &m_state_variables);

}

/**@brief WPT Service event handler.
 *
 * @details This function handles events from the WPT Service.
 *
 * @param[in]   p_wpts  WPT Service context.
 * @param[in]   p_evt    Event.
 */
static void m_wpts_evt_handler(ble_wpts_t * p_wpts, ble_wpts_evt_t * p_evt)
{
    uint32_t err_code;

    switch(p_evt->evt_type)
    {
        case  BLE_WPTS_EVT_PRU_CONTROL_COMMAND:
            // Handle PRU Control Command
            if(PRU_CCP_ADJUST_POWER_CAPABILITY)
            {
                err_code = pru_tmux_adj_power_set(p_evt->data.pru_control.adj_power);
                APP_ERROR_CHECK(err_code);
            }
            m_on_sm_event(p_evt->data.pru_control.enable_pru_output ? PRU_SM_SIGNAL_CHARGE_START : PRU_SM_SIGNAL_CHARGE_STOP);
            break;

       default:
            break;
    }

#ifdef DEBUG_OUT_ENABLE
    wpts_test_evt_handle_cb(p_wpts, p_evt);
#endif //DEBUG_OUT_ENABLE
}


/**@brief BLE stack GAP event handler.
 *
 * @details This function handles BLE GAP events.
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 */
static void on_ble_evt(ble_evt_t * p_ble_evt)
{
    uint32_t err_code = NRF_SUCCESS;

    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_TIMEOUT:
            if (p_ble_evt->evt.gap_evt.params.timeout.src == BLE_GAP_TIMEOUT_SRC_ADVERTISING)
                m_advertising_start();
            break;

        case BLE_GATTS_EVT_SYS_ATTR_MISSING:
            err_code = sd_ble_gatts_sys_attr_set(m_conn_handle, NULL, 0, 0);
            APP_ERROR_CHECK(err_code);
            break;

        case BLE_GAP_EVT_CONNECTED:
            m_wpts.reboot_bit = 1;
            m_conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
            break;

        case BLE_GAP_EVT_DISCONNECTED:
            m_conn_handle = BLE_CONN_HANDLE_INVALID;
            m_advertising = false;
            m_on_sm_event(PRU_SM_SIGNAL_DISCONNECTED); // Treat disconnect in same way as power removed.
            break;

        case BLE_GAP_EVT_SEC_PARAMS_REQUEST:
        {
            ble_gap_sec_params_t sec_params;

            memset(&sec_params, 0, sizeof(ble_gap_sec_params_t));
            sec_params.bond = 0;                        // No bonding
            sec_params.io_caps = BLE_GAP_IO_CAPS_NONE;  // No IO capabilities
            sec_params.min_key_size = PRU_SEC_MIN_KEYSIZE; // Min allowed key size
            sec_params.max_key_size = PRU_SEC_MAX_KEYSIZE; // Max allowed key size
            sec_params.mitm = 0;                        // No man in the middle protection
            sec_params.oob = 0;                         // No out of band data available

            err_code = sd_ble_gap_sec_params_reply(p_ble_evt->evt.gap_evt.conn_handle, BLE_GAP_SEC_STATUS_SUCCESS, &sec_params, NULL);
        }
        break;

        case BLE_GAP_EVT_AUTH_STATUS:
            ASSERT(p_ble_evt->evt.gap_evt.params.auth_status.auth_status == BLE_GAP_SEC_STATUS_SUCCESS);
            break;

        default:
            break;
    }

    APP_ERROR_CHECK(err_code);
}

/**@brief Initialize the WPT Service.
 */
static void m_ble_wpts_init(void)
{
    uint32_t            err_code;
    ble_wpts_init_t    wpts_init;

    // Initialize PRU Service
    memset(&wpts_init, 0, sizeof(wpts_init));
    memset(&m_wpts, 0, sizeof(ble_wpts_t));

    wpts_init.evt_handler = m_wpts_evt_handler;

    // Here the sec level for the WPT Service can be changed/increased.
    PRU_SECURITY(&wpts_init.wpts_attr_md.read_perm);
    PRU_SECURITY(&wpts_init.wpts_attr_md.write_perm);

    // Configure PRU Static Parameters.
    wpts_init.pru_static.option_fields_r1_valid = PRU_DELTA_R1_VALID;
    wpts_init.pru_static.delta_r1 = PRU_DELTA_R1;
    wpts_init.pru_static.protocol_rev = PRU_PROTOCOL_REVISION;
    wpts_init.pru_static.pru_category = PRU_CATEGORY;
    wpts_init.pru_static.info_nfc_receiver = PRU_NFC_RECEIVER_SUPPORTED;
    wpts_init.pru_static.info_sep_btle_radio = PRU_SEPARATE_BTLE_RADIO_SUPORTED;
    wpts_init.pru_static.info_pc_alg_pref = PRU_POWER_CONTROL_ALG_PREFERENCE;
    wpts_init.pru_static.info_adj_power_capability = PRU_CCP_ADJUST_POWER_CAPABILITY;
    wpts_init.pru_static.info_ccp_connected_mode = PRU_CCP_CONNECT_MODE_SUPPORTED;
    wpts_init.pru_static.info_ptu_test_mode = BLE_WPTS_PRU_INFO_PTU_TEST_MODE_DISABLED_VAL;
    wpts_init.pru_static.hw_revision = PRU_HW_REVISION;
    wpts_init.pru_static.fw_revision = PRU_FW_REVISION;
    wpts_init.pru_static.prect_max = PRU_PRECT_MAX;
    wpts_init.pru_static.vrect_min_static = PRU_VRECT_MIN;
    wpts_init.pru_static.vrect_high_static = PRU_VRECT_HIGH;
    wpts_init.pru_static.vrect_set = PRU_VRECT_SET;

    err_code = ble_wpts_init(&m_wpts, &wpts_init);
    APP_ERROR_CHECK(err_code);
}


/**@brief Set PRU dynamic data.
 *
 * @details Updates the characteristic attributes with p_pru_sensor_data.
 *
 * @param[in]   p_wpts  WPT Service context.
 *
 * @param[in]   p_pru_sensor_data  PRU sensor data
 *
 * @returns NRF_SUCCESS/ERROR status code
 */
static uint32_t m_ble_wpts_pru_dynamic_set(ble_wpts_t * p_wpts, const pru_sensor_data_t * p_pru_sensor_data)
{
  pru_dynamic_t pru_dynamic;

    memset(&pru_dynamic,0,sizeof(pru_dynamic_t));

    pru_dynamic.optional_fields = ((1 << BLE_WPTS_PRU_DYNAMIC_OPTION_VOUT_BITPOS)          |
                                   (1 << BLE_WPTS_PRU_DYNAMIC_OPTION_IOUT_BITPOS)          |
                                   (1 << BLE_WPTS_PRU_DYNAMIC_OPTION_TEMPERATURE_BITPOS)   |
                                   (1 << BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_MIN_DYN_BITPOS) |
                                   (1 << BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_SET_DYN_BITPOS) |
                                   (1 << BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_HIGH_DYN_BITPOS));

    pru_dynamic.vrect                       = p_pru_sensor_data ->vrect;
    pru_dynamic.irect                       = p_pru_sensor_data ->irect;
    pru_dynamic.vout                        = p_pru_sensor_data ->vout;
    pru_dynamic.iout                        = p_pru_sensor_data ->iout;
    pru_dynamic.temperature                 = p_pru_sensor_data ->temperature;
    pru_dynamic.vrect_min_dyn               = PRU_VRECT_MIN;
    pru_dynamic.vrect_set_dyn               = PRU_VRECT_SET;
    pru_dynamic.vrect_high_dyn              = PRU_VRECT_HIGH;
    pru_dynamic.alerts.pru_over_voltage     = p_pru_sensor_data ->over_voltage;
    pru_dynamic.alerts.pru_over_current     = p_pru_sensor_data ->over_current;
    pru_dynamic.alerts.pru_over_temperature = p_pru_sensor_data ->over_temperature;
    pru_dynamic.alerts.charge_complete      = p_pru_sensor_data ->charge_complete;
    pru_dynamic.alerts.wired_charge_detect  = p_pru_sensor_data ->wired_charge_detect;
    pru_dynamic.alerts.pru_self_protection  = 0;
    pru_dynamic.adjust_power_response       = p_pru_sensor_data -> power_adjusted;
    pru_dynamic.pru_charge_port             = p_pru_sensor_data->charge_port;
    pru_dynamic.tester_command              = WPTS_PTU_TEST_CMD_NORMAL_OPERATION_POWER_CTL;

    return ble_wpts_pru_dynamic_set(p_wpts, &pru_dynamic);
}


/**@brief Sensors measurement timer timeout handler.
 *
 * @details This function will be called each time the sensors measurement timer expires.
 *
 * @param[in]   p_context   Pointer used for passing some arbitrary information (context) from the
 *                          app_start_timer() call to the timeout handler.
 */
static void m_sensors_timer_handler(void * p_context)
{
    uint32_t err_code;
    pru_alert_t alerts;
    bool new_alert_occured = false;

    UNUSED_PARAMETER(p_context);
    pru_sensors_read_all();

    err_code = m_ble_wpts_pru_dynamic_set(&m_wpts, pru_sensors_data_get());
    APP_ERROR_CHECK(err_code);
    err_code = pru_sensor_alerts_get(&new_alert_occured, &alerts);
    APP_ERROR_CHECK(err_code);
    if(new_alert_occured)
    {
        err_code = ble_wpts_alert_send(&m_wpts, &alerts);
        // If not disconnected or CCCD not yet received.
        if((err_code != BLE_ERROR_INVALID_CONN_HANDLE) && (err_code != NRF_ERROR_INVALID_STATE) && (err_code != BLE_ERROR_GATTS_SYS_ATTR_MISSING))
        {
            APP_ERROR_CHECK(err_code);
        }
    }
    pru_sensors_gen_evts();
}

/**@brief Start timer for reading sensors one time.
 */
static void m_sensors_timer_trigger()
{
    uint32_t err_code;
    err_code = app_timer_start(m_pru_sensors_timer_single_shot_id, APP_TIMER_MIN_TIMEOUT_TICKS, NULL);
    APP_ERROR_CHECK(err_code);
}

/**@brief Start timer for reading sensors periodically.
 */
static void m_sensors_timer_start(uint16_t period_ms)
{
    uint32_t err_code;
    err_code = app_timer_start(m_p_pru_sensors_timer_id, APP_TIMER_TICKS(period_ms, APP_TIMER_PRESCALER), NULL);
    APP_ERROR_CHECK(err_code);
}

/**@brief Instanicate timers used in this file.
 */
static void m_timers_init(void)
{
    uint32_t err_code;

    err_code = app_timer_create(&m_p_pru_sensors_timer_id, APP_TIMER_MODE_REPEATED, m_sensors_timer_handler);
    APP_ERROR_CHECK(err_code);
    err_code = app_timer_create(&m_pru_sensors_timer_single_shot_id, APP_TIMER_MODE_SINGLE_SHOT, m_sensors_timer_handler);

    APP_ERROR_CHECK(err_code);
}

/** @} */

/******************************************************************************/
/** @name API implementation
 *  @{ */
/******************************************************************************/

void pru_init(app_sm_evt_handler_t sm_handler)
{
    m_app_sm_evt_handler = sm_handler;
    m_timers_init();
    pru_sensors_init(m_on_sm_event);
}

void pru_start(void)
{
    uint32_t err_code;

    m_conn_handle = BLE_CONN_HANDLE_INVALID;

    m_gap_params_init();
    m_ble_wpts_init();

    m_wpts.reboot_bit = 0;         // We have just rebooted

    err_code = tmux_init();    // Initializes HAL as well
    APP_ERROR_CHECK(err_code);

    pru_sensors_read_all();

    err_code = m_ble_wpts_pru_dynamic_set(&m_wpts, pru_sensors_data_get());
    APP_ERROR_CHECK(err_code);

    m_sensors_timer_start(PRU_SENSORS_TIMER_PERIOD_MS);
    m_sensors_timer_trigger();
}

void pru_on_ble_evt(ble_evt_t * p_ble_evt)
{
    on_ble_evt(p_ble_evt);
    ble_wpts_on_ble_evt(&m_wpts, p_ble_evt);
}

#ifdef DFU_SUPPORT

void terminate(void)
{
    // No action needed to tear down the PRU application.
}

bool dfu_check(void)
{
    const pru_sensor_data_t * sensor_data;
    sensor_data = pru_sensors_data_get();
    return sensor_data->vrect == 0;
}

#endif // DFU_SUPPORT
/** @} */
