/* Copyright (c) Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

 /** @file */

#include "pru_adv.h"
#include "ble_advdata.h"
#include "pru_sensors.h"
#include "ble_wpts.h"
#include "pru_hw_config.h"
#include "pru_config.h"

/******************************************************************************/
/** @name API implementation
 *  @{ */
/******************************************************************************/

uint32_t pru_advdata_set(ble_wpts_t * p_wpts)
{
    uint32_t        err_code;
    ble_advdata_t   advdata;
    uint8_t         flags = BLE_GAP_ADV_FLAGS_LE_ONLY_LIMITED_DISC_MODE;
    uint8_t         service_data_array[BLE_WPTS_ADV_SERVICE_DATA_LEN];
    bool            ovp_status;
    const pru_sensor_data_t * p_pru_sensor_data;

    ble_advdata_service_data_t service_data;

    // Build and set advertising data
    memset(&advdata, 0, sizeof(advdata));

    advdata.name_type               = BLE_ADVDATA_NO_NAME;
    advdata.include_appearance      = false;
    advdata.flags                   = flags;
    advdata.service_data_count      = 1;
    advdata.p_service_data_array    = &service_data;

    // Set Service data
    service_data.service_uuid = BLE_WPTS_UUID16;
    service_data.data.size    = BLE_WPTS_ADV_SERVICE_DATA_LEN;
    service_data.data.p_data  = service_data_array;
    memset(service_data_array, 0, BLE_WPTS_ADV_SERVICE_DATA_LEN);

    // Set GATT primary service handle.
    (void)uint16_encode(p_wpts->service_handle, &service_data_array[BLE_WPTS_ADV_SERVICE_HANDLE_POS]);

    // Set PRU RSSI parameters. We do not check whether the values are in bounds.
    // Set PRU antenna gain field.
    if (PRU_ANTENNA_GAIN_KNOWN)
    {
        BLE_WPTS_BITFIELD_WRITE(service_data_array[BLE_WPTS_ADV_PRU_RSSI_PARAMS_POS], \
            BLE_WPTS_ADV_PRU_RSSI_PRU_GAIN_BITMSK, \
            BLE_WPTS_ADV_PRU_RSSI_PRU_GAIN_BITPOS, \
            (uint8_t)(BLE_WPTS_ADV_ANTENNA_GAIN_OFFSET + PRU_ANTENNA_GAIN));
    }
    else
    {
        BLE_WPTS_BITFIELD_WRITE(service_data_array[BLE_WPTS_ADV_PRU_RSSI_PARAMS_POS], \
            BLE_WPTS_ADV_PRU_RSSI_PRU_GAIN_BITMSK, \
            BLE_WPTS_ADV_PRU_RSSI_PRU_GAIN_BITPOS, \
            BLE_WPTS_ADV_ANTENNA_GAIN_UNKNOWN_VAL);
    }

    // Set PRU TX power field.
    if (PRU_TX_POWER_KNOWN)
    {
        BLE_WPTS_BITFIELD_WRITE(service_data_array[BLE_WPTS_ADV_PRU_RSSI_PARAMS_POS], \
            BLE_WPTS_ADV_PRU_RSSI_PRU_PWR_BITMSK, \
            BLE_WPTS_ADV_PRU_RSSI_PRU_PWR_BITPOS,
            (uint8_t)(BLE_WPTS_ADV_PRU_TX_POWER_LEVEL_OFFSET + PRU_TX_POWER_LEVEL));
    }
    else
    {
        BLE_WPTS_BITFIELD_WRITE(service_data_array[BLE_WPTS_ADV_PRU_RSSI_PARAMS_POS], \
            BLE_WPTS_ADV_PRU_RSSI_PRU_PWR_BITMSK, \
            BLE_WPTS_ADV_PRU_RSSI_PRU_PWR_BITPOS,
            BLE_WPTS_ADV_POWER_UNKNOWN_VAL);
    }

    // Get OVP status
    p_pru_sensor_data = pru_sensors_data_get();
    ovp_status = p_pru_sensor_data->over_voltage;

    // Set advertising flags
    BLE_WPTS_BITFIELD_WRITE(service_data_array[BLE_WPTS_ADV_FLAGS_POS], BLE_WPTS_ADV_FLAG_IMPEDANCE_SHIFT_BITMSK, BLE_WPTS_ADV_FLAG_IMPEDANCE_SHIFT_BITPOS, PRU_CATEGORY);
    BLE_WPTS_BITFIELD_WRITE(service_data_array[BLE_WPTS_ADV_FLAGS_POS], BLE_WPTS_ADV_FLAG_REBOOT_INDICATOR_BITMSK, BLE_WPTS_ADV_FLAG_REBOOT_INDICATOR_BITPOS, p_wpts->reboot_bit);
    BLE_WPTS_BITFIELD_WRITE(service_data_array[BLE_WPTS_ADV_FLAGS_POS], BLE_WPTS_ADV_FLAG_OVP_STATUS_BITMSK, BLE_WPTS_ADV_FLAG_OVP_STATUS_BITPOS, ovp_status);
    BLE_WPTS_BITFIELD_WRITE(service_data_array[BLE_WPTS_ADV_FLAGS_POS], BLE_WPTS_ADV_FLAG_TIME_SET_SUPPORT_BITMSK, BLE_WPTS_ADV_FLAG_TIME_SET_SUPPORT_BITPOS, PRU_TIME_SET_SUPPORT);

    err_code = ble_advdata_set(&advdata, NULL);
    return err_code;
}

/** @} */
