/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file */

#include <string.h>
#include <stdio.h>

#include "simple_uart.h"
#include "app_print.h"
#include "wpt.h"
#include "pru_sm.h"
#include "pru_hw_config.h"
#include "pru_test_mux.h"
#include "debug.h"
#include "ble_wpts.h"
#include "pru.h"
#include "ble.h"
#include "nrf_soc.h"
#include "softdevice_handler.h"
#include "nrf_sdm.h"
#include "pru_sensors.h"
#include "bsp.h"

#define EVT_STRING_LENGTH_MAX                  20        /**< Max length of event message string */
#define DEBUG_LINE_LENGTH                      80        /**< Width of debug output table */
#define VRECT_INCREMENT                       500        /**< Value by which VRECT is incremented/decremented for each call to the simulated measurement function. */
#define IRECT_INCREMENT                       100        /**< Value by which IRECT is incremented/decremented for each call to the simulated measurement function. */
#define VOUT_INCREMENT                        500        /**< Value by which VOUT is incremented/decremented for each call to the simulated measurement function. */
#define IOUT_INCREMENT                        100        /**< Value by which IOUT is incremented/decremented for each call to the simulated measurement function. */
#define TEMPERATURE_INCREMENT                   5        /**< Value by which the Temperature is incremented/decremented for each call to the simulated measurement function. */

#define DEBUG_CONN_HANDLE                       0        /**< Assumed Connection handle during debug. */
#define DEBUG_ALERT_VALUE_HANDLE             0x0f        /**< Assumed handle of Alert characteristic value during debug. */

/******************************************************************************/
/** @name Static functions and variables
 *  @{ */
/******************************************************************************/

/**@struct debug_output_t
 * @brief Collection of debug output
 */
typedef struct
{
    pru_sm_state_t state;
    pru_sensor_data_t sensor_data;
    pru_control_t control_data;
} debug_output_t;

static debug_output_t m_debug_output;                        /**< Debug output collection */
static char m_evt_evt_string[EVT_STRING_LENGTH_MAX] = "";    /**< Event message string */
static pru_tmux_override_t override_values;             /**< Local copy of override values */
static pru_control_t m_latest_control_data_received;    /**< Local latest received control data */
static bool m_uart_initialized = false;                      /**< UART initalized flag. Used to initialize UART on first scharacter sent. */
static bool m_transmission_stopped = false;                  /**< Is transmission stopped? */
static const ble_gap_addr_t MODE_TRANS_ADDRESS =
            {.addr_type = BLE_GAP_ADDR_TYPE_RANDOM_STATIC,
             .addr = {0xee, 0xee, 0xff, 0xff, 0xee, 0xee}};  /**< Address to use in Mode Transition if address is non-0 */




/**@brief Override weak wpts_test_evt_handle_cb()
 * @param p_wpts WPT Service
 * @param p_evt WPT Service event
 */
void wpts_test_evt_handle_cb(ble_wpts_t * p_wpts, ble_wpts_evt_t * p_evt)
{
    if(p_evt->evt_type == BLE_WPTS_EVT_PRU_CONTROL_COMMAND)
    {
        memcpy(&m_latest_control_data_received, &p_evt->data.pru_control, sizeof(pru_control_t));
    }
}

/**@brief Function for processing board specific test commands.
 *
 * @param[in]  cmd   Command code to process.

 * @retval true Command code was processed by this function.
 * @retval false Command code was not processed by this function.
 *
 */
bool pru_process_command_board(uint8_t cmd);


/** @brief Check if there are any characters in the UART receive buffer
 *
 * This function returns the true if there is a character available for reading
 * in the UART receive buffer.
 *
 * @retval true There is a character available for reading
 * @retval false there is not a character available for reading
 */
static bool m_uart_chars_available(void)
{
  return (NRF_UART0->EVENTS_RXDRDY == 1);
}

/**@brief Send Mode Transition alert to PTU
 *
 * @param include_address Should the address be non-null?
 * @param mode_trans_val Value to use in Mode Transition Value field of PRU alert.
 */
static void m_send_mode_transition_alert(uint8_t mode_trans_val, bool include_address)
{
    uint32_t err_code;
    uint8_t alert[BLE_WPTS_PRU_ALERT_MAX_LEN];
    uint16_t hvx_len = BLE_WPTS_PRU_ALERT_MAX_LEN;
    ble_gatts_hvx_params_t params;

    memset(alert, 0, BLE_WPTS_PRU_ALERT_MAX_LEN);
    params.offset = 0;
    params.handle = DEBUG_ALERT_VALUE_HANDLE;
    params.type = BLE_GATT_HVX_INDICATION;
    params.p_data = alert;
    params.p_len = &hvx_len;

    BLE_WPTS_BITFIELD_WRITE(params.p_data[BLE_WPTS_PRU_ALERT_POS],\
                        BLE_WPTS_PRU_ALERT_MODE_TRANSITION_BITMSK, \
                        BLE_WPTS_PRU_ALERT_MODE_TRANSITION_BITPOS, \
                        mode_trans_val);

    if(include_address)
    {
        memcpy(&params.p_data[1], MODE_TRANS_ADDRESS.addr, 6); // Copy mode transition address

        err_code = sd_ble_gap_address_set(BLE_GAP_ADDR_CYCLE_MODE_NONE, &MODE_TRANS_ADDRESS);
        APP_ERROR_CHECK(err_code);
    }


    else
        memset(&params.p_data[1], 0, 6);


    err_code = sd_ble_gatts_hvx(DEBUG_CONN_HANDLE, &params);
    APP_ERROR_CHECK(err_code);
}


/**@brief Function for processing common test commands.
 *
 * @param[in]  cmd   Command code to process.

 * @retval true Command code was processed by this function.
 * @retval false Command code was not processed by this function.
 *
 */
static bool m_pru_process_command_common(uint8_t cmd)
{
    bool b_cmd_processed = true;
    uint32_t err_code;

    pru_tmux_override_values_get(&override_values);

    switch (cmd)
    {
        case 'q':
            print_string("Command: Power adjusted bit set");
            override_values.power_adjusted = true;
            print_string("\n\r");
            break;

        case 'a':
            print_string("Command: Power adjusted bit cleared");
            override_values.power_adjusted = false;
            print_string("\n\r");
            break;

        case 'w':
            print_string("Command: Inc IOUT -> ");
            override_values.iout += IOUT_INCREMENT;
            print_hex(override_values.iout, 2);
            print_string("\n\r");
            break;

        case 's':
            print_string("Command: Dec IOUT -> ");
            override_values.iout -= IOUT_INCREMENT;
            print_hex(override_values.iout, 2);
            print_string("\n\r");
            break;

        case 'r':
            print_string("Command: Inc VOUT -> ");
            override_values.vout += VOUT_INCREMENT;
            print_hex(override_values.vout, 2);
            print_string("\n\r");
            break;

        case 'f':
            print_string("Command: Dec VOUT -> ");
            override_values.vout -= VOUT_INCREMENT;
            print_hex(override_values.vout, 2);
            print_string("\n\r");
            break;

        case 't':
            print_string("Command: Wired charger on\n\r");
            override_values.wired_charge_detect = true;
            break;

        case 'g':
            print_string("Command: Wired charger off\n\r");
            override_values.wired_charge_detect = false;
            break;

        case 'y':
            print_string("Command: Inc VRECT -> ");
            override_values.vrect += VRECT_INCREMENT;
            print_hex(override_values.vrect, 2);
            print_string("\n\r");
            break;

        case 'h':
            print_string("Command: Dec VRECT -> ");
            override_values.vrect -= VRECT_INCREMENT;
            print_hex(override_values.vrect, 2);
            print_string("\n\r");
            break;

        case 'u':
            print_string("Command: Inc IRECT -> ");
            override_values.irect += IRECT_INCREMENT;
            print_hex(override_values.irect, 2);
            print_string("\n\r");
            break;

        case 'j':
            print_string("Command: Dec IRECT -> ");
            override_values.irect -= IRECT_INCREMENT;
            print_hex(override_values.irect, 2);
            print_string("\n\r");
            break;

        case 'i':
            print_string("Command: Inc TEMP -> ");
            override_values.temperature += TEMPERATURE_INCREMENT;
            print_hex(override_values.temperature, 2);
            print_string("\n\r");
            break;

        case 'k':
            print_string("Command: Dec TEMP -> ");
            override_values.temperature -= TEMPERATURE_INCREMENT;
            print_hex(override_values.temperature, 2);
            print_string("\n\r");
            break;

        case 'o':
            print_string("Command: Test-mode enabled\n\r");
            tmux_get_select(TMUX_SELECT_OVERRIDE);
            tmux_set_select(TMUX_SELECT_OVERRIDE);
            break;

        case 'p':
            print_string("SDK Version ");
            print_string(SDK_VERSION);
            print_string("\n\r");
            break;

        case 'l':
            print_string("Command: Test-mode disabled\n\r");
            tmux_get_select(TMUX_SELECT_NO_OVERRIDE);
            tmux_set_select(TMUX_SELECT_NO_OVERRIDE);
            break;

        case 'z':
            print_string("Command: Reset\r\n");
            NVIC_SystemReset();
            break;

        case 'x':
            print_string("Command: Send Slave Security Request\n\r");

            ble_gap_sec_params_t sec_params;
            memset(&sec_params, 0, sizeof(ble_gap_sec_params_t));
            sec_params.mitm = 0;
            sec_params.bond = 0;

            err_code = sd_ble_gap_authenticate(DEBUG_CONN_HANDLE, &sec_params);
            APP_ERROR_CHECK(err_code);

            break;

        case 'n':
            print_string("Command: Charge complete\n\r");
            override_values.charge_complete = true;
            break;

        case 'm':
            print_string("Command: Charge incomplete\n\r");
            override_values.charge_complete = false;
            break;

        case '1':
            print_string("Command: Vrect to PRU_VRECT_UVLO\n\r");
            override_values.vrect = PRU_VRECT_UVLO;
            break;

        case '2':
            print_string("Command: Vrect to PRU_VRECT_BOOT\n\r");
            override_values.vrect =  PRU_VRECT_BOOT;
            break;

        case '3':
            print_string("Command: Vrect to PRU_VRECT_MIN\n\r");
            override_values.vrect  = PRU_VRECT_MIN;
            break;

        case '4':
            print_string("Command: Vrect to PRU_VRECT_HIGH\n\r");
            override_values.vrect = (PRU_VRECT_HIGH);
            break;

        case '5':
            print_string("Command: Vrect to PRU_VRECT_MAX\n\r");
            override_values.vrect = (PRU_VRECT_MAX);
            break;

        case 'e':
            m_send_mode_transition_alert(BLE_WPTS_PRU_ALERT_MODE_3S_MODE_TRANSITION_VAL, true);
            break;

        case 'd':
            m_send_mode_transition_alert(BLE_WPTS_PRU_ALERT_MODE_3S_MODE_TRANSITION_VAL, false);
            break;

        case 'E':
            m_send_mode_transition_alert(BLE_WPTS_PRU_ALERT_MODE_6S_MODE_TRANSITION_VAL, true);
            break;

        case 'D':
            m_send_mode_transition_alert(BLE_WPTS_PRU_ALERT_MODE_6S_MODE_TRANSITION_VAL, false);
            break;

        case 'v':
            if(!m_transmission_stopped)
            {
                print_string("Command: Stop packet transmission\n\r");
                m_transmission_stopped = !m_transmission_stopped; // Need to do this before disabling the softdevice
                (void)sd_softdevice_disable();
            }

            else
            {
                print_string("Command: Start packet transmission\n\r");
                m_transmission_stopped = !m_transmission_stopped;
                nrf_clock_lf_cfg_t clock_lf_cfg = NRF_CLOCK_LFCLKSRC;
                SOFTDEVICE_HANDLER_INIT(&clock_lf_cfg, NULL);
            }
            break;
        case '6':
            print_string("\n\rpru\n\r");
            break;
        case '7':
        case '8':
        #ifdef DFU_SUPPORT
            print_string("Entering DFU mode ...\n\r");
            dfu_start();
            break;
        #endif // DFU_SUPPORT

        case '9':
        case 'c':
        case '0':
        case ',':
        case '.':
        default:
            b_cmd_processed = false;
            break;
    }

    pru_tmux_override_values_set(&override_values);

    return b_cmd_processed;
}

/**@brief Get and process uart commands
 */
static void sm_uart_rx_poll(void)
{
    uint8_t cmd;
    bool b_cmd_processed;

    if (!m_uart_chars_available())
    {
        return;
    }

    cmd = simple_uart_get();

    #ifdef PRU_TEST_HW_UART_CMD_ENABLE
    b_cmd_processed = pru_process_command_board(cmd);
    if (!b_cmd_processed) b_cmd_processed = m_pru_process_command_common(cmd);
    #else
    b_cmd_processed = m_pru_process_command_common(cmd);
    #endif

    if (!b_cmd_processed)
    {
        print_string("\n\rCommand not implemented\n\r");
    }
}

/**@brief Print a DEBUG_LINE_LENGTH character long "horizontal line"
 * to UART.
 */
static void line_print(void)
{
    int i;

    for(i = 0; i < DEBUG_LINE_LENGTH; i++)
    {
        print_string("-");
    }
    print_string("\r\n");
}

/**@brief Update m_debug_output variable.
 * @return true if debug output has changed.
 */
static bool debug_output_changed(void)
{
    const pru_sensor_data_t * sensor_data;
    bool changed = false;

    sensor_data = pru_sensors_data_get();

    if(m_debug_output.state != pru_sm_execute(PRU_SM_SIGNAL_NULL, NULL))
    {
        changed = true;
        m_debug_output.state = pru_sm_execute(PRU_SM_SIGNAL_NULL, NULL);
    }

    if(memcmp(&m_debug_output.sensor_data, sensor_data, sizeof(pru_sensor_data_t)) != 0)
    {
        changed = true;
        memcpy(&m_debug_output.sensor_data, sensor_data, sizeof(pru_sensor_data_t));
    }

    if(memcmp(&m_debug_output.control_data, &m_latest_control_data_received, sizeof(pru_control_t)))
    {
        changed = true;
        memcpy(&m_debug_output.control_data, &m_latest_control_data_received, sizeof(pru_control_t));
    }

    if(strlen(m_evt_evt_string) > 0)
    {
        changed = true;
    }

    return changed;
}

/**@brief Convert state enum to string.
 * @param state The state enum to convert
 * @return Pointer to char.
 */
static char * m_state_string_get(pru_sm_state_t state)
{
    switch(state)
    {
        case PRU_SM_STATE_NULL:
            return "PRU Null";
        case PRU_SM_STATE_BOOT:
            return "PRU Boot";
        case PRU_SM_STATE_PRU_ON:
            return "PRU On";
        case PRU_SM_STATE_SYSTEM_ERROR:
            return "PRU System Er-or";
        default:
            return "Invalid state";
    }

}

/**@brief Print contents of m_debug_output to UART.
 */
static void debug_output_print(void)
{
    char temp_string[DEBUG_LINE_LENGTH];                    /**< Debug output collection */

    // Status
    line_print();
    print_string("StatusStart\r\n");
    line_print();
    print_string("State\r\n");
    sprintf(temp_string, "%s\r\n", m_state_string_get(m_debug_output.state));
    print_string(temp_string);

    // Dynamic parameters
    line_print();
    print_string("Dynamic parameters\r\n");
    line_print();
    print_string("Vrect\tIrect\tVout\tIout\tTemp\tCharged\tP_adj\r\n");
    sprintf(temp_string, "%u\t%u\t%u\t%u\t%d\t%u\t%u\r\n",
            m_debug_output.sensor_data.vrect,
            m_debug_output.sensor_data.irect,
            m_debug_output.sensor_data.vout,
            m_debug_output.sensor_data.iout,
            m_debug_output.sensor_data.temperature,
            m_debug_output.sensor_data.charge_complete,
            m_debug_output.sensor_data.power_adjusted);

    print_string(temp_string);

    // Static parameters
    line_print();
    print_string("Static parameters\r\n");
    line_print();
    print_string("Vrset\tVrmin\tVrhigh\tTempmax\tVrmax\tPrmax\r\n");
    sprintf(temp_string, "%u\t%u\t%u\t%u\t%u\t%u\r\n",
            PRU_VRECT_SET,
            PRU_VRECT_MIN,
            PRU_VRECT_HIGH,
            PRU_TEMP_MAX,
            PRU_VRECT_MAX,
            PRU_PRECT_MAX);
    print_string(temp_string);

    // Previous received control data
    line_print();
    print_string("Control packet\r\n");
    line_print();

    print_string("En_out\tEn_Chin\tAdj_pow\tPerm\tTset\r\n");
    sprintf(temp_string, "%u\t%u\t%u\t%u\t%u\r\n",
            m_debug_output.control_data.enable_pru_output,
            m_debug_output.control_data.enable_pru_charge_indicator,
            m_debug_output.control_data.adj_power,
            m_debug_output.control_data.permissions,
            m_debug_output.control_data.time_set);
    print_string(temp_string);

    // Print event message
    if(strlen(m_evt_evt_string) > 0)
    {
        print_string("\r\n");
        print_string(m_evt_evt_string);
        sprintf(m_evt_evt_string, "");
    }

    print_string("StatusEnd\r\n");
}

/** @} */

/******************************************************************************/
/** @name API implementation
 *  @{ */
/******************************************************************************/

void debug(void)
{
    if (!m_uart_initialized)
    {
        simple_uart_config(UART_BAUDRATE_BAUDRATE_Baud460800, RTS_PIN_NUMBER, TX_PIN_NUMBER, CTS_PIN_NUMBER, RX_PIN_NUMBER, HWFC);
        m_uart_initialized = true;
    }
    if(debug_output_changed())
    {
        debug_output_print();
    }
    sm_uart_rx_poll();
}

void debug_add_string(const char *s)
{
    sprintf(m_evt_evt_string, "%s", s);
}

/** @} */
