/* Copyright (c) 2012 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file */

#ifdef PRU_PSTORAGE_ENABLE

#include "pru_pstorage.h"
#include "pru.h"
#include "nrf_error.h"
#include "nrf_soc.h"
#include "string.h"

#define PRU_PSTORAGE_EMPTY_FLASH_PATTERN 0xffffffff                        /**< Empty flash pattern */
#define PRU_PSTORAGE_FLASH_PAGE_SIZE ((uint16_t)NRF_FICR->CODEPAGESIZE)    /**< Page size in bytes */
#define PRU_PSTORAGE_FLASH_START_PAGE (NRF_FICR->CODESIZE - 4)             /**< Start page for PRU PSTORAGE */
#define PRU_PSTORAGE_FLASH_START_ADDR \
    (PRU_PSTORAGE_FLASH_PAGE_SIZE * PRU_PSTORAGE_FLASH_START_PAGE)    /**< Start address of PRU PSTORAGE flash */

/******************************************************************************/
/** @name Static functions and variables
 *  @{ */
/******************************************************************************/

/**@brief Pstorage state */
typedef enum {
    PRU_PSTORAGE_STATE_IDLE,
    PRU_PSTORAGE_STATE_ERASE,
    PRU_PSTORAGE_STATE_WRITE
} pru_pstorage_state;

static volatile bool m_operation_success;                                   /**< Was the last operation successful? */
static volatile pru_pstorage_state m_state;                            /**< Current state of this module */

/** @} */

/******************************************************************************/
/** @name API implementation
 *  @{ */
/******************************************************************************/

void pru_pstorage_sys_evt_handler(uint32_t event)
{
    m_state = PRU_PSTORAGE_STATE_IDLE;
    switch(event)
    {
        case NRF_EVT_FLASH_OPERATION_ERROR:
            m_operation_success = false;
            break;

        case NRF_EVT_FLASH_OPERATION_SUCCESS:
            m_operation_success = true;
           break;

        default:
            break;
    }
}

void pru_pstorage_write(pru_pstorage_appdata_t* p_appdata)
{
    uint32_t err_code = NRF_ERROR_BUSY;
    m_operation_success = false;

    if(p_appdata == NULL)
    {
        APP_ERROR_HANDLER(NRF_ERROR_INVALID_PARAM);
        return;
    }

    // Wait for pending PSTORAGE action to complete
    while(m_state != PRU_PSTORAGE_STATE_IDLE)
        ;

    while(err_code == NRF_ERROR_BUSY)
    {
        m_state = PRU_PSTORAGE_STATE_WRITE;
        err_code  = sd_flash_write((uint32_t*) PRU_PSTORAGE_FLASH_START_ADDR,
                                  (uint32_t *) p_appdata,
                                  sizeof(pru_pstorage_appdata_t) / sizeof(uint32_t));
    }
    APP_ERROR_CHECK(err_code);
}

bool pru_pstorage_read(pru_pstorage_appdata_t* p_appdata)
{
    if(p_appdata != NULL)
        memcpy(p_appdata, (void*) PRU_PSTORAGE_FLASH_START_ADDR, sizeof(pru_pstorage_appdata_t));

    return p_appdata != NULL && (*(uint32_t *) p_appdata != PRU_PSTORAGE_EMPTY_FLASH_PATTERN);
}

void pru_pstorage_erase_appdata(void)
    {
    uint32_t err_code = NRF_ERROR_BUSY;

    m_state = PRU_PSTORAGE_STATE_ERASE;

    while(err_code == NRF_ERROR_BUSY)
    {
        err_code = sd_flash_page_erase(PRU_PSTORAGE_FLASH_START_PAGE);
    }

    APP_ERROR_CHECK(err_code);
}

void pru_pstorage_status(bool * p_complete, bool * p_success)
{
    *p_complete = (m_state == PRU_PSTORAGE_STATE_IDLE);
    *p_success = m_operation_success;
}

/** @} */

#endif
