/* Copyright (c) 2012 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file */

#include <stdint.h>
#include <string.h>

#include "pru_sensors.h"
#include "pru_hw_config.h"
#include "pru_test_mux.h"
#include "nrf_error.h"
#include "nrf_assert.h"


/******************************************************************************/
/** @name Static functions and variables
 *  @{ */
/******************************************************************************/

static pru_sensor_data_t   m_pru_sensor_data;                         /**< The latest version of the sensor data */
static uint32_t                 m_hv_counter = 0;                          /**< Counter for how many sensor measurements have ticked over [ms]. */
static pru_alert_t         m_prev_alerts;                             /**< Previously returned alerts from pru_sensor_alerts_get(). */
static pru_sm_handler_t    m_signal_handler;                          /**< State machine event handler. */

/** @} */

/******************************************************************************/
/** @name API implementation
 *  @{ */
/******************************************************************************/

void pru_sensors_init(pru_sm_handler_t sm_handler)
{
    // Initialize 'static' PRU dynamic values
    memset(&m_pru_sensor_data,0,sizeof(m_pru_sensor_data));
    memset(&m_prev_alerts, 0, sizeof(pru_alert_t));
    m_signal_handler = sm_handler;
    m_hv_counter = 0;
}

void pru_sensors_read_all(void)
{
#ifndef STATIC_POWER_ADJUSTED_BIT
    static uint16_t nof_calls_since_power_adjusted_set = 0;
#endif
    uint32_t err_code = NRF_SUCCESS;
    uint16_t vrect, irect, vout, iout;
    int16_t temperature;
    bool wired;
    bool charge_port;
    bool charge_complete;
    bool last_power_adjusted;

    // Update sensor values if reading was successful.
    err_code =  pru_tmux_vrect_get(&vrect);
    APP_ERROR_CHECK(err_code);
    m_pru_sensor_data.vrect = vrect;

    err_code = pru_tmux_irect_get(&irect);
    APP_ERROR_CHECK(err_code);
    m_pru_sensor_data.irect = irect;

    err_code = pru_tmux_vout_get(&vout);
    APP_ERROR_CHECK(err_code);
    m_pru_sensor_data.vout = vout;

    err_code = pru_tmux_iout_get(&iout);
    APP_ERROR_CHECK(err_code);
    m_pru_sensor_data.iout = iout;

    err_code = pru_tmux_temperature_get(&temperature);
    APP_ERROR_CHECK(err_code);
    m_pru_sensor_data.temperature = temperature;

    err_code = pru_tmux_wired_charge_detect_status_get(&wired);
    APP_ERROR_CHECK(err_code);
    m_pru_sensor_data.wired_charge_detect = wired;

    err_code = pru_tmux_charge_complete_status_get(&charge_complete);
    APP_ERROR_CHECK(err_code);
    m_pru_sensor_data.charge_complete = charge_complete;

    err_code = pru_tmux_vout_enable_get(&charge_port);
    APP_ERROR_CHECK(err_code);
    m_pru_sensor_data.charge_port = charge_port;

    err_code = pru_tmux_power_adjusted_get(&last_power_adjusted);
    APP_ERROR_CHECK(err_code);

#ifndef STATIC_POWER_ADJUSTED_BIT
    if(m_pru_sensor_data.power_adjusted)
    {
        nof_calls_since_power_adjusted_set++;

         // See if we need to reset the 'Adjusted Power Response' value.
        if((PRU_SENSORS_TIMER_PERIOD_MS * nof_calls_since_power_adjusted_set) >= PRU_RESET_POWER_ADJUSTED_MS)
        {
            m_pru_sensor_data.power_adjusted = false;
            nof_calls_since_power_adjusted_set = 0;
        }
    }
    else if(last_power_adjusted)
    {
        // The HAL has reported that power has been adjusted. Reset the HAL value, and start the 1 second reset functionality
        // for the 'Adjusted Power Response' bit.
        err_code = pru_tmux_power_adjusted_set(false);
        APP_ERROR_CHECK(err_code);

        m_pru_sensor_data.power_adjusted = true;
        nof_calls_since_power_adjusted_set = 0;
    }
#else
    m_pru_sensor_data.power_adjusted  = last_power_adjusted;
#endif

    m_pru_sensor_data.over_voltage = (m_pru_sensor_data.vrect > PRU_VRECT_MAX) ? 1:0;
    m_pru_sensor_data.over_current = (((uint32_t)m_pru_sensor_data.vrect*m_pru_sensor_data.irect/(100*1000)) >= PRU_PRECT_MAX) ? 1:0;
    m_pru_sensor_data.over_temperature = (m_pru_sensor_data.temperature >= (int16_t)PRU_TEMP_MAX) ? 1: 0;

    if(m_pru_sensor_data.vrect > PRU_VRECT_HIGH)
    {
        m_hv_counter += PRU_SENSORS_TIMER_PERIOD_MS;
    }
    else
    {
        m_hv_counter = 0;
    }
    m_pru_sensor_data.sustained_high_voltage = (m_hv_counter > PRU_VRECT_HIGH_SUSTAIN_MS) ? 1:0;

    #ifdef USE_BOARD_SPECIFIC_SENSORS
    err_code = pru_sensors_board_read_all(&m_pru_sensor_data);
    APP_ERROR_CHECK(err_code);
    #endif
}

const pru_sensor_data_t * pru_sensors_data_get(void)
{
   return &m_pru_sensor_data;
}

uint32_t pru_sensor_alerts_get(bool * new_alert_occured, pru_alert_t * p_pru_alert)
{
    ASSERT(p_pru_alert != NULL);
    ASSERT(new_alert_occured != NULL);

    if((p_pru_alert != NULL) && (new_alert_occured != NULL))
    {
        memset(p_pru_alert, 0, sizeof(pru_alert_t));
        p_pru_alert->alerts.pru_over_voltage        = m_pru_sensor_data.over_voltage;
        p_pru_alert->alerts.pru_over_current        = m_pru_sensor_data.over_current;
        p_pru_alert->alerts.pru_over_temperature    = m_pru_sensor_data.over_temperature;
        p_pru_alert->alerts.charge_complete         = m_pru_sensor_data.charge_complete;
        p_pru_alert->alerts.wired_charge_detect     = m_pru_sensor_data.wired_charge_detect;
        p_pru_alert->alerts.pru_self_protection     = BLE_WPTS_PRU_ALERT_NO_SELF_PROTECTION_VAL;

        if((p_pru_alert->alerts.pru_over_voltage ||
           p_pru_alert->alerts.pru_over_current ||
           p_pru_alert->alerts.pru_over_temperature ||
           p_pru_alert->alerts.charge_complete ||
           p_pru_alert->alerts.wired_charge_detect) &&
           (memcmp(&m_prev_alerts, p_pru_alert, sizeof(pru_alert_t) != 0)))
        {
            *new_alert_occured = true;
        }
        else
        {
            *new_alert_occured = false;
        }

        memcpy(&m_prev_alerts, p_pru_alert, sizeof(pru_alert_t));

        return NRF_SUCCESS;
    }
    else
    {
        return NRF_ERROR_NULL;
    }
}

void pru_sensors_gen_evts(void)
{
    if(m_signal_handler != NULL)
    {
        const pru_sensor_data_t * p_sensor_data = pru_sensors_data_get();
        bool system_error = (p_sensor_data->over_voltage || p_sensor_data->over_current || p_sensor_data->over_temperature);

        // System Error or System OK Event
        if(system_error)
            m_signal_handler(PRU_SM_SIGNAL_SYSTEM_ERROR);

        if(p_sensor_data->vrect >= PRU_VRECT_BOOT)
            m_signal_handler(PRU_SM_SIGNAL_POWER_APPLIED);

        if(p_sensor_data->vrect < PRU_VRECT_UVLO)
            m_signal_handler(PRU_SM_SIGNAL_POWER_REMOVED);

        if(p_sensor_data->charge_complete || p_sensor_data->wired_charge_detect)
            m_signal_handler(PRU_SM_SIGNAL_CHARGE_COMPLETE);
    }
}

/** @} */
