/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file */

#include <string.h>
#include <stdio.h>

#include "pru_hw_config.h"
#include "pru_test_mux.h"
#include "softdevice_handler.h"
#include "wireless_debug.h"
#include "nrf_gzll.h"
#include "pru_sensors.h"
#include "nrf_assert.h"
#include "bsp.h"

#define EVT_STRING_LENGTH_MAX                  20        /**< Max length of event message string */
#define DEBUG_LINE_LENGTH                      80        /**< Width of debug output table */
#define VRECT_INCREMENT                       500        /**< Value by which VRECT is incremented/decremented for each call to the simulated measurement function. */
#define IRECT_INCREMENT                       100        /**< Value by which IRECT is incremented/decremented for each call to the simulated measurement function. */
#define VOUT_INCREMENT                        500        /**< Value by which VOUT is incremented/decremented for each call to the simulated measurement function. */
#define IOUT_INCREMENT                        100        /**< Value by which IOUT is incremented/decremented for each call to the simulated measurement function. */
#define TEMPERATURE_INCREMENT                   5        /**< Value by which the Temperature is incremented/decremented for each call to the simulated measurement function. */
#define PIPE_NUMBER                             0        /**< Use pipe 0 for the wireless debug interface (WDI) */
#define TX_PAYLOAD_LENGTH                      32        /**< Maximum PL length used by WDI */


/**@struct debug_output_t
 * @brief Collection of debug output
 */
typedef struct
{
    pru_sm_state_t state;
    pru_sensor_data_t sensor_data;
    pru_control_t control_data;
} debug_output_t;

static char temp_string[DEBUG_LINE_LENGTH];                  /**< Debug string */
static debug_output_t m_debug_output;                        /**< Debug output collection */
static char m_evt_evt_string[EVT_STRING_LENGTH_MAX] = "";    /**< Event message string */
static pru_tmux_override_t override_values;             /**< Local copy of override values */
static pru_control_t m_latest_control_data_received;    /**< Local latest received control data */
static bool m_transmission_stopped = false;                  /**< Is transmission stopped? */
static bool m_debug_print = false;                           /**< Used to signal that we are sending a string over air. */
static uint8_t debug_pkt[TX_PAYLOAD_LENGTH];                 /**< WDI packet */
extern bool gzll_initialized;                                /**< true when gzll is initialized */

/**@brief Override weak wpts_test_evt_handle_cb()
 * @param p_wpts WPT Service
 * @param p_evt WPT Service event
 */
void wpts_test_evt_handle_cb(ble_wpts_t * p_wpts, ble_wpts_evt_t * p_evt)
{
    if(p_evt->evt_type == BLE_WPTS_EVT_PRU_CONTROL_COMMAND)
    {
        memcpy(&m_latest_control_data_received, &p_evt->data.pru_control, sizeof(pru_control_t));
    }
}

/**@brief Function for writing a string to the debug packet
 *
 * @param[in]  str String to write
 *
*/
static void debug_print_string(const char *str)
{
    size_t len = strlen(str);
    ASSERT(len < (TX_PAYLOAD_LENGTH-1));
    if(len < (TX_PAYLOAD_LENGTH-1))
    {
        debug_pkt[0] = DEBUG_STR;
        strncpy((char *)&debug_pkt[1], str, len);
        debug_pkt[len+1] = 0;
        m_debug_print = true;
    }
}


/**@brief Function for processing board specific test commands.
 *
 * @param[in]  cmd   Command code to process.

 * @retval true Command code was processed by this function.
 * @retval false Command code was not processed by this function.
 *
 */
bool pru_process_command_board(uint8_t cmd);


/**@brief Function for processing common test commands.
 *
 * @param[in]  cmd   Command code to process.

 * @retval true Command code was processed by this function.
 * @retval false Command code was not processed by this function.
 *
 */
bool m_pru_process_command_common(uint8_t cmd)
{
    bool b_cmd_processed = true;
    uint32_t err_code;

    pru_tmux_override_values_get(&override_values);

    switch (cmd)
    {
        // Inc/dec , enable/disable commands

        case 'q':
            debug_print_string("Power adjusted bit set\n\r");
            override_values.power_adjusted = true;
            break;

        case 'a':
            debug_print_string("Power adjusted bit cleared\n\r");
            override_values.power_adjusted = false;
            break;

        case 'w':
            override_values.iout += IOUT_INCREMENT;
            sprintf(temp_string, "Inc IOUT -> %04X\n\r", override_values.iout);
            debug_print_string(temp_string);
            break;

        case 's':
            override_values.iout -= IOUT_INCREMENT;
            sprintf(temp_string, "Dec IOUT -> %04X\n\r", override_values.iout);
            debug_print_string(temp_string);
            break;

        case 'e':
            debug_print_string("NOT SUPPORTED\n\r");
            break;

        case 'd':
            debug_print_string("NOT SUPPORTED\n\r");
            break;

        case 'r':
            override_values.vout += VOUT_INCREMENT;
            sprintf(temp_string, "Inc VOUT -> %04X\n\r", override_values.vout);
            debug_print_string(temp_string);
            break;

        case 'f':
            override_values.vout -= VOUT_INCREMENT;
            sprintf(temp_string, "Dec VOUT -> %04X\n\r", override_values.vout);
            debug_print_string(temp_string);
            break;

        case 't':
            debug_print_string("Wired charger on\n\r");
            override_values.wired_charge_detect = true;
            break;
        case 'g':
            debug_print_string("Wired charger off\n\r");
            override_values.wired_charge_detect = false;
            break;

        case 'y':
            override_values.vrect += VRECT_INCREMENT;
            sprintf(temp_string, "Inc VRECT -> %4X\n\r", override_values.vrect);
            debug_print_string(temp_string);
            break;
        case 'h':
            override_values.vrect -= VRECT_INCREMENT;
            sprintf(temp_string, "Dec VRECT -> %04X\n\r", override_values.vrect);
            debug_print_string(temp_string);
            break;

        case 'u':
            override_values.irect += IRECT_INCREMENT;
            sprintf(temp_string, "Inc IRECT -> %04X\n\r", override_values.irect);
            debug_print_string(temp_string);
            break;
        case 'j':
            override_values.irect -= IRECT_INCREMENT;
            sprintf(temp_string, "Dec IRECT -> %04X\n\r", override_values.irect);
            debug_print_string(temp_string);
            break;

        case 'i':
            override_values.temperature += TEMPERATURE_INCREMENT;
            sprintf(temp_string, "Inc TEMP -> %04X\n\r", override_values.temperature);
            debug_print_string(temp_string);
            break;
        case 'k':
            override_values.temperature -= TEMPERATURE_INCREMENT;
            sprintf(temp_string, "Dec TEMP -> %04X\n\r", override_values.temperature);
            debug_print_string(temp_string);
            break;

        case 'o':
            debug_print_string("Test-mode enabled\n\r");
            tmux_get_select(TMUX_SELECT_OVERRIDE);
            tmux_set_select(TMUX_SELECT_OVERRIDE);
            break;

        case 'p':
            debug_print_string("SDK Version ");
            debug_print_string(SDK_VERSION);
            debug_print_string("\n\r");
            break;

        case 'l':
            debug_print_string("Test-mode disabled\n\r");
            tmux_get_select(TMUX_SELECT_NO_OVERRIDE);
            tmux_set_select(TMUX_SELECT_NO_OVERRIDE);
            break;

        // Single button commands
        case 'z':
            debug_print_string("Reset\r\n");
            NVIC_SystemReset();
            break;

        case 'x':
            debug_print_string("Send Slave Security Request\n\r");

            ble_gap_sec_params_t sec_params;
            memset(&sec_params, 0, sizeof(ble_gap_sec_params_t));
            sec_params.mitm = 0;
            sec_params.bond = 0;

            err_code = sd_ble_gap_authenticate(0, &sec_params); // Use 0 as we will only be connected to one device.
            APP_ERROR_CHECK(err_code);

            break;

        case 'n':
            debug_print_string("Charge complete\n\r");
            override_values.charge_complete = true;
            break;

        case 'm':
            debug_print_string("Charge incomplete\n\r");
            override_values.charge_complete = false;
            break;

        case '1':
            debug_print_string("Vrect to PRU_VRECT_UVLO\n\r");
            override_values.vrect = PRU_VRECT_UVLO;
            break;

        case '2':
            debug_print_string("Vrect to PRU_VRECT_BOOT\n\r");
            override_values.vrect =  PRU_VRECT_BOOT;
            break;

        case '3':
            debug_print_string("Vrect to PRU_VRECT_MIN\n\r");
            override_values.vrect  = PRU_VRECT_MIN;
            break;

        case '4':
            debug_print_string("Vrect to PRU_VRECT_HIGH\n\r");
            override_values.vrect = (PRU_VRECT_HIGH);
            break;

        case '5':
            debug_print_string("Vrect to PRU_VRECT_MAX\n\r");
            override_values.vrect = (PRU_VRECT_MAX);
            break;

        case 'c':
        case 'v':
            if(!m_transmission_stopped)
            {
                debug_print_string("Stop packet transmission\n\r");
                m_transmission_stopped = !m_transmission_stopped; // Need to do this before disabling the softdevice
                (void)sd_softdevice_disable();
            }

            else
            {
                debug_print_string("Start packet transmission\n\r");
                m_transmission_stopped = !m_transmission_stopped;
                nrf_clock_lf_cfg_t clock_lf_cfg = NRF_CLOCK_LFCLKSRC;
                SOFTDEVICE_HANDLER_INIT(&clock_lf_cfg, NULL);
            }
            break;
        case '6':
            debug_print_string("\n\rpru\n\r");
            break;
        case '7':
        case '8':
        #ifdef DFU_SUPPORT
            debug_print_string("Entering DFU mode ...\n\r");
            dfu_start();
            break;
        #endif // DFU_SUPPORT

        case '9':
        case ',':
        case '.':
        case '0':
        default:
            b_cmd_processed = false;
            break;
    }

    pru_tmux_override_values_set(&override_values);

    return b_cmd_processed;
}

/**@brief Get and process uart commands
 */
static void cmd_poll(void)
{
    uint8_t cmd;
    bool b_cmd_processed;

    if (!debug_cmd_available())
    {
        return;
    }
    cmd = get_debug_cmd();
    b_cmd_processed = m_pru_process_command_common(cmd);
    if (!b_cmd_processed)
    {
        debug_print_string("\n\rCommand not implemented\n\r");
    }
}


/**@brief Update m_debug_output variable.
 * @return true if debug output has changed.
 */
static bool debug_output_changed(void)
{
    const pru_sensor_data_t * sensor_data;
    bool changed = false;

    sensor_data = pru_sensors_data_get();

    if(m_debug_output.state != pru_sm_execute(PRU_SM_SIGNAL_NULL, NULL))
    {
        changed = true;
        m_debug_output.state = pru_sm_execute(PRU_SM_SIGNAL_NULL, NULL);
    }

    if(memcmp(&m_debug_output.sensor_data, sensor_data, sizeof(pru_sensor_data_t)) != 0)
    {
        changed = true;
        memcpy(&m_debug_output.sensor_data, sensor_data, sizeof(pru_sensor_data_t));
    }

    if(memcmp(&m_debug_output.control_data, &m_latest_control_data_received, sizeof(pru_control_t)))
    {
        changed = true;
        memcpy(&m_debug_output.control_data, &m_latest_control_data_received, sizeof(pru_control_t));
    }

    if(strlen(m_evt_evt_string) > 0)
    {
        changed = true;
    }

    return changed;
}

void debug(void)
{
    uint32_t pkt_len = 1;

    if (m_debug_print)
    {
        m_debug_print = false;
        pkt_len = strlen((const char *)debug_pkt) + 1;
    }
    else if(debug_output_changed())
    {
        //debug_output_print();
        debug_pkt[0] = DEBUG_PKT;
        debug_pkt[1] = (uint8_t)m_debug_output.state;
        debug_pkt[2] = (uint8_t)(m_debug_output.sensor_data.vrect & 0xff);
        debug_pkt[3] = (uint8_t)(m_debug_output.sensor_data.vrect >> 8);
        debug_pkt[4] = (uint8_t)(m_debug_output.sensor_data.irect & 0xff);
        debug_pkt[5] = (uint8_t)(m_debug_output.sensor_data.irect >> 8);
        debug_pkt[6] = (uint8_t)(m_debug_output.sensor_data.vout & 0xff);
        debug_pkt[7] = (uint8_t)(m_debug_output.sensor_data.vout >> 8);
        debug_pkt[8] = (uint8_t)(m_debug_output.sensor_data.iout & 0xff);
        debug_pkt[9] = (uint8_t)(m_debug_output.sensor_data.iout >> 8);
        debug_pkt[10] = (uint8_t)(m_debug_output.sensor_data.temperature & 0xff);
        debug_pkt[11] = (uint8_t)(m_debug_output.sensor_data.temperature >> 8);
        debug_pkt[12] = (uint8_t)m_debug_output.sensor_data.charge_complete;
        debug_pkt[13] = (uint8_t)(PRU_VRECT_SET & 0xff);
        debug_pkt[14] = (uint8_t)(PRU_VRECT_SET >> 8);
        debug_pkt[15] = (uint8_t)(PRU_VRECT_MIN & 0xff);
        debug_pkt[16] = (uint8_t)(PRU_VRECT_MIN >> 8);
        debug_pkt[17] = (uint8_t)(PRU_VRECT_MAX & 0xff);
        debug_pkt[18] = (uint8_t)(PRU_VRECT_MAX >> 8);
        debug_pkt[19] = (uint8_t)PRU_TEMP_MAX ;
        debug_pkt[20] = (uint8_t)(PRU_VRECT_MAX & 0xff);
        debug_pkt[21] = (uint8_t)(PRU_VRECT_MAX >> 8);
        debug_pkt[22] = (uint8_t)(PRU_PRECT_MAX & 0xff);
        //lint -e572
        debug_pkt[23] = (uint8_t)(PRU_PRECT_MAX >> 8);
        if(strlen(m_evt_evt_string) > 0)
        {
            debug_pkt[24] = 1;
        }
        else
        {
            debug_pkt[24] = 0;
        }
        debug_pkt[25] = (uint8_t)m_debug_output.sensor_data.power_adjusted;
        debug_pkt[26] = (uint8_t)m_debug_output.control_data.enable_pru_output;
        debug_pkt[27] = (uint8_t)m_debug_output.control_data.enable_pru_charge_indicator;
        debug_pkt[28] = (uint8_t)m_debug_output.control_data.adj_power;
        debug_pkt[29] = (uint8_t)m_debug_output.control_data.permissions;
        debug_pkt[30] = (uint8_t)m_debug_output.control_data.time_set;
        pkt_len = TX_PAYLOAD_LENGTH;
    }
    while(!nrf_gzll_add_packet_to_tx_fifo(PIPE_NUMBER, debug_pkt, pkt_len))
        ;
    debug_pkt[0] = DEBUG_POLL;
    cmd_poll();
}
