/* Copyright (c) Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file */

#include <stdint.h>
#include <string.h>

#include "ptu.h"
#include "nrf_assert.h"
#include "ptu_test_mux.h"
#include "ptu_sensors.h"
#include "ptu_conn_man.h"
#include "ptu_power_ctl.h"
#include "ptu_power_sharing.h"
#include "ptu_beacons.h"
#include "ptu_latching_fault.h"
#include "ptu_distant_list_handler.h"

/******************************************************************************/
/** @name Static functions and variables
 *  @{ */
/******************************************************************************/

APP_TIMER_DEF(m_pru_dyn_read_timer_id);                     /**< PRU dynamic parameter read timer ID. */
APP_TIMER_DEF(m_ptu_sensor_timer_id);                       /**< Sensor timer */
static ptu_sm_state_vars_t m_state;                    /**< State variables. */
static app_sm_evt_handler_t m_app_sm_evt_handler = NULL;    /**< Application state machine event handler. */

/**
 * @brief Used when executing state machine events.
 */
static void m_ptu_sm_execute(ptu_sm_signal_type_t signal)
{
    uint32_t err_code;

    (void)ptu_sm_execute(signal, &m_state);

    // Notify application of state machine signal
    if(m_app_sm_evt_handler != NULL)
        m_app_sm_evt_handler(signal, &m_state);

    // If state changed
    if(m_state.current_state != m_state.prev_state)
    {
        switch(m_state.current_state)
        {
            case PTU_SM_STATE_CONFIGURATION:
                m_ptu_sm_execute(PTU_SM_SIGNAL_CONFIGURATION_COMPLETE);
                break;

            case PTU_SM_STATE_POWER_SAVE:
                if(ptu_reg_n_entries_get() > 0)
                    ptu_cm_disconnect_all();

                ptu_cm_scan_enable();
                ptu_beacons_start();
                break;

            case PTU_SM_STATE_LOW_POWER:
                err_code = ptu_power_ctrl_set_poweramp_input(PTU_POWERAMP_INPUT_START);
                APP_ERROR_CHECK(err_code);

                err_code = ptu_tmux_poweramp_enable_set(true);
                APP_ERROR_CHECK(err_code);
                break;

            case PTU_SM_STATE_POWER_TRANSFER:
                err_code = ptu_power_ctrl_set_poweramp_input(PTU_POWERAMP_INPUT_NOMINAL);
                APP_ERROR_CHECK(err_code);

                err_code = ptu_tmux_poweramp_enable_set(true);
                APP_ERROR_CHECK(err_code);
                break;

            case PTU_SM_STATE_LOCAL_FAULT:
                err_code = ptu_tmux_poweramp_enable_set(false);
                APP_ERROR_CHECK(err_code);

                ptu_cm_disconnect_all();
                ptu_cm_scan_disable();
                break;

            case PTU_SM_STATE_LATCH_FAULT:
                err_code = ptu_tmux_poweramp_enable_set(false);
                APP_ERROR_CHECK(err_code);

                ptu_cm_disconnect_all();
                ptu_cm_scan_disable();

                err_code = ptu_latching_fault_entered(m_state.prev_state);
                APP_ERROR_CHECK(err_code);
                break;

            default:
                break;
        }
    }
}

/**
 * PRU dynamic read timer handler.
* @param p_context context data.
 */
static void m_ptu_pru_dyn_read_timer_handler(void * p_context)
{
    uint32_t err_code;

    ptu_cm_dynamic_read_all();
    if(!ptu_cm_mode_trans_in_progress())
    {
        ptu_power_ctrl_adjust();

        #if(PTU_CCA_ROGUE_DETECT_ENABLE == true)
        const ptu_sensor_data_t * sensors_data;
        err_code = ptu_sensors_data_get(&sensors_data);
        APP_ERROR_CHECK(err_code);

        if(sensors_data->rogue_obj_detected)
        {
            m_ptu_sm_execute(PTU_SM_SIGNAL_ROGUE_OBJECT_DETECTED);
        }
        #endif

        if(ptu_reg_n_entries_get() == 0)
        {
            m_ptu_sm_execute(PTU_SM_SIGNAL_ALL_DEVICES_DISCONNECTED);
        }
    }
}

/**@brief Handler of sensor timer
 *
 * @param[in]  p_context Generic context to sent to handler.
 *
 */
static void m_ptu_sensor_timer_handler(void * p_context)
{
    const ptu_sensor_data_t * sensors_data;

    uint32_t err_code;
    err_code = ptu_sensors_read();
    APP_ERROR_CHECK(err_code);
    err_code = ptu_sensors_data_get(&sensors_data);
    APP_ERROR_CHECK(err_code);
    ptu_power_sharing_temperature_warning_set(sensors_data->temperature_warning);
}

/** @brief Handle PRU alert.
 */
static void m_alert_handler(ble_wpts_c_t * p_wpts_c, ble_wpts_c_evt_t * const p_wpts_c_evt)
{
    ptu_reg_item_t *p_reg_item;
    pru_alert_bits_t * p_pru_alert;

    p_reg_item = ptu_reg_item_get_from_conn_handle(p_wpts_c->conn_handle);

    if(p_reg_item != NULL)
    {
        ASSERT( (p_wpts_c_evt->type  == BLE_WPTS_C_EVT_PRU_DYNAMIC_READ_RESP) ||
                (p_wpts_c_evt->type  == BLE_WPTS_C_EVT_PRU_ALERT) );

        if (p_wpts_c_evt->type  == BLE_WPTS_C_EVT_PRU_DYNAMIC_READ_RESP)
        {
            p_pru_alert = &(p_wpts_c_evt->data.pru_dynamic.alerts);
        }
        else
        {
            p_pru_alert = &(p_wpts_c_evt->data.pru_alert.alerts);
        }

        p_reg_item->charged = p_pru_alert->charge_complete || p_pru_alert->wired_charge_detect;

        // System error
        if(p_pru_alert->pru_over_voltage ||
           p_pru_alert->pru_over_current ||
           p_pru_alert->pru_over_temperature )
        {
            m_ptu_sm_execute(PTU_SM_SIGNAL_SYSTEM_ERROR);
            return;
        }
    }
}

/** @brief Handle WPT client service event.
 */
static void m_wpts_evt_handler(ble_wpts_c_t * p_wpts_c, ble_wpts_c_evt_t * const p_wpts_c_evt)
{
    static bool m_test_mode = false;
    uint32_t calc_var;

    if(p_wpts_c_evt->gatt_status == BLE_GATT_STATUS_SUCCESS)
    {
        ptu_reg_item_t * p_reg_item = ptu_reg_item_get_from_conn_handle(p_wpts_c->conn_handle);
        ASSERT(p_reg_item != NULL);

        switch (p_wpts_c_evt->type)
        {
            case BLE_WPTS_C_EVT_PRU_ALERT_ENABLE_WRITE_RESP:
                break;

            case BLE_WPTS_C_EVT_PRU_STATIC_READ_RESP:
                m_test_mode = (p_wpts_c_evt->data.pru_static.info_ptu_test_mode == BLE_WPTS_PRU_INFO_PTU_TEST_MODE_ENABLED_VAL);
                break;

            case BLE_WPTS_C_EVT_PTU_STATIC_WRITE_RESP:
                break;

            case BLE_WPTS_C_EVT_PRU_DYNAMIC_READ_RESP:
                // Copy received parameters to registry
                memcpy(&p_reg_item->prev_pru_dynamic, &p_wpts_c_evt->data.pru_dynamic, sizeof(pru_dynamic_t));

                // Update VRECT_STATIC_MIN, VRECT_STATIC_HIGH and VRECT_SET parameters if required.
                if(BLE_WPTS_BITFIELD_READ(p_reg_item->prev_pru_dynamic.optional_fields,
                   BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_MIN_DYN_BITMSK,
                   BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_MIN_DYN_BITPOS))
                {
                    p_reg_item->prev_pru_static.vrect_min_static = p_reg_item->prev_pru_dynamic.vrect_min_dyn;
                }

                if(BLE_WPTS_BITFIELD_READ(p_reg_item->prev_pru_dynamic.optional_fields,
                   BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_HIGH_DYN_BITMSK,
                   BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_HIGH_DYN_BITPOS))
                {
                    p_reg_item->prev_pru_static.vrect_high_static = p_reg_item->prev_pru_dynamic.vrect_high_dyn;
                }

                if(BLE_WPTS_BITFIELD_READ(p_reg_item->prev_pru_dynamic.optional_fields,
                   BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_SET_DYN_BITMSK,
                   BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_SET_DYN_BITPOS))
                {
                    p_reg_item->prev_pru_static.vrect_set = p_reg_item->prev_pru_dynamic.vrect_set_dyn;
                }

                if(m_test_mode) // If PTU was told to enter test mode during registration
                    ptu_power_ctrl_set_itx_val(p_wpts_c_evt->data.pru_dynamic.tester_command);

                // Derive prect from vrect * irect and write to registry
                calc_var = p_reg_item->prev_pru_dynamic.irect;            // [mA]
                calc_var *= (uint32_t)p_reg_item->prev_pru_dynamic.vrect; // [mA] * [mV] -> [uW]
                p_reg_item->prev_pru_dynamic.prect = (calc_var / 100000); // [(1/10)W];

                // Derive PRU power utilization and write to registry
                p_reg_item->prev_pru_dynamic.putil =
                (uint8_t)(((uint32_t)p_reg_item->prev_pru_dynamic.prect * 100)/  // * 100 to make it %
                          ((uint32_t)p_reg_item->prev_pru_static.prect_max));

                // Process alert field
                m_alert_handler(p_wpts_c, p_wpts_c_evt);

                p_reg_item->pending_dyn_read = false;
                break;

            case BLE_WPTS_C_EVT_PRU_CONTROL_WRITE_RESP:
                if(p_reg_item->pending_charge_disable)
                {
                    p_reg_item->pending_charge_disable = false;
                    ptu_cm_remove_device(p_reg_item);
                }
                break;

            case BLE_WPTS_C_EVT_PRU_ALERT:
                m_alert_handler(p_wpts_c, p_wpts_c_evt);
                break;

            default:
                break;
        }
        ptu_cm_on_wpt_service_evt(p_wpts_c, p_wpts_c_evt);
        ptu_power_sharing_on_wpt_service_evt(p_wpts_c, p_wpts_c_evt);
    }
}

/** @brief Initialize Client Service.
 */
static void m_services_init(void)
{
    uint32_t err_code;
    ble_wpts_c_init_t  wpts_c_init;

    wpts_c_init.evt_handler = m_wpts_evt_handler;

    err_code = ble_wpts_c_init(&wpts_c_init);
    APP_ERROR_CHECK(err_code);
}


/** @brief Initialize timers.
 */
static void m_timers_init(void)
{
    uint32_t err_code;

    // Instantiate sensor reading timer
    err_code = app_timer_create(&m_ptu_sensor_timer_id, APP_TIMER_MODE_REPEATED, m_ptu_sensor_timer_handler);
    APP_ERROR_CHECK(err_code);

    // Instantiate PRU dyn read timer
    err_code = app_timer_create(&m_pru_dyn_read_timer_id, APP_TIMER_MODE_REPEATED, m_ptu_pru_dyn_read_timer_handler);
    APP_ERROR_CHECK(err_code);
}

/** @} */

/******************************************************************************/
/** @name API implementation
 *  @{ */
/******************************************************************************/

void ptu_on_ble_evt(ble_evt_t * p_ble_evt)
{
    ptu_reg_item_t * reg_item;

    ptu_cm_on_ble_evt(p_ble_evt);

    if ((reg_item = ptu_reg_item_get_from_conn_handle(p_ble_evt->evt.common_evt.conn_handle)) != NULL)
    {
        ble_wpts_c_on_ble_evt(&reg_item->ble_wpts_c, p_ble_evt);
    }
}

void ptu_init(app_sm_evt_handler_t sm_evt_handler)
{
    uint32_t err_code;

    m_app_sm_evt_handler = sm_evt_handler;

    m_timers_init();

    m_services_init();

    ptu_reg_init();

    err_code = ptu_cm_init(m_ptu_sm_execute);
    APP_ERROR_CHECK(err_code);

    err_code = ptu_power_sharing_init(m_ptu_sm_execute);
    APP_ERROR_CHECK(err_code);

    err_code = ptu_sensors_init(m_ptu_sm_execute);
    APP_ERROR_CHECK(err_code);

    ptu_beacons_init();

    err_code = ptu_dlh_init(&(ptu_dlh_init_t) {.ignore_latency_ms = PTU_MS_FROM_BEACON_STOP_TO_ACTIVATE_DISTANT_LIST});
    APP_ERROR_CHECK(err_code);

    err_code = ptu_latching_fault_init(m_ptu_sm_execute);
    APP_ERROR_CHECK(err_code);

    err_code = tmux_init();  // Initializes HAL as well
    APP_ERROR_CHECK(err_code);
}

void ptu_start(void)
{
    uint32_t err_code;

    err_code = TIMER_START(m_ptu_sensor_timer_id, PTU_SENSORS_READ_INTERVAL_MS, NULL);
    APP_ERROR_CHECK(err_code);
    err_code = TIMER_START(m_pru_dyn_read_timer_id, PTU_PRU_DYN_READ_INTERVAL_MS, NULL);
    APP_ERROR_CHECK(err_code);

    m_ptu_sm_execute(PTU_SM_SIGNAL_CONFIGURATION_COMPLETE);
}


#ifdef DFU_SUPPORT
// Implement functions required by the DFU functionality

void terminate(void)
{
    ptu_cm_disconnect_all();
}

bool dfu_check(void)
{
    return (ptu_reg_n_entries_get() == 0);
}
#endif //DFU_SUPPORT

/** @} */
