/* Copyright (c) 2015 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file */

#include "ptu_beacons.h"
#include "ptu_sensors.h"
#include "ptu_conn_man.h"
#include "ptu_distant_list_handler.h"
#include "ptu_power_ctl.h"
#include "ptu_test_mux.h"


/******************************************************************************/
/** @name Static functions and variables
 *  @{ */
/******************************************************************************/

/**@brief Beacon status. */
typedef enum
{
    PTU_BEACON_STATUS_OFF,         /**< Beacon off. */
    PTU_BEACON_STATUS_SHORT,       /**< Short beacon on. */
    PTU_BEACON_STATUS_LONG,        /**< Long beacon on. */
    PTU_BEACON_STATUS_EXTENDED,    /**< Extended beacon on. */
} ptu_beacon_status_t;

/**@brief Beacon control struct. */
typedef struct
{
    uint32_t period;                     /**< Beacon period [ms]. */
    ptu_beacon_status_t status;     /**< Beacon status. */
} ptu_beacon_ctl_t;


static ptu_beacon_ctl_t m_beacon_ctl;                      /**< Beacon control structure */
APP_TIMER_DEF(m_beacon_timer_id);                               /**< Beacon timer */
APP_TIMER_DEF(m_extended_beacon_load_var_timer_id);             /**< Extended beacon timer */
static bool long_beacon_extension = false;                      /**< Should long beacon be extended? */

/**
 * @brief Handler for extended beacon load variation timer
 */
static void m_extended_beacon_load_var_handler(void * p_context)
{
    uint32_t err_code;

    if(!long_beacon_extension)
    {
        if(ptu_sensors_valid_long_beacon_extension_load_variation_found())
        {
            long_beacon_extension = true;
        }
    }

    // If PTU is not in Power Save state, stop the repeated extended beacon load variation timer
    if(PTU_SM_CURRENT_STATE() != PTU_SM_STATE_POWER_SAVE)
    {
        err_code = app_timer_stop(m_extended_beacon_load_var_timer_id);
        APP_ERROR_CHECK(err_code);
    }
}

 /**@brief Function for controlling beacon behaviour.
 *  @param[in] context Interrupt context.
 */
static void m_beacon_timer_handler(void * context)
{
    uint32_t err_code;

    // Ensure beacon is only switched on/off in power save state
    if(PTU_SM_CURRENT_STATE() == PTU_SM_STATE_POWER_SAVE)
    {
        switch(m_beacon_ctl.status)
        {
            const ptu_sensor_data_t * sensor_data;

            case PTU_BEACON_STATUS_OFF:      // Start of beacon period

                err_code = ptu_power_ctrl_set_poweramp_input(PTU_POWERAMP_INPUT_BEACON_SHORT);
                APP_ERROR_CHECK(err_code);

                err_code = ptu_tmux_poweramp_enable_set(true);
                APP_ERROR_CHECK(err_code);

                err_code = ptu_sensors_read();
                APP_ERROR_CHECK(err_code);

                err_code = ptu_sensors_data_get(&sensor_data);
                APP_ERROR_CHECK(err_code);

                err_code = TIMER_START(m_beacon_timer_id, PTU_BEACON_DURATION_SHORT_MS, NULL);
                APP_ERROR_CHECK(err_code);

                m_beacon_ctl.status = PTU_BEACON_STATUS_SHORT;
                break;

            case PTU_BEACON_STATUS_SHORT:    // Short beacon end
                m_beacon_ctl.period++;

                err_code = ptu_sensors_data_get(&sensor_data);
                APP_ERROR_CHECK(err_code);

                if((m_beacon_ctl.period < PTU_BEACON_CYCLES_LONG) && !sensor_data->load_var_detected)
                {
                    err_code = ptu_tmux_poweramp_enable_set(false); // Switch coil off
                    APP_ERROR_CHECK(err_code);
                    err_code = TIMER_START(m_beacon_timer_id, (PTU_BEACON_CYCLE_PERIOD_MS - PTU_BEACON_DURATION_SHORT_MS), NULL);
                    APP_ERROR_CHECK(err_code);
                    m_beacon_ctl.status = PTU_BEACON_STATUS_OFF;
                }
                else // Start long beacon
                {
                    ptu_dlh_notify_beacon_active();
                    err_code = ptu_power_ctrl_set_poweramp_input(PTU_POWERAMP_INPUT_BEACON_LONG);
                    APP_ERROR_CHECK(err_code);
                    err_code = TIMER_START(m_beacon_timer_id, PTU_BEACON_DURATION_LONG_MS, NULL);
                    APP_ERROR_CHECK(err_code);
                    err_code = TIMER_START( m_extended_beacon_load_var_timer_id, PTU_POWER_SAVE_LONG_BEACON_PTX_READ_INTERVAL_MS, NULL);
                    APP_ERROR_CHECK(err_code);
                    m_beacon_ctl.status = PTU_BEACON_STATUS_LONG;
                    m_beacon_ctl.period = 0;
                }
                break;

            case PTU_BEACON_STATUS_LONG:     // Long beacon end
                err_code = ptu_sensors_data_get(&sensor_data);
                APP_ERROR_CHECK(err_code);

                err_code = app_timer_stop(m_extended_beacon_load_var_timer_id);
                APP_ERROR_CHECK(err_code);

                ptu_sensors_clear_long_beacon_extension_load_variation_buffer();

                if(long_beacon_extension)
                {
                    // Start extended beacon
                    long_beacon_extension = false;
                    err_code = TIMER_START(m_beacon_timer_id,(PTU_BEACON_DURATION_EXTENDED_MS - PTU_BEACON_DURATION_LONG_MS), NULL);
                    APP_ERROR_CHECK(err_code);
                    m_beacon_ctl.status = PTU_BEACON_STATUS_EXTENDED;
                }

                else
                {
                    ptu_dlh_notify_beacon_disabled();
                    err_code = ptu_tmux_poweramp_enable_set(false);
                    APP_ERROR_CHECK(err_code);
                    err_code = TIMER_START(m_beacon_timer_id,
                                                (PTU_BEACON_CYCLE_PERIOD_MS - (PTU_BEACON_DURATION_SHORT_MS + PTU_BEACON_DURATION_LONG_MS)),
                                                 NULL);
                    APP_ERROR_CHECK(err_code);
                    m_beacon_ctl.status = PTU_BEACON_STATUS_OFF;
                }
                break;

            case PTU_BEACON_STATUS_EXTENDED: // Extended Long Beacon end
                ptu_dlh_notify_beacon_disabled();
                err_code = ptu_tmux_poweramp_enable_set(false);
                APP_ERROR_CHECK(err_code);
                m_beacon_ctl.status = PTU_BEACON_STATUS_OFF;
                m_beacon_timer_handler(NULL); // As we have executed extended beacon, we know that the cycle time has passed anyway, so execute beacon state function
                break;

           default:
                APP_ERROR_HANDLER(NRF_ERROR_INVALID_STATE);
                break;
        }
    }
}
/** @} */

/******************************************************************************/
/** @name API implementation
 *  @{ */
/******************************************************************************/

void ptu_beacons_start(void)
{
    m_beacon_timer_handler(NULL);
}

void ptu_beacons_init(void)
{
    uint32_t err_code;

    err_code = app_timer_create(&m_beacon_timer_id, APP_TIMER_MODE_SINGLE_SHOT, m_beacon_timer_handler);
    APP_ERROR_CHECK(err_code);

    err_code = app_timer_create(&m_extended_beacon_load_var_timer_id, APP_TIMER_MODE_REPEATED, m_extended_beacon_load_var_handler);
    APP_ERROR_CHECK(err_code);

    m_beacon_ctl.period = 0;
    m_beacon_ctl.status = PTU_BEACON_STATUS_OFF;
}

/** @} */
