/* Copyright (c) 2012 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file */

#include <stdint.h>
#include <string.h>
#include <stdlib.h>

#include "nrf_assert.h"
#include "ble_hci.h"
#include "ptu_sensors.h"
#include "ptu_conn_man.h"
#include "ptu_power_ctl.h"
#include "ptu_distant_list_handler.h"


/******************************************************************************/
/** @name Static functions and variables
 *  @{ */
/******************************************************************************/

/** Substates during recognition of load variance shift "time set pattern" */
typedef enum
{
    PTU_CM_TIME_SET_STATE_IDLE = 0,                           /**< No search for time set pattern. */
    PTU_CM_TIME_SET_STATE_P_TX_LOW,                           /**< Expecting low P_TX shift. */
    PTU_CM_TIME_SET_STATE_P_TX_HIGH,                          /**< Expecting P_TX to return to initial value. */
} ptu_cm_time_set_state_t;

typedef struct
{
    ble_gap_addr_t prev_received_adv_packet_addr;                  /**< Address of previously received adv packet */
    ble_gap_addr_t curr_device_connecting;                         /**< Address of device being connected */
    ble_gap_scan_params_t scan_params;                             /**< Scan parameters. */
    uint32_t time_set_p_tx_in_reference;                           /**< Measured power at PTU resonator before controle packet is sent */
    ptu_cm_time_set_state_t time_set_state;                        /**< Impedance shift search state */
    bool scanning_reqested;                                        /**< Scan requested from above layers. */
    bool scanning;                                                 /**< Scan requested from above layers. */
    bool connecting;                                               /**< COnnect in progress. */
    bool mode_trans_in_progess;                                    /**< Mode transition in progress. */
    uint8_t ignored_adv_from_same_pru_cnt;                         /**< Consecutive adv packets being discarded from same PRU. */
    uint8_t nof_connections;                                       /**< Number of current connections. */
    uint32_t ticks_at_mt_end;                                      /**< Current ticks value when Mode Transition has ended. */
} ptu_cm_status_t;

typedef struct {
    uint8_t nof_samples_taken;  /**< Number of samples taken since start if time set procedure. */
    uint8_t variation_duration; /**< Duration of current variation. */
    bool    load_variation;     /**< Is a valid variation active? */
} ptu_cm_timeset_data_t;

APP_TIMER_DEF(m_mode_trans_timer_id);                               /**< Mode transition timer ID. */
APP_TIMER_DEF(m_time_set_check_timer_id);                           /**< "Time set check" timer ID. */
static ptu_cm_status_t     m_status;                                /**< Connection manager status variables */
static ptu_static_t        m_ptu_static = {
    .option_valid_max_imp       = BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_VALID,
    .option_valid_max_res       = BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_VALID,
    .ptu_power                  = BLE_WPTS_PTU_POWER,
    .ptu_max_source_impedance   = BLE_WPTS_PTU_STATIC_MAX_SOURCE_IMPEDANCE_VAL,
    .ptu_max_load_resistance    = BLE_WPTS_PTU_STATIC_MAX_LOAD_RESISTANCE_VAL,
    .ptu_class                  = PTU_CLASS,
    .hardware_rev               = PTU_HW_REVISION,
    .firmware_rev               = PTU_FW_REVISION,
    .protocol_rev               = PTU_PROTOCOL_REVISION,
    .devices_supported_bitval   = BLE_WPTS_PTU_STATIC_NUM_DEVICES_BITVAL}; /**< PTU Static parameter. */

static ptu_sm_handler_t m_sm_handler;

static const ble_gap_conn_params_t m_conn_params = {
                .min_conn_interval = PTU_MIN_CONN_INTERVAL,
                .max_conn_interval = PTU_MAX_CONN_INTERVAL,
                .slave_latency     = PTU_SLAVE_LATENCY,
                .conn_sup_timeout  = PTU_CONN_SUP_TIMEOUT};       /**< PTU Connection parameters. */

/**
 * Wrapper function for calling sd_ble_gap_connect().
 * Will update current state of m_status.
 * @param[in] p_addr Pointer to GAP address to connect to
 */
static void m_connect(ble_gap_addr_t const * p_addr)
{
    ble_gap_scan_params_t scan_params;
    uint32_t err_code, diff;

    if(m_status.connecting || m_status.nof_connections >= PTU_MAX_CONNECTIONS || p_addr == NULL)
        return;

    m_status.connecting = true;
    m_status.scanning = false;

    memcpy(&m_status.curr_device_connecting, p_addr, sizeof(ble_gap_addr_t));

    memcpy(&scan_params, &m_status.scan_params, sizeof(scan_params));
    scan_params.timeout = 1;

    if(m_status.mode_trans_in_progess)
    {
        // Can safely assume that ticks now is lower than TICKS_AT_MT
        uint32_t ticks_now;

        err_code = app_timer_cnt_get(&ticks_now);
        APP_ERROR_CHECK(err_code);

        diff = ticks_diff(m_status.ticks_at_mt_end, ticks_now) / 32000 ;
        scan_params.timeout += diff;
    }

    do
    {
        err_code = sd_ble_gap_connect(p_addr, &scan_params, &m_conn_params);
    }while(err_code == NRF_ERROR_BUSY);

    APP_ERROR_CHECK(err_code);
}


/**
 * Disconnect from conn_handle.
 * Will not fail if disconnection has already been initialized.
 * @param conn_handle Connection handle to disconnect from.
 */
static void m_disconnect(uint16_t conn_handle)
{
    uint32_t err_code = sd_ble_gap_disconnect(conn_handle, BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);

    if((err_code != NRF_ERROR_INVALID_STATE) && (err_code != BLE_ERROR_INVALID_CONN_HANDLE))
        APP_ERROR_CHECK(err_code);
}


static void m_on_fully_accepted_adv_packet(ptu_reg_item_t * reg_item_p, ble_gap_evt_adv_report_t * p_adv_report);

/**
 * Initiate connection and registration of PRU.
 * Shall be called after adv. report only.
 * @param reg_item PRU for which to start connection/registration.
 */
static void m_connection_init(ptu_reg_item_t * reg_item)
{
    uint32_t err_code;

    ASSERT(reg_item != NULL);
    reg_item->state = REG_ITEM_STATE_WAITING_TO_CONNECT;

    err_code = app_timer_stop(reg_item->timer_id);
    APP_ERROR_CHECK(err_code);

    // Start registrartion timer for this registry item
    err_code = app_timer_start(reg_item->timer_id, APP_TIMER_TICKS(PTU_REG_TIMEOUT_MS, APP_TIMER_PRESCALER), reg_item);
    APP_ERROR_CHECK(err_code);
}


// On registration timeout
static void m_on_reg_timeout(ptu_reg_item_t * reg_item_p)
{
    uint32_t err_code;

    ptu_cm_remove_device(reg_item_p);

    if(m_status.connecting)
    {
        err_code = sd_ble_gap_connect_cancel();
        APP_ERROR_CHECK(err_code);

        m_status.connecting = false;
    }

    m_sm_handler(PTU_SM_SIGNAL_REGISTRATION_TIMEOUT);

    // This is needed to get PTU from Power Transfer to Power Save fast when a reconnection times out.
    if(m_status.nof_connections == 0)
    {
        m_sm_handler(PTU_SM_SIGNAL_ALL_DEVICES_DISCONNECTED);
    }
}

static void m_mode_trans_timer_handle(void * p_context)
{
    if(m_status.nof_connections == 0)
    {
        m_sm_handler(PTU_SM_SIGNAL_ALL_DEVICES_DISCONNECTED);
    }

    m_status.ticks_at_mt_end = 0;
    m_status.mode_trans_in_progess = false;
}

static void m_set_reg_item_fully_accepted(ptu_reg_item_t * reg_item_p)
{
    #if PTU_REVOKE_ACCEPTED_STATE
        uint32_t err_code;
        uint32_t ptu_revoke_accepted_state_timeout_ms = PTU_REVOKE_ACCEPTED_STATE_NON_POWER_SAVE_MS;

        if(PTU_SM_CURRENT_STATE() == PTU_SM_STATE_POWER_SAVE)
        {
            ptu_revoke_accepted_state_timeout_ms = PTU_REVOKE_ACCEPTED_STATE_POWER_SAVE_MS;
        }

        err_code = TIMER_START( reg_item_p->timer_id, ptu_revoke_accepted_state_timeout_ms, reg_item_p);
        APP_ERROR_CHECK(err_code);
    #endif // PTU_REVOKE_ACCEPTED_STATE == true

    reg_item_p->state = REG_ITEM_STATE_FULLY_ACCEPTED;
}

static void m_reg_timer_handle(void * p_context)
{
    ptu_reg_item_t * reg_item_p;
    reg_item_p = p_context;

    if(reg_item_p->state == REG_ITEM_STATE_PRE_CONNECT) // 1700 ms Timeout, accept advertisement.
    {
        m_set_reg_item_fully_accepted(reg_item_p);
    }

    else // Registration timeout
    {
        m_on_reg_timeout(p_context);
    }
}

static void m_send_ctrl_denied_due_to_cc(ptu_reg_item_t * reg_item_p)
{
    // Send control packet stating that PRU will not be allowed to charge due to cross connection
    memset(&reg_item_p->prev_ctl, 0, sizeof(pru_control_t));
    reg_item_p->prev_ctl.permissions = CTL_PERMISSION_DENIED_CROSS_CONNECTION;
    reg_item_p->pending_ctl_write = true;
    reg_item_p->pending_charge_disable = true;

    while(ble_wpts_c_send_pru_control(&reg_item_p->ble_wpts_c, &reg_item_p->prev_ctl) == BLE_ERROR_NO_TX_PACKETS)
    {
        ;
    }
    // Device will now be removed from registry and disconnected once control write response arrives.
}

static void m_stop_time_set_procedure(ptu_cm_timeset_data_t * p_ts)
{
    uint32_t err_code;

    err_code = app_timer_stop(m_time_set_check_timer_id);
    APP_ERROR_CHECK(err_code);

    ptu_power_ctrl_set_disable_power_amplifier_adjustments(false); // Re-enable adjustments of I_TX_COIL

    p_ts->nof_samples_taken = 0;
    p_ts->variation_duration = 0;
    p_ts->load_variation = false;
}

/**
 * @brief Timeout handler for the time set functionality.
 * @details Will be called every PTU_TIME_SET_CHECK_INTERVAL_MS after receiving
 *          CONTROL RSP from PRU that the PTU will perform time set on.
            To handle artifacts and jitter in the timer, the upper and lower limits are expanded
            compared to the BSS.
 * @param p_context Timeout context, contains pointer to registry item for time set target PRU.
 */
static void m_ptu_time_set_check_timer_handler(void * p_context)
{
    static ptu_cm_timeset_data_t ts;
    uint32_t err_code;
    const ptu_sensor_data_t * sensor_data;
    ptu_reg_item_t * reg_item_p = p_context;

    err_code = ptu_sensors_read();
    APP_ERROR_CHECK(err_code);

    err_code = ptu_sensors_data_get(&sensor_data);
    APP_ERROR_CHECK(err_code);

    if(abs(sensor_data->p_tx_in - m_status.time_set_p_tx_in_reference) >= PTU_P_TX_IN_LOAD_DETECT)
    {
        ts.variation_duration+= 2;
        ts.load_variation = true;
    }
    // load variation has stopped, check length
    else if(ts.load_variation)
    {
        if(ts.variation_duration >= PTU_TIME_SET_LOWER_LIMIT_MS &&
           ts.variation_duration <= PTU_TIME_SET_UPPER_LIMIT_MS)
        {
            m_stop_time_set_procedure(&ts);
        }
        // reset time set data in case valid variation is found later
        ts.variation_duration = 0;
        ts.load_variation = false;
    }

    // > 500 ms has passed with no valid time set load found.
    if(ts.nof_samples_taken++ >= PTU_TIME_SET_500_MS_OF_SAMPLES)
    {
        m_stop_time_set_procedure(&ts);
        m_send_ctrl_denied_due_to_cc(reg_item_p);
    }
}

/**
 * Function evaluating flag m_scanning_requested and starts / stops
 * scanning accordingly.
 * Updates flag m_scanning_on.
 */
static void m_scan_maintain(void)
{
    uint32_t err_code;

    // Shall we start scanning?
    if(m_status.scanning_reqested && !m_status.scanning && !m_status.connecting)
    {
        if(ptu_reg_n_entries_get() == 0)
        {
            m_status.scan_params.window = PTU_POWER_SAVE_STATE_SCAN_WINDOW;
            m_status.scan_params.interval = PTU_POWER_SAVE_STATE_SCAN_INTERVAL;
        }
        else
        {
            m_status.scan_params.window = PTU_NORMAL_SCAN_WINDOW;
            m_status.scan_params.interval = PTU_NORMAL_SCAN_INTERVAL;
        }

        do{
            err_code = sd_ble_gap_scan_start(&m_status.scan_params);
        }while(err_code != NRF_SUCCESS);

        //m_reconnect_enable = true;
        m_status.scanning = true;
    }

    // Shall we stop scanning?
    if(!m_status.scanning_reqested && m_status.scanning)
    {
        if(m_status.scanning)
        {
            do{
                err_code = sd_ble_gap_scan_stop();
            }while(err_code != NRF_SUCCESS);
            m_status.scanning = false;
        }
    }
}

static bool m_imp_shift_required(uint8_t adv_flags)
{
    uint8_t pru_category;
    pru_category = BLE_WPTS_BITFIELD_READ(adv_flags , BLE_WPTS_ADV_FLAG_IMPEDANCE_SHIFT_BITMSK, BLE_WPTS_ADV_FLAG_IMPEDANCE_SHIFT_BITPOS);
    return !(pru_category == BLE_WPTS_PRU_CATEGORY_NO_IMP_SHIFT);
}

static void m_update_pre_connect_adv_cnt_for_reg_item(ptu_reg_item_t * reg_item_p)
{
    uint32_t err_code;

    // If this is the first adv packet from this PRU
    if(reg_item_p->pre_connect_cnt == 0)
    {
        err_code = app_timer_start(reg_item_p->timer_id, APP_TIMER_TICKS(PTU_ADV_IGNORE_TIMEOUT_MS, APP_TIMER_PRESCALER), reg_item_p);
        APP_ERROR_CHECK(err_code);
    }

    if(reg_item_p->pre_connect_cnt < PTU_ADV_PRE_CONNECT_ALLOW)
    {
        reg_item_p->pre_connect_cnt++;
    }

    // If PTU has received enough advertisements from PRU connect to it
    if(reg_item_p->pre_connect_cnt >= (PTU_ADV_PRE_CONNECT_ALLOW - 1)) // Do "-1" so that the 11th packet is actually accepted.
    {
        m_set_reg_item_fully_accepted(reg_item_p);
    }
}

static void m_ensure_device_is_in_registry(ptu_reg_item_t ** reg_item_p, ble_gap_evt_adv_report_t * p_adv_report, ptu_reg_item_state_t init_state)
{
    if (*reg_item_p == NULL)
        *reg_item_p = ptu_reg_item_add(p_adv_report, init_state);

      // If PRU not added to registry so far
    if(*reg_item_p == NULL)
    {
        // Delete oldest PRU with REG_ITEM_STATE_PRE_CONNECT from registry to make room.
        *reg_item_p = ptu_reg_item_oldest_get(REG_ITEM_STATE_PRE_CONNECT);
        if(*reg_item_p != NULL)
        {
            ptu_reg_item_delete(*reg_item_p);
        }
        else // No registry item with state PRE_CONNECT, look for FULLY ACCEPTED.
        {
            *reg_item_p = ptu_reg_item_oldest_get(REG_ITEM_STATE_FULLY_ACCEPTED);
            if(*reg_item_p != NULL)
                ptu_reg_item_delete(*reg_item_p);
            else
                return;
        }
        // Attempt again adding the PRU to the registry
        *reg_item_p = ptu_reg_item_add(p_adv_report, init_state);
    }
}

static bool m_load_var_detected(void)
{
    uint32_t err_code;
    const ptu_sensor_data_t * sensor_data;
    err_code = ptu_sensors_data_get(&sensor_data);
    APP_ERROR_CHECK(err_code);
    return sensor_data->load_var_detected;
}

static void m_on_partially_accepted_adv_packet(ptu_reg_item_t * reg_item_p, ble_gap_evt_adv_report_t * p_adv_report)
{
    m_ensure_device_is_in_registry(&reg_item_p, p_adv_report, REG_ITEM_STATE_PRE_CONNECT);

    // If PRU has been added to registry
    if(reg_item_p != NULL && reg_item_p->state == REG_ITEM_STATE_PRE_CONNECT)
    {
        m_update_pre_connect_adv_cnt_for_reg_item(reg_item_p);
    }
}

static void m_on_fully_accepted_adv_packet(ptu_reg_item_t * reg_item_p, ble_gap_evt_adv_report_t * p_adv_report)
{
    if(p_adv_report != NULL)
        m_ensure_device_is_in_registry(&reg_item_p, p_adv_report, REG_ITEM_STATE_WAITING_TO_CONNECT);

    if(reg_item_p != NULL)
    {
        m_connection_init(reg_item_p);
        m_sm_handler(PTU_SM_SIGNAL_PRU_ADV_RCVD);
    }
}

/**
 * Function to be when advertise packet with WPT UUID has been received.
 * @param p_ble_evt event data forwarded from softdevice.
 * @param adv_flags the adv flags from the adv packet.
 */
static void m_on_wpt_adv_report(ble_evt_t * p_ble_evt, uint8_t adv_flags)
{
    ptu_reg_item_t * reg_item_p;
    bool imp_shift_ok, rssi_ok, fully_accepted, state_ok;

#ifndef PTU_DISABLE_DISTANT_LIST
    ptu_dlh_on_wpt_adv_report(&p_ble_evt->evt.gap_evt.params.adv_report.peer_addr, adv_flags);
#endif //PTU_DISABLE_DISTANT_LIST

    if( !ptu_dlh_device_is_in_distant_list(&p_ble_evt->evt.gap_evt.params.adv_report.peer_addr) &&
        (ptu_sm_execute(PTU_SM_SIGNAL_NULL, NULL) != PTU_SM_STATE_LOW_POWER || ptu_reg_n_entries_get() == 0))
    {
        reg_item_p = ptu_reg_item_get_from_address(&p_ble_evt->evt.gap_evt.params.adv_report.peer_addr);

        if(reg_item_p != NULL && reg_item_p->reconnect)
        {
            reg_item_p->reconnect = false;
            m_connect(&reg_item_p->address);
        }

        else
        {
            state_ok = (reg_item_p == NULL || reg_item_p->state == REG_ITEM_STATE_PRE_CONNECT || reg_item_p->state == REG_ITEM_STATE_FULLY_ACCEPTED);
            rssi_ok = ((p_ble_evt->evt.gap_evt.params.adv_report.rssi > PTU_ADV_PWR_MIN) || !PTU_CCA_RSSI_ENABLE);
            imp_shift_ok = ((m_imp_shift_required(adv_flags) && m_load_var_detected()) || (!m_imp_shift_required(adv_flags) || !PTU_CCA_ADV_IMP_SHIFT_ENABLE));
            fully_accepted = (rssi_ok && imp_shift_ok) || reg_item_p->state == REG_ITEM_STATE_FULLY_ACCEPTED;

            // If all connection conditions are fulfilled
            if(fully_accepted && state_ok)
            {
                m_on_fully_accepted_adv_packet(reg_item_p, &p_ble_evt->evt.gap_evt.params.adv_report);
            }
            // If one of the connection conditions are fulfilled
            else if(rssi_ok || imp_shift_ok)
            {
                m_on_partially_accepted_adv_packet(reg_item_p, &p_ble_evt->evt.gap_evt.params.adv_report);
            }
        }
    }
}

/**
 * Function to be called after "scan response".
 * @param p_ble_evt event data forwarded from softdevice.
 */
static void m_on_scan_response(ble_evt_t * p_ble_evt)
{
    ptu_reg_item_t * reg_item_p;

    reg_item_p = ptu_reg_item_get_from_address(&p_ble_evt->evt.gap_evt.params.adv_report.peer_addr);

    // Only care about scan response from devices added to registry
    if( reg_item_p != NULL &&
        reg_item_p->state==REG_ITEM_STATE_WAITING_TO_CONNECT)
    {
        m_connect(&reg_item_p->address);

        reg_item_p->state = REG_ITEM_STATE_CONNECTING;
    }
}

static void m_device_registration_complete(ptu_reg_item_t * reg_item_p)
{
     uint32_t err_code;

     err_code = app_timer_stop(reg_item_p->timer_id);
     APP_ERROR_CHECK(err_code);

     // Mode transition handling
     m_status.mode_trans_in_progess = false;
     err_code = app_timer_stop(m_mode_trans_timer_id);
     APP_ERROR_CHECK(err_code);

     reg_item_p->state = REG_ITEM_STATE_REGISTERED;
}

/**
 * Function to be called after connection to a device is established.
 * on BLE_GAP_EVT_CONNECTED.
 * @param p_ble_evt event data forwarded from softdevice.
 */
static void m_on_evt_connected(ble_evt_t * p_ble_evt)
{
    ptu_reg_item_t * reg_item_p;
    uint32_t err_code;
    reg_item_p = ptu_reg_item_get_from_address(&p_ble_evt->evt.gap_evt.params.connected.peer_addr);

    ASSERT(m_status.connecting);
    m_status.connecting = false;
    m_status.nof_connections++;

    if(reg_item_p != NULL)
    {
        // State is 'REGISTERED' if it is a reconnection, otherwise it is 'CONNECTING'
        ASSERT(reg_item_p->state == REG_ITEM_STATE_CONNECTING || reg_item_p->state == REG_ITEM_STATE_REGISTERED);
        reg_item_p->ble_wpts_c.conn_handle = p_ble_evt->evt.gap_evt.conn_handle;

        memset(&m_status.curr_device_connecting, 0, sizeof(ble_gap_addr_t));

        // Connection is of a new device, perform registration
        if(reg_item_p->state == REG_ITEM_STATE_CONNECTING)
        {
            reg_item_p->state = REG_ITEM_STATE_REGISTERING_PRU_STAT_RD;
            err_code = ble_wpts_c_read_pru_static(&(reg_item_p -> ble_wpts_c));
            APP_ERROR_CHECK_ALLOW_DISCONNECT(err_code);
        }

        // Connection is a reconnection
        else if(reg_item_p->state == REG_ITEM_STATE_REGISTERED)
        {
            // Stop reconnection timeout timer
            err_code = app_timer_stop(reg_item_p->timer_id);
            APP_ERROR_CHECK(err_code);

            // Re-send PRU control packet to ensure it is set to same state as before reconnection
            err_code = ble_wpts_c_send_pru_control(&reg_item_p->ble_wpts_c, &reg_item_p->prev_ctl);
            APP_ERROR_CHECK(err_code);
        }

    }
    else
    {
        m_disconnect(p_ble_evt->evt.common_evt.conn_handle);
    }
}

/**
 * Function to be called on BLE_GAP_EVT_TIMEOUT event.
 * @param p_ble_evt event data forwarded from softdevice.
 */
static void m_on_evt_timeout(ble_evt_t * p_ble_evt)
{
    ptu_reg_item_t * reg_item_p;

    // If connection timeout
    if(p_ble_evt->evt.gap_evt.params.timeout.src == BLE_GAP_TIMEOUT_SRC_CONN)
    {
        ASSERT(m_status.connecting);

        m_status.connecting = false;

        reg_item_p = ptu_reg_item_get_from_address(&m_status.curr_device_connecting);

        if(reg_item_p == NULL)
            return;

        ptu_reg_item_delete(reg_item_p);

        if(m_status.nof_connections == 0)
        {
            m_sm_handler(PTU_SM_SIGNAL_ALL_DEVICES_DISCONNECTED);
        }
    }
    else if(p_ble_evt->evt.gap_evt.params.timeout.src == BLE_GAP_TIMEOUT_SRC_SECURITY_REQUEST)
    {
        // We will not perform any handling if security request times out. This is done to be compatible with wrongly implemented devices.
    }
}

/**
 * Function to be called on BLE_GAP_EVT_DISCONNECTED event.
 * @param p_ble_evt event data forwarded from softdevice.
 */
static void m_on_evt_disconnected(ble_evt_t * p_ble_evt)
{
    const ptu_sensor_data_t * p_sensor_data;
    ptu_reg_item_t * reg_item_p;
        uint32_t err_code;
    uint8_t zero_address[BLE_GAP_ADDR_LEN];

    memset(zero_address, 0, BLE_GAP_ADDR_LEN);

    m_status.nof_connections--;

    reg_item_p = ptu_reg_item_get_from_conn_handle(p_ble_evt->evt.gap_evt.conn_handle);

    if(memcmp(reg_item_p->address.addr, zero_address, BLE_GAP_ADDR_LEN) == 0)
    {
        ptu_reg_item_delete(reg_item_p); // Mode transition with Zero-address, delete device to ensure that registration performed.
    }

    else if(reg_item_p != NULL)
    {
        reg_item_p->pending_dyn_read = false;

        err_code = ptu_sensors_data_get(&p_sensor_data);
        APP_ERROR_CHECK(err_code);

        if( !p_sensor_data->load_var_detected && reg_item_p->state == REG_ITEM_STATE_REGISTERED)
        {
            if(!m_status.mode_trans_in_progess)
            {
                reg_item_p->reconnect = true;

                err_code = app_timer_start(reg_item_p->timer_id, APP_TIMER_TICKS(PTU_RECONNECT_TIMEOUT_MS, APP_TIMER_PRESCALER), reg_item_p);
                APP_ERROR_CHECK(err_code);
            }

            else // Mode transition with non-zero address
            {
                m_connect(&reg_item_p->address);
            }

        }

        else
        {
            ptu_reg_item_delete(reg_item_p);

            if(m_status.nof_connections == 0)
            {
                m_sm_handler(PTU_SM_SIGNAL_ALL_DEVICES_DISCONNECTED);
            }
        }

    }
}

/**
 * Function starting pairing sequence.
 * @param conn_handle connection handle.
 */
static void m_pairing_start(uint16_t conn_handle)
{
    uint32_t err_code;

    ble_gap_sec_params_t sec_params;
    memset(&sec_params, 0, sizeof(ble_gap_sec_params_t));
    sec_params.io_caps = BLE_GAP_IO_CAPS_NONE;
    sec_params.min_key_size = PTU_SEC_MIN_KEYSIZE;
    sec_params.max_key_size = PTU_SEC_MAX_KEYSIZE;

    err_code = sd_ble_gap_authenticate(conn_handle, &sec_params);
    APP_ERROR_CHECK(err_code);
}

/**
 * Function to be called on BLE_GATTC_EVT_READ_RSP and BLE_GATTC_EVT_WRITE_RSP events.
 * @param p_ble_evt event data forwarded from softdevice.
 */
static void m_on_evt_rw_response(ble_evt_t * p_ble_evt)
{
    if(p_ble_evt->evt.gattc_evt.gatt_status == BLE_GATT_STATUS_ATTERR_INSUF_AUTHENTICATION)
    {
        m_pairing_start(p_ble_evt->evt.gattc_evt.conn_handle);
    }
}

/**
 * Function to be called on BLE_GAP_EVT_SEC_REQUEST event.
 * @param p_ble_evt event data forwarded from softdevice.
 */
static void m_on_evt_seq_request(ble_evt_t * p_ble_evt)
{
    m_pairing_start(p_ble_evt->evt.gattc_evt.conn_handle);
}

/**
 * Function to be called on BLE_GAP_EVT_SEC_PARAMS_REQUEST event.
 * @param p_ble_evt event data forwarded from softdevice.
 */
static void m_on_evt_sec_params_request(ble_evt_t * p_ble_evt)
{
    uint32_t err_code;
    err_code = sd_ble_gap_sec_params_reply(p_ble_evt->evt.gap_evt.conn_handle, BLE_GAP_SEC_STATUS_SUCCESS, NULL, NULL);
    APP_ERROR_CHECK(err_code);
}

/**
 * Function to be called on BLE_GAP_EVT_CONN_SEC_UPDATE event.
 * @param p_ble_evt event data forwarded from softdevice.
 */
static void m_on_evt_conn_sec_update(ble_evt_t * p_ble_evt)
{
    uint32_t err_code;
    ptu_reg_item_t * reg_item_p = ptu_reg_item_get_from_conn_handle(p_ble_evt->evt.gap_evt.conn_handle);

    if(reg_item_p != NULL)
    {
        // If we were in the middle of registration process -> continue registration.
        switch(reg_item_p->state)
        {
            case REG_ITEM_STATE_REGISTERING_EN_ALERT:
                err_code = ble_wpts_c_enable_pru_alert_notification(&(reg_item_p -> ble_wpts_c));
                APP_ERROR_CHECK_ALLOW_DISCONNECT(err_code);
                break;

            case REG_ITEM_STATE_REGISTERING_PRU_STAT_RD:
                err_code = ble_wpts_c_read_pru_static(&(reg_item_p -> ble_wpts_c));
                APP_ERROR_CHECK_ALLOW_DISCONNECT(err_code);
                break;

            case REG_ITEM_STATE_REGISTERING_PTU_STAT_WR:
                err_code = ble_wpts_c_write_ptu_static(&(reg_item_p -> ble_wpts_c), &m_ptu_static);
                APP_ERROR_CHECK_ALLOW_DISCONNECT(err_code);
                break;

            case REG_ITEM_STATE_REGISTERING_PRU_DYN_RD:
                err_code = ble_wpts_c_read_pru_dynamic(&(reg_item_p -> ble_wpts_c));
                APP_ERROR_CHECK_ALLOW_DISCONNECT(err_code);
                break;

            case REG_ITEM_STATE_PRU_CTL_SEND:
                // ptu_power_ctl.c module will send control packet complpeting registration.
                break;

            default:
                break;
        }

        // Security mode 1 level 2 is expected when IO capabilities = BLE_GAP_IO_CAPS_NONE
        ASSERT(p_ble_evt->evt.gap_evt.params.conn_sec_update.conn_sec.sec_mode.sm == 1);
        ASSERT(p_ble_evt->evt.gap_evt.params.conn_sec_update.conn_sec.sec_mode.lv == 2);

        reg_item_p->link_encrypted = true;
    }
}

static void m_on_wpt_alert(ble_wpts_c_t * p_wpts_c, ble_wpts_c_evt_t * const p_wpts_c_evt)
{
    ptu_reg_item_t *p_reg_item;
    uint8_t mode_trans_val;
    uint32_t err_code;

    err_code = app_timer_cnt_get(&m_status.ticks_at_mt_end);
    APP_ERROR_CHECK(err_code);

    ASSERT(p_wpts_c_evt->type == BLE_WPTS_C_EVT_PRU_ALERT);

    mode_trans_val = p_wpts_c_evt->data.pru_alert.mode_transition;

    if(mode_trans_val != BLE_WPTS_PRU_ALERT_MODE_NO_MODE_TRANSITION_VAL)
    {
        p_reg_item = ptu_reg_item_get_from_conn_handle(p_wpts_c->conn_handle);
        ASSERT(p_reg_item != NULL);

        switch(mode_trans_val)
        {
            case BLE_WPTS_PRU_ALERT_MODE_2S_MODE_TRANSITION_VAL:
                err_code = app_timer_start(m_mode_trans_timer_id, APP_TIMER_TICKS(2000, APP_TIMER_PRESCALER), NULL);
                APP_ERROR_CHECK(err_code);
                m_status.ticks_at_mt_end += APP_TIMER_TICKS(2000, APP_TIMER_PRESCALER);
                break;

            case BLE_WPTS_PRU_ALERT_MODE_3S_MODE_TRANSITION_VAL:
                err_code = app_timer_start(m_mode_trans_timer_id, APP_TIMER_TICKS(3000, APP_TIMER_PRESCALER), NULL);
                APP_ERROR_CHECK(err_code);
                m_status.ticks_at_mt_end += APP_TIMER_TICKS(3000, APP_TIMER_PRESCALER);
                break;

            case BLE_WPTS_PRU_ALERT_MODE_6S_MODE_TRANSITION_VAL:
                err_code = app_timer_start(m_mode_trans_timer_id, APP_TIMER_TICKS(6000, APP_TIMER_PRESCALER), NULL);
                APP_ERROR_CHECK(err_code);
                m_status.ticks_at_mt_end += APP_TIMER_TICKS(6000, APP_TIMER_PRESCALER);
                break;
        }
        m_status.mode_trans_in_progess = true;
        memcpy(p_reg_item->address.addr, p_wpts_c_evt->data.pru_alert.device_address, BLE_GAP_ADDR_LEN);
    }
}

/** @brief Set pending dynamic read for PRU.
 */
static void ptu_char_pru_dynamic_read(uint16_t conn_handle)
{
    ptu_reg_item_t * reg_entry;

    reg_entry = ptu_reg_item_get_from_conn_handle(conn_handle);

    // Ignore return value as device could allready be disconnected
    if(reg_entry != NULL)
    {
        if(!reg_entry->pending_dyn_read)
        {
            if(ble_wpts_c_read_pru_dynamic(&(reg_entry -> ble_wpts_c)) == NRF_SUCCESS)
            {
                reg_entry->pending_dyn_read = true;
            }
        }
    }
}

/** @} */

/******************************************************************************/
/** @name API implementation
 *  @{ */
/******************************************************************************/

void ptu_cm_remove_device(ptu_reg_item_t * reg_item_p)
{
    uint32_t err_code;

    err_code = app_timer_stop(reg_item_p->timer_id);
    APP_ERROR_CHECK(err_code);

    // If a connection exists for this item.
    if((reg_item_p->state != REG_ITEM_STATE_UNUSED) &&
       (reg_item_p->state != REG_ITEM_STATE_PRE_CONNECT) &&
       (reg_item_p->state != REG_ITEM_STATE_WAITING_TO_CONNECT) &&
       (reg_item_p->state != REG_ITEM_STATE_FULLY_ACCEPTED))
    {
        m_disconnect(reg_item_p->ble_wpts_c.conn_handle);
    }

    ptu_reg_item_delete(reg_item_p);
}

bool ptu_cm_mode_trans_in_progress(void)
{
    return m_status.mode_trans_in_progess;
}


uint32_t ptu_cm_init(ptu_sm_handler_t sm_handler)
{
    int i;
    uint32_t err_code;
    ptu_reg_item_t *reg_item;

    if(sm_handler == NULL)
        return NRF_ERROR_INVALID_PARAM;

    m_sm_handler = sm_handler;
    memset(&m_status, 0, sizeof(ptu_cm_status_t));
    m_status.scan_params.active = 1;

    // Instantiate registration and connection timers
    for(i = 0; i < PTU_MAX_CONNECTIONS; i++)
    {
        err_code = ptu_reg_item_get_from_index(i, &reg_item);
        APP_ERROR_CHECK(err_code);

        err_code = app_timer_create(&(reg_item->timer_id), APP_TIMER_MODE_SINGLE_SHOT, m_reg_timer_handle);
        APP_ERROR_CHECK(err_code);
    }

    // Initiate time set impedance shift detect timer
    err_code = app_timer_create(&m_time_set_check_timer_id, APP_TIMER_MODE_REPEATED, m_ptu_time_set_check_timer_handler);
    APP_ERROR_CHECK(err_code);

    err_code = app_timer_create(&m_mode_trans_timer_id, APP_TIMER_MODE_SINGLE_SHOT, m_mode_trans_timer_handle);
    APP_ERROR_CHECK(err_code);

    return NRF_SUCCESS;
}

void ptu_cm_on_ble_evt(ble_evt_t * p_ble_evt)
{
    ble_wpts_service_data_t curr_service_data;

    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_ADV_REPORT:
            // If AD received (not scan response)
            if( p_ble_evt->evt.gap_evt.params.adv_report.scan_rsp == 0)
            {
                // If WPT advertisement
                if (ble_wpts_service_data_read(&p_ble_evt->evt.gap_evt.params.adv_report, &curr_service_data) == NRF_SUCCESS &&
                    curr_service_data.uuid16 == BLE_WPTS_UUID16)
                    m_on_wpt_adv_report(p_ble_evt, curr_service_data.adv_flags);
            }
            else // If scan response received
            {
                m_on_scan_response(p_ble_evt);
            }
            break;

        case BLE_GAP_EVT_CONNECTED:
            m_on_evt_connected(p_ble_evt);
            break;

        case BLE_GAP_EVT_TIMEOUT:
            m_on_evt_timeout(p_ble_evt);
            break;

        case BLE_GAP_EVT_DISCONNECTED:
            m_on_evt_disconnected(p_ble_evt);
            break;

        case BLE_GATTC_EVT_READ_RSP:
        case BLE_GATTC_EVT_WRITE_RSP:
            m_on_evt_rw_response(p_ble_evt);
            break;

        case BLE_GAP_EVT_SEC_REQUEST:
            m_on_evt_seq_request(p_ble_evt);
            break;

        case BLE_GAP_EVT_SEC_PARAMS_REQUEST:
            m_on_evt_sec_params_request(p_ble_evt);
            break;

        case BLE_GAP_EVT_CONN_SEC_UPDATE:
            m_on_evt_conn_sec_update(p_ble_evt);
            break;

        default:
            break;
    }

    m_scan_maintain();
}

// New device requesting power
void ptu_cm_on_wpt_service_evt(ble_wpts_c_t * p_wpts_c, ble_wpts_c_evt_t * const p_wpts_c_evt)
{
    ptu_reg_item_t * reg_item_p = ptu_reg_item_get_from_conn_handle(p_wpts_c->conn_handle);
    uint32_t err_code;

    if (reg_item_p != NULL)
    {
        switch (p_wpts_c_evt->type)
        {
            case BLE_WPTS_C_EVT_PRU_STATIC_READ_RESP:
                reg_item_p->state = REG_ITEM_STATE_REGISTERING_PTU_STAT_WR;
                err_code = ble_wpts_c_write_ptu_static(&(reg_item_p -> ble_wpts_c), &m_ptu_static);
                APP_ERROR_CHECK_ALLOW_DISCONNECT(err_code);
                // Make a local copy in registry of PRU static parameters
                reg_item_p->prev_pru_static = p_wpts_c_evt->data.pru_static;
                break;

            case BLE_WPTS_C_EVT_PTU_STATIC_WRITE_RESP:
                reg_item_p->state = REG_ITEM_STATE_REGISTERING_PRU_DYN_RD;
                err_code = ble_wpts_c_read_pru_dynamic(&(reg_item_p -> ble_wpts_c));
                APP_ERROR_CHECK_ALLOW_DISCONNECT(err_code);
                reg_item_p->pending_dyn_read = true;
                break;

            case BLE_WPTS_C_EVT_PRU_DYNAMIC_READ_RESP:
                if(reg_item_p->state != REG_ITEM_STATE_REGISTERED)
                {
                    const ptu_sensor_data_t * sensor_data;
                    err_code = ptu_sensors_data_get(&sensor_data);
                    APP_ERROR_CHECK(err_code);
                    m_status.time_set_p_tx_in_reference = sensor_data->p_tx_in;

                    //ASSERT(reg_item_p != 0);
                    reg_item_p->state = REG_ITEM_STATE_PRU_CTL_SEND;

                    // Info: Now, a control packet will be sent by the "power adjust"
                    // part of the code. So we can always rely on getting a
                    // BLE_WPTS_C_EVT_PRU_CONTROL_WRITE_RESP that will end registration.
                }
                break;

            case BLE_WPTS_C_EVT_PRU_CONTROL_WRITE_RESP:
                if(reg_item_p->state != REG_ITEM_STATE_REGISTERED)
                {
                    m_device_registration_complete(reg_item_p);
                    // Stop registration timer.
                    err_code = app_timer_stop(reg_item_p->timer_id);
                    APP_ERROR_CHECK(err_code);

                    // If this PRU has "Time set" support, and we just enabled PRU Charge Output
                    if(BLE_WPTS_BITFIELD_READ(reg_item_p->adv_flags, BLE_WPTS_ADV_FLAG_TIME_SET_SUPPORT_BITMSK, BLE_WPTS_ADV_FLAG_TIME_SET_SUPPORT_BITPOS) != 0 &&
                       reg_item_p->prev_ctl.enable_pru_output == 1)
                    {
                        err_code = app_timer_start(m_time_set_check_timer_id, APP_TIMER_TICKS(PTU_TIME_SET_CHECK_INTERVAL_MS, APP_TIMER_PRESCALER), reg_item_p);
                        APP_ERROR_CHECK(err_code);
                        ptu_power_ctrl_set_disable_power_amplifier_adjustments(true); // Disable adjustments of I_TX_COIL during time set.
                    }

                    err_code = ble_wpts_c_enable_pru_alert_notification(&(reg_item_p -> ble_wpts_c));
                    APP_ERROR_CHECK_ALLOW_DISCONNECT(err_code);

                    // We need to test permissions, so save the control packet data.
                    if(reg_item_p->prev_ctl.permissions == CTL_PERMISSION_PERMITTED)
                    {
                        m_sm_handler(PTU_SM_SIGNAL_CHARGE_START);
                    }
                }
                break;

            case BLE_WPTS_C_EVT_PRU_ALERT:
                m_on_wpt_alert(p_wpts_c, p_wpts_c_evt);
                break;

            default:
                break;
        }
    }
    m_scan_maintain();
}

void ptu_cm_scan_disable(void)
{
    m_status.scanning_reqested = false;
    m_scan_maintain();
}

void ptu_cm_scan_enable(void)
{
    m_status.scanning_reqested = true;
    m_scan_maintain();
}

void ptu_cm_disconnect_all(void)
{
    int i;
    ptu_reg_item_t * reg_item;
    uint32_t err_code;

    for(i = 0; i < PTU_MAX_CONNECTIONS; i++)
    {
        err_code = ptu_reg_item_get_from_index(i, &reg_item);
        APP_ERROR_CHECK(err_code);
        ptu_cm_remove_device(reg_item);
    }
}


void ptu_cm_dynamic_read_all(void)
{
    ptu_reg_item_t* registered_devices[PTU_MAX_CONNECTIONS];
    uint8_t n_reg_devices = ptu_reg_registered_devices_get(registered_devices);

    for(uint8_t i = 0; i < n_reg_devices; i++)
    {
        ptu_char_pru_dynamic_read(registered_devices[i]->ble_wpts_c.conn_handle);
    }
}

/** @} */
