/* Copyright (c) Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

#include <string.h>
#include <stdio.h>

#include "app_print.h"
#include "simple_uart.h"
#include "ptu_test_mux.h"
#include "ptu_registry.h"
#include "ptu_sensors.h"
#include "bsp.h"

/**
 * @file
 * @brief Source file for PTU UART test interface.
 */


#define DEBUG_LINE_LENGTH                      80   /**< Width of debug output table */
#define EVT_STRING_LENGTH_MAX                  20   /**< Max length of event message string */
#define VTX_INCREMENT                        1000   /**< Value by which the Vtx is incremented/decremented for each call to the simulated measurement function. */
#define ITX_INCREMENT                         100   /**< Value by which the Itx is incremented/decremented for each call to the simulated measurement function. */
#define TEMPERATURE_INCREMENT                  10   /**< Value by which the Temperature is incremented/decremented for each call to the simulated measurement function. */
#define EXTENDED_BEACON_THRESHOLD_MS          500   /**< The amount of time a beacon lasts before the debug functionality determines it is an extended beacon. */
#define MAX_RTC_COUNTER_VAL            0x00FFFFFF   /**< Maximum value of the RTC counter. */

/******************************************************************************/
/** @name Static functions and variables
 *  @{ */
/******************************************************************************/

/**@struct debug_output_t
 * @brief Collection of debug output
 */
typedef struct
{
    ptu_sm_state_t       current_state;
    ptu_reg_item_state_t reg_item_state[PTU_MAX_CONNECTIONS];
    bool                 link_encrypted[PTU_MAX_CONNECTIONS];
    ptu_sensor_data_t    sensor_data;
    pru_dynamic_t        dyn_params[PTU_MAX_CONNECTIONS];
    uint8_t              extended_beacon_currently_running;
} debug_output_t;

static debug_output_t m_debug_output;                     /**< Debug output collection */
static char m_evt_evt_string[EVT_STRING_LENGTH_MAX] = ""; /**< Event message string */
static ptu_tmux_override_t override_values;               /**< Local copy of override values */
static bool m_uart_initialized = false;                   /**< UART initalized flag. Used to initialize UART on first scharacter sent. */
static bool m_simplified_beacon_print = false;            /**< Simplify print flag. Used to decide if beacon debug data should be short. */
static bool force_debug_print = false;                    /**< Flag to force debug print even though debug information has not changed. */

/** @brief Will syntesize a load representing a load variation to the sensors module.
 *         This will be set to true for 1.5 seconds.
 */
static void m_set_load_variation_detected(void)
{
    override_values.power_amp_enable = true;

    for(uint8_t i = 0; i < PTU_Z_MEAS_BUF_LENGTH; ++i)
    {
        if(i == 0) // First call
            override_values.vtx = 10 * (PTU_ITX_MIN * PTU_Z_TX_IN_LOAD_DETECT) + 1; // Make sure v/i > Z_TX_IN_LOAD_DETECTED

        else // non - first call
            override_values.vtx = 1;

        override_values.itx = PTU_ITX_MIN;

        ptu_tmux_override_values_set(&override_values);

        (void)ptu_sensors_read(); // Triggers call to load variation detection
    }

    override_values.power_amp_enable = false;

    ptu_tmux_override_values_set(&override_values);

    (void)ptu_sensors_read(); // Triggers call to load variation detection

}

/**
 * @brief Convert state to string.
 * @param state The state to translate
 * @return String representation of state.
 */
static char * sm_state_string_get(ptu_sm_state_t state)
{
    switch(state)
    {
        case PTU_SM_STATE_CONFIGURATION:
            return ("Conf");
        case PTU_SM_STATE_POWER_SAVE:
            return ("Psave");
        case PTU_SM_STATE_LOW_POWER:
            return ("LowP");
        case PTU_SM_STATE_POWER_TRANSFER:
            return ("Ptrans");
        case PTU_SM_STATE_LATCH_FAULT:
            return ("LatchF");
        case PTU_SM_STATE_LOCAL_FAULT:
            return ("LocF");
        default:
            return "Invalid state";
    }
}

/** @brief Check if there are any characters in the UART receive buffer
 *
 * This function returns the true if there is a character available for reading
 * in the UART receive buffer.
 *
 * @return Number of characters available
 */
static bool m_uart_chars_available(void)
{
  return (NRF_UART0->EVENTS_RXDRDY == 1);
}

/**@brief Function for processing common test commands.
 *
 * @param[in]  cmd   Command code to process.
 *
 * @retval true Command code was processed by this function.
 * @retval false Command code was not processed by this function.
 *
 */
static bool m_ptu_process_command_common(uint8_t cmd)
{
    bool cmd_processed = true;
    uint32_t err_code;

    ptu_tmux_override_values_get(&override_values);

    switch (cmd)
    {
        case 'z':
            // Reset
            print_string("\n\rCommand: Reset\r\n");
            NVIC_SystemReset();
            break;

        case 'b':
            // Beacon debug print
            if(m_simplified_beacon_print)
                print_string("\n\rCommand: Beacon debug print disabled\r\n");
            else
                print_string("\n\rCommand: Beacon debug print enabled\r\n");
            m_simplified_beacon_print = !m_simplified_beacon_print;
            break;

        case 'y':
            print_string("\n\rCommand: Inc VTX -> ");
            override_values.vtx += VTX_INCREMENT;
            print_hex(override_values.vtx, 2);
            print_string("\n\r");
            break;

        case 'h':
            print_string("\n\rCommand: Dec VTX -> ");
            override_values.vtx -= VTX_INCREMENT;
            print_hex(override_values.vtx, 2);
            print_string("\n\r");
            break;

        case 'u':
            print_string("\n\rCommand: Inc ITX -> ");
            override_values.itx += ITX_INCREMENT;
            print_hex(override_values.itx, 2);
            print_string("\n\r");
            break;

        case 'j':
            print_string("\n\rCommand: Dec ITX -> ");
            override_values.itx -= ITX_INCREMENT;
            print_hex(override_values.itx, 2);
            print_string("\n\r");
            break;

        case 'i':
            print_string("\n\rCommand: Inc temperature -> ");
            override_values.temperature += TEMPERATURE_INCREMENT;
            print_hex(override_values.temperature, 2);
            print_string("\n\r");
            break;

        case 'k':
            print_string("\n\rCommand: Dec temperature -> ");
            override_values.temperature -= TEMPERATURE_INCREMENT;
            print_hex(override_values.temperature, 2);
            print_string("\n\r");
            break;

        case 'o':
            print_string("\n\rCommand: Test-mode enabled\n\r");
            tmux_set_select(TMUX_SELECT_OVERRIDE);
            tmux_get_select(TMUX_SELECT_OVERRIDE);
            force_debug_print=true;
            break;

        case 'p':
            print_string("SDK Version ");
            print_string(SDK_VERSION);
            print_string("\n\r");
            break;

        case 'l':
            print_string("\n\rCommand: Test-mode disabled\n\r");
            tmux_set_select(TMUX_SELECT_NO_OVERRIDE);
            tmux_get_select(TMUX_SELECT_NO_OVERRIDE);
            force_debug_print=true;
            break;

        case 'n':
            print_string("\n\rCommand: TX power set to -40dBm\n\r");
            err_code = sd_ble_gap_tx_power_set(-40);
            APP_ERROR_CHECK(err_code);
            break;

        case 'm':
            print_string("\n\rCommand: TX power set to PTU_OUTPUT_TX_POWER\n\r");
            err_code = sd_ble_gap_tx_power_set(PTU_OUTPUT_TX_POWER);
            APP_ERROR_CHECK(err_code);
            break;

        case 'v':
            print_string("\n\rCommand: Set Load Variation TRUE\n\r");
            m_set_load_variation_detected();
            break;


        case '2':
            print_string("\n\rCommand: Resonator ON\n\r");
            override_values.power_amp_enable = true;
            break;

        case '3':
            print_string("\n\rCommand: Resonator OFF\n\r");
            override_values.power_amp_enable = false;
            break;

        case '4':
            print_string("\n\rCommand: Inc poweramp level -> ");
            override_values.power_amp_level += POWERAMP_TEST_VAR_STEP_SIZE;
            print_hex(override_values.power_amp_level, 2);
            print_string("\n\r");
            break;

        case '5':
            print_string("\n\rCommand: Dec poweramp level -> ");
            override_values.power_amp_level -= POWERAMP_TEST_VAR_STEP_SIZE;
            print_hex(override_values.power_amp_level, 2);
            print_string("\n\r");
            break;

        case '6':
            print_string("\n\rptu\n\r");
            break;

        case '7':
            print_string("ptu_reg_n_entries_get=");
            print_hex(ptu_reg_n_entries_get(),1);
            print_string("\n\r");
            break;

        case '8':
        #ifdef DFU_SUPPORT
            print_string("Entering DFU mode ...\n\r");
            dfu_start();
            break;
        #endif //DFU_SUPPORT

        default:
            cmd_processed = false;
            break;
    }

    ptu_tmux_override_values_set(&override_values);

    return cmd_processed;
}

/**@brief Function for getting the time, in ms, that have passed since the argument system tick.
 *
 * @param[in]  tick   System tick we are measureing time from.
 *
 * @retval Time, in ms, that have passes since the argument system tick.
 *
 */
static uint32_t  ms_since(uint32_t tick){
    uint32_t current_tick;
    uint32_t delta_ticks;

    (void) app_timer_cnt_get(&current_tick);
    if(tick>current_tick) //Overflow occured
    {
        delta_ticks = (((MAX_RTC_COUNTER_VAL-tick)+current_tick) & MAX_RTC_COUNTER_VAL);
    }
    else
    {
        delta_ticks = ((current_tick-tick) & MAX_RTC_COUNTER_VAL);
    }
    return (1000*delta_ticks)/APP_TIMER_CLOCK_FREQ;
}


/** @} */

/******************************************************************************/
/** @name API implementation
 *  @{ */
/******************************************************************************/

/**@brief Function for processing board related test commands.
 *
 * @param[in]  cmd   Command code to process.
 *
 * @retval true Command code was processed by this function.
 * @retval false Command code was not processed by this function.
 *
 */
bool ptu_process_command_board(uint8_t cmd);

/**@brief Get and process uart commands
 */
void sm_uart_rx_poll(void)
{
    uint8_t cmd = 0;
    bool b_cmd_processed;
    if (!m_uart_chars_available())
    {
        return;
    }

    cmd = simple_uart_get();

    #ifdef PTU_TEST_HW_UART_CMD_ENABLE
    b_cmd_processed = ptu_process_command_board(cmd);
    if (!b_cmd_processed) b_cmd_processed = m_ptu_process_command_common(cmd);
    #else
    b_cmd_processed = m_ptu_process_command_common(cmd);
    #endif

    if (!b_cmd_processed)
    {
        print_string("\n\rCommand not implemented");
    }
}

/**@brief Print a DEBUG_LINE_LENGTH character long "horizontal line"
 * to UART.
 */
void line_print(void)
{
    int i;

    for(i = 0; i < DEBUG_LINE_LENGTH; i++)
    {
        print_string("-");
    }
    print_string("\r\n");
}

/**@brief Update m_debug_output variable.
 * @return true if debug output has changed.
 */
static bool debug_output_changed(void)
{
    int i;
    bool changed = false;
    uint32_t err_code;
    const ptu_sensor_data_t* sensor_data;
    uint16_t current_powerlevel;
    bool current_poweramp_status;
    static bool prev_poweramp_status;
    static uint32_t prev_powerlevel;
    static uint32_t ticks_at_last_beacon_changed;
    static uint8_t currently_in_long_beacon = false;
    ptu_reg_item_t * p_reg_item;

    err_code = ptu_sensors_data_get(&sensor_data);
    APP_ERROR_CHECK(err_code);


    (void)ptu_tmux_poweramp_enable_get(&current_poweramp_status);
    (void)ptu_tmux_poweramp_level_get(&current_powerlevel);

    uint32_t delta=ms_since(ticks_at_last_beacon_changed);

    // Beacon status has changed
    if(prev_poweramp_status != current_poweramp_status ||
        prev_powerlevel != current_powerlevel)
    {
        err_code = app_timer_cnt_get(&ticks_at_last_beacon_changed);
        APP_ERROR_CHECK(err_code);

        if(current_poweramp_status && current_powerlevel == PTU_POWERAMP_INPUT_BEACON_SHORT){
            currently_in_long_beacon = false;
            m_debug_output.extended_beacon_currently_running=false;
           if(m_simplified_beacon_print)
               print_string("s");
        }
        else if(current_poweramp_status && current_powerlevel == PTU_POWERAMP_INPUT_BEACON_LONG)
        {
            currently_in_long_beacon = true;
            m_debug_output.extended_beacon_currently_running=false;
            if(m_simplified_beacon_print){
                print_string("e\n\r");
                print_string("l");
            }
        }

        if(!current_poweramp_status){
            currently_in_long_beacon = false;
            m_debug_output.extended_beacon_currently_running=false;
            if(m_simplified_beacon_print)
                print_string("e\n\r");
        }
    }
    // Extended beacon
    else if(currently_in_long_beacon && delta > EXTENDED_BEACON_THRESHOLD_MS)
    {
        currently_in_long_beacon = false;
        m_debug_output.extended_beacon_currently_running=true;
        if(m_simplified_beacon_print){
            print_string("e\n\r");
            print_string("x");
        }
    }
    prev_poweramp_status = current_poweramp_status;
    prev_powerlevel = current_powerlevel;

    if(m_simplified_beacon_print)
        return false;



    if(m_debug_output.current_state != ptu_sm_execute(PTU_SM_SIGNAL_NULL, NULL))
    {
        changed = true;
        m_debug_output.current_state = ptu_sm_execute(PTU_SM_SIGNAL_NULL, NULL);
    }

    if(memcmp(&m_debug_output.sensor_data, sensor_data, sizeof(ptu_sensor_data_t)) != 0)
    {
        changed = true;
        memcpy(&m_debug_output.sensor_data, sensor_data, sizeof(ptu_sensor_data_t));
    }

    for(i = 0; i < PTU_MAX_CONNECTIONS; i++)
    {
        (void)ptu_reg_item_get_from_index(i, &p_reg_item);

        if(m_debug_output.reg_item_state[i] != p_reg_item->state)
        {
            changed = true;
            m_debug_output.reg_item_state[i] = p_reg_item->state;
        }

        if(m_debug_output.link_encrypted[i] != p_reg_item->link_encrypted)
        {
            changed = true;
            m_debug_output.link_encrypted[i] = p_reg_item->link_encrypted;
        }


        if(memcmp(&m_debug_output.dyn_params[i], &p_reg_item->prev_pru_dynamic, sizeof(pru_dynamic_t)) != 0)
        {
           changed = true;
           memcpy(&m_debug_output.dyn_params[i], &p_reg_item->prev_pru_dynamic, sizeof(pru_dynamic_t));
        }
    }

    if(strlen(m_evt_evt_string) > 0)
    {
        changed = true;
    }

    return changed;
}

/**@brief Print contents of m_debug_output to UART.
 */
static void debug_output_print(void)
{
    int i;
    char temp_string[DEBUG_LINE_LENGTH];                    /**< Debug output collection */
    uint32_t curr_ticks;

    (void) app_timer_cnt_get(&curr_ticks);

    line_print();

    // PTU main heading
    print_string("StatusStart ");
    print_hex((curr_ticks/32), 4); // Output current ticks value in ms.
    print_string("\r\n");

    line_print();

    // Table heading 1
    print_string("State\tSubst\tLddet\tLoadvar\tRobj\tResimp\tVina\tIina\tTemp\tX Beacon\r\n");

    // Table contents 1
    sprintf(temp_string, "%s\t", sm_state_string_get(m_debug_output.current_state));
    print_string(temp_string);

    print_string(temp_string);

    char * x_beacon_status[]={"Off", "On"};
    sprintf(temp_string, "%u\t%u\t%u\t%u\t%u\t%u\t%d\t%s\r\n",
            m_debug_output.sensor_data.load_detected,
            m_debug_output.sensor_data.load_var_detected,
            m_debug_output.sensor_data.rogue_obj_detected,
            m_debug_output.sensor_data.res_impedance,
            m_debug_output.sensor_data.v_ina,
            m_debug_output.sensor_data.i_ina,
            m_debug_output.sensor_data.temperature,
            x_beacon_status[m_debug_output.extended_beacon_currently_running]);
    print_string(temp_string);
    line_print();

    // Table heading 2
    print_string("Registered PRUs\r\n");
    line_print();

    // Table contents 2
    print_string("#\tVrect\tIrect\tVout\tIout\tPrect\tPutil\tTemp\tCharged\tEncrypt\r\n");
    for(i = 0; i < PTU_MAX_CONNECTIONS; i++)
    {
        if(m_debug_output.reg_item_state[i] == REG_ITEM_STATE_REGISTERED)
        {
            sprintf(temp_string, "%u\t%u\t%u\t%u\t%u\t%u\t%u\t%i\t%u\t%u\t",
            i,
            m_debug_output.dyn_params[i].vrect,
            m_debug_output.dyn_params[i].irect,
            m_debug_output.dyn_params[i].vout,
            m_debug_output.dyn_params[i].iout,
            m_debug_output.dyn_params[i].prect,
            m_debug_output.dyn_params[i].putil,
            m_debug_output.dyn_params[i].temperature,
            m_debug_output.dyn_params[i].alerts.charge_complete,
            m_debug_output.link_encrypted[i]);
        }
        else
        {
            sprintf(temp_string, "%u\t-\t-\t-\t-\t-\t-\t-\t-\t-", i);
        }
        print_string(temp_string);
        print_string("\r\n");
    }

    // Print event message
    if(strlen(m_evt_evt_string) > 0)
    {
        print_string("\r\n");
        print_string(m_evt_evt_string);
        sprintf(m_evt_evt_string, "");
    }

    print_string("StatusEnd\r\n");
}

void debug(void)
{
    if (!m_uart_initialized)
    {
        simple_uart_config(UART_BAUDRATE_BAUDRATE_Baud460800, RTS_PIN_NUMBER, TX_PIN_NUMBER, CTS_PIN_NUMBER, RX_PIN_NUMBER, HWFC);
        m_uart_initialized = true;
    }
    if(debug_output_changed() || force_debug_print)
    {
        debug_output_print();
        force_debug_print=false;
    }
    sm_uart_rx_poll();
}

void debug_add_string(const char *s)
{
    sprintf(m_evt_evt_string, "%s", s);
}

/** @} */
