/* Copyright (c) Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
 */

#include <string.h>

#include "nrf_assert.h"
#include "app_timer.h"
#include "ptu.h"
#include "ptu_config.h"
#include "ptu_distant_list_handler.h"

APP_TIMER_DEF(m_ignore_latency_timer_id);                   /**< Timer ID for ensuring latency from notification of beacon being disabled to starting to add devices . */
static uint8_t m_distant_list_index;                        /**< Current index in distant list array. */
static bool m_ignoring_devices;                             /**< Is devices currently being added to the distant list? I.E. is the Long Beacon NOT active? */
static ble_gap_addr_t m_distant_list[PTU_DISTANT_LIST_LEN]; /**< Distant list, list of GAP addresses to be ignored. */
static uint32_t m_ignore_latency_ms;                        /**< How long after getting notified of beacon disabled should the distant
                                                                 list handler start adding devices to the distant list? */

/** @brief Check if two ble_gap_addr_t addresses are equal
  * @details Will assert if either of the address pointers are NULL
  * @param p_a1 First address to check
  * @param p_a2 Address to check against
  *
  * @return True if the addresses match, false otherwise
  */
static bool m_address_equals(ble_gap_addr_t const * p_a1, ble_gap_addr_t const * p_a2)
{
    APP_ERROR_CHECK_BOOL(p_a1 != NULL);
    APP_ERROR_CHECK_BOOL(p_a2 != NULL);
    return memcmp(p_a1, p_a2, sizeof(ble_gap_addr_t)) == 0;
}

/** @brief Get next index in distant list array.
  * @details Works as a circular buffer where values are overwritten when it is full.
  * @param current_index The current index
  *
  * @return Next valid index of the distant list array.
  */
static uint8_t m_next_index(uint8_t current_index)
{
    uint8_t incremented_index = current_index + 1;
    return incremented_index >= PTU_DISTANT_LIST_LEN ? 0 : incremented_index;
}

/** @brief Hander for the ignore latency timer id.
  * @details Will set 'm_ignoring_devices' to false, I.E. stop adding devices to the distant list
  * @param p_context The given context
  *
  */
static void m_ignore_latency_timer_id_handler(void * p_context)
{
    m_ignoring_devices = false;
}

/** @brief Check if given address is in the distant list.
  * @param p_addr Address to check
  *
  * @return True if the addresses is in distant list, false otherwise
  */
static bool m_address_in_distant_list(ble_gap_addr_t const * p_addr)
{
    APP_ERROR_CHECK_BOOL(p_addr != NULL);

    for(uint8_t i = 0; i < PTU_DISTANT_LIST_LEN; ++i)
    {
        if(m_address_equals(&m_distant_list[i], p_addr))
            return true;
    }

    return false;
}

/** @brief Removes given address from distant list
  * @details Will assert if p_addr is NULL
  * @param p_addr Address of device to remove
  *
  * @return True if the addresses match, false otherwise
  */
static void m_remove_device(ble_gap_addr_t const * p_addr)
{
    APP_ERROR_CHECK_BOOL(p_addr != NULL);

    for(uint8_t i = 0; i < PTU_DISTANT_LIST_LEN; ++i)
    {
        if(m_address_equals(&m_distant_list[i], p_addr))
        {
            memset(&m_distant_list[i], 0, sizeof(ble_gap_addr_t));
        }
    }
}

void ptu_dlh_notify_beacon_active(void)
{
    uint32_t err_code;

    err_code = TIMER_START(m_ignore_latency_timer_id, m_ignore_latency_ms, NULL);
    APP_ERROR_CHECK(err_code);
}

void ptu_dlh_notify_beacon_disabled(void)
{
    m_ignoring_devices = true;
}

void ptu_dlh_on_wpt_adv_report(ble_gap_addr_t const * p_addr, uint8_t adv_flags)
{
    // Check reboot bit and remove device if reboot bit is set, and the device is currently in the distance list
    if(BLE_WPTS_BITFIELD_READ( adv_flags,
                                BLE_WPTS_ADV_FLAG_REBOOT_INDICATOR_BITMSK,
                                BLE_WPTS_ADV_FLAG_REBOOT_INDICATOR_BITPOS) == 0)
    {
        if(m_address_in_distant_list(p_addr))
        {
            m_remove_device(p_addr);
        }
    }

    else if(m_ignoring_devices && !m_address_in_distant_list(p_addr))
    {
        m_distant_list_index = m_next_index(m_distant_list_index);
        memcpy(&m_distant_list[m_distant_list_index], p_addr, sizeof(ble_gap_addr_t));
    }
}

bool ptu_dlh_device_is_in_distant_list(ble_gap_addr_t const * p_addr)
{
    APP_ERROR_CHECK_BOOL(p_addr != NULL);

    return m_address_in_distant_list(p_addr);
}

uint32_t ptu_dlh_init(ptu_dlh_init_t const * p_init)
{
    uint32_t err_code;

    if(p_init == NULL)
    {
        return NRF_ERROR_INVALID_PARAM;
    }

    m_ignore_latency_ms = p_init->ignore_latency_ms;

    m_distant_list_index = 0;

    m_ignoring_devices = true;

    memset(m_distant_list, 0, sizeof(m_distant_list));

    err_code = app_timer_create(&m_ignore_latency_timer_id, APP_TIMER_MODE_SINGLE_SHOT, m_ignore_latency_timer_id_handler);
    if(err_code != NRF_SUCCESS)
    {
        return err_code;
    }

    return NRF_SUCCESS;
}
