/* Copyright (c) Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
*/
#include <stdint.h>
#include <stdbool.h>
#include <string.h>

#include "ptu_latching_fault.h"
#include "ptu_config.h"
#include "ptu_sensors.h"
#include "ptu_power_ctl.h"
#include "ptu_test_mux.h"

/******************************************************************************/
/** @name Static functions and variables
 *  @{ */
/******************************************************************************/

/**
 * @enum lf_timer_state_t
 * @brief Enumerator describing what m_short_beacon_timer_id is being used for.
 */
typedef enum
{
    LF_TIMER_STATE_BEACON_OFF,        /**< Multiple latching faults beacon sequence */
    LF_TIMER_STATE_BEACON_SHORT,      /**< Multiple latching faults beacon sequence  */
} lf_timer_state_t;

/**
 * @struct lf_status_t
 * @brief Latching fault status variables.
 */
typedef struct
{
    lf_timer_state_t timer_state;     /**< State of m_short_beacon_timer_id timer */
    uint32_t successive_faults;       /**< Number of successive auto cleared latching faults */
    bool persistent;                  /**< Is the PTU in persistent latching fault? */
} lf_status_t;


static ptu_sm_handler_t m_ptu_sm_execute;                          /**< Profiles state machine handler */
APP_TIMER_DEF(m_clear_latching_fault_timer_id);                         /**< Clear latching fault timer */
APP_TIMER_DEF(m_short_beacon_timer_id);                                 /**< Beacon timer */
APP_TIMER_DEF(m_reset_nof_consecutive_latching_fault_count_timer_id);   /**< Reset number of consecutive latching faults timer */
static lf_status_t m_lf_status;                                         /**< Latching fault status */

 /**@brief Clear number of consecutive latching faults timer handler.
 *
 * @param[in]  p_context  Generic context to sent to handler.
 *
 */
static void m_reset_nof_consecutive_latching_fault_count_timer_handler(void * p_context)
{
    m_lf_status.successive_faults = 0;
    m_lf_status.persistent = false;
}

 /**@brief Latch fault entry handler.
 *
 * @param[in]  p_context  Generic context to sent to handler.
 *
 */
static void m_latch_fault_attempt_to_clear_timer_handler(void * p_context)
{
    m_ptu_sm_execute(PTU_SM_SIGNAL_LATCHING_FAULT_CLEARED_BY_USER);
}


/**@brief Check if no load variation has been seen for PTU_PRU_NO_LOAD_VAR_NO_DEVICES_MS milliseconds,
 *        implying that no devices are on the pad.
 *
 * @param p_sensor_data Pointer to PTU sensor data.
 * @return true if no devices on pad, false otherwise.
 */
static bool m_no_devices_on_pad(const ptu_sensor_data_t * p_sensor_data)
{
    static uint8_t load_var_detect_count = 0;

    if((p_sensor_data->res_impedance < PTU_Z_TX_IN_NO_LOAD))
    {
        load_var_detect_count = 0;
        return false;
    }

    if(++load_var_detect_count >= (PTU_LATCH_FAULT_NO_LOAD_VAR_NO_DEVICES_MS / PTU_BEACON_CYCLE_PERIOD_MS))
    {
        load_var_detect_count = 0;
        return true;
    }

    return false;
}

/**@brief Latch fault short beacon timer handler.
 *
 * @param[in]  p_context  Generic context to sent to handler.
 */
static void m_latch_fault_short_beacon_timer_handler(void * p_context)
{
    uint32_t err_code;
    const ptu_sensor_data_t * p_sensor_data;


    if(PTU_SM_CURRENT_STATE() != PTU_SM_STATE_LATCH_FAULT)
    {
        return;
    }

    switch(m_lf_status.timer_state)
    {
        case LF_TIMER_STATE_BEACON_OFF: // start short beacon

            err_code = ptu_power_ctrl_set_poweramp_input(PTU_POWERAMP_INPUT_BEACON_SHORT);
            APP_ERROR_CHECK(err_code);

            err_code = ptu_tmux_poweramp_enable_set(true);
            APP_ERROR_CHECK(err_code);

            err_code = TIMER_START(m_short_beacon_timer_id, PTU_BEACON_DURATION_SHORT_MS, NULL);
            APP_ERROR_CHECK(err_code);

            m_lf_status.timer_state = LF_TIMER_STATE_BEACON_SHORT;
            break;

       case LF_TIMER_STATE_BEACON_SHORT: // end short beacon
            // sensors needs to be read before disabling poweramp in order for load variation detection to be performed.
            err_code = ptu_sensors_read();
            APP_ERROR_CHECK(err_code);

            err_code = ptu_tmux_poweramp_enable_set(false);
            APP_ERROR_CHECK(err_code);

            err_code = TIMER_START(m_short_beacon_timer_id, (PTU_BEACON_CYCLE_PERIOD_MS - PTU_BEACON_DURATION_SHORT_MS), NULL);
            APP_ERROR_CHECK(err_code);

            m_lf_status.timer_state = LF_TIMER_STATE_BEACON_OFF;

            err_code = ptu_sensors_data_get(&p_sensor_data);
            APP_ERROR_CHECK(err_code);

            // Non-persistent latching fault has been cleared by seeing a load variation
            if(!m_lf_status.persistent && p_sensor_data->load_var_detected)
            {
                m_ptu_sm_execute(PTU_SM_SIGNAL_LATCHING_FAULT_CLEARED_BY_USER);
            }

            // User has cleared persistent latching fault by removing all devices from pad
            else if(m_lf_status.persistent && m_no_devices_on_pad(p_sensor_data))
            {
                m_lf_status.successive_faults = 0;
                m_lf_status.persistent = false;
                m_ptu_sm_execute(PTU_SM_SIGNAL_LATCHING_FAULT_CLEARED_BY_USER);
            }
            break;

       default:
           break;
    }
}

/** @} */

/******************************************************************************/
/** @name API implementation
 *  @{ */
/******************************************************************************/

uint32_t ptu_latching_fault_entered(ptu_sm_state_t previous)
{
    uint32_t err_code;

    if(PTU_SM_CURRENT_STATE() != PTU_SM_STATE_LATCH_FAULT)
    {
        return NRF_ERROR_INVALID_STATE;
    }

    m_lf_status.timer_state = LF_TIMER_STATE_BEACON_OFF;

    err_code = TIMER_START(m_short_beacon_timer_id, PTU_LATCH_FAULT_ENTRY_DELAY_MS, NULL);
    APP_ERROR_CHECK(err_code);

    err_code = TIMER_START(m_reset_nof_consecutive_latching_fault_count_timer_id, PTU_LATCH_FAULT_NO_ERR_CLR_TIMEOUT_MS, NULL);
    APP_ERROR_CHECK(err_code);

    if(m_lf_status.successive_faults ++ < PTU_LATCH_FAULT_MAX_AUTO_CLR_ATTEMPTS)
    {
        err_code = TIMER_START(m_clear_latching_fault_timer_id, PTU_LATCH_FAULT_EXIT_ATTEMPT_DELAY_MS, NULL);
        APP_ERROR_CHECK(err_code);
    }

    else
    {
        // Count of successive latching faults indicates that PTU is now in persistent latching fault
        m_lf_status.persistent = true;
    }

    return NRF_SUCCESS;
}


uint32_t ptu_latching_fault_init(ptu_sm_handler_t sm_handler)
{
    uint32_t err_code;

    if(sm_handler == NULL)
    {
        return NRF_ERROR_INVALID_PARAM;
    }

    m_ptu_sm_execute = sm_handler;

    memset(&m_lf_status, 0, sizeof(lf_status_t));

    err_code = app_timer_create(&m_clear_latching_fault_timer_id,
                                APP_TIMER_MODE_SINGLE_SHOT,
                                m_latch_fault_attempt_to_clear_timer_handler);
    APP_ERROR_CHECK(err_code);

    err_code = app_timer_create(&m_short_beacon_timer_id,
                            APP_TIMER_MODE_SINGLE_SHOT,
                            m_latch_fault_short_beacon_timer_handler);
    APP_ERROR_CHECK(err_code);

    err_code = app_timer_create(&m_reset_nof_consecutive_latching_fault_count_timer_id,
                            APP_TIMER_MODE_SINGLE_SHOT,
                            m_reset_nof_consecutive_latching_fault_count_timer_handler);
    APP_ERROR_CHECK(err_code);

    return NRF_SUCCESS;
}

/** @} */
