/* Copyright (c) Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file */

#include <math.h>

#include "ptu_power_ctl.h"
#include "ptu_test_mux.h"
#include "ptu_registry.h"
#include "nrf_assert.h"

/******************************************************************************/
/** @name Static functions and variables
 *  @{ */
/******************************************************************************/

/**@brief Definition of power transfer substates */
typedef enum
{
   POWER_TRANSFER_SUBSTATE_1_OPTIMAL,        /**< All PRUs in optimal substate       */
   POWER_TRANSFER_SUBSTATE_2_LOW,            /**< >= 1 PRU in low voltage substate   */
   POWER_TRANSFER_SUBSTATE_3_HIGH            /**< >= 1 PRU in high voltage substate  */
} power_transfer_substates_t;


static bool                         m_itx_adj_disabled = false; /**< Have we been told by PRU to only adjust poweramp input by test commands? */

static power_transfer_substates_t   m_substate;                 /**< Current substate. Updated by pru_scan_all() */

static uint16_t                     m_conn_handle_dominant;     /**< Handle to dominant PRU. Updated by pru_scan_all() */

/** @brief Inspect registry item and find the current value of VRECT_MIN and VRECT_HIGH
 *  @details If the registry item has set the optional dynamic values of these variables, these should be used.
 *
 *  @param[in]  p_reg_item Pointer to registry item to inspect.
 *  @param[out] p_vrect_high Pointer to vrect_high value.
 *  @param[out] p_vrect_min Pointer to vrect_min value.
 */
static void m_find_vrect_high_and_min(ptu_reg_item_t const * p_reg_item, uint16_t * p_vrect_high, uint16_t * p_vrect_min)
{
    bool has_dynamic_vrect_min = BLE_WPTS_BITFIELD_READ(p_reg_item->prev_pru_dynamic.optional_fields,
                                                        BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_MIN_DYN_BITMSK,
                                                        BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_MIN_DYN_BITPOS);

    bool has_dynamic_vrect_high = BLE_WPTS_BITFIELD_READ(p_reg_item->prev_pru_dynamic.optional_fields,
                                                         BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_HIGH_DYN_BITMSK,
                                                         BLE_WPTS_PRU_DYNAMIC_OPTION_VRECT_HIGH_DYN_BITPOS);

    *p_vrect_high = has_dynamic_vrect_high ?
                        p_reg_item->prev_pru_dynamic.vrect_high_dyn :
                        p_reg_item->prev_pru_static.vrect_high_static;

    *p_vrect_min = has_dynamic_vrect_min ?
                        p_reg_item->prev_pru_dynamic.vrect_min_dyn :
                        p_reg_item->prev_pru_static.vrect_min_static;
}

/** @brief Inspect registry item and see if reduced step size is allowed for increase and/or decrease.
 *  @details Will only set input parameters to true, never false.
 *  @param[in]  p_reg_item Pointer to registry item to inspect.
 *  @param[out] p_reduced_positive_step_size Pointer to bool saying whether positive step size should be reduced.
 *  @param[out] p_reduced_negative_step_size Pointer to bool saying whether negative step size should be reduced.
 */
static void m_update_power_amplifier_adjustment_step_size(ptu_reg_item_t const * p_reg_item,
                                                          bool * p_reduced_positive_step_size,
                                                          bool * p_reduced_negative_step_size)
{
    uint16_t vrect_high;
    uint16_t vrect_min;

    m_find_vrect_high_and_min(p_reg_item, &vrect_high, &vrect_min);

    if(p_reg_item->prev_pru_dynamic.vrect > round(vrect_high * PTU_ALLOWED_REDUCED_POSITIVE_ITX_COIL_ADJUSTMENT_MULTIPLIER))
    {
        *p_reduced_positive_step_size = true;
    }

    if(p_reg_item->prev_pru_dynamic.vrect < round(vrect_min * PTU_ALLOWED_REDUCED_NEGATIVE_ITX_COIL_ADJUSTMENT_MULTIPLIER))
    {
        *p_reduced_negative_step_size = true;
    }
}


/** @brief Inspect registry item for dominant PRU and see if reduced step size is allowed for negative and positive changes.
 *  @details Will only set input parameters to true, never false.
 *  @param[in]  p_dominant_pru_reg_item Pointer to dominant PRU registry item to inspect.
 *  @param[out] p_reduced_positive_step_size Pointer to bool saying whether positive step size should be reduced.
 *  @param[out] p_reduced_negative_step_size Pointer to bool saying whether negative step size should be reduced.
 */
static void m_update_power_amplifier_adjustment_step_size_for_dominant_pru(ptu_reg_item_t const * p_dominant_pru_reg_item,
                                                          bool * p_reduced_positive_step_size,
                                                          bool * p_reduced_negative_step_size)
{
    uint16_t vrect      = p_dominant_pru_reg_item->prev_pru_dynamic.vrect;
    uint16_t vrect_set  = p_dominant_pru_reg_item->prev_pru_static.vrect_set;

    // If V_RECT qualifies PTU for step size reduction
    if((vrect < round(vrect_set * PTU_DOMINANT_PRU_UPPER_LIMIT_VRECTSET_MULTIPLIER_FOR_REDUCTION_IN_STEP_SIZE)) &&
       (vrect > round(vrect_set * PTU_DOMINANT_PRU_LOWER_LIMIT_VRECTSET_MULTIPLIER_FOR_REDUCTION_IN_STEP_SIZE)))
    {
        *p_reduced_negative_step_size = true;
        *p_reduced_positive_step_size = true;
    }
}


/** @brief Scan through dynamic and static parameters received from all connected PRUs.
 *  @details
 * Update following variables:
 *  - m_substate
 *  - m_conn_handle_dominant
 *  - p_reduced_positive_step_size and p_reduced_negative_step_size
 *
 *  @param[out] p_reduced_positive_step_size Pointer to bool saying whether positive step size should be reduced
 *  @param[out] p_reduced_negative_step_size Pointer to bool saying whether negative step size should be reduced
 *  @param[out] p_increased_step_size Pointer to bool saying whether increased step size is allowed
 */
static void m_pru_scan_all(bool * p_reduced_positive_step_size, bool * p_reduced_negative_step_size, bool * p_increased_step_size)
{
    ptu_reg_item_t* registered_devices[PTU_MAX_CONNECTIONS];
    uint8_t n_reg_devices = ptu_reg_registered_devices_get(registered_devices);
    bool high_voltage_f = false;
    bool low_voltage_f = false;
    uint8_t l_putil_max = 0;
    int8_t dominant_pru_index = -1;

    *p_reduced_positive_step_size = false;
    *p_reduced_negative_step_size = false;
    *p_increased_step_size = true;

    for(uint8_t i = 0; i < n_reg_devices; ++i)
    {
        m_update_power_amplifier_adjustment_step_size(registered_devices[i], p_reduced_positive_step_size, p_reduced_negative_step_size);

        if(registered_devices[i]->prev_pru_dynamic.putil > l_putil_max)
        {
            l_putil_max = registered_devices[i]->prev_pru_dynamic.putil;
            dominant_pru_index = i;
        }

        if(registered_devices[i]->prev_pru_static.pru_category < BLE_WPTS_PRU_CATEGORY_4)
        {
            *p_increased_step_size = false;
        }

        // Note that the PRU Static bitfields vrect_min_static and vrect_high_static
        // are updated to values from the last read PRU Dynamic characteristic if the
        // respective optional fields are enabled.
        if(registered_devices[i]->prev_pru_dynamic.vrect > registered_devices[i]->prev_pru_static.vrect_high_static  - PTU_VRECT_HIGH_OFFSET)
        {
            high_voltage_f = true;
        }

        else if(registered_devices[i]->prev_pru_dynamic.vrect < registered_devices[i]->prev_pru_static.vrect_min_static)
        {
            low_voltage_f = true;
        }
    }

    if(high_voltage_f)
    {
        m_substate = POWER_TRANSFER_SUBSTATE_3_HIGH;
    }
    else if(low_voltage_f)
    {
        m_substate = POWER_TRANSFER_SUBSTATE_2_LOW;
    }
    else
    {
        m_substate = POWER_TRANSFER_SUBSTATE_1_OPTIMAL;
    }

    if(dominant_pru_index >= 0)
    {
        m_conn_handle_dominant = registered_devices[dominant_pru_index]->ble_wpts_c.conn_handle;

        m_update_power_amplifier_adjustment_step_size_for_dominant_pru(registered_devices[dominant_pru_index],
                                                                       p_reduced_positive_step_size,
                                                                       p_reduced_negative_step_size);
    }

}

/** @brief Find step size to be used given permissions to use reduced and/or increased step size.
 *  @param reduced_step_size Is reduced step size allowed?
 *  @param increased_step_size Is increased step size allowed?
 *  @return Resulting POWERAMP step size
 */
static uint32_t m_get_step_size(bool reduced_step_size, bool increased_step_size)
{
    if(reduced_step_size)
    {
        return PTU_POWERAMP_INPUT_STEP_SIZE_REDUCED;
    }

    else if(increased_step_size)
    {
        return PTU_POWERAMP_INPUT_STEP_SIZE_INCREASED;
    }

    return PTU_POWERAMP_INPUT_STEP_SIZE;
}


/** @brief Increase PTU poweramp output.
 *  @param reduced_positive_step_size Is step size reduced?
 *  @param increased_step_size Is step size increased?
 *  @return NRF_SUCCESS if successful, otherwise an error code will be returned.
 */
static uint32_t m_poweramp_input_step_up(bool reduced_positive_step_size, bool increased_step_size)
{
    uint32_t err_code;
    uint16_t level_to_set;
    uint16_t current_level;
    uint16_t step_size;

    err_code = ptu_tmux_poweramp_level_get(&current_level);

    if(err_code != NRF_SUCCESS)
    {
        return err_code;
    }

    step_size = m_get_step_size(reduced_positive_step_size, increased_step_size);

    if (current_level + step_size > PTU_POWERAMP_INPUT_MAX)
    {
        level_to_set = PTU_POWERAMP_INPUT_MAX;
    }

    else
    {
        level_to_set = current_level + step_size;
    }

    return ptu_power_ctrl_set_poweramp_input(level_to_set);
}

/** @brief Decrease PTU poweramp output.
 *  @param reduced_negative_step_size Is step size reduced?
 *  @param increased_step_size Is step size increased?
 *  @retval NRF_SUCCESS if successful, otherwise an error code will be returned.
 */
static uint32_t m_poweramp_input_step_down(bool reduced_negative_step_size, bool increased_step_size)
{
    uint32_t err_code;
    uint16_t level_to_set;
    uint16_t current_level;
    uint16_t step_size;

    err_code = ptu_tmux_poweramp_level_get(&current_level);

    if(err_code != NRF_SUCCESS)
    {
        return err_code;
    }

    step_size = m_get_step_size(reduced_negative_step_size, increased_step_size);

    if (((int32_t)current_level - (int32_t)step_size) < PTU_POWERAMP_INPUT_NOMINAL)
    {
        level_to_set = PTU_POWERAMP_INPUT_NOMINAL;
    }

    else
    {
        level_to_set = current_level - step_size;
    }

    return ptu_power_ctrl_set_poweramp_input(level_to_set);
}

/** @} */

/******************************************************************************/
/** @name API implementation
 *  @{ */
/******************************************************************************/


void ptu_power_ctrl_adjust(void)
{
    uint32_t err_code;
    bool reduced_positive_step_size;
    bool reduced_negative_step_size;
    bool increased_step_size;

    m_pru_scan_all(&reduced_positive_step_size, &reduced_negative_step_size, &increased_step_size);

    const pru_static_t *  l_pru_static  = &(ptu_reg_item_get_from_conn_handle(m_conn_handle_dominant)->prev_pru_static);
    const pru_dynamic_t * l_pru_dynamic = &(ptu_reg_item_get_from_conn_handle(m_conn_handle_dominant)->prev_pru_dynamic);

    switch(m_substate)
    {
        case POWER_TRANSFER_SUBSTATE_1_OPTIMAL:
            if(l_pru_dynamic -> vrect < (l_pru_static -> vrect_set - (PTU_POWER_TRANSFER_HYSTERISIS / 2)))
            {
                err_code = m_poweramp_input_step_up(reduced_positive_step_size, increased_step_size);
                APP_ERROR_CHECK(err_code);
            }
            else if(l_pru_dynamic -> vrect > (l_pru_static -> vrect_set + (PTU_POWER_TRANSFER_HYSTERISIS / 2)))
            {
                err_code = m_poweramp_input_step_down(reduced_negative_step_size, increased_step_size);
                APP_ERROR_CHECK(err_code);
            }
            break;

        case POWER_TRANSFER_SUBSTATE_2_LOW:
            // false as argument since modified step size is not permitted in substate 2
            err_code = m_poweramp_input_step_up(false, false);
            APP_ERROR_CHECK(err_code);
            break;

        case POWER_TRANSFER_SUBSTATE_3_HIGH:
            // false as argument since modified step size is not permitted in substate 3
            err_code = m_poweramp_input_step_down(false, false);
            APP_ERROR_CHECK(err_code);
            break;

        default:
            break;
    }
}

uint32_t ptu_power_ctrl_set_poweramp_input(uint16_t level)
{
    if(!m_itx_adj_disabled)
        return ptu_tmux_poweramp_level_set(level);
    return NRF_SUCCESS;
}

void ptu_power_ctrl_set_disable_power_amplifier_adjustments(bool val)
{
    m_itx_adj_disabled = val;
}

void ptu_power_ctrl_set_itx_val(ptu_tester_command_t cmd)
{
    uint16_t tmp_itx;

    switch(cmd)
    {
        case WPTS_PTU_TEST_CMD_NORMAL_OPERATION_POWER_CTL:
            break;

        case WPTS_PTU_TEST_CMD_INCREASE_ITX:
            APP_ERROR_CHECK(ptu_tmux_poweramp_level_get(&tmp_itx));
            APP_ERROR_CHECK(ptu_tmux_poweramp_level_set(tmp_itx  + PTU_POWERAMP_INPUT_TEST_STEP_SIZE));
            break;

        case WPTS_PTU_TEST_CMD_DECREASE_ITX:
            APP_ERROR_CHECK(ptu_tmux_poweramp_level_get(&tmp_itx));
            APP_ERROR_CHECK(ptu_tmux_poweramp_level_set(tmp_itx  -PTU_POWERAMP_INPUT_TEST_STEP_SIZE));
            break;

        case WPTS_PTU_TEST_CMD_NO_ADJ_ITX:
            break;

        default:
            ASSERT(false);
            break;
    }
}

/** @} */
