/* Copyright (c) 2012 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file */

#include <stdint.h>
#include <string.h>
#include <math.h>

#include "nrf_assert.h"
#include "ptu_conn_man.h"
#include "ptu_power_sharing.h"

/******************************************************************************/
/** @name Static funcitons and variables.
 *  @{ */
/******************************************************************************/

/** @brief Check if a device is capable of receiving power adjust commands. */
#define REG_DEVICE_CAN_ADJUST_POWER(DEV)                 \
     (DEV->prev_pru_static.info_adj_power_capability &&  \
     !DEV->p_adj_disabled &&                             \
     !DEV->active_p_adj_cooldown)


/**@struct reg_devices_prect_status_t.
 */
typedef struct
{
    uint32_t granted_prect_sum;    // Sum of granted Prect to PRUs (including PRUs waiting to allow to draw power)
    int32_t  available_power;      // Max available PTU power - granted_prect_sum
    bool     pending_power_adjust; // We are waiting for a "power adjusted response" from one or more PRUs
    bool     prus_waiting;         // One or more PRUs are waiting to be allowed to use power
    bool     temperature_warning;  // PTU temperature warning active.
} ptu_pctl_status_t;

static ptu_pctl_status_t m_pctl_status; // Status variables
static ptu_sm_handler_t m_sm_handler;

/**@brief Sends the control packet in prev_ctl to a PRU.
 *
 * @param reg_item_p registry entry for PRU to which to send control packet.
 */
static void m_ble_wpts_c_send_pru_control_wait_buffer(ptu_reg_item_t * reg_item_p)
{
    uint32_t err_code;

    if(BLE_WPTS_BITFIELD_READ(reg_item_p->adv_flags, BLE_WPTS_ADV_FLAG_TIME_SET_SUPPORT_BITMSK, BLE_WPTS_ADV_FLAG_TIME_SET_SUPPORT_BITPOS)== 1) // Supports time-set
    {
        reg_item_p -> prev_ctl.time_set = BLE_WPTS_TIME_SET_BITVAL;
    }

    else
    {
        reg_item_p -> prev_ctl.time_set = CTL_TIME_SET_DENIED;
    }

    do{
        reg_item_p->pending_ctl_write = true;
        err_code = ble_wpts_c_send_pru_control(&reg_item_p->ble_wpts_c, &reg_item_p->prev_ctl);
    } while(err_code == BLE_ERROR_NO_TX_PACKETS);
}

/**@brief Get truncated value for adjusted power.
 *
 * @param prect_max The PRUs PRECT_MAX value.
 *
 * @param power_adjust The power adjustment to apply.
 *
 * @retval Truncated value of the resulting PRECT.
 */
static uint8_t m_get_prect(uint8_t prect_max, ctl_adj_power_t power_adjust)
{
    uint8_t retval = 0;

    switch(power_adjust)
    {
           case CTL_ADJ_POWER_MAX_VAL:
                retval = prect_max;
                break;
            case CTL_ADJ_POWER_66_PCT_VAL:
                retval = (uint8_t)ceil(prect_max * 0.66);
                break;
            case CTL_ADJ_POWER_33_PCT_VAL:
                retval = (uint8_t)ceil(prect_max * 0.33);
                break;
            case CTL_ADJ_POWER_2_5_W_VAL:
                retval = 25;
                break;
            default:
                ASSERT(false);
                break;
    }
    return retval;
}

/**@brief Get the maximum Prect in [100 mW] for a PRU.
 *
 * @param reg_item registry item for the PRU.
 *
 * @retval The maximum Prect in [100 mW] for the PRU.
 */
static uint8_t m_prect_max_get(ptu_reg_item_t * reg_item)
{
    ASSERT(reg_item != NULL);
    uint8_t retval = 0;

    if(reg_item->prev_pru_static.info_adj_power_capability)
        retval = m_get_prect(reg_item->prev_pru_static.prect_max, reg_item->prev_ctl.adj_power);
    else
        retval = reg_item->prev_pru_static.prect_max;

    return retval;
}

/**@brief Get the maximum increased Prect if we step up the allowed
 * Prect for the PRU with one step.
 *
 * @param reg_item registry item for the PRU.
 *
 * @retval Increase in Prect in [100 mW].
 */
static uint8_t m_prect_diff_after_inc_get(ptu_reg_item_t * reg_item)
{
    ASSERT(reg_item != NULL);
    ctl_adj_power_t incremented_power_level = reg_item->prev_ctl.adj_power;
    switch(reg_item->prev_ctl.adj_power)
    {
        case CTL_ADJ_POWER_MAX_VAL:
            return 0;
        case CTL_ADJ_POWER_66_PCT_VAL:
            incremented_power_level = CTL_ADJ_POWER_MAX_VAL;
            break;
        case CTL_ADJ_POWER_33_PCT_VAL:
            incremented_power_level = CTL_ADJ_POWER_66_PCT_VAL;
            break;
        case CTL_ADJ_POWER_2_5_W_VAL:
            incremented_power_level = CTL_ADJ_POWER_33_PCT_VAL;
            break;
        default:
            ASSERT(false);
    }

    return m_get_prect(reg_item->prev_pru_static.prect_max, incremented_power_level)
           - m_get_prect(reg_item->prev_pru_static.prect_max, reg_item->prev_ctl.adj_power);
}



/**@brief Transmits a PRU control packet, and performs all required
 * bookkeeping related to performing the power ajust process.
 *
 * @param reg_item registry item for the PRU.
 *
 */
static void m_send_power_adjust_command(ptu_reg_item_t * reg_item)
{
    uint32_t err_code;

    m_ble_wpts_c_send_pru_control_wait_buffer(reg_item);
    reg_item->active_p_adj_cooldown = true;
    err_code = app_timer_cnt_get(&(reg_item->p_adj_time));
    APP_ERROR_CHECK(err_code);
    reg_item->adjusted_power = false;
    reg_item->p_rect_before_p_adj = reg_item->prev_pru_dynamic.vrect * reg_item->prev_pru_dynamic.irect;
}

/**@brief Search for PRUs having "adjust power" capability and send control
 * packet instructing the PRU to reduce power.
 *
 * @retval true if a PRU was instructed to reduce power.
 */
static bool m_pru_decrease_power()
{
    ptu_reg_item_t* match_reg_item;
    ptu_reg_item_t* registered_devices[PTU_MAX_CONNECTIONS];
    uint8_t n_reg_devices = ptu_reg_registered_devices_get(registered_devices);

    match_reg_item = NULL;

    for(uint8_t i = 0; i < n_reg_devices; i++)
    {

        if(REG_DEVICE_CAN_ADJUST_POWER(registered_devices[i]) && \
           (registered_devices[i]->prev_ctl.adj_power == CTL_ADJ_POWER_MAX_VAL))
        {
            match_reg_item = registered_devices[i];
            match_reg_item->prev_adj_power = CTL_ADJ_POWER_MAX_VAL;
            match_reg_item->prev_ctl.adj_power = CTL_ADJ_POWER_66_PCT_VAL;
            break;
        }
    }

    if(match_reg_item == NULL)
    {
        for(uint8_t i = 0; i < n_reg_devices; i++)
        {
            if(REG_DEVICE_CAN_ADJUST_POWER(registered_devices[i]) && \
               (registered_devices[i]->prev_ctl.adj_power == CTL_ADJ_POWER_66_PCT_VAL))
             {
                 match_reg_item = registered_devices[i];
                 match_reg_item->prev_adj_power = CTL_ADJ_POWER_66_PCT_VAL;
                 match_reg_item->prev_ctl.adj_power = CTL_ADJ_POWER_33_PCT_VAL;
                 break;
             }
         }
    }

    if(match_reg_item == NULL)
    {
        for(uint8_t i = 0; i < n_reg_devices; i++)
        {
            if(REG_DEVICE_CAN_ADJUST_POWER(registered_devices[i]) && \
               (registered_devices[i]->prev_ctl.adj_power == CTL_ADJ_POWER_33_PCT_VAL))
             {
                 match_reg_item = registered_devices[i];
                 match_reg_item->prev_adj_power = CTL_ADJ_POWER_33_PCT_VAL;
                 match_reg_item->prev_ctl.adj_power = CTL_ADJ_POWER_2_5_W_VAL;
                 break;
             }
         }
    }

    if(match_reg_item != NULL)
    {
        m_send_power_adjust_command(match_reg_item);
        return true;
    }
    else
    {
        return false;
    }
}

/**@brief Search for PRUs having "adjust power" capability and send control
 * packet allowing the PRU to increase power.
 *
 * @retval true if a PRU was allowed to increase power.
 */
static bool m_pru_increase_power()
{
    ptu_reg_item_t* match_reg_item;
    ptu_reg_item_t* registered_devices[PTU_MAX_CONNECTIONS];
    uint8_t n_reg_devices = ptu_reg_registered_devices_get(registered_devices);

    match_reg_item = NULL;

    for(uint8_t i = 0; i < n_reg_devices; i++)
    {
        if(REG_DEVICE_CAN_ADJUST_POWER(registered_devices[i]) && \
          (registered_devices[i]->prev_ctl.adj_power == CTL_ADJ_POWER_2_5_W_VAL) && \
          (m_prect_diff_after_inc_get(registered_devices[i]) <= m_pctl_status.available_power))
        {
            match_reg_item = registered_devices[i];
            match_reg_item->prev_adj_power = CTL_ADJ_POWER_2_5_W_VAL;
            match_reg_item->prev_ctl.adj_power = CTL_ADJ_POWER_33_PCT_VAL;
            break;
        }
    }

    if(match_reg_item == NULL)
    {
        for(uint8_t i = 0; i < n_reg_devices; i++)
        {
            if(REG_DEVICE_CAN_ADJUST_POWER(registered_devices[i]) && \
              (registered_devices[i]->prev_ctl.adj_power == CTL_ADJ_POWER_33_PCT_VAL) && \
              (m_prect_diff_after_inc_get(registered_devices[i]) <= m_pctl_status.available_power))
            {
                match_reg_item = registered_devices[i];
                match_reg_item->prev_adj_power = CTL_ADJ_POWER_33_PCT_VAL;
                match_reg_item->prev_ctl.adj_power = CTL_ADJ_POWER_66_PCT_VAL;
                break;
            }
        }
    }

    if(match_reg_item == NULL)
    {
        for(uint8_t i = 0; i < n_reg_devices; i++)
        {
            if(REG_DEVICE_CAN_ADJUST_POWER(registered_devices[i]) && \
               (registered_devices[i]->prev_ctl.adj_power == CTL_ADJ_POWER_66_PCT_VAL) &&\
               (m_prect_diff_after_inc_get(registered_devices[i]) <= m_pctl_status.available_power))
             {
                 match_reg_item = registered_devices[i];
                 match_reg_item->prev_adj_power = CTL_ADJ_POWER_66_PCT_VAL;
                 match_reg_item->prev_ctl.adj_power = CTL_ADJ_POWER_MAX_VAL;
                 break;
             }
         }
    }

    if(match_reg_item != NULL)
    {
        m_send_power_adjust_command(match_reg_item);
        return true;
    }
    else
    {
        return false;
    }
}

/**@brief Search for PRUs being permitted to enable charge output
 * and send control packet denying the PRU to do so.
 *
 * @retval true if a PRU was instructed disable charge output.
 */
static bool m_pru_deny_power()
{
    ptu_reg_item_t * match_reg_item;
    ptu_reg_item_t* registered_devices[PTU_MAX_CONNECTIONS];
    uint8_t n_reg_devices = ptu_reg_registered_devices_get(registered_devices);

    for(uint8_t i = 0; i < n_reg_devices; i++)
    {
        if(registered_devices[i]->prev_ctl.permissions == CTL_PERMISSION_PERMITTED)
        {
            match_reg_item = registered_devices[i];
            match_reg_item->prev_ctl.enable_pru_output = 0;
            match_reg_item->prev_ctl.enable_pru_charge_indicator = 0;
            match_reg_item->prev_ctl.permissions = m_pctl_status.temperature_warning ? CTL_PERMISSION_DENIED_HIGH_TEMP : CTL_PERMISSION_DENIED_POWER;
            m_ble_wpts_c_send_pru_control_wait_buffer(match_reg_item);
            return true;
         }
     }

    return false;
}

/**@brief Check if the maximum number of PRUs is being charged.
 *
 * @retval true if a PRU was instructed disable charge output.
 */
static bool m_max_number_of_devices_is_being_charged(void)
{
    ptu_reg_item_t* registered_devices[PTU_MAX_CONNECTIONS];
    uint8_t nof_devices = ptu_reg_registered_devices_get(registered_devices);
    uint8_t nof_devices_being_charged = 0;

    for(uint8_t i = 0; i < nof_devices; ++i)
    {
        if(registered_devices[i]->prev_ctl.enable_pru_output == 1 || registered_devices[i]->pending_ctl_write)
            nof_devices_being_charged++;
    }

    ASSERT(nof_devices_being_charged <= PTU_MAX_DEVICES);

    return nof_devices_being_charged == PTU_MAX_DEVICES;
}

/**@brief Search for PRUs not being permitted to enable charge output
 * and send control packet allowing the PRU to do so if the PTU can
 * tolerate this increased power draw.
 *
 * @retval true if a PRU was allowed to enable charge output.
 */
static bool m_pru_allow_power()
{
    ptu_reg_item_t* registered_devices[PTU_MAX_CONNECTIONS];
    ptu_reg_item_t * match_reg_item;
    uint8_t n_reg_devices = ptu_reg_registered_devices_get(registered_devices);

    for(uint8_t i = 0; i < n_reg_devices; i++)
    {
        if(registered_devices[i]->prev_ctl.permissions != CTL_PERMISSION_PERMITTED && \
           m_prect_max_get(registered_devices[i]) <= m_pctl_status.available_power && \
           registered_devices[i]->adjusted_power)
        {
            match_reg_item = registered_devices[i];
            match_reg_item->prev_ctl.enable_pru_output = 1;
            match_reg_item->prev_ctl.enable_pru_charge_indicator = 1;
            match_reg_item->prev_ctl.permissions = CTL_PERMISSION_PERMITTED;
            m_ble_wpts_c_send_pru_control_wait_buffer(match_reg_item);
            return true;
         }
     }

    return false;
}


/**@brief See if it is possible to adjust power drain from a not
 * yet registerred PRU so that it fits within the current remaining
 * power of the PTU.
 *
 * @param new_reg_item_p registry item for the PRU.
 *
 * @retval The power adjustment required for PRU to be within
 * power limit. CTL_ADJ_POWER_MAX_VAL if it is not possible
 * to get PRU within limit.
 */
static ctl_adj_power_t m_attempt_to_get_prect_within_available_resources(ptu_reg_item_t * new_reg_item_p)
{
    // If we can not adjust power on new device (of any reason), return MAX_VAL, and regular Power Sharing between
    // already connected PRUs will take place.
    ctl_adj_power_t power_adjust = CTL_ADJ_POWER_MAX_VAL;
    uint8_t current_adjust_val = 0;
    uint8_t prect_33_percent = (uint8_t)(new_reg_item_p->prev_pru_static.prect_max * 0.33);
    uint8_t prect_66_percent = prect_33_percent * 2;
    // See wether PTU has enough power for 66% or 33% of P_RECT_MAX
    if(prect_66_percent <= m_pctl_status.available_power)
    {
        current_adjust_val = prect_66_percent;
        power_adjust = CTL_ADJ_POWER_66_PCT_VAL;
    }

    else if(prect_33_percent <= m_pctl_status.available_power)
    {
        current_adjust_val = prect_33_percent;
        power_adjust = CTL_ADJ_POWER_33_PCT_VAL;
    }

    // What is biggest, our fraction of the P_RECT_MAX, or 2.5W ?
    // In any case, make sure PTU has enough to power left
    if(current_adjust_val < 25 && m_pctl_status.available_power >= 25)
        power_adjust = CTL_ADJ_POWER_2_5_W_VAL;

    return power_adjust;
}


/**@brief Collect status information from all PRUs and update m_pctl_status.
 */
static void m_status_update(void)
{
    ptu_reg_item_t* registered_devices[PTU_MAX_CONNECTIONS];
    uint8_t n_reg_devices = ptu_reg_registered_devices_get(registered_devices);
    m_pctl_status.granted_prect_sum = 0;
    m_pctl_status.pending_power_adjust = false;
    m_pctl_status.prus_waiting = false;

    for(uint8_t i = 0; i < n_reg_devices; i++)
    {
        if(registered_devices[i]->prev_pru_static.info_adj_power_capability &&
            !(registered_devices[i]->p_adj_disabled) &&
            !(registered_devices[i]->prev_pru_dynamic.adjust_power_response) &&
              registered_devices[i]->prev_ctl.adj_power != CTL_ADJ_POWER_MAX_VAL)
        {
            m_pctl_status.pending_power_adjust = true;
        }

        // Potential power adjust
        if(registered_devices[i]->pending_ctl_write)
        {
            m_pctl_status.pending_power_adjust = true;
        }

        // If any PRUs are denied to start charging.
        if(registered_devices[i]->prev_ctl.permissions != BLE_WPTS_PERMISSION_PERMITTED)
        {
            m_pctl_status.prus_waiting = true;
        }

        if(registered_devices[i]->prev_ctl.permissions == BLE_WPTS_PERMISSION_PERMITTED)
        {
            m_pctl_status.granted_prect_sum += m_prect_max_get(registered_devices[i]);
        }
    }

    m_pctl_status.available_power = (int32_t) (PTU_PTX_IN_MAX * PTU_POWER_EFFICIENCY);

    // If we have a temperature warning we reduce the total power the PTU can source
    // by PTU_PTX_IN_TEMP_WARNING_DEC %.
    if(m_pctl_status.temperature_warning)
    {
        m_pctl_status.available_power *= (100 - PTU_PTX_IN_TEMP_WARNING_DEC);
        m_pctl_status.available_power /= 100;
    }

    m_pctl_status.available_power -= m_pctl_status.granted_prect_sum;
}


/**@brief Disable PRU output for all connected PRUs.
 */
static void m_charge_disable_all()
{
    ptu_reg_item_t* registered_devices[PTU_MAX_CONNECTIONS];
    uint8_t n_reg_devices = ptu_reg_registered_devices_get(registered_devices);

    for(uint8_t i = 0; i < n_reg_devices ; i++)
    {
        registered_devices[i]->prev_ctl.enable_pru_output = 0;
        registered_devices[i]->prev_ctl.enable_pru_charge_indicator = 0;
        m_ble_wpts_c_send_pru_control_wait_buffer(registered_devices[i]);
        registered_devices[i]->pending_charge_disable = true;
    }
}

/**@brief Check if all PRUs are done charging, in which case all will be disabled.
 */
static void m_charge_complete_update(ptu_reg_item_t * p_reg_item, ble_wpts_c_evt_t * const p_wpts_c_evt)
{
    if(p_reg_item->charged && ptu_reg_all_charged())
    {
        m_charge_disable_all();
        m_sm_handler(PTU_SM_SIGNAL_PRU_ALL_CHARGE_COMPLETE);
    }
}

/**@brief See if PRU has adjusted power by looking at the I_RECT and V_RECT
 * values from PRU dynamic params.
 *
 * @param item The PRU item to check.
 *
 * @retval Has PRU adjusted power.
 */
static bool m_adjusted_power_sensed(ptu_reg_item_t * item)
{
    int32_t diff, required_diff;

    required_diff = ((int32_t)m_get_prect(item->prev_pru_static.prect_max, item->prev_adj_power)) \
                    - ((int32_t)m_get_prect(item->prev_pru_static.prect_max, item->prev_ctl.adj_power));
    diff = (int32_t)(item->prev_pru_dynamic.vout * item->prev_pru_dynamic.iout) - item->p_rect_before_p_adj;

    return (item->prev_pru_dynamic.optional_fields & BLE_WPTS_PRU_DYNAMIC_OPTION_VOUT_BITPOS) && \
           (item->prev_pru_dynamic.optional_fields & BLE_WPTS_PRU_DYNAMIC_OPTION_IOUT_BITPOS)  && \
            diff == required_diff;
}

/**@brief Update cooldown values for all PRUs which has received a power adjust command recently.
 * Also checks for PRUs that does not respond with power adjust bit correctly after being sent a power adjust command.
 */
static void m_update_pru_p_adj_status(void)
{
    uint32_t err_code, diff;
    ptu_reg_item_t* registered_devices[PTU_MAX_CONNECTIONS];
    uint8_t n_reg_devices = ptu_reg_registered_devices_get(registered_devices);
    uint32_t current_time;

    err_code = app_timer_cnt_get(&current_time);
    APP_ERROR_CHECK(err_code);

    for(uint8_t i = 0; i < n_reg_devices; ++i)
    {
        if(registered_devices[i] != NULL)
        {
            diff = ticks_diff(current_time, registered_devices[i]->p_adj_time);

            // Is power adjust cooldown period finished?
            if(registered_devices[i]->active_p_adj_cooldown && \
               diff >= APP_TIMER_TICKS(PTU_P_ADJ_CMD_COOLDOWN_MS, APP_TIMER_PRESCALER))
                registered_devices[i]->active_p_adj_cooldown = false;

            // Is the PRU not responding to the last power adjust command?
            if(diff >= APP_TIMER_TICKS(PTU_ADJUSTED_POWER_BIT_TIMEOUT_MS, APP_TIMER_PRESCALER) \
                && !(registered_devices[i]->adjusted_power ||   m_adjusted_power_sensed(registered_devices[i])))
            {
                registered_devices[i]->prev_ctl.adj_power = registered_devices[i]->prev_adj_power; // Assume PRU will remain at previous power level.
                registered_devices[i]->p_adj_disabled = true; // Do not send any more power adjust commands to PRU.

                // If this was the initial power adjust command and PRU does not respond, disconnect it.
                if(registered_devices[i]->prev_ctl.permissions == CTL_PERMISSION_PERMITTED_WAITING && registered_devices[i]->prev_ctl.adj_power != CTL_ADJ_POWER_MAX_VAL)
                    ptu_cm_remove_device(registered_devices[i]);
            }
        }
    }
}

/**@brief Update registry item and issue control packet for allowing PRU to charge.
 *
 * @param[in] reg_item_p Pointer to registry item that should be allowed to charge.
 */
static void m_allow_pru_to_charge(ptu_reg_item_t * reg_item_p)
{
    reg_item_p -> prev_ctl.enable_pru_output = 1;
    reg_item_p -> prev_ctl.enable_pru_charge_indicator = 1;
    reg_item_p -> prev_ctl.adj_power = CTL_ADJ_POWER_MAX_VAL;
    reg_item_p -> prev_ctl.permissions = CTL_PERMISSION_PERMITTED;
    m_ble_wpts_c_send_pru_control_wait_buffer(reg_item_p);
}

/**@brief Update registry item and issue control packet for denying PRU charge.
 * Will remove device from registry.
 *
 * @param[in] reg_item_p Pointer to registry item that should be denied charge.
 * @param permissions Reason for being denied.
 */
static void m_deny(ptu_reg_item_t * reg_item_p, ctl_perm_t permissions)
{
    reg_item_p -> prev_ctl.enable_pru_output = 0;

    // It is only in the case where PRU is too large Category that PRU charge could NEVER happen
    if(permissions == BLE_WPTS_PERMISSION_DENIED_CLASS)
        reg_item_p -> prev_ctl.enable_pru_charge_indicator = 0;
    else
        reg_item_p -> prev_ctl.enable_pru_charge_indicator = 1;

    reg_item_p -> prev_ctl.permissions = permissions;
    m_ble_wpts_c_send_pru_control_wait_buffer(reg_item_p);
    ptu_cm_remove_device(reg_item_p);
}

/**@brief Allow PRU to charge with adjusted power draw. Update registry item
 * and issue control packet indicating the required adjustment.
 *
 * @param[in] reg_item_p Pointer to registry item that should be allowed adjusted charge.
 */
static void m_permit_with_waiting_time_and_adjust(ptu_reg_item_t * reg_item_p)
{
    reg_item_p -> prev_ctl.enable_pru_output = 0;
    reg_item_p -> prev_ctl.enable_pru_charge_indicator = 1;
    reg_item_p -> prev_ctl.permissions = CTL_PERMISSION_PERMITTED_WAITING;
    reg_item_p -> prev_ctl.adj_power = m_attempt_to_get_prect_within_available_resources(reg_item_p);
    m_send_power_adjust_command(reg_item_p);
}

/**@brief Handle the power sharing of a new PRU.
 * @param[in] reg_item_p Pointer to registry item for new PRU.
 */
static void m_handle_new_device(ptu_reg_item_t * reg_item_p)
{
   if(reg_item_p -> prev_pru_static.pru_category > (PTU_CLASS + 2)) // PTU Class 2 = 1, Class 2 supports Category 3 PRU, hence " + 2 "
        m_deny(reg_item_p, CTL_PERMISSION_DENIED_CLASS);

    else if(m_max_number_of_devices_is_being_charged())
        m_deny(reg_item_p, CTL_PERMISSION_DENIED_DEVICES);

    else if(m_pctl_status.available_power < reg_item_p->prev_pru_static.prect_max || m_pctl_status.pending_power_adjust)
    {
        // If new device supports power adjustment, make an attempt to adjust its power.
        if(reg_item_p->prev_pru_static.info_adj_power_capability == 1)
            m_permit_with_waiting_time_and_adjust(reg_item_p);

        // Otherwise, adjust PTU can never fulfill the PRUs power demand, and it should be rejected.
        else
            m_deny(reg_item_p, CTL_PERMISSION_DENIED_POWER);
    }

    else
        m_allow_pru_to_charge(reg_item_p);
}

/** @} */

/******************************************************************************/
/** @name API implementation.
 *  @{ */
/******************************************************************************/
uint32_t ptu_power_sharing_init(ptu_sm_handler_t sm_handler)
{
    if(sm_handler == NULL)
        return NRF_ERROR_INVALID_PARAM;

    m_sm_handler = sm_handler;
    memset(&m_pctl_status, 0, sizeof(ptu_pctl_status_t));

    return NRF_SUCCESS;
}

void ptu_power_sharing_temperature_warning_set(bool warning)
{
    m_pctl_status.temperature_warning = warning;
}

void ptu_power_sharing_on_wpt_service_evt(ble_wpts_c_t * p_wpts_c, ble_wpts_c_evt_t * const p_wpts_c_evt)
{
    ptu_reg_item_t * new_reg_item_p = ptu_reg_item_get_from_conn_handle(p_wpts_c->conn_handle);

    m_update_pru_p_adj_status();

    if (new_reg_item_p != NULL)
    {
        // Perform charge_complete control
        if(p_wpts_c_evt->type == BLE_WPTS_C_EVT_PRU_ALERT ||
            p_wpts_c_evt->type == BLE_WPTS_C_EVT_PRU_DYNAMIC_READ_RESP)
        {
            m_charge_complete_update(new_reg_item_p, p_wpts_c_evt);
            m_status_update();

            if(p_wpts_c_evt->data.pru_dynamic.adjust_power_response == 1)
                new_reg_item_p->adjusted_power = true;
        }

        // Whenever we receive a PRU dynamic update we might wanna adjust power.
        switch(p_wpts_c_evt->type)
        {
            case BLE_WPTS_C_EVT_PRU_DYNAMIC_READ_RESP:

                // If we have a new device to register.
                if(new_reg_item_p->state != REG_ITEM_STATE_REGISTERED)
                {
                    m_handle_new_device(new_reg_item_p);
                }

                else
                {
                    if(!m_pctl_status.pending_power_adjust)
                    {
                        // If we have in sum have granted PRUs to consume more power than we can give
                        if(m_pctl_status.available_power < 0)
                        {
                            // Attempt to instruct PRUs to reduce power
                            if(!m_pru_decrease_power())
                            {
                                // If not possible to reduce PRU power, we have to deny some PRUs to use power.
                                (void)m_pru_deny_power();
                            }
                        }
                        // If any PRUs has been denied to start charging
                        else if(m_pctl_status.prus_waiting)
                        {
                            // Attempt to allow denied PRUs start charging.
                            if(!m_pru_allow_power())
                            {
                               // If this was not possible we try to reduce power to the other PRUs.
                               (void)m_pru_decrease_power();
                            }
                        }
                        else
                        {
                            // We have power "left". We allow PRUs to consume more power.
                            (void)m_pru_increase_power();
                        }
                    }
                }
                break;
            case BLE_WPTS_C_EVT_PRU_CONTROL_WRITE_RESP:
                new_reg_item_p->pending_ctl_write = false;
                break;
            default:
                break;
        }
    }
}

/** @} */
