/* Copyright (c) Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file */

#include <stdint.h>
#include <string.h>

#include "nrf_assert.h"
#include "ptu_registry.h"

/******************************************************************************/
/** @name Static funcitons and variables.
 *  @{ */
/******************************************************************************/

#if PTU_MAX_CONNECTIONS > 0
APP_TIMER_DEF(m_reg_timer_0);
#endif
#if PTU_MAX_CONNECTIONS > 1
APP_TIMER_DEF(m_reg_timer_1);
#endif
#if PTU_MAX_CONNECTIONS > 2
APP_TIMER_DEF(m_reg_timer_2);
#endif
#if PTU_MAX_CONNECTIONS > 3
APP_TIMER_DEF(m_reg_timer_3);
#endif
#if PTU_MAX_CONNECTIONS > 4
APP_TIMER_DEF(m_reg_timer_4);
#endif
#if PTU_MAX_CONNECTIONS > 5
APP_TIMER_DEF(m_reg_timer_5);
#endif
#if PTU_MAX_CONNECTIONS > 6
APP_TIMER_DEF(m_reg_timer_6);
#endif
#if PTU_MAX_CONNECTIONS > 7
APP_TIMER_DEF(m_reg_timer_7);
#endif

static ptu_reg_t m_registry;   /**< Registry database. */

/**@brief Set all values in a registry item to default values. 'timer_id' will be retained.
 * @param item is a pointer to the item to initialize.
 */
static void m_reg_item_init(ptu_reg_item_t * item)
{
    app_timer_id_t temp;
    temp = item->timer_id;
    memset(item, 0, sizeof(ptu_reg_item_t));
    item->ble_wpts_c.conn_handle = BLE_CONN_HANDLE_INVALID;
    item->timer_id = temp;
}

/**@brief Get unused (available) registry item.
 *
 * @return pointer to registry item. NULL if no item available.
 */
static ptu_reg_item_t * m_reg_item_unused_get(void)
{
    int i;

    for(i = 0; i < PTU_MAX_CONNECTIONS; i++)
    {
        if(m_registry.items[i].state == REG_ITEM_STATE_UNUSED)
        {
            return &m_registry.items[i];
        }
    }

    return NULL;
}

/** @} */

/******************************************************************************/
/** @name API implementation.
 *  @{ */
/******************************************************************************/

void ptu_reg_init(void)
{
    int i;
    m_registry.n_entries = 0;

    for(i = 0; i < PTU_MAX_CONNECTIONS; i++)
    {
        m_reg_item_init(&m_registry.items[i]);
    }
#if PTU_MAX_CONNECTIONS > 0
    m_registry.items[0].timer_id = m_reg_timer_0;
#endif
#if PTU_MAX_CONNECTIONS > 1
    m_registry.items[1].timer_id = m_reg_timer_1;
#endif
#if PTU_MAX_CONNECTIONS > 2
    m_registry.items[2].timer_id = m_reg_timer_2;
#endif
#if PTU_MAX_CONNECTIONS > 3
    m_registry.items[3].timer_id = m_reg_timer_3;
#endif
#if PTU_MAX_CONNECTIONS > 4
    m_registry.items[4].timer_id = m_reg_timer_4;
#endif
#if PTU_MAX_CONNECTIONS > 5
    m_registry.items[5].timer_id = m_reg_timer_5;
#endif
#if PTU_MAX_CONNECTIONS > 6
    m_registry.items[6].timer_id = m_reg_timer_6;
#endif
#if PTU_MAX_CONNECTIONS > 7
    m_registry.items[7].timer_id = m_reg_timer_7;
#endif
}

uint32_t ptu_reg_item_get_from_index(uint8_t index, ptu_reg_item_t ** item)
{
    if(index < PTU_MAX_CONNECTIONS)
    {
        *item = &m_registry.items[index];
        return NRF_SUCCESS;
    }

    return NRF_ERROR_INVALID_PARAM;
}

ptu_reg_item_t * ptu_reg_item_get_from_conn_handle(uint16_t conn_handle)
{
    if(m_registry.n_entries > 0 && conn_handle != BLE_CONN_HANDLE_INVALID)
    {
        for(uint8_t i = 0; i < PTU_MAX_CONNECTIONS; i++)
        {
            if( m_registry.items[i].ble_wpts_c.conn_handle == conn_handle && \
                m_registry.items[i].state != REG_ITEM_STATE_UNUSED)
            {
                return &m_registry.items[i];
            }
        }
    }
    return NULL;
}

ptu_reg_item_t * ptu_reg_item_get_from_address(ble_gap_addr_t * address)
{
    if(m_registry.n_entries > 0)
    {
        for(uint8_t i = 0; i < PTU_MAX_CONNECTIONS; i++)
        {
            if(memcmp(m_registry.items[i].address.addr, address->addr, BLE_GAP_ADDR_LEN) == 0 && \
               m_registry.items[i].state != REG_ITEM_STATE_UNUSED)
            {
                return &m_registry.items[i];
            }
        }
    }
    return NULL;
}

ptu_reg_item_t * ptu_reg_item_oldest_get(ptu_reg_item_state_t state)
{
    uint8_t age_max = 0;
    ptu_reg_item_t * retval = NULL;

    ASSERT(state != REG_ITEM_STATE_UNUSED);

    for(uint8_t i = 0; i < PTU_MAX_CONNECTIONS; i++)
    {
        if(m_registry.items[i].state == state && m_registry.items[i].age >= age_max)
        {
            age_max = m_registry.items[i].age;
            retval = &m_registry.items[i];
        }
    }
    return retval;
}

ptu_reg_item_t * ptu_reg_item_add(ble_gap_evt_adv_report_t * p_adv_report, ptu_reg_item_state_t init_state)
{
    ptu_reg_item_t * reg_item = ptu_reg_item_get_from_address(&p_adv_report->peer_addr);
    ble_wpts_service_data_t curr_service_data;
    uint32_t err_code;

    ASSERT(reg_item == NULL); // Assert this device not registered allready
    reg_item = m_reg_item_unused_get();

    if(reg_item != NULL)
    {
        // We are about to add a new item. Increase age of all other items in registry.
        for(uint8_t i = 0; i < PTU_MAX_CONNECTIONS; i++)
        {
            if(m_registry.items[i].state != REG_ITEM_STATE_UNUSED)
            {
                m_registry.items[i].age++;
            }
        }

        // Need to extract advdata here to get primary service handle
        err_code = ble_wpts_service_data_read(p_adv_report, &curr_service_data);
        APP_ERROR_CHECK(err_code);
        ble_wpts_char_handles_update(curr_service_data.gatt_prim_srv_handle, &reg_item->ble_wpts_c);
        memcpy(&reg_item->address, &p_adv_report->peer_addr, sizeof(ble_gap_addr_t));
        reg_item->adv_flags = curr_service_data.adv_flags;
        reg_item->state = init_state;
        m_registry.n_entries++;
        return reg_item;
    }
    return NULL;
}

void ptu_reg_item_delete(ptu_reg_item_t * item)
{
    uint8_t i;

    for(i = 0; i < PTU_MAX_CONNECTIONS; i++)
    {
        if(item == &m_registry.items[i])
        {
            break;
        }
    }

    // Assert that this is a valid registry address.
    ASSERT(i < PTU_MAX_CONNECTIONS);

    if(item->state != REG_ITEM_STATE_UNUSED)
    {
        ASSERT(m_registry.n_entries != 0);

        // Decrease age of items older than the one being deleted.
        for(i = 0; i < PTU_MAX_CONNECTIONS; i++)
        {
            if(m_registry.items[i].state != REG_ITEM_STATE_UNUSED && \
               m_registry.items[i].age > item->age)
            {
                m_registry.items[i].age--;
            }
        }

        m_registry.n_entries--;
    }
    m_reg_item_init(item);
}

uint8_t ptu_reg_registered_devices_get(ptu_reg_item_t** registered_devices)
{
    uint8_t n_devices = 0;

    for(uint8_t i = 0; i < PTU_MAX_CONNECTIONS; i++)
    {
        if(m_registry.items[i].state == REG_ITEM_STATE_REGISTERED)
        {
            registered_devices[n_devices++] = &m_registry.items[i];
        }
    }
    return n_devices;
}

uint8_t ptu_reg_n_entries_get(void)
{
    return m_registry.n_entries;
}

bool ptu_reg_all_charged(void)
{
    int i;

    for(i = 0; i < PTU_MAX_CONNECTIONS; i++)
    {
        if((m_registry.items[i].state != REG_ITEM_STATE_UNUSED) && (!m_registry.items[i].charged) )
        {
            return false;
        }
    }
    return true;
}

uint32_t ptu_reg_set_device_waiting_to_connect(ptu_reg_item_t * reg_item_p, uint8_t * address)
{
    uint8_t zero_address[BLE_GAP_ADDR_LEN] = {0};
    if (memcmp(zero_address, address, BLE_GAP_ADDR_LEN) == 0)
        return NRF_ERROR_INVALID_ADDR;

    reg_item_p->state = REG_ITEM_STATE_WAITING_TO_CONNECT;
    reg_item_p->ble_wpts_c.conn_handle = BLE_CONN_HANDLE_INVALID;
    memcpy(reg_item_p->address.addr, address, BLE_GAP_ADDR_LEN);

    return NRF_SUCCESS;
}
/** @} */
