/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
 */

//lint -save -esym(438,res_impedance_running_average)
//lint -save -esym(550,res_impedance_running_average)

#include <stdbool.h>
#include <stdint.h>
#include <string.h>

#include "ptu_sensors.h"
#include "nrf_assert.h"
#include "ptu_registry.h"
#include "ptu_test_mux.h"


#define PTU_SIZEOF_LOAD_VAL_BUF 8                                              ///< Size of P_TX_IN load variation buffer
#define PTU_SENSORS_PTX_LOAD_VARIATION_MAX_PERIOD_MS 15                        ///< Maximum allowed period for having a 100Hz +/- 20% load variation
#define PTU_SENSORS_PTX_LOAD_VARIATION_MIN_PERIOD_MS 6                         ///< Minimum allowed period for having a 100Hz +/- 20% load variation
#define PTU_SENSORS_PTX_LOAD_VARIATION_MAX_PERIOD_TICKS APP_TIMER_TICKS(PTU_SENSORS_PTX_LOAD_VARIATION_MAX_PERIOD_MS, APP_TIMER_PRESCALER) ///< Max allowed period in ticks
#define PTU_SENSORS_PTX_LOAD_VARIATION_MIN_PERIOD_TICKS APP_TIMER_TICKS(PTU_SENSORS_PTX_LOAD_VARIATION_MIN_PERIOD_MS, APP_TIMER_PRESCALER) ///< Min allowed period in ticks
#define PREV(val) ((val - 1) < 0 ? PTU_SIZEOF_LOAD_VAL_BUF - 1 : (val - 1))    ///< Get previous index in circular buffer
#define NEXT(val) ((val + 1)  % PTU_SIZEOF_LOAD_VAL_BUF)                       ///< Get next index in circular buffer
#define GOTO_NEXT_IDX(idx) (idx = NEXT(idx))                                        ///< Circular buffer index update
#define GET_BIT(val, idx) ((val >> idx) & 1)                                        ///< Read bit value

/******************************************************************************/
/** @name Static functions and variables
 *  @{ */
/******************************************************************************/


typedef struct ptu_load_var_measurement_s {
    uint32_t ticks;
    uint16_t val;
} ptu_sensors_load_variation_sample_t;                                                 ///< Struct representing a single P_TX sample

static ptu_sensors_load_variation_sample_t m_lbe_samples[PTU_SIZEOF_LOAD_VAL_BUF];       ///< Buffer used for storing P_TX_IN load variation data


static ptu_sensor_data_t m_ptu_sensor_data;                                            ///< The latest version of the sensor data
APP_TIMER_DEF(m_load_var_detected_timer_id);                                                ///< Load variation detect timer ID
static ptu_sm_handler_t m_sm_handler;                                                  ///< State machine event handler


/**
 * Function storing measured resonator impedance to buffer and using max/min values
 * of this buffer to determine whether a load variation has been detected.
 */
static void m_load_var_detect()
{
    static bool m_res_imp_buf_filled_once = false;
    static uint8_t m_res_imp_buf_index = 0;
    static uint16_t m_res_imp_buf[PTU_Z_MEAS_BUF_LENGTH];
    uint16_t res_impedance_max = 0;
    uint16_t res_impedance_min = 0xffff;
    uint16_t res_impedance_var_max = 0;
    uint32_t err_code;

    // Save samples to array
    m_res_imp_buf[m_res_imp_buf_index] = m_ptu_sensor_data.res_impedance;

    m_res_imp_buf_index++;

    if(m_res_imp_buf_index == PTU_Z_MEAS_BUF_LENGTH)
    {
        m_res_imp_buf_filled_once = true;
    }

    m_res_imp_buf_index = (m_res_imp_buf_index % PTU_Z_MEAS_BUF_LENGTH);

    // Only detect load variation after buffer has been filled at least once
    if(m_res_imp_buf_filled_once)
    {
        // Find max and min values in buffer
        for (uint8_t i=0; i < PTU_Z_MEAS_BUF_LENGTH; i++)
        {
            if(m_res_imp_buf[i] > res_impedance_max)
            {
                res_impedance_max = m_res_imp_buf[i];
            }

            if(m_res_imp_buf[i] < res_impedance_min)
            {
                res_impedance_min = m_res_imp_buf[i];
            }
        }

        res_impedance_var_max = (res_impedance_max - res_impedance_min);

        if(res_impedance_var_max > PTU_Z_TX_IN_LOAD_DETECT)
        {
            m_ptu_sensor_data.load_var_detected = true;
            err_code = app_timer_stop(m_load_var_detected_timer_id);
            APP_ERROR_CHECK(err_code);
            err_code = app_timer_start(m_load_var_detected_timer_id, APP_TIMER_TICKS(PTU_LOAD_VARIATION_DETECT_LIFETIME_MS, APP_TIMER_PRESCALER), NULL);
            APP_ERROR_CHECK(err_code);
        }
    }
}

/**
 * Function detecting rogue object based on power delievered to power amplifier and
 * reported power consumption from registered PRUs.
 */
static void m_rogue_object_detect(void)
{
    uint32_t i;
    uint32_t err_code;
    ptu_reg_item_t * p_reg_item = NULL;
    uint32_t reported_power_sum = 0;
    uint32_t poweramp_input_power = (m_ptu_sensor_data.v_ina*m_ptu_sensor_data.i_ina)/ 100000 ;//[1/10 W/]

    for(i=0;i<PTU_MAX_CONNECTIONS; i++)
    {
        err_code = ptu_reg_item_get_from_index(i, &p_reg_item);
        APP_ERROR_CHECK(err_code);

        if(p_reg_item->state == REG_ITEM_STATE_REGISTERED)
        {
            reported_power_sum += p_reg_item->prev_pru_dynamic.prect;
        }
    }

    if((poweramp_input_power * PTU_POWER_EFFICIENCY) > (reported_power_sum + PTU_ROGUE_OBJECT_DETECT_THRESHOLD))
    {
        m_ptu_sensor_data.rogue_obj_detected = true;
    }
    else
    {
        m_ptu_sensor_data.rogue_obj_detected  = false;
    }
}

/**
 * Load variation detect timer handler.
 * @param p_context context data.
 */
static void m_load_var_detected_timer_handler(void * p_context)
{
    m_ptu_sensor_data.load_var_detected = false;
}

/**
 * Locate the first sample where the the next value is smaller.
 * @param p_first_top Pointer to where the resulting index will be stored.
 * @param last_updated_idx Tells what index was last updated.
 * @return true if top is found.
 */
static bool m_find_first_top(uint8_t * p_first_top, uint8_t last_updated_idx)
{
    uint8_t i;

    for ( i = NEXT(last_updated_idx); i != last_updated_idx ; GOTO_NEXT_IDX(i))
    {
        // Find index where previous value was smaller, and next value is smaller or equal to current
        if(m_lbe_samples[i].val > m_lbe_samples[NEXT(i)].val)
        {
            *p_first_top = i;
            return true;
        }
    }

    return false;
}

/**
 * Locate the first sample where PTX delta from first top is sufficient for Long Beacon Extension.
 * @param p_bottom Pointer to where the resulting index will be stored.
 * @param first_top Index of first top sample.
 * @return true if bottom is found.
 */
static bool m_find_bottom(uint8_t * p_bottom, uint8_t first_top)
{
    uint32_t i;

    for( i = NEXT(first_top);  (i != first_top) && (m_lbe_samples[first_top].val - m_lbe_samples[i].val < PTU_P_TX_IN_LOAD_DETECT); GOTO_NEXT_IDX(i))
    {
        if(ticks_diff(m_lbe_samples[i].ticks, m_lbe_samples[first_top].ticks) > PTU_SENSORS_PTX_LOAD_VARIATION_MAX_PERIOD_TICKS)
            return false;
    }

    *p_bottom = i;

    return true;
}

/**
 * Locate the second top in the extended beacon sample list, and check if the load variation is valid.
 *
 * This function will start looking at the first sample AFTER the given first top.
 * The initial bottom is assumed to be the first sample AFTER the given first top.
 * However, if the current sample has lower value than the previous one, it will be
 * assigned as the new bottom.
 *
 * The function will return immediately when it sees a valid load variation.
 *
 * @param first_top Index of first top sample.
 * @param bottom Pointer to current bottom sample..
 * @return true if valid load variation is found
 */
static bool m_find_second_top_and_validate(uint8_t first_top, uint8_t bottom)
{
    uint32_t diff;
    int8_t i;

    i = bottom;

    // Find first point that has large enough P_TX diff from bottom, and that has the correct distance from the first top
    do
    {
        GOTO_NEXT_IDX(i);

        if(m_lbe_samples[i].val < m_lbe_samples[bottom].val) // we have found a lower bottom, this will not be a second top
        {
            bottom = i;

            // The bottom can not be the second top, jump to next sample.
            continue;
        }

        // If we see a sample with value lower than the previous sample OR (the sample has the same value
        // as the previous sample AND it is not the same value as the bottom sample), return 'false'
        // as the sample we are looking at might be part of different load variation period or the duty cycle
        // of the current period is incorrect.
        else if( m_lbe_samples[PREV(i)].val > m_lbe_samples[i].val ||
                (m_lbe_samples[PREV(i)].val == m_lbe_samples[i].val && m_lbe_samples[i].val != m_lbe_samples[bottom].val))
            return false;

        diff = ticks_diff(m_lbe_samples[i].ticks, m_lbe_samples[first_top].ticks);

        if((diff <= PTU_SENSORS_PTX_LOAD_VARIATION_MAX_PERIOD_TICKS && // Timing is correct
            diff >= PTU_SENSORS_PTX_LOAD_VARIATION_MIN_PERIOD_TICKS) &&
           m_lbe_samples[i].val - m_lbe_samples[bottom].val >= PTU_P_TX_IN_LOAD_DETECT)// Variation from bottom is sufficient
        {
            return true;
        }

        // Make sure we only iterate once over the circular buffer.
        // We can not compare the ticks to see if we have gone through the circular buffer as you might have overflow situations in the ticks value.
        }while(i != first_top);

    return false;
}

/**
 * Store current ticks and ptx value in a list of samples
 *
 * @param last_updated_index Where should we store the sample
 */
static void m_save_ptx_sample(uint8_t last_updated_index)
{
    uint32_t err_code, curr_ticks;

    err_code = ptu_tmux_ptx_get(&(m_ptu_sensor_data.p_tx_in));
    APP_ERROR_CHECK(err_code);

    err_code = app_timer_cnt_get(&curr_ticks);
    APP_ERROR_CHECK(err_code);

    // Store sample with current PTX value and timestamp
    m_lbe_samples[last_updated_index].val  = m_ptu_sensor_data.p_tx_in;
    m_lbe_samples[last_updated_index].ticks = curr_ticks;
}
/** @} */

/******************************************************************************/
/** @name API implementation
 *  @{ */
/******************************************************************************/

uint32_t ptu_sensors_init(ptu_sm_handler_t sm_handler)
{
    uint32_t err_code;

    if(sm_handler == NULL)
        return NRF_ERROR_INVALID_PARAM;

    m_sm_handler = sm_handler;

    err_code = app_timer_create(&m_load_var_detected_timer_id,
                            APP_TIMER_MODE_SINGLE_SHOT,
                            m_load_var_detected_timer_handler);
    APP_ERROR_CHECK(err_code);

    memset(&m_ptu_sensor_data, 0, sizeof(ptu_sensor_data_t));

    return NRF_SUCCESS;
}

uint32_t ptu_sensors_read(void)
{
    uint32_t err_code;
    bool poweramp_on;

    // Read all sensors
    err_code = ptu_tmux_vtx_get(&m_ptu_sensor_data.v_ina);
    if (err_code != NRF_SUCCESS) return err_code;

    err_code = ptu_tmux_itx_get(&m_ptu_sensor_data.i_ina);
    if (err_code != NRF_SUCCESS) return err_code;

    err_code = ptu_tmux_ptx_get(&m_ptu_sensor_data.p_tx_in);
    if (err_code != NRF_SUCCESS) return err_code;

    err_code = ptu_tmux_temperature_get(&m_ptu_sensor_data.temperature);
    if (err_code != NRF_SUCCESS) return err_code;

    if(!m_ptu_sensor_data.temperature_warning)
    {
        if(m_ptu_sensor_data.temperature > PTU_TEMPERATURE_WARNING_SET)
        {
            m_ptu_sensor_data.temperature_warning = true;
        }
    }
    else
    {
        if(m_ptu_sensor_data.temperature < PTU_TEMPERATURE_WARNING_CLEAR)
        {
            m_ptu_sensor_data.temperature_warning = false;
        }
    }

    // Derive load impedance from vtx and itx if resonator is on
    err_code = ptu_tmux_poweramp_enable_get(&poweramp_on);
    if (err_code != NRF_SUCCESS) return err_code;

    if(poweramp_on)
    {
        if(m_ptu_sensor_data.i_ina > 0)
        {
            m_ptu_sensor_data.res_impedance = m_ptu_sensor_data.v_ina / m_ptu_sensor_data.i_ina;
        }
        else
        {
            m_ptu_sensor_data.res_impedance = 0xffff; // "infinite"
        }

        // Load variatio detection
        m_load_var_detect();

        // Rogue object detection
        m_rogue_object_detect();

        // Load detection
        if(m_ptu_sensor_data.res_impedance < PTU_Z_TX_IN_NO_LOAD)
        {
            m_ptu_sensor_data.load_detected = true;
        }
        else
        {
            m_ptu_sensor_data.load_detected = false;
        }
    }

    // Check for event conditions:
    bool b_under_temperature = false;
    bool b_over_temperature = false;
    bool b_itx_over_current = false;
    bool b_itx_under_current = false;

    if (m_ptu_sensor_data.temperature < PTU_TEMPERATURE_MIN)
    {
        b_under_temperature = true;
    }
    else if (m_ptu_sensor_data.temperature > PTU_TEMPERATURE_MAX)
    {
        b_over_temperature = true;
    }

    if (m_ptu_sensor_data.i_ina < PTU_ITX_MIN)
    {
        b_itx_under_current = true;
    }
    else if (m_ptu_sensor_data.i_ina > PTU_ITX_MAX)
    {
        b_itx_over_current = true;
    }

    if (b_over_temperature || b_under_temperature || b_itx_over_current || b_itx_under_current)
    {
        m_sm_handler(PTU_SM_SIGNAL_LOCAL_FAULT);
    }
    else
    {
        m_sm_handler(PTU_SM_SIGNAL_LOCAL_FAULT_CLEARED);
    }

    return NRF_SUCCESS;
}

uint32_t ptu_sensors_data_get(const ptu_sensor_data_t ** sensors_data)
{
    ASSERT(sensors_data != NULL);

    *sensors_data = &m_ptu_sensor_data;
    return NRF_SUCCESS;
}

void ptu_sensors_clear_long_beacon_extension_load_variation_buffer(void)
{
    memset(m_lbe_samples, 0, sizeof(ptu_sensors_load_variation_sample_t) * PTU_SIZEOF_LOAD_VAL_BUF);
}

bool ptu_sensors_valid_long_beacon_extension_load_variation_found(void)
{
    static uint8_t last_updated_index = 0;
    uint8_t first_top, bottom;

    // Jump to next index of sample buffer and update its value
    GOTO_NEXT_IDX(last_updated_index);

    m_save_ptx_sample(last_updated_index);

    // Find first top of wave
    if(!m_find_first_top(&first_top, last_updated_index))
        return false;

    // Find the bottom
    if(!m_find_bottom(&bottom, first_top))
        return false;

    // Find the second top and validate load variation
    return m_find_second_top_and_validate(first_top, bottom);
}

/** @} */

//lint -restore
