/* Copyright (c) Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
*/
#include "ptu_sm.h"

/******************************************************************************/
/** @name Static funcitons and variables.
 *  @{ */
/******************************************************************************/

/**@brief Find the next state given the current state and the state machine event.
  *
  *@param signal  State machine signal.
  *@param p_state Current state variables.
  *
  *@return The resulting state.
 */
static ptu_sm_state_t m_get_next_state(ptu_sm_signal_type_t signal, ptu_sm_state_vars_t * p_state)
{
    ptu_sm_state_t next_state = p_state->current_state;

    switch(p_state->current_state)
    {
        case PTU_SM_STATE_CONFIGURATION:
            if(     signal == PTU_SM_SIGNAL_CONFIGURATION_COMPLETE)    next_state = PTU_SM_STATE_POWER_SAVE;
            else if(signal == PTU_SM_SIGNAL_LOCAL_FAULT)               next_state = PTU_SM_STATE_LOCAL_FAULT;
            break;

        case PTU_SM_STATE_POWER_SAVE:
            if(     signal == PTU_SM_SIGNAL_LOCAL_FAULT)               next_state = PTU_SM_STATE_LOCAL_FAULT;
            else if(signal == PTU_SM_SIGNAL_SYSTEM_ERROR)              next_state = PTU_SM_STATE_LATCH_FAULT;
            else if(signal == PTU_SM_SIGNAL_ROGUE_OBJECT_DETECTED)     next_state = PTU_SM_STATE_LATCH_FAULT;
            else if(signal == PTU_SM_SIGNAL_PRU_ADV_RCVD)              next_state = PTU_SM_STATE_LOW_POWER;
            break;

        case PTU_SM_STATE_LOW_POWER:
            if(     signal == PTU_SM_SIGNAL_LOCAL_FAULT)               next_state = PTU_SM_STATE_LOCAL_FAULT;
            else if(signal == PTU_SM_SIGNAL_SYSTEM_ERROR)              next_state = PTU_SM_STATE_LATCH_FAULT;
            else if(signal == PTU_SM_SIGNAL_ROGUE_OBJECT_DETECTED)     next_state = PTU_SM_STATE_LATCH_FAULT;
            else if(signal == PTU_SM_SIGNAL_REGISTRATION_TIMEOUT ||
                    signal == PTU_SM_SIGNAL_ALL_DEVICES_DISCONNECTED)  next_state = PTU_SM_STATE_POWER_SAVE;
            else if(signal == PTU_SM_SIGNAL_CHARGE_START)              next_state = PTU_SM_STATE_POWER_TRANSFER;
            break;

        case PTU_SM_STATE_POWER_TRANSFER:
            if(     signal == PTU_SM_SIGNAL_LOCAL_FAULT)               next_state = PTU_SM_STATE_LOCAL_FAULT;
            else if(signal == PTU_SM_SIGNAL_SYSTEM_ERROR)              next_state = PTU_SM_STATE_LATCH_FAULT;
            else if(signal == PTU_SM_SIGNAL_ROGUE_OBJECT_DETECTED)     next_state = PTU_SM_STATE_LATCH_FAULT;
            else if(signal == PTU_SM_SIGNAL_ALL_DEVICES_DISCONNECTED ||
                    signal == PTU_SM_SIGNAL_PRU_ALL_CHARGE_COMPLETE)   next_state = PTU_SM_STATE_POWER_SAVE;
            break;

        case PTU_SM_STATE_LATCH_FAULT:
            if(     signal == PTU_SM_SIGNAL_LOCAL_FAULT)               next_state = PTU_SM_STATE_LOCAL_FAULT;
            else if(signal == PTU_SM_SIGNAL_LATCHING_FAULT_CLEARED_BY_USER)
                                                                            next_state = PTU_SM_STATE_CONFIGURATION;
            break;

        case PTU_SM_STATE_LOCAL_FAULT:
            if(     signal == PTU_SM_SIGNAL_LOCAL_FAULT_CLEARED)
            {
                if(p_state->prev_state == PTU_SM_STATE_LATCH_FAULT)
                    next_state = PTU_SM_STATE_LATCH_FAULT;
                else
                    next_state = PTU_SM_STATE_CONFIGURATION;
            }
            break;

        default:
            break;
    }

    return next_state;
}

/** @} */

/******************************************************************************/
/** @name API implementation.
 *  @{ */
/******************************************************************************/

ptu_sm_state_t ptu_sm_execute(ptu_sm_signal_type_t signal, ptu_sm_state_vars_t * p_state)
{
    static ptu_sm_state_t current_state = PTU_SM_STATE_CONFIGURATION;

    if(signal != PTU_SM_SIGNAL_NULL && p_state != 0) // used for debugging
    {
        p_state->next_state = m_get_next_state(signal, p_state);
        p_state->prev_state = p_state->current_state;
        p_state->current_state = p_state->next_state;
        current_state = p_state->current_state;
    }

    return current_state;
}

/** @} */
