/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file */

#include <stdbool.h>
#include <stdint.h>
#include <string.h>

#include "ptu_test_mux.h"
#include "ptu_hal.h"
#include "nrf_assert.h"
#include "ptu_hw_config.h"
#include "app_error.h"

/******************************************************************************/
/** @name Static funcitons and variables.
 *  @{ */
/******************************************************************************/

static tmux_select_t get_mux_select;           /**< Multiplexer setting for **hal**get() functions  */
static tmux_select_t set_mux_select;           /**< Multiplexer setting for **hal**set() functions  */
static ptu_tmux_override_t override_values;    /**< Collection of override values */

/**@brief Write all override values to **hal**set() functions
 * if overriding of **set functions is enabled.
 */
static void m_update_hal_from_override_values(void)
{
    uint32_t err_code;

    if(set_mux_select == TMUX_SELECT_OVERRIDE)
    {
        err_code = ptu_hal_poweramp_enable_set(override_values.power_amp_enable);
        APP_ERROR_CHECK(err_code);
        err_code = ptu_hal_poweramp_level_set(override_values.power_amp_level);
        APP_ERROR_CHECK(err_code);
    }
}

/** @} */

/******************************************************************************/
/** @name API implementation.
 *  @{ */
/******************************************************************************/

uint32_t tmux_init(void)
{
    override_values.itx = PTU_ITX_DEFAULT;
    override_values.vtx = PTU_VTX_DEFAULT;
    override_values.temperature = PTU_TEMPERATUR_DEFAULT;
    override_values.power_amp_enable = false;
    override_values.power_amp_level = PTU_POWERAMP_INPUT_DEFAULT;
    get_mux_select = TMUX_SELECT_NO_OVERRIDE;
    set_mux_select = TMUX_SELECT_NO_OVERRIDE;
    return ptu_hal_init();
}

uint32_t ptu_tmux_poweramp_enable_set(bool enable)
{
    TMUX_SET(ptu_hal_poweramp_enable_set, enable);
}

uint32_t ptu_tmux_poweramp_level_set(uint16_t level)
{
    TMUX_SET(ptu_hal_poweramp_level_set, level);
}

uint32_t ptu_tmux_poweramp_enable_get(bool * enable)
{
    if (enable == NULL)
        return NRF_ERROR_INVALID_PARAM;
    TMUX_GET(ptu_hal_poweramp_enable_get, enable, override_values.power_amp_enable);
}

uint32_t ptu_tmux_poweramp_level_get(uint16_t * p_level)
{
    if (p_level == NULL)
        return NRF_ERROR_INVALID_PARAM;
    TMUX_GET(ptu_hal_poweramp_level_get, p_level, override_values.power_amp_level);
}

uint32_t ptu_tmux_itx_get(uint16_t * p_itx)
{
    if (p_itx == NULL)
        return NRF_ERROR_INVALID_PARAM;
    TMUX_GET(ptu_hal_itx_get, p_itx, override_values.itx);
}

uint32_t ptu_tmux_vtx_get(uint16_t * p_vtx)
{
    if (p_vtx == NULL)
        return NRF_ERROR_INVALID_PARAM;
    TMUX_GET(ptu_hal_vtx_get, p_vtx, override_values.vtx);
}

uint32_t ptu_tmux_ptx_get(uint16_t * p_ptx)
{
    if (p_ptx == NULL)
        return NRF_ERROR_INVALID_PARAM;
    TMUX_GET(ptu_hal_ptx_get, p_ptx, override_values.ptx);
}

uint32_t ptu_tmux_temperature_get(int16_t * p_temperature)
{
    if (p_temperature == NULL)
        return NRF_ERROR_INVALID_PARAM;
    TMUX_GET(ptu_hal_temperature_get, p_temperature, override_values.temperature);
}

/** @} */

/******************************************************************************
 * Test API
 ******************************************************************************/

void tmux_get_select(tmux_select_t select)
{
    get_mux_select = select;
}

void tmux_set_select(tmux_select_t select)
{
    set_mux_select = select;
    m_update_hal_from_override_values();
}

void ptu_tmux_override_values_set(ptu_tmux_override_t * values)
{
    memcpy(&override_values, values, sizeof(ptu_tmux_override_t));
    m_update_hal_from_override_values();
}

void ptu_tmux_override_values_get(ptu_tmux_override_t * values)
{
    memcpy(values, &override_values, sizeof(ptu_tmux_override_t));
}
