/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 * $LastChangedRevision$
 */

#include <string.h>
#include <stdio.h>

#include "wpt.h"
#include "nrf_soc.h"
#include "app_util_platform.h"
#include "ptu_sm.h"
#include "ptu_test_mux.h"
#include "ptu_registry.h"
#include "ptu_sensors.h"
#include "bsp.h"
#include "wireless_debug.h"
#include "nrf_gzll.h"
#include "nrf_assert.h"

/**
 * @file
 * @brief Source file for PTU UART test interface.
 */

#define DEBUG_LINE_LENGTH                      80        /**< Width of debug output table */
#define EVT_STRING_LENGTH_MAX                  20        /**< Max length of event message string */
#define VTX_INCREMENT                        1000        /**< Value by which the Vtx is incremented/decremented for each call to the simulated measurement function. */
#define ITX_INCREMENT                         100        /**< Value by which the Itx is incremented/decremented for each call to the simulated measurement function. */
#define TEMPERATURE_INCREMENT                  10        /**< Value by which the Temperature is incremented/decremented for each call to the simulated measurement function. */
#define PIPE_NUMBER                             0        /**< Use pipe 0 for the wireless debug interface (WDI) */
#define TX_PAYLOAD_LENGTH                      32        /**< Maximum PL length used by WDI */
#define UPDATE_INTERVAL                       200        /**< ms between WDI packets */

/**@struct debug_output_t
 * @brief Collection of debug output
 */
typedef struct
{
    ptu_sm_state_t current_state;
    ptu_reg_item_state_t reg_item_state[PTU_MAX_CONNECTIONS];
    bool link_encrypted[PTU_MAX_CONNECTIONS];
    ptu_sensor_data_t sensor_data;
    pru_dynamic_t dyn_params[PTU_MAX_CONNECTIONS];
    uint8_t ptu_dbg_flags;
    uint8_t registered_prus;
} debug_output_t;


static bool pru_changed[PTU_MAX_CONNECTIONS];             /**< Set to true if PRU params changed */
static char temp_string[DEBUG_LINE_LENGTH];               /**< Debug string */
static debug_output_t m_debug_output;                     /**< Debug output collection */
static char m_evt_evt_string[EVT_STRING_LENGTH_MAX] = ""; /**< Event message string */
static ptu_tmux_override_t override_values;               /**< Local copy of override values */
static bool m_simplified_beacon_print = false;            /**< Simplify print flag. Used to decide if beacon debug data should be short. */
static bool m_debug_print = false;                        /**< Used to signal that we are sending a string over air. */
static uint8_t debug_pkt[TX_PAYLOAD_LENGTH];              /**< WDI packet */
extern bool gzll_initialized;                             /**< true when gzll is initialized */

/**@brief Function for writing a string to the debug packet
 *
 * @param[in]  str String to write
 *
*/
static void debug_print_string(const char *str)
{
    size_t len = strlen(str);
    ASSERT(len < (TX_PAYLOAD_LENGTH-1));
    if(len < (TX_PAYLOAD_LENGTH-1))
    {
        strncpy((char *)&debug_pkt[1], str, len);
        debug_pkt[len+1] = 0;
        m_debug_print = true;
    }
}


/** @brief Will syntesize a load representing a load variation to the sensors module.
 *         This will be set to true for 1.5 seconds.
 */
static void m_set_load_variation_detected(void)
{
    override_values.power_amp_enable = true;

    for(uint8_t i = 0; i < PTU_Z_MEAS_BUF_LENGTH; ++i)
    {
        if(i == 0) // First call
            override_values.vtx = 10 * (PTU_ITX_MIN * PTU_Z_TX_IN_LOAD_DETECT) + 1; // Make sure v/i > Z_TX_IN_LOAD_DETECTED

        else // non - first call
            override_values.vtx = 1;

        override_values.itx = PTU_ITX_MIN;

        ptu_tmux_override_values_set(&override_values);

        (void)ptu_sensors_read(); // Triggers call to load variation detection
    }

    override_values.power_amp_enable = false;

    ptu_tmux_override_values_set(&override_values);

    (void)ptu_sensors_read(); // Triggers call to load variation detection

}

/**@brief Function for processing common test commands.
 *
 * @param[in]  cmd   Command code to process.
 *
 * @retval true Command code was processed by this function.
 * @retval false Command code was not processed by this function.
 *
 */
static bool m_ptu_process_command_common(uint8_t cmd)
{
    bool cmd_processed = true;
    uint32_t err_code;

    ptu_tmux_override_values_get(&override_values);

    switch (cmd)
    {
        case 'z':
            // Reset
            debug_print_string("\n\rCommand: Reset\r\n");
            NVIC_SystemReset();
            break;

        case 'b':
            // Beacon debug print
            if(m_simplified_beacon_print)
                debug_print_string("\n\rCommand: Beacon debug print disabled\r\n");
            else
                debug_print_string("\n\rCommand: Beacon debug print enabled\r\n");
            m_simplified_beacon_print = !m_simplified_beacon_print;
            break;

        case 'y':
            override_values.vtx += VTX_INCREMENT;
            sprintf(temp_string, "\n\rCommand: Inc VTX -> %04X\r\n", override_values.vtx);
            debug_print_string(temp_string);
            break;

        case 'h':
            override_values.vtx -= VTX_INCREMENT;
            sprintf(temp_string, "\n\rCommand: Dec VTX -> %04X\r\n", override_values.vtx);
            debug_print_string(temp_string);
            break;

        case 'u':
            override_values.itx += ITX_INCREMENT;
            sprintf(temp_string, "\n\rCommand: Inc ITX -> %04X\r\n", override_values.itx);
            debug_print_string(temp_string);
            break;

        case 'j':
            override_values.itx -= ITX_INCREMENT;
            sprintf(temp_string, "\n\rCommand: Dec ITX -> %04X\r\n", override_values.itx);
            debug_print_string(temp_string);
            break;

        case 'i':
            override_values.temperature += TEMPERATURE_INCREMENT;
            sprintf(temp_string, "\n\rCommand: Inc temperature -> %04X\r\n", override_values.temperature);
            debug_print_string(temp_string);
            break;

        case 'k':
            override_values.temperature -= TEMPERATURE_INCREMENT;
            sprintf(temp_string, "\n\rCommand: Dec temperature -> %04X\r\n", override_values.temperature);
            debug_print_string("\n\r");
            break;

        case 'o':
            debug_print_string("\n\rCommand: Test-mode enabled\n\r");
            tmux_set_select(TMUX_SELECT_OVERRIDE);
            tmux_get_select(TMUX_SELECT_OVERRIDE);
            break;

        case 'p':
            debug_print_string("SDK Version ");
            debug_print_string(SDK_VERSION);
            debug_print_string("\n\r");
            break;

        case 'l':
            debug_print_string("\n\rCommand: Test-mode disabled\n\r");
            tmux_set_select(TMUX_SELECT_NO_OVERRIDE);
            tmux_get_select(TMUX_SELECT_NO_OVERRIDE);
            break;

        case 'n':
            debug_print_string("\n\rCommand: TX power set to -40dBm\n\r");
            err_code = sd_ble_gap_tx_power_set(-40);
            APP_ERROR_CHECK(err_code);
            break;

        case 'm':
            debug_print_string("\n\rCommand: TX power set to PTU_OUTPUT_TX_POWER\n\r");
            err_code = sd_ble_gap_tx_power_set(PTU_OUTPUT_TX_POWER);
            APP_ERROR_CHECK(err_code);
            break;

        case 'v':
            debug_print_string("\n\rCommand: Set Load Variation TRUE\n\r");
            m_set_load_variation_detected();
            break;


        case '2':
            debug_print_string("\n\rCommand: Resonator ON\n\r");
            override_values.power_amp_enable = true;
            break;

        case '3':
            debug_print_string("\n\rCommand: Resonator OFF\n\r");
            override_values.power_amp_enable = true;
            break;

        case '4':
            override_values.power_amp_level += POWERAMP_TEST_VAR_STEP_SIZE;
            sprintf(temp_string, "\n\rCommand: Inc poweramp level -> %04X\r\n", override_values.power_amp_level);
            debug_print_string(temp_string);
            break;

        case '5':
            override_values.power_amp_level -= POWERAMP_TEST_VAR_STEP_SIZE;
            sprintf(temp_string, "\n\rCommand: Dec poweramp level -> %04X\r\n", override_values.power_amp_level);
            debug_print_string(temp_string);
            break;

        case '6':
            debug_print_string("\n\rptu\n\r");
            break;

        case '7':
            sprintf(temp_string, "ptu_reg_n_entries_get=%02X\r\n", ptu_reg_n_entries_get());
            debug_print_string(temp_string);
            break;

        case '8':
        #ifdef DFU_SUPPORT
            debug_print_string("Entering DFU mode ...\n\r");
            dfu_start();
            break;
        #endif //DFU_SUPPORT

        default:
            cmd_processed = false;
            break;
    }

    ptu_tmux_override_values_set(&override_values);

    return cmd_processed;
}

/**@brief Function for processing board related test commands.
 *
 * @param[in]  cmd   Command code to process.
 *
 * @retval true Command code was processed by this function.
 * @retval false Command code was not processed by this function.
 *
 */
bool ptu_process_command_board(uint8_t cmd);

/**@brief Get and process uart commands
 */
void cmd_poll(void)
{
    uint8_t cmd = 0;
    bool b_cmd_processed;
    if (!debug_cmd_available())
    {
        return;
    }

    cmd = get_debug_cmd();
    b_cmd_processed = m_ptu_process_command_common(cmd);
    if (!b_cmd_processed)
    {
        debug_print_string("\n\rCommand not implemented");
    }
}

/**@brief Print a DEBUG_LINE_LENGTH character long "horizontal line"
 * to UART.
 */
void line_print(void)
{
    int i;

    for(i = 0; i < DEBUG_LINE_LENGTH; i++)
    {
        debug_print_string("-");
    }
    debug_print_string("\r\n");
}

/**@brief Update m_debug_output variable.
 * @return true if debug output has changed.
 */
static bool debug_output_changed(void)
{
    int i;
    bool changed = false;
    uint32_t err_code;
    const ptu_sensor_data_t* sensor_data;
    uint16_t current_powerlevel;
    bool current_poweramp_status;
    static bool prev_poweramp_status;
    static uint32_t prev_powerlevel;
    ptu_reg_item_t * p_reg_item;

    err_code = ptu_sensors_data_get(&sensor_data);
    APP_ERROR_CHECK(err_code);

    if(m_simplified_beacon_print && m_debug_output.current_state == PTU_SM_STATE_POWER_SAVE)
    {
        (void)ptu_tmux_poweramp_enable_get(&current_poweramp_status);
        (void)ptu_tmux_poweramp_level_get(&current_powerlevel);

        // Beacon status has changed
        if((prev_poweramp_status != current_poweramp_status) || (prev_powerlevel != current_powerlevel))
        {
            if(current_poweramp_status && current_powerlevel == PTU_POWERAMP_INPUT_BEACON_SHORT)
                debug_print_string("s");
            else if(current_poweramp_status && current_powerlevel == PTU_POWERAMP_INPUT_BEACON_LONG)
            {
                debug_print_string("e\n\r");
                debug_print_string("l");
            }

            if(!current_poweramp_status)
                    debug_print_string("e\n\r");
        }
        prev_poweramp_status = current_poweramp_status;
        prev_powerlevel = current_powerlevel;
    }


    if(m_debug_output.current_state != ptu_sm_execute(PTU_SM_SIGNAL_NULL, NULL))
    {
        changed = true;
        m_debug_output.current_state = ptu_sm_execute(PTU_SM_SIGNAL_NULL, NULL);
    }

    if(memcmp(&m_debug_output.sensor_data, sensor_data, sizeof(ptu_sensor_data_t)) != 0)
    {
        changed = true;
        memcpy(&m_debug_output.sensor_data, sensor_data, sizeof(ptu_sensor_data_t));
        m_debug_output.ptu_dbg_flags = (uint8_t)sensor_data->load_detected
                                     | (uint8_t)sensor_data->load_var_detected << 1
                                     | (uint8_t)sensor_data->rogue_obj_detected << 2;
    }

    uint8_t registered_prus = 0;
    for(i = 0; i < PTU_MAX_CONNECTIONS; i++)
    {
        pru_changed[i] = false;
        (void)ptu_reg_item_get_from_index(i, &p_reg_item);

        if (p_reg_item->state == REG_ITEM_STATE_REGISTERED)
        {
            registered_prus |= 1 << i;
        }
        if(m_debug_output.reg_item_state[i] != p_reg_item->state)
        {
            if (p_reg_item->state == REG_ITEM_STATE_REGISTERED)
                changed = pru_changed[i] = true;
            m_debug_output.reg_item_state[i] = p_reg_item->state;
        }

        if(m_debug_output.link_encrypted[i] != p_reg_item->link_encrypted)
        {
            if (p_reg_item->state == REG_ITEM_STATE_REGISTERED)
                changed = pru_changed[i] = true;
            m_debug_output.link_encrypted[i] = p_reg_item->link_encrypted;
        }

        if(memcmp(&m_debug_output.dyn_params[i], &p_reg_item->prev_pru_dynamic, sizeof(pru_dynamic_t)) != 0)
        {
            if (p_reg_item->state == REG_ITEM_STATE_REGISTERED)
                changed = pru_changed[i] = true;
            memcpy(&m_debug_output.dyn_params[i], &p_reg_item->prev_pru_dynamic, sizeof(pru_dynamic_t));
        }
    }
    if (m_debug_output.registered_prus != registered_prus)
    {
        m_debug_output.registered_prus = registered_prus;
        changed = true;
    }

    if(strlen(m_evt_evt_string) > 0)
    {
        changed = true;
    }

    return changed;
}

/**@brief Read PRU debug data
 * @param n PRU index
 * @param[in] buf Pointer to debug buffer
 */
void get_pru_data(int n, uint8_t *buf)
{
    buf[0] = (uint8_t)(m_debug_output.dyn_params[n].vrect >> 8);
    buf[1] = (uint8_t)m_debug_output.dyn_params[n].vrect;
    buf[2] = (uint8_t)(m_debug_output.dyn_params[n].irect >> 8);
    buf[3] = (uint8_t)m_debug_output.dyn_params[n].irect;
    buf[4] = (uint8_t)(m_debug_output.dyn_params[n].vout >> 8);
    buf[5] = (uint8_t)m_debug_output.dyn_params[n].vout;
    buf[6] = (uint8_t)(m_debug_output.dyn_params[n].iout >> 8);
    buf[7] = (uint8_t)m_debug_output.dyn_params[n].iout;
    buf[8] = (uint8_t)(m_debug_output.dyn_params[n].prect >> 8);
    buf[9] = (uint8_t)m_debug_output.dyn_params[n].prect;
    buf[10]= (uint8_t)m_debug_output.dyn_params[n].putil;
    buf[11]= (uint8_t)(m_debug_output.dyn_params[n].temperature >> 8);
    buf[12]= (uint8_t)m_debug_output.dyn_params[n].temperature;
    buf[13]= (uint8_t)m_debug_output.dyn_params[n].alerts.charge_complete | (uint8_t)m_debug_output.link_encrypted[n] << 1;
}

void debug(void)
{
    int i;
    uint32_t pkt_len = 0;
    static bool more_data;
    static bool gzll_ok = true;
    static uint32_t old_ticks;

    if (gzll_ok)
    {
        uint32_t curr_ticks;
        (void)app_timer_cnt_get(&curr_ticks);
        pkt_len = 0;
        if (debug_pkt[0] == DEBUG_STR)
            m_debug_print = false;
        if (!more_data)
        {
            more_data = debug_output_changed();
        }
        if (m_debug_print)
        {
            debug_pkt[0] = DEBUG_STR;
            m_debug_print = false;
            pkt_len = strlen((const char *)debug_pkt) + 1;
        } else if(more_data)
        {
            debug_pkt[0] = DEBUG_PTU;
            debug_pkt[1] = (uint8_t)(curr_ticks >> 24);
            debug_pkt[2] = (uint8_t)(curr_ticks >> 16);
            debug_pkt[3] = (uint8_t)(curr_ticks >> 8);
            debug_pkt[4] = (uint8_t)curr_ticks;
            debug_pkt[5] = (uint8_t)m_debug_output.current_state;
            debug_pkt[6] = m_debug_output.ptu_dbg_flags;
            debug_pkt[7] = (uint8_t)(m_debug_output.sensor_data.res_impedance >> 8);
            debug_pkt[8] = (uint8_t)m_debug_output.sensor_data.res_impedance;
            debug_pkt[9] = (uint8_t)(m_debug_output.sensor_data.v_ina >> 8);
            debug_pkt[10]= (uint8_t)m_debug_output.sensor_data.v_ina;
            debug_pkt[11]= (uint8_t)(m_debug_output.sensor_data.i_ina >> 8);
            debug_pkt[12]= (uint8_t)m_debug_output.sensor_data.i_ina;
            debug_pkt[13]= (uint8_t)(m_debug_output.sensor_data.temperature >> 8);
            debug_pkt[14]= (uint8_t)m_debug_output.sensor_data.temperature;
            debug_pkt[15] = m_debug_output.registered_prus;
            debug_pkt[16] = 0;
            for(i = 0; i < PTU_MAX_CONNECTIONS; i++)
            {
                if (pru_changed[i])
                {
                    debug_pkt[16] = (uint8_t)(DEBUG_PRU0 + i);
                    get_pru_data(i, &debug_pkt[17]);
                    pru_changed[i] = false;
                    break;
                }
            }
            more_data = false;
            for (i=0;i < PTU_MAX_CONNECTIONS; i++)
            {
                if (pru_changed[i])
                {
                    more_data = true;
                    break;
                }
            }
            pkt_len = TX_PAYLOAD_LENGTH;
        }
        else
        {
            if (ticks_diff(curr_ticks, old_ticks)/32 > UPDATE_INTERVAL)
            {
                debug_pkt[0] = DEBUG_POLL;
                pkt_len = 1;
                old_ticks = curr_ticks;
            }
        }
    }
    if (pkt_len > 0)
    {
        gzll_ok = nrf_gzll_add_packet_to_tx_fifo(PIPE_NUMBER, debug_pkt, pkt_len);
    }
    else
    {
        gzll_ok = true;
    }
    cmd_poll();
}
