/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file */

#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include "nrf_assert.h"
#include "nrf_soc.h"
#include "nrf_gzll.h"
#include "app_error.h"
#include "debug.h"
#include "wireless_debug.h"
#include "wireless_debug_config.h"
#include "softdevice_handler.h"

//void notification_cb(nrf_impl_notification_t notification);
/*lint -e526 "Symbol RADIO_IRQHandler not defined" */
void RADIO_IRQHandler(void);                            ///< Radio IRQ Handler

#define PIPE_NUMBER                     0               ///< Use pipe 0 for the wireless debug interface (WDI)
#define TX_PAYLOAD_LENGTH              32               ///< GZLL Data payload length
#define ACK_PAYLOAD_LENGTH             32               ///< GZLL Ack payload length
#ifdef WDBGPTU
#define WDBG_TIMESLOT_LEN           15000               ///< Radio timeslot length (PTU)
#else
#define WDBG_TIMESLOT_LEN           25000               ///< Radio timeslot length (PRU)
#endif
#define WDBG_TIMESLOT_SUSPEND_GZLL   4000               ///< Suspend GZLL 400us before radio timeslot ends
#define WDBG_TIMESLOT_REQUEST_END    1000               ///< Request end 1000us before radio timeslot ends
#define WDBG_TIMESLOT_TIMEOUT     1000000               ///< Radio timeslot timeout (us)

static nrf_radio_request_t  m_timeslot_request;         ///< Passed to the sd_radio_request function
static uint32_t             m_slot_length;              ///< Radio timeslot length
static volatile bool        m_cmd_received = false;     ///< Set to true when a command is received
static volatile bool        m_gzll_initialized = false; ///< Set to true when gzll is initialized

static nrf_radio_signal_callback_return_param_t signal_callback_return_param; ///< Return value for the Radio session callback function
static uint8_t ack_payload[ACK_PAYLOAD_LENGTH];         ///< Buffer for the ACK payload

/**
 * @brief Fill in m_timeslot_request for next event
 */
static void m_configure_next_event(void)
{
    m_slot_length                                  = WDBG_TIMESLOT_LEN;
    m_timeslot_request.request_type                = NRF_RADIO_REQ_TYPE_EARLIEST;
    m_timeslot_request.params.earliest.hfclk       = NRF_RADIO_HFCLK_CFG_XTAL_GUARANTEED;
    m_timeslot_request.params.earliest.priority    = NRF_RADIO_PRIORITY_NORMAL;
    m_timeslot_request.params.earliest.length_us   = m_slot_length;
    m_timeslot_request.params.earliest.timeout_us  = WDBG_TIMESLOT_TIMEOUT;
}

/**
 * @brief SOC events handler
 */
void sys_evt_dispatch(uint32_t evt_id)
{
    uint32_t err_code;

    switch (evt_id)
    {
        case NRF_EVT_RADIO_SIGNAL_CALLBACK_INVALID_RETURN:
            ASSERT(false);
            break;

        case NRF_EVT_RADIO_SESSION_IDLE:
            ASSERT(false);
            break;

        case NRF_EVT_RADIO_SESSION_CLOSED:
            ASSERT(false);
            break;

        case NRF_EVT_RADIO_BLOCKED:
            m_configure_next_event();
            err_code = sd_radio_request(&m_timeslot_request);
            APP_ERROR_CHECK(err_code);
            break;

        case NRF_EVT_RADIO_CANCELED:
            m_configure_next_event();
            err_code = sd_radio_request(&m_timeslot_request);
            APP_ERROR_CHECK(err_code);
            break;

        default:
            break;
    }
}

/**
 * @brief Called when NRF_RADIO_CALLBACK_SIGNAL_TYPE_START received
 */
static void m_on_start(void)
{
    volatile bool res; // Volatile to stop compiler warning
    uint8_t channels[NRF_GZLL_CHANNEL_TABLE_SIZE] = NRF_GZLL_CHANNELS;

    if (!m_gzll_initialized)
    {
        res = nrf_gzll_init(NRF_GZLL_MODE_DEVICE);
        ASSERT(res);
        res = nrf_gzll_set_device_channel_selection_policy(NRF_GZLL_DEVICE_CHANNEL_SELECTION_POLICY_USE_CURRENT);
        ASSERT(res);
        res = nrf_gzll_set_xosc_ctl(NRF_GZLL_XOSC_CTL_MANUAL);
        ASSERT(res);
        res = nrf_gzll_set_max_tx_attempts(0);
        ASSERT(res);
        res = nrf_gzll_set_base_address_0(NRF_GZLL_BASE_ADDRESS_0);
        ASSERT(res);
        res = nrf_gzll_set_channel_table(channels, NRF_GZLL_CHANNEL_TABLE_SIZE);
        ASSERT(res);
        res = nrf_gzll_enable();
        ASSERT(res);
        m_gzll_initialized = true;
    }
    else
    {
        res = nrf_gzll_set_mode(NRF_GZLL_MODE_DEVICE);
        ASSERT(res);
    }
    NRF_TIMER0->INTENSET = TIMER_INTENSET_COMPARE0_Msk;
    NRF_TIMER0->CC[0] = m_slot_length - WDBG_TIMESLOT_SUSPEND_GZLL;
    NVIC_EnableIRQ(TIMER0_IRQn);
}

/**
 * @brief Called when NRF_RADIO_CALLBACK_SIGNAL_TYPE_TIMER0 received
 */
static void m_on_multitimer(void)
{
    if (NRF_TIMER0->EVENTS_COMPARE[0] == 1)
    {
        NRF_TIMER0->EVENTS_COMPARE[0] = 0;
        if (nrf_gzll_get_mode() != NRF_GZLL_MODE_SUSPEND)
        {
            (void)nrf_gzll_set_mode(NRF_GZLL_MODE_SUSPEND);
            NRF_TIMER0->INTENSET = TIMER_INTENSET_COMPARE0_Msk;
            NRF_TIMER0->CC[0] = m_slot_length - WDBG_TIMESLOT_REQUEST_END;
        }
        else
        {
            ASSERT(nrf_gzll_get_mode() == NRF_GZLL_MODE_SUSPEND);
            m_configure_next_event();
            signal_callback_return_param.params.request.p_next = &m_timeslot_request;
            signal_callback_return_param.callback_action = NRF_RADIO_SIGNAL_CALLBACK_ACTION_REQUEST_AND_END;
        }
    }
}

/**
 * @brief Radio session callback
 */
nrf_radio_signal_callback_return_param_t * m_radio_callback(uint8_t signal_type)
{
    signal_callback_return_param.params.request.p_next = NULL;
    signal_callback_return_param.callback_action = NRF_RADIO_SIGNAL_CALLBACK_ACTION_NONE;

    switch(signal_type)
    {
       case NRF_RADIO_CALLBACK_SIGNAL_TYPE_START:
            m_on_start();
            break;

        case NRF_RADIO_CALLBACK_SIGNAL_TYPE_RADIO:
            RADIO_IRQHandler();
            break;

        case NRF_RADIO_CALLBACK_SIGNAL_TYPE_TIMER0:
            m_on_multitimer();
            break;
    }
    return (&signal_callback_return_param);
}

/**
 * @brief Initialize this module
 */
uint32_t sd_radio_init(void)
{
    uint32_t err_code;
    err_code = softdevice_sys_evt_handler_set(sys_evt_dispatch);
    APP_ERROR_CHECK(err_code);
    err_code = sd_radio_session_open(m_radio_callback);
    if (err_code != NRF_SUCCESS)
        return err_code;
    m_configure_next_event();
    err_code = sd_radio_request(&m_timeslot_request);
    if (err_code != NRF_SUCCESS)
    {
        (void)sd_radio_session_close();
        return err_code;
    }
    return NRF_SUCCESS;
}

/**
 * @brief GZLL ACK received callback
 */
void nrf_gzll_device_tx_success(uint32_t pipe, nrf_gzll_device_tx_info_t tx_info)
{
    uint32_t ack_payload_length = ACK_PAYLOAD_LENGTH;
    if (tx_info.payload_received_in_ack)
    {
        if (nrf_gzll_fetch_packet_from_rx_fifo(pipe, ack_payload, &ack_payload_length))
        {
            ASSERT(ack_payload_length == 1);
            m_cmd_received = true;
        }
    }
}

/**
 * @brief GZLL Transmission failed callback
 */
void nrf_gzll_device_tx_failed(uint32_t pipe, nrf_gzll_device_tx_info_t tx_info)
{
}

/**
 * @brief GZLL data packet received callback
 */
void nrf_gzll_host_rx_data_ready(uint32_t pipe, nrf_gzll_host_rx_info_t rx_info)
{
}

/**
 * @brief GZLL Disabled callback.
 */
void nrf_gzll_disabled(void)
{
}

/**
 * @brief Call this function to check if a command is received
 */
bool debug_cmd_available(void)
{
    return m_cmd_received;
}

/**
 * @brief Clear the m_cmd_received and return the received command
 */
char get_debug_cmd(void)
{
    char cmd = ack_payload[0];
    m_cmd_received = false;
    return cmd;
}
